package ch.ige.adr.server;

import java.util.regex.*;
import java.util.*;


/**
 * simple functions to format user input fields
 * <p>
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 * @author Thomas Iten
 * @version $Id: TextFormat.java,v 1.1 2007/07/03 15:20:54 nydegger Exp $
 */
public class TextFormat
{

  //---------------------------------------------------------------------------
  // replace methods
  //---------------------------------------------------------------------------

  /**
   * replace all old values within the given text with the new value
   *
   * @param   text      text to replace old with new values
   * @param   oldValue  old character to replace
   * @param   newValue  new character to insert
   * @return  String    new text with replaced values
   */
  public static String replace(String text, char oldValue, char newValue)
  {
    return replace(text,Character.toString(oldValue),Character.toString(newValue));
  }


  /**
   * replace all old values within the given text with the new value
   *
   * @param   text      text to replace old with new values
   * @param   oldValue  old string to replace
   * @param   newValue  new string to insert
   * @return  String    new text with replaced values
   */
  public static String replace(String text, String oldValue, String newValue)
  {
    StringBuffer  buf = new StringBuffer(text);
    int           start,end;

    start = buf.toString().indexOf(oldValue);

    while (start > -1)
    {
      end   = start + oldValue.length();
      buf.replace(start,end,newValue);

      start = buf.toString().indexOf(oldValue,start + newValue.length());
    }

    return buf.toString();
  }



  /**
   * replace all old values within the given text with the new value. the
   * comparsion ignores the case.
   *
   * @param   text      text to replace old with new values
   * @param   oldValue  old string to replace
   * @param   newValue  new string to insert
   * @return  String    new text with replaced values
   */
  public static String replaceIgnoreCase(String text, String oldValue, String newValue)
  {
    StringBuffer  buf       = new StringBuffer(text);
    String        oldLower  = oldValue.toLowerCase();
    int           start,end;

    start = buf.toString().toLowerCase().indexOf(oldLower);

    while (start > -1)
    {
      end   = start + oldValue.length();
      buf.replace(start,end,newValue);

      start = buf.toString().toLowerCase().indexOf(oldLower,start + newValue.length());
    }

    return buf.toString();
  }


  /**
   * replace old values with new value expect for character within a group,
   * marked by the given group delimiter.
   *
   * @param   text        text to replace old with new values
   * @param   groupDelim  groups delimiter
   * @param   oldValue    old value to replace
   * @param   newValue    new value to insert
   * @return  String      formatted result
   */
  public static String replaceIgnoringGroups (String text, char groupDelim, char oldValue, char newValue)
  {
   char[]    chars   = text.toCharArray();
   int       start   = 0;
   int       end     = 0;
   boolean   ignore  = false;

    for (int i = 0; i < chars.length; i++)
    {
      // check ignore flag
      if (chars[i] == groupDelim)
      {
        ignore = ! ignore;
      }

      // copy characters
      if (ignore)
      {
        // ignore characters within a group (i.e. copy 1 to 1)
        chars[end++] = chars[i];
      }
      else
      {
        // replace old with new value or copy 1 to 1
        if ( chars[i] == oldValue )
        {
          chars[end++] = newValue;
        }
        else
        {
          chars[end++] = chars[i];
        }
      }

    } // end for

    // return formatted string
    return new String(chars,start,end);

  }


  /**
   * replace each whitespace sign with the deliverd charachter
   *
   * @param   text      input to format
   * @param   newValue  new value for the whitespace characters
   * @return  String    string with replaced whitespaces
   */
  public static String replaceWhitespace(String text, char newValue)
  {
    char[]  chars   = text.toCharArray();

    for (int i = 0; i < chars.length; i++)
    {
      if ( Character.isWhitespace(chars[i]) )
      {
        chars[i] = newValue;
      }
    } // end for

    return new String(chars);
  }


  /**
   * replace each whitespace sign with the deliverd string
   *
   * @param   text      input to format
   * @param   newValue  new value for the whitespace characters
   * @return  String    string with replaced whitespaces
   */
  public static String replaceWhitespace(String text, String newValue)
  {
    char[]        chars = text.toCharArray();
    StringBuffer  buf   = new StringBuffer();

    for (int i = 0; i < chars.length; i++)
    {
      if ( Character.isWhitespace(chars[i]) )
      {
        buf.append(newValue);
      }
      else
      {
        buf.append(chars[i]);
      }
    } // end for

    return buf.toString();
  }


  /**
   * remove leading and trailing whitespaces and reduce all n whitespaces
   * occurrences (in a row) to 1 replaceChar within the given string, expect
   * for all groups within the given delimiter.
   *
   * sample with ' as delimter and , as replace character
   *  input  = "hello   'tom   mike   pete'    how    are you   ?  "
   *  output = "hello,'tom   mike   pete',how,are,you,?"
   *
   * @param   text        input to format
   * @param   groupDelim  groups delimiter to ignore
   * @param   newValue    replace character for whitespace sign
   * @return  String      formatted output
   */
  public static String replaceWhitespaceIgnoringGroups(String text, char groupDelim, char newValue)
  {
    char[]    chars   = text.trim().toCharArray();
    int       start   = 0;
    int       end     = 0;
    int       occ     = 0;
    boolean   ignore  = false;

    for (int i = 0; i < chars.length; i++)
    {
      // check ignore flag
      if (chars[i] == groupDelim)
      {
        ignore = ! ignore;
      }

      // copy characters
      if (ignore)
      {
        // ignore whitespaces
        chars[end++] = chars[i];
      }
      else
      {
        // replace n occurrences with one
        if ( Character.isWhitespace(chars[i]) )
        {
          occ++;
        }
        else
        {
          occ=0;
        }

        if (occ == 0)
        {
          chars[end++] = chars[i];
        }
        else if (occ == 1)
        {
          chars[end++] = newValue;
        }
      }

    } // end for

    // return formatted string
    return new String(chars,start,end);
  }

  //---------------------------------------------------------------------------
  // reduce methods
  //---------------------------------------------------------------------------

  /**
   * remove leading and trailing whitespaces and reduce all n whitespaces
   * occurrences (in a row) to 1 whitespace within the given string.
   *
   * @param   text    input to format
   * @return  String  string with single whitespaces
   */
  public static String reduceWhitespace(String text)
  {
    char[]    chars   = text.trim().toCharArray();
    int       start   = 0;
    int       end     = 0;
    int       occ     = 0;

    // replace n occurrences with one
    for (int i = 0; i < chars.length; i++)
    {
      if ( Character.isWhitespace(chars[i]) )
      {
        occ++;
      }
      else
      {
        occ=0;
      }

      if (occ<2)
      {
        chars[end++] = chars[i];
      }
    } // end for

    // return formatted string
    return new String(chars,start,end);
  }

  /**
   * remove leading and trailing whitespaces and reduce all n whitespaces
   * occurrences (in a row) to 1 whitespace within the given string, expect
   * for all groups items within the given group delimiter.
   *
   * sample with ' as delimter:
   *  input  = "hello   'tom   mike   pete', how    are you   ?  "
   *  output = "hello 'tom   mike   pete', how are you ?"
   *
   * @param   text        input to format
   * @param   groupDelim  groups delimiter to ignore
   * @return  String      formatted output
   */
  public static String reduceWhitespaceIgnoringGroups(String text, char groupDelim)
  {
    char[]    chars   = text.trim().toCharArray();
    int       start   = 0;
    int       end     = 0;
    int       occ     = 0;
    boolean   ignore  = false;

    for (int i = 0; i < chars.length; i++)
    {
      // check ignore flag
      if (chars[i] == groupDelim)
      {
        ignore = ! ignore;
      }

      // copy characters
      if (ignore)
      {
        // ignore whitespaces
        chars[end++] = chars[i];
      }
      else
      {
        // replace n occurrences with one
        if ( Character.isWhitespace(chars[i]) )
        {
          occ++;
        }
        else
        {
          occ=0;
        }

        if (occ<2)
        {
          chars[end++] = chars[i];
        }
      }

    } // end for

    // return formatted string
    return new String(chars,start,end);
  }


  //---------------------------------------------------------------------------
  // compress methods
  //---------------------------------------------------------------------------

  /**
   * remove leading and trailing whitespaces and reduce all n whitespaces
   * occurrences (in a row) to 1 blank within the given string
   *
   * @param   text        input to format
   * @return  String      formatted output
   */
  public static String compressWhitespaceToBlank (String text)
  {
    char[]    chars   = text.trim().toCharArray();
    int       start   = 0;
    int       end     = 0;
    int       occ     = 0;

    // replace n occurrences with one
    for (int i = 0; i < chars.length; i++)
    {
      if ( Character.isWhitespace(chars[i]) )
      {
        occ++;
      }
      else
      {
        occ=0;
      }

      if (occ == 0)
      {
        chars[end++] = chars[i];
      }
      else if (occ == 1)
      {
        chars[end++] = ' ';
      }
    } // end for

    // return formatted string
    return new String(chars,start,end);
  }


  /**
   * Trimms a given String to the maximal defined length
   * @param str String
   * @param len int
   * @param strHasMore String
   * @return String
   */
  public static String trimmLength(String str, int len, String strHasMore)
  {
    String result = str;
    if(str != null && str.length() > len)
    {
      result = str.substring(0, len - strHasMore.length()) + strHasMore;
    } // if(str != null && str.length() > len) //

    return result;
  }


  /**
   * Trims a given String to the defined maximum length, but ignores
   * tags (anything within <code>&lt;</code> and <code>&gt;</code>).<br>
   * Note that this doesn't keep the correct tag nesting. So for instance
   * trimming <code>&lt;b&gt;Hello&lt;/b&gt;</code> to <code>5</code> with
   * <code>..</code> at the end will result in <code>&lt;b&gt;Hel..</code>, which is invalid
   * HTML.
   * @param string The String to trim
   * @param maximumLength The maximum length of the resulting <strong>including</strong>
   *        the length of <code>hasMore</code>, but <strong>excluding</strong> any tags
   * @param hasMore The String to append to the end if the original String is trimmed
   * @return The String trimmed to the maximum length with the <code>hasMore</code> String at the end
   * TODO lee - This must also work with HTML entities like &auml; which must not be split
   */
  public static String trimIgnoringTags( String string, int maximumLength, String hasMore )
  {
    final String TAG_REGEXP = "<.*?>";

    // Just return the 'hasMore' for a length that can't be taken from the String anyway
    if ( maximumLength <= 0 || maximumLength < hasMore.length() )
    {
      return hasMore;
    }

    String content = string.replaceAll( TAG_REGEXP, "" );
    // If the String doesn't contain tags, it's easy...
    if ( content.equals( string ) )
    {
      return trimmLength( string, maximumLength, hasMore );
    }
    // If the pure content is too small anyway, just return the 'hasMore' String
    if ( content.length() <= hasMore.length() )
    {
      return hasMore;
    }
    // If only the 'hasMore' String would be returned anyway, then do so
    if ( content.length() >= hasMore.length() && maximumLength == hasMore.length() )
    {
      return hasMore;
    }
    // If the maximum length is greater than the content, just return everything
    if ( maximumLength >= content.length() )
    {
      return string;
    }

    Pattern pattern = Pattern.compile( TAG_REGEXP, Pattern.MULTILINE );
    Matcher matcher = pattern.matcher( string );
    ArrayList chunks = new ArrayList();

    int effectiveMaximumLength = maximumLength - hasMore.length();
    int previousMatchEnd = 0;
    int contentLength = 0;
    boolean matchedWholeString = true;
    int hasMoreIndex = 0;

    matcher:while ( matcher.find() )
    {
      String contentPart = string.substring( previousMatchEnd, matcher.start() );
      if ( contentLength + contentPart.length() < effectiveMaximumLength )
      {
        chunks.add( contentPart );
        contentLength += contentPart.length();
      }
      else
      {
        for ( int i = 0; i < contentPart.length(); i++ )
        {
          if ( contentLength < effectiveMaximumLength )
          {
            chunks.add( contentPart.substring( i, i + 1 ) );
            contentLength++;
          }
          else if ( contentLength == maximumLength )
          {
            matchedWholeString = false;
            break matcher;
          }
          else
          {
            chunks.add( hasMore.substring( hasMoreIndex, hasMoreIndex + 1 ) );
            contentLength++;
            hasMoreIndex++;
            if ( contentLength == maximumLength )
            {
              matchedWholeString = false;
            }
          }
        }
      }

      String tag = string.substring( matcher.start(), matcher.end() );
      chunks.add( tag );
      previousMatchEnd = matcher.end();
    }
    if ( matchedWholeString )
    {
      chunks.add( string.substring( previousMatchEnd ) );
    }

    StringBuffer result = new StringBuffer();
    Iterator chunkIterator = chunks.iterator();
    while ( chunkIterator.hasNext() )
    {
      result.append( chunkIterator.next() );
    }

    /*
         LOG.warn("Trying to trim '" + string + "' to a length of " + maximumLength +
             ", but it's too short");*/
    return result.toString();
  }

  /**
   * check if the given text consists only of whitespaces. if so return null
   * otherwise return original value.
   *
   * @param   paramText    text to examine
   * @return  String  original text or null
   */
  public static String trimToNull(String paramText)
  
  {
    String text = paramText;
    
    if (text==null) return null;

    String test = compressWhitespaceToBlank(text);

    if (test.equals("") || test.equals(" "))
    {
      text = null;
    }

    return text;
  }


  //---------------------------------------------------------------------------
  // misc
  //---------------------------------------------------------------------------

  /**
   * add leading zero's to a given string
   *
   * @param   text    given string                  (sample = "2")
   * @param   len     total length of new string    (sample = 3)
   * @return  String  new string with leading zeros (sample = "002")
   */
  public static String addLeadingZero (int text, int len)
  {
    StringBuffer buf = new StringBuffer(len);

    buf.append(text);
    while (buf.length() < len)
    {
      buf.insert(0,'0');
    }
    return buf.toString();
  }


  /**
   * add leading zero's to a given string
   *
   * @param   text    given string                  (sample = "2")
   * @param   len     total length of new string    (sample = 3)
   * @return  String  new string with leading zeros (sample = "002")
   */
  public static String addLeadingZero (String text, int len)
  {
    StringBuffer buf = new StringBuffer(len);

    buf.append(text);
    while (buf.length() < len)
    {
      buf.insert(0,'0');
    }
    return buf.toString();
  }


  /**
   * removes all substrings within the given text
   *
   * @param   text        input text
   * @param   removeText  substring to remove
   * @return  String      output text
   */
  public static String remove (String text, String removeText)
  {
    StringBuffer  buf = new StringBuffer(text);
    int idx = buf.indexOf(removeText);

    while (idx >= 0)
    {
      buf.delete(idx,idx+removeText.length());
      idx = buf.indexOf(removeText);
    }

    return buf.toString();
  }



  /**
   * remove leading zero of a given string
   *
   * @param   text    input text  (sample = "003")
   * @return  String  output text (sample = "3")
   */
  public static String removeLeadingZero (String text)
  {
    int start = 0;

    while (text.charAt(start) == '0')
      start++;

    return text.substring(start,text.length());
  }


  /**
   * Replace the braces with the corresponding values
   * @param i18nValue String
   * @param values Object[]
   * @return String
   */
  public static String i18NValuesReplace(String i18nValue, Object[] values)
  {
    String value = i18nValue;
    for ( int i = 0; i < values.length; i++ )
    {
      value = value.replaceFirst( "\\{" + i + "\\}", "" + values[i]);
    }
    return value;
  }


} // end class
