package ch.ige.adr.server.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import ch.ige.adr.vo.PostOrtAVO;
import ch.ige.adr.vo.PostOrtVO;
import ch.ige.util.log.LogLevelHelper;
import ch.ige.util.server.DBHelper;

/**
 * Datenbankzugriff auf PostOrt.
 * <p>
 * Copyright (c) 2001, Eidgenössisches Institut für Geistiges Eigentum
 * @author  Ueli Ryser
 * @version $Id: PostOrtDAO.java,v 1.16 2007/07/03 15:20:54 nydegger Exp $
 */
public class PostOrtDAO
{

  /**
   * Selektiert alle Ortschaften und die dazugehörigen PLZ
   * @return Hashtable mit allen Orten und PLZ
   * @throws SQLException Unerwarteter Fehler
   */
  public Map select() throws SQLException
  {
   String query = null;
   Connection con = null;
   PreparedStatement ps = null;
   ResultSet rs = null;
   Map hashMap = new HashMap(5000);

   try
   {
     query = "SELECT trim(plz) plz, trim(ort) ort FROM post_ort";
     con = DBHelper.getConnection(DBHelper.BAGIS);
     ps = con.prepareStatement(query);
     rs = ps.executeQuery();

     while (rs.next())
      {
        String plz = rs.getString("plz");
        String ort = rs.getString("ort");
        String plzOrtKey = plz + ort;
        hashMap.put(plzOrtKey, "");
      }
      ps.close();
      return hashMap;
   }
   catch (SQLException ex)
   {
     LogLevelHelper.log(ex, PostOrtDAO.class, query);
     throw ex;
   }
   finally
   {
     DBHelper.close(ps, con);
   }
  }


  /**
   * Selektiert ein Eintrag von PostOrt
   * @param ordNrPost Ordnungsnummer der Post
   * @return Value Object mit PostOrt
   * @throws SQLException
   */
  public PostOrtVO select(int ordNrPost) throws SQLException
  {
    String query = null;
    Connection con = null;
    PreparedStatement ps = null;
    ResultSet rs = null;
    try
    {
      query = "SELECT trim(plz_typ) plz_typ, trim(plz) plz, trim(plz_zusatz) plz_zusatz, "+
              "trim(ort) ort, ort_norm, trim(kanton) kanton, trim(land_cd) land_cd, version " +
              "FROM   post_ort " +
              "WHERE  ord_nr_post = ? ";
      con = DBHelper.getConnection(DBHelper.BAGIS);
      ps = con.prepareStatement(query);
      ps.setInt(1, ordNrPost);
      rs = ps.executeQuery();
      if (!rs.next())
      {
        return null;
      }
      PostOrtVO postOrtVO = new PostOrtVO();
      postOrtVO.setPlzTyp(rs.getString("plz_typ"));
      postOrtVO.setPlz(rs.getString("plz"));
      postOrtVO.setPlzZusatz(rs.getString("plz_zusatz"));
      postOrtVO.setOrt(rs.getString("ort"));
      postOrtVO.setOrtNorm(rs.getString("ort_norm"));
      postOrtVO.setKanton(rs.getString("kanton"));
      postOrtVO.setLandCd(rs.getString("landCd"));
      postOrtVO.setVersion(rs.getInt("version"));
      return postOrtVO;
    }
    catch (SQLException ex)
    {
      LogLevelHelper.log(ex, PostOrtDAO.class, query + " ord_nr_post = " + ordNrPost);
      throw ex;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }

/**
   * Fügt einen Datensatz in die Tabelle POST_ORT ein
   * @param  vo  ValueObject
   * @exception SQLException   Unerwarteter SQL Fehler
   */
  public void insert(PostOrtVO vo) throws SQLException
  {
    String            query = null;
    Connection        con   = null;
    PreparedStatement ps    = null;

    try
    {
      query = "INSERT INTO post_ort(" +
                      "ord_nr_post, plz_typ, plz, plz_zusatz, " +
                      "ort, ort_norm, kanton, land_cd, version) " +
          "VALUES    ( ?,?,?,?,?,?,?,?,? )";
      con  = DBHelper.getConnection(DBHelper.BAGIS);
      ps = con.prepareStatement(query);
      ps.setInt(1, vo.getOrdNrPost());
      ps.setString(2, vo.getPlzTyp());
      ps.setString(3, vo.getPlz());
      ps.setString(4, vo.getPlzZusatz());
      ps.setString(5, vo.getOrt());
      ps.setString(6, vo.getOrtNorm());
      ps.setString(7, vo.getKanton());
      ps.setString(8, vo.getLandCd());
      ps.setInt(9, vo.getVersion());

      ps.executeUpdate();
    }
    catch (SQLException ex)
    {
      LogLevelHelper.log(ex, PostOrtDAO.class, query + vo.toString());
      throw ex;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Löscht alle Datensatz aus der Tabelle POST_ORT
   * @return  int            Anzahl gelöschter Datensätze
   * @exception SQLException Unerwarteter SQL Fehler
   */
  public int delete() throws SQLException
  {
    String            query = null;
    Connection        con   = null;
    PreparedStatement ps    = null;

    try
    {
      query = "DELETE FROM post_ort";
      con  = DBHelper.getConnection(DBHelper.BAGIS);
      ps = con.prepareStatement(query);
      return ps.executeUpdate();
    }
    catch (SQLException ex)
    {
      LogLevelHelper.log(ex, PostOrtDAO.class, query);
      throw ex;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }
  /**
   * Findet alle Ortschaften mit plz oder ort.
   * Anfangsziffern bzw. Anfangsbuchstaben sind ausreichend,
   * wenn nur plz oder ort angegeben wird.
   * Wird plz und ort angegeben, muss die ganze PLZ bzw. der ganze Ort angegeben
   * werden.
   * @param plz Postleitzahl der Ortschaft
   * @param ort Name der Ortschaft
   * @return ArrayList mit PostOrtAVO-Objekten mit allen gefundenen Ortschaften
   * @throws SQLException
   */
  public ArrayList select(String plz, String ort) throws SQLException
  {
    int count = 0;
    String query = null;
    ArrayList al = new ArrayList(100);
    Connection con = null;
    PreparedStatement ps = null;
    ResultSet rs = null;
    try
    {
      con = DBHelper.getConnection(DBHelper.BAGIS);
      if (plz != null && plz.length() > 0 && ort != null && ort.length() > 0)
      {
        query = "SELECT plz, ort FROM post_ort WHERE plz = ? AND ort_norm = ?";
        // mit PLZ und normalisiertem Ort suchen
        ps = con.prepareStatement(query);
        ps.setString(1, plz);
        ps.setString(2, ort);
      }
      else if (plz != null && plz.length() > 0)
      {
        query = "SELECT plz, ort FROM post_ort WHERE plz LIKE ?";
        // mit PLZ suchen
        ps = con.prepareStatement(query);
        ps.setString(1, plz + '%');
      }
      else if (ort != null && ort.length() > 0)
      {
        // mit normalisiertem Ort suchen
        query = "SELECT plz, ort FROM post_ort WHERE ort_norm LIKE ?";
        ps = con.prepareStatement(query);
        ps.setString(1, ort + '%');
      }
      else
      {
        return null;
      }
      rs = ps.executeQuery();
      while (rs.next() && count < 100)
      {
        PostOrtAVO postOrt = new PostOrtAVO();
        postOrt.plz = rs.getString("plz");
        postOrt.ort = rs.getString("ort");
        al.add(postOrt);
        count++;
      }
    }
    catch (SQLException ex)
    {
      LogLevelHelper.log(ex, PostOrtDAO.class, query + " plz = " + plz + ", ort_norm = " + ort);
      throw ex;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
    return al;
  }
}
