package ch.ige.art.server.dao;

import java.sql.*;
import ch.ige.art.vo.ArtbezVO;
import ch.ige.util.log.LogLevelHelper;
import ch.ige.util.server.DBHelper;

/**
 * DataAccessObject für artbez
 * <p>
 * Copyright (c) 2002, Eidgenössisches Institut für Geistiges Eigentum
 * @author  Stefan Nydegger
 * @version $Id: ArtbezDAO.java,v 1.6 2006/12/13 10:08:31 ruegsegger Exp $
 */

public class ArtbezDAO
{

  /**
   * Selektiert einen Datensatz aus der Tabelle artbez
   *
   * @param artNr         ArtikelNr
   * @return              ArtbezVO  Value Object
   * @throws SQLException Unerwarteter SQL Fehler
   */
  public ArtbezVO select( String artNr ) throws SQLException
  {
    Connection con = null;
    PreparedStatement ps = null;
    String query = null;
    ResultSet rs = null;
    ArtbezVO artbezVO = null;

    try
    {
      con = DBHelper.getConnection( DBHelper.BAGIS );
      query =
          "SELECT trim(art_nr) as art_nr, " +
          "trim(sprach_cd) as sprach_cd, " +
          "trim(bez_art) as bez_art, " +
          "trim(dim) as dim, " +
          "erst_dat, " +
          "erst_ben, " +
          "geaend_dat, " +
          "geaend_ben, " +
          "u_version " +
          "FROM artbez WHERE art_nr = RPAD(?, 18,' ') ";

      ps = con.prepareStatement(query);
      ps.setString( 1, artNr );
      rs = ps.executeQuery();

      if ( rs.next() )
      {
        artbezVO = new ArtbezVO();
        artbezVO.artNr = rs.getString( "art_nr" );
        artbezVO.sprachCd = rs.getString( "sprach_cd" );
        artbezVO.bezArt = rs.getString( "bez_art" );
        artbezVO.dim = rs.getString( "dim" );
        artbezVO.erstDat = rs.getDate( "erst_dat" );
        artbezVO.erstBen = rs.getString( "erst_ben" );
        artbezVO.geaendDat = rs.getDate( "geaend_dat" );
        artbezVO.geaendBen = rs.getString( "geaend_ben" );
        artbezVO.uVersion = rs.getString( "u_version" );
      }
      return artbezVO;
    }
    catch ( SQLException ex )
    {
      LogLevelHelper.log( ex, ArtbezDAO.class, query + artNr);
      throw ex;
    }
    finally
    {
      DBHelper.close( ps, con );
    }
  }

  /**
   * Fügt einen Datensatz in die Tabelle artbez ein
   *
   * @param     artbezVO     ValueObject ArtbezVO
   * @exception SQLException Unerwarteter SQL Fehler
   */
  public void insert( ArtbezVO artbezVO ) throws SQLException
  {
    Connection con = null;
    PreparedStatement ps = null;
    String query = null;

    try
    {
      con = DBHelper.getConnection( DBHelper.BAGIS );
      query =
          "INSERT INTO artbez(" +
          "art_nr, " +
          "sprach_cd, " +
          "bez_art, " +
          "dim, " +
          "erst_dat, " +
          "erst_ben, " +
          "geaend_dat, " +
          "geaend_ben, " +
          "u_version) " +
          "VALUES( ?,?,?,?,?,?,?,?,?) ";

      ps = con.prepareStatement(query);
      ps.setString( 1, artbezVO.artNr );
      ps.setString( 2, artbezVO.sprachCd );
      ps.setString( 3, artbezVO.bezArt );
      ps.setString( 4, artbezVO.dim );
      if ( artbezVO.erstDat == null )
      {
        ps.setNull( 5, java.sql.Types.DATE );
      }
      else
      {
        ps.setDate( 5, new java.sql.Date( artbezVO.erstDat.getTime() ) );
      }
      ps.setString( 6, artbezVO.erstBen );
      if ( artbezVO.geaendDat == null )
      {
        ps.setNull( 7, java.sql.Types.DATE );
      }
      else
      {
        ps.setDate( 7, new java.sql.Date( artbezVO.geaendDat.getTime() ) );
      }
      ps.setString( 8, artbezVO.geaendBen );
      ps.setString( 9, artbezVO.uVersion );

      ps.executeUpdate();
    }
    catch ( SQLException ex )
    {
      LogLevelHelper.log( ex, ArtbezDAO.class, query + artbezVO.toString());
      throw ex;
    }
    finally
    {
      DBHelper.close( ps, con );
    }
  }

  /**
   * Aktualisiert einen Datensatz in der Tabelle artbez
   *
   * @param     artbezVO      ArtbezVO ValueObject
   * @return    int           Anzahl aktualisierter Datensätze
   * @exception SQLException  Unerwarteter SQL Fehler
   */
  public int update( ArtbezVO artbezVO ) throws SQLException
  {
    Connection con = null;
    PreparedStatement ps = null;
    String query = null;

    try
    {
      con = DBHelper.getConnection( DBHelper.BAGIS );
      query =
          "UPDATE artbez SET " +
          "bez_art = ?, " +
          "geaend_dat = ?, " +
          "geaend_ben = ?, " +
          "u_version = ? " +
          "WHERE art_nr = RPAD(?, 18,' ') " +
          "AND sprach_cd = RPAD(?, 3, ' ') ";

      ps = con.prepareStatement(query);
      ps.setString( 1, artbezVO.bezArt );
      if ( artbezVO.geaendDat == null )
      {
        ps.setNull( 2, java.sql.Types.DATE );
      }
      else
      {
        ps.setDate( 2, new java.sql.Date( artbezVO.geaendDat.getTime() ) );
      }
      ps.setString( 3, artbezVO.geaendBen );
      ps.setString( 4, artbezVO.uVersion );
      ps.setString( 5, artbezVO.artNr );
      ps.setString( 6, artbezVO.sprachCd );

      return ps.executeUpdate();
    }
    catch ( SQLException ex )
    {
      LogLevelHelper.log( ex, ArtbezDAO.class, query + artbezVO.toString());
      throw ex;
    }
    finally
    {
      DBHelper.close( ps, con );
    }
  }

  /**
   * Löscht einen Datensatz aus der Tabelle artbez
   *
   * @param   artNr          Primary Key
   * @return  int            Anzahl gelöschter Datensätze
   * @exception SQLException Unerwarteter SQL Fehler
   */
  public int delete( String artNr ) throws SQLException
  {
    Connection con = null;
    PreparedStatement ps = null;
    String query = null;

    try
    {
      con = DBHelper.getConnection( DBHelper.BAGIS );

      query = "DELETE FROM artbez WHERE art_nr = RPAD(?, 18,' ') ";
      ps = con.prepareStatement(query);
      ps.setString( 1, artNr );

      return ps.executeUpdate();
    }
    catch ( SQLException ex )
    {
      LogLevelHelper.log( ex, ArtbezDAO.class, query + artNr );
      throw ex;
    }
    finally
    {
      DBHelper.close( ps, con );
    }
  }
}

