package ch.ige.image;

import java.awt.Dimension;
import java.awt.image.RenderedImage;
import java.io.Externalizable;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.WritableByteChannel;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Base class for encoded images.
 * <p>
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 * 
 * @author Michael Locher
 * @version $Id: EncodedBitmapImage.java,v 1.17 2009/04/02 17:30:41 locher Exp $
 */
public abstract class EncodedBitmapImage extends BitmapImage implements Image, Externalizable, Cloneable, EncodedImage
{
  private transient ByteBuffer data;

  public synchronized void dispose()
  {
    this.flushCache();
    this.data = null;
  }

  private synchronized void assureNotDisposed()
  {
    if ( this.data == null )
    {
      throw new IllegalStateException( "image already disposed" );
    }
  }

  @Override
  protected void finalize() throws Throwable
  {
    this.dispose();
    super.finalize();
  }

  @Override
  protected Object clone() throws CloneNotSupportedException
  {
    return this.cloneImage();
  }

  public synchronized EncodedBitmapImage cloneImage() throws CloneNotSupportedException
  {
    assureNotDisposed();
    EncodedBitmapImage clone = (EncodedBitmapImage) super.clone();

    final int size = this.data.remaining();
    final ByteBuffer copy = this.data.isDirect() ? ByteBuffer.allocateDirect( size ) : ByteBuffer.allocate( size );
    copy.put( this.data.duplicate() );
    copy.flip();
    synchronized ( clone )
    {
      clone.data = copy;
    }
    clone.releaseClonedLazyReferences();
    return clone;
  }

  protected abstract void releaseClonedLazyReferences();

  @Override
  public String toString()
  {
    return super.toString() + " - Filesize: " + this.getFilesize() / 1024 + " kB";
  }

  /** For deserialization only. */
  protected EncodedBitmapImage()
  {
    super();
  }

  protected EncodedBitmapImage( ByteBuffer data )
  {
    this.data = data;
  }

  @Deprecated
  public synchronized byte[] getData()
  {
    assureNotDisposed();
    if ( this.data.isDirect() )
    {
      byte[] d = new byte[this.data.remaining()];
      this.data.duplicate().get( d );
      return d;
    }
    return this.data.array();
  }

  // @see ch.ige.image.EncodedImage#getDataBuffer()
  public synchronized ByteBuffer getDataBuffer()
  {
    assureNotDisposed();
    return this.data.duplicate();
  }

  // @see ch.ige.image.BitmapImage#asRenderedImage()
  @Override
  public RenderedImage asRenderedImage()
  {
    // TODO Auto-generated method stub
    return null;
  }

  public String getDigest()
  {
    return EncodedBitmapImage.imageDigest( this.getDataBuffer() );
  }

  /**
   * Calculates the SHA-1 Digest.
   * 
   * @param data the image data
   * @return the 40 character hex digest
   */
  @Deprecated
  public static String imageDigest( byte[] data )
  {
    return imageDigest( ByteBuffer.wrap( data ) );
  }

  /**
   * Calculates the SHA-1 Digest.
   * 
   * @param data the image data
   * @return the 40 character hex digest
   */
  public static String imageDigest( ByteBuffer data )
  {
    try
    {
      MessageDigest md = MessageDigest.getInstance( "SHA-1" );
      md.reset();
      if ( data != null )
      {
        md.update( data.duplicate() );
      }
      byte[] digest = md.digest();
      StringBuffer result = new StringBuffer();
      for ( byte element: digest )
      {
        int next = 0xFF & element;
        if ( next < 0x10 )
        {
          result.append( '0' );
        }
        result.append( Integer.toHexString( next ) );
      }
      return result.toString();
    }
    catch ( NoSuchAlgorithmException e )
    {
      throw new RuntimeException( e );
    }
  }

  public synchronized int getFilesize()
  {
    assureNotDisposed();
    return this.data.remaining();
  }

  public synchronized void write( File destFile ) throws IOException
  {
    assureNotDisposed();
    write( destFile, this.data );
  }

  public static void write( File destFile, ByteBuffer data ) throws IOException
  {
    FileOutputStream fos = new FileOutputStream( destFile );
    try
    {
      FileChannel channel = fos.getChannel();
      try
      {
        if ( data != null )
        {
          ByteBuffer outBuffer = data.duplicate();
          while ( outBuffer.hasRemaining() )
          {
            channel.write( outBuffer );
          }
        }
      }
      finally
      {
        channel.close();
      }
    }
    finally
    {
      fos.close();
    }
  }

  static void write( File destFile, byte[] data ) throws IOException
  {
    write( destFile, data == null ? null : ByteBuffer.wrap( data ) );
  }

  public synchronized void write( OutputStream os ) throws IOException
  {
    assureNotDisposed();
    WritableByteChannel out = Channels.newChannel( os );
    ByteBuffer outData = data.duplicate();
    do
    {
      out.write( outData );
    } while ( outData.hasRemaining() );
    os.flush();
  }

  public synchronized void writeBase64( OutputStream os ) throws IOException
  {
    assureNotDisposed();
    new sun.misc.BASE64Encoder().encodeBuffer( this.data.duplicate(), os );
    os.flush();
  }

  public void writeSVG( OutputStream os ) throws IOException
  {
    PrintWriter out;
    try
    {
      out = new PrintWriter( new OutputStreamWriter( os, "UTF-8" ) );
    }
    catch ( UnsupportedEncodingException e )
    {
      out = new PrintWriter( os );
    }
    Dimension size = this.getMetadata().getSize();
    String width = Integer.toString( size.width );
    String height = Integer.toString( size.height );

    out.println( "<?xml version='1.0' encoding='UTF-8'?>\n"
        + "<!DOCTYPE svg PUBLIC '-//W3C//DTD SVG 1.0//EN' 'http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd'>\n"
        + "<svg version='1.0' xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink' width='" + width + "' height='" + height + "' >" );

    out.print( "<image xlink:type='simple' xlink:actuate='onLoad' preserveAspectRatio='none' xlink:show='embed' x='0' y='0' width='" + width + "' height='"
        + height + "' xlink:href='data:" );
    out.print( this.getMetadata().getFormat().getMIMEType() + ";base64," );
    out.flush();
    this.writeBase64( os );
    os.flush();
    out.println( "' />\n</svg>\n" );
    out.flush();
  }

  // @see java.io.Externalizable#readExternal(java.io.ObjectInput)
  @Override
  public synchronized void readExternal( ObjectInput in ) throws IOException, ClassNotFoundException
  {
    super.readExternal( in );
    this.data = IOUtil.readBuffer( in );
  }

  // @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
  @Override
  public synchronized void writeExternal( ObjectOutput out ) throws IOException
  {
    super.writeExternal( out );
    IOUtil.writeBuffer( out, this.data );
  }

}
