package ch.ige.image;

import java.awt.Dimension;
import java.awt.color.ICC_Profile;

import ch.ige.image.format.Format;

/**
 * Datacontainer for Image Metadata.
 * <p>
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 * 
 * @see ch.ige.image.Image#getWritableMetadata()
 * @author Michael Locher
 * @version $Id: MetadataBean.java,v 1.19 2009/03/19 13:26:22 locher Exp $
 */
public class MetadataBean implements WritableMetadata
{
  private static final long serialVersionUID = 1L;

  private Format format = null;

  private String colorspace = null;

  private Dimension size;

  private int dpi = -1;

  public MetadataBean( java.awt.Image sizeFromImage )
  {
    this.setSize( new Dimension( sizeFromImage.getWidth( null ), sizeFromImage.getHeight( null ) ) );
  }

  public MetadataBean( Dimension size )
  {
    this.setSize( size );
  }

  public MetadataBean( Metadata template )
  {
    super();
    this.extractFrom( template );
  }

  private void extractFrom( Metadata template )
  {
    this.format = template.getFormat();

    this.setSize( template.getSize() );

    try
    {
      this.colorspace = template.getColorspaceType();
    }
    catch ( MetadataException e )
    {
      // ignore
    }

    try
    {
      this.setDPI( template.getDPI() );
    }
    catch ( MetadataException e )
    {
      // ignore
    }

  }

  private void setSize( Dimension size )
  {
    if ( size.width < 1 || size.height < 1 )
    {
      throw new IllegalArgumentException( "width and height must be greater than 0" );
    }
    this.size = new Dimension( size );
  }

  @Override
  public boolean equals( Object obj )
  {
    if ( obj instanceof MetadataBean )
    {
      MetadataBean other = (MetadataBean) obj;
      return this.format.equals( other.format ) && this.sameSizeAndResolution( other )
          && ( this.colorspace == null ? other.colorspace == null : this.colorspace.equals( other.colorspace ) );
    }
    return super.equals( obj );
  }

  /**
   * Tests if the other has the same resolution and size.
   * 
   * @param other the other to be examined
   * @return true if size and resolution are equal
   */
  public boolean sameSizeAndResolution( MetadataBean other )
  {
    return this.size.equals( other.size ) && this.dpi == other.dpi;
  }

  @Override
  public int hashCode()
  {
    //TODO Implement hashcode !
    return super.hashCode();
  }

  public String getColorspaceType() throws MetadataException
  {
    if ( this.colorspace == null )
    {
      throw new MetadataException( "Colorspace Metadata not available" );
    }
    return colorspace;
  }

  public void setColorspaceType( String colorspace )
  {
    this.colorspace = colorspace;
  }

  public int getDPI() throws MetadataException
  {
    if ( this.dpi < 1 )
    {
      throw new MetadataException( "DPI Metadata not available" );
    }
    return dpi;
  }

  public void setDPI( int dpi )
  {
    this.dpi = dpi;
  }

  public Format getFormat()
  {
    return format;
  }

  public void setFormat( Format format )
  {
    this.format = format;
  }

  public Dimension getSize()
  {
    return this.size;
  }

  public Dimension getSize( String unit ) throws MetadataException
  {
    if ( Image.UNIT_PIXEL.equals( unit ) )
    {
      return this.size;
    }

    if ( Image.UNIT_MILLIMETER.equals( unit ) )
    {
      return BitmapImage.convertPixelToMillimeter( this.size, this.getDPI() );
    }

    throw new MetadataException( unit + " not supported" );
  }

  public void setHeight( int height )
  {
    this.size.height = height;
  }

  public void setWidth( int width )
  {
    this.size.width = width;
  }

  public boolean hasEmbeddedICCProfile()
  {
    return false;
  }

  public ICC_Profile getEmbeddedICCProfile()
  {
    return null;
  }

  @Override
  public String toString()
  {
    StringBuffer sb = new StringBuffer( this.format.toString() );

    sb.append( " (" );
    sb.append( this.size.width );
    sb.append( "px X " );
    sb.append( this.size.height );
    sb.append( "px)" );

    if ( this.dpi > 0 )
    {

      try
      {
        Dimension sizeInMillis = this.getSize( Image.UNIT_MILLIMETER );
        sb.append( " = (" );

        sb.append( sizeInMillis.width );
        sb.append( "mm X " );
        sb.append( sizeInMillis.height );
        sb.append( "mm)" );

      }
      catch ( MetadataException e )
      {
        // strange...
      }

      sb.append( " @ " );
      sb.append( this.dpi );
      sb.append( " DPI" );
    }

    if ( this.colorspace != null )
    {
      sb.append( " - " );
      sb.append( this.colorspace );
    }

    return sb.toString();
  }

  /**
   * Gets a header line for the supported CSV format.
   * 
   * @see #getCSV()
   * @return the header line
   */
  public static String getCSVHeader()
  {
    return Format.getCSVHeader() + ";width;height;dpi;colorspace";
  }

  /**
   * Gets the metadata in CSV format.
   * 
   * @return a CSV line
   */
  public String getCSV()
  {
    StringBuffer sb = new StringBuffer( this.format.getCSV() );
    sb.append( ';' );
    sb.append( this.size.width );
    sb.append( ';' );
    sb.append( this.size.height );
    sb.append( ';' );
    sb.append( this.dpi > 0 ? this.dpi : '?' );
    sb.append( ';' );
    sb.append( ( this.colorspace != null ) ? this.colorspace : "?" );
    return sb.toString();
  }

  // @see ch.ige.image.Metadata#getConversionRuleName()
  public String getConversionRuleName()
  {
    return null;
  }

  // @see ch.ige.image.Metadata#getMAC()
  public String getMAC()
  {
    return null;
  }

  // @see ch.ige.image.Metadata#getOrigMetadata()
  public String getOrigMetadata()
  {
    return null;
  }
}
