package ch.ige.image.test;

import java.awt.Dimension;
import java.nio.ByteBuffer;
import java.util.Random;

import ch.ige.image.EncodedImage;
import ch.ige.image.Image;
import ch.ige.image.ImageException;
import ch.ige.image.ImageFactory;
import ch.ige.image.ImageFactoryImpl;
import ch.ige.image.MetadataBean;
import ch.ige.image.format.Format;

/**
 * Test the Image Factory.
 * <p>
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 * 
 * @author Michael Locher
 * @version $Id: ImageFactoryTest.java,v 1.14 2009/03/23 13:22:23 locher Exp $
 */
public class ImageFactoryTest extends AbstractImageTester
{

  public ImageFactoryTest( String name )
  {
    super( name );
  }

  private EncodedImage load( String name ) throws Exception
  {
    assertNotNull( ImageFactoryTest.class.getResource( name ) );
    return new ImageFactoryImpl().loadURL( ImageFactoryTest.class.getResource( name ) );
  }

  public void testWrapAWTImage() throws Exception
  {
    if ( !java.awt.GraphicsEnvironment.isHeadless() )
    {
      java.awt.Image img = new javax.swing.ImageIcon( ImageFactoryTest.class.getResource( "circle.jpg" ) ).getImage();
      assertTrue( !( img instanceof java.awt.image.RenderedImage ) );
      ImageFactory factory = new ImageFactoryImpl();
      Image wrapped = factory.wrapImage( img );
      MetadataBean meta = new MetadataBean( wrapped.getMetadata() );

      assertTrue( "Format", meta.getFormat().isKindOf( "java" ) );
      assertEquals( "Size", new Dimension( 64, 64 ), meta.getSize() );
      assertEquals( "DEFAULT DPI Resolution", 72, meta.getDPI() );
      assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
    }
  }

  public void testParsePNG() throws Exception
  {
    Image img = this.load( "circle.png" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", format.isKindOf( "png" ) );
    assertEquals( "Compression", "deflate", format.getCompressionType() );
    assertTrue( "Lossless information", format.isLossless() );
    assertEquals( "Size", new Dimension( 64, 64 ), meta.getSize() );
    assertEquals( "Resolution", 72, meta.getDPI() );
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testParseGIF() throws Exception
  {
    Image img = this.load( "circle.gif" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "gif" ) );
    assertEquals( "Compression", "lzw", format.getCompressionType() );
    assertTrue( "Lossless information", format.isLossless() );
    assertEquals( "Size", new Dimension( 64, 64 ), meta.getSize() );
    // NB: No resolution information in GIF Format
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testParseJPEG() throws Exception
  {
    Image img = this.load( "circle.jpg" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "jpeg" ) );
    assertEquals( "Compression", "JPEG", format.getCompressionType() );
    assertTrue( "Lossless information", !format.isLossless() );
    assertEquals( "Size", new Dimension( 64, 64 ), meta.getSize() );
    assertEquals( "Resolution", 72, meta.getDPI() );
    assertEquals( "Colorspace", "YCbCr", meta.getColorspaceType() );
  }

  public void testParseTIFFnoCompress() throws Exception
  {
    Image img = this.load( "circle_noCompress.tiff" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "tiff" ) );
    assertEquals( "Compression", "None", format.getCompressionType() );
    assertTrue( "Lossless information", format.isLossless() );
    assertEquals( "Size", new Dimension( 64, 64 ), meta.getSize() );
    assertEquals( "Resolution", 72, meta.getDPI() );
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testParseTIFFdctCompress() throws Exception
  {
    Image img = this.load( "circle_dctCompress.tiff" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "tiff" ) );
    assertEquals( "Compression", "JPEG", format.getCompressionType() );
    assertTrue( "Lossless information", !format.isLossless() );
    assertEquals( "Size", new Dimension( 64, 64 ), meta.getSize() );
    assertEquals( "Resolution", 72, meta.getDPI() );
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testParseTIFFlzwCompress() throws Exception
  {
    Image img = this.load( "circle_lzwCompress.tiff" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "tiff" ) );
    assertEquals( "Compression", "LZW", format.getCompressionType() );
    assertTrue( "Lossless information", format.isLossless() );
    assertEquals( "Size", new Dimension( 64, 64 ), meta.getSize() );
    assertEquals( "Resolution", 72, meta.getDPI() );
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testParseBMP() throws Exception
  {
    Image img = this.load( "circle.bmp" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "bmp" ) );
    assertEquals( "Compression", "BI_RGB", format.getCompressionType() );
    assertTrue( "Lossless information", format.isLossless() );
    assertEquals( "Size", new Dimension( 64, 64 ), meta.getSize() );
    assertEquals( "Resolution", 72, meta.getDPI() );
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testA6png() throws Exception
  {
    Image img = this.load( "a6_300.png" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "png" ) );
    assertEquals( "Compression", "deflate", format.getCompressionType() );
    assertTrue( "Lossless information", format.isLossless() );
    assertEquals( "Size", new Dimension( 1240, 1754 ), meta.getSize() );
    assertEquals( "Resolution", 300, meta.getDPI() );
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testA6tiff() throws Exception
  {
    Image img = this.load( "a6_300.tiff" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "tiff" ) );
    assertEquals( "Compression", "LZW", format.getCompressionType() );
    assertTrue( "Lossless information", format.isLossless() );
    assertEquals( "Size", new Dimension( 1240, 1754 ), meta.getSize() );
    assertEquals( "Resolution", 300, meta.getDPI() );
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testA6gif() throws Exception
  {
    Image img = this.load( "a6_300.gif" );
    MetadataBean meta = new MetadataBean( img.getMetadata() );
    Format format = meta.getFormat();

    assertTrue( "Format", meta.getFormat().isKindOf( "gif" ) );
    assertEquals( "Compression", "lzw", format.getCompressionType() );
    assertTrue( "Lossless information", format.isLossless() );
    assertEquals( "Size", new Dimension( 1240, 1754 ), meta.getSize() );
    // NB: No resolution information in GIF Format
    assertEquals( "Colorspace", "RGB", meta.getColorspaceType() );
  }

  public void testNotAnImage() throws Exception
  {
    Random rnd = new Random( 1 );
    byte[] small = new byte[256];
    byte[] medium = new byte[4096];
    byte[] bigger = new byte[64 * 1024];
    rnd.nextBytes( small );
    rnd.nextBytes( medium );
    rnd.nextBytes( bigger );

    this.parseInvalidBytes( ByteBuffer.wrap( small ) );
    this.parseInvalidBytes( ByteBuffer.wrap( medium ) );
    this.parseInvalidBytes( ByteBuffer.wrap( bigger ) );
  }

  public void testCorruptPNG() throws Exception
  {
    EncodedImage img = this.load( "a6_300.png" );
    corruptImageAndParse( img, 2000, 2050 );
    corruptImageAndParse( img, 300, 400 );
    corruptImageAndParse( img, 10, 30 );
  }

  public void testCorruptTIFF() throws Exception
  {
    EncodedImage img = this.load( "a6_300.tiff" );
    // corruptImageAndParse(img, 2000, 2050); //Corruption not detectable!
    // corruptImageAndParse(img, 300, 400); //Corruption not detectable!
    corruptImageAndParse( img, 4, 30 );
  }

  public void testCorruptGIF() throws Exception
  {
    EncodedImage img = this.load( "a6_300.gif" );
    // corruptImageAndParse(img, 2000, 2050); //Corruption not detectable!
    // corruptImageAndParse(img, 70, 90); //Corruption not detectable!
    corruptImageAndParse( img, 10, 30 );
  }

  public void testCorruptBMP() throws Exception
  {
    EncodedImage img = this.load( "circle.bmp" );
    // corruptImageAndParse(img, 2000, 2050); //Corruption not detectable!
    // corruptImageAndParse(img, 70, 90); //Corruption not detectable!
    corruptImageAndParse( img, 10, 30 );
  }

  private void corruptImageAndParse( EncodedImage img, final int startPosCorruption, final int stopPosCorruption )
  {
    ByteBuffer data = img.getDataBuffer();
    this.insertRandomBytes( data, startPosCorruption, stopPosCorruption );
    this.parseInvalidBytes( data );
  }

  private void insertRandomBytes( ByteBuffer data, int start, int stop )
  {
    Random rnd = new Random( 1 );
    int length = data.remaining();
    int a = start > length ? length : start;
    int b = stop > length ? length : stop;
    for ( int i = a; i < b; i++ )
    {
      data.put( i, (byte) rnd.nextInt( 256 ) );
    }
  }

  private void parseInvalidBytes( ByteBuffer data )
  {
    try
    {
      ImageFactoryImpl factory = new ImageFactoryImpl();
      factory.loadBytes( data );
      fail( "random image should not be parsable" );
    }
    catch ( ImageException e )
    {
      // OK
    }
  }

}
