package ch.ige.imagehandler.batch;

import java.io.File;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;

import ch.ige.image.EncodedImage;
import ch.ige.image.Image;

/**
 * 
 *
 * <p>Copyright (c) 2008, Swiss Federal Institute of Intellectual Property</p>
 * @author   Michael Locher
 * @version  $Id: ConversionReview.java,v 1.7 2009/04/16 09:36:06 locher Exp $
 */
public class ConversionReview extends AbstractFileConsumer
{
  private static final Logger LOG = Logger.getLogger( ConversionReview.class );

  private final AtomicInteger count = new AtomicInteger( 0 );

  private final OutputStreamWriter manifest;

  private final boolean withMetadata;

  private volatile Index index;

  public ConversionReview( File destLocation, boolean withMetadata )
  {
    super( destLocation );
    this.withMetadata = withMetadata;
    this.manifest = relativeWriter( "manifest.txt" );
    this.index = new Index( 1, 100 );
  }

  // @see ch.ige.imagehandler.batch.Consumer#shutdown()
  public void shutdown()
  {
    try
    {
      this.index.shutdown();
    }
    catch ( Exception e )
    {
      LOG.error( "failed to close index", e );
    }

    try
    {
      this.manifest.close();
    }
    catch ( IOException e )
    {
      LOG.error( "failed to close manifest", e );
    }
  }

  // @see ch.ige.imagehandler.batch.Consumer#wantMetadata()
  public boolean wantMetadata()
  {
    return this.withMetadata;
  }

  // @see ch.ige.imagehandler.batch.Consumer#handleConverted(ch.ige.imagehandler.batch.JobDescription, ch.ige.imagehandler.batch.ImageWithContext, java.util.Map)
  public void handleConverted( JobDescription job, ImageWithContext<EncodedImage> img, Map<String, Future<ImageWithContext<Image>>> results )
  {
    int num = this.count.incrementAndGet();
    Index idx = null;
    try
    {
      idx = this.getIndex();
      idx.startEntry( num );
    }
    catch ( IllegalArgumentException e )
    {
      LOG.error( "failed to create index", e );
    }

    try
    {
      writeContext( idx, img, num );
      job.getSource().writeManifestEntry( this.manifest, img );
    }
    catch ( IOException e )
    {
      LOG.error( "failed to write context", e );
    }
    try
    {
      write( idx, "orig", img, num );
    }
    catch ( IOException e )
    {
      LOG.error( "failed to write original", e );
    }
    for ( Map.Entry<String, Future<ImageWithContext<Image>>> conv: results.entrySet() )
    {
      String rule = conv.getKey();
      try
      {
        try
        {
          ImageWithContext<Image> ic = conv.getValue().get();
          if ( ic.isEncoded() )
          {
            write( idx, rule, ic.asEncoded(), num );
          }
          ic.dispose();
        }
        catch ( ExecutionException e )
        {
          writeException( idx, rule, e.getCause(), num );
        }
      }
      catch ( Exception e )
      {
        LOG.error( rule + " failed", e );
      }
    }
    if ( idx != null )
    {
      idx.endEntry();
    }
    img.dispose();
  }

  private void writeContext( Index idx, ImageWithContext<EncodedImage> img, int num ) throws IOException
  {
    final Set<Entry<String, Object>> entries = img.getContextEntries();
    if ( entries.isEmpty() )
    {
      return;
    }
    Writer out = relativeWriter( String.format( "%06d_context.txt", num ) );
    try
    {
      for ( Entry<String, Object> entry: entries )
      {
        String key = entry.getKey();
        if ( !ImageWithContext.METADATA_CONTEXT_KEY.equals( key ) )
        {
          String val = entry.getValue().toString();
          out.write( key );
          out.write( ';' );
          out.write( val );
          out.write( '\n' );
          if ( idx != null )
          {
            idx.addContext( key, val );
          }
        }
      }
    }
    finally
    {
      out.close();
    }
  }

  private void writeException( Index idx, String rule, Throwable e, int num ) throws IOException
  {
    String fn = String.format( "%06d_error_%s.txt", num, rule );
    Writer out = relativeWriter( fn );
    try
    {
      e.printStackTrace( new PrintWriter( out ) );
    }
    finally
    {
      out.close();
    }
    if ( idx != null )
    {
      idx.addException( rule, fn );
    }
  }

  private void write( Index idx, String rule, ImageWithContext<EncodedImage> imageWithMeta, int num ) throws IOException
  {
    String extension = "img";
    EncodedImage image = imageWithMeta.getImage();
    try
    {
      extension = image.getMetadata().getFormat().getName().toLowerCase();
    }
    catch ( Exception e )
    {
      LOG.info( "could not determine format", e );
    }
    String fn = String.format( "%06d_%s.%s", num, rule, extension );
    File imgFile = relativeFile( fn );
    image.write( imgFile );
    String fnMetadata = writeMetadata( imageWithMeta, num, rule );
    if ( idx != null )
    {
      idx.addImage( rule, fn, fnMetadata );
    }
  }

  private String writeMetadata( ImageWithContext<EncodedImage> img, int num, String rule )
  {
    if ( img.hasMetadata() )
    {
      Writer out = null;
      try
      {
        CharSequence meta = img.getMetadata();
        if ( meta.length() > 0 )
        {
          String fn = String.format( "%06d_%s.txt", num, rule );
          out = relativeWriter( fn );
          out.append( meta );
          out.flush();
          return fn;
        }
      }
      catch ( IOException e )
      {
        LOG.error( "metadata failed", e );
      }
      finally
      {
        if ( out != null )
        {
          try
          {
            out.close();
          }
          catch ( IOException e )
          {
            LOG.error( "close failed", e );
          }
        }
      }
    }
    return null;
  }

  //-----------------------------------------------------------------

  private synchronized Index getIndex()
  {
    if ( this.index.isFull() )
    {
      Index old = this.index;
      this.index = old.createSuccessor();
      old.shutdown();
    }
    return this.index;
  }

  private class Index
  {
    private final int num;

    private final int capacity;

    private final PrintWriter out;

    private final AtomicInteger remaining;

    public Index( int num, int capacity )
    {
      super();
      this.num = num;
      this.capacity = capacity;
      this.remaining = new AtomicInteger( capacity );
      this.out = new PrintWriter( relativeWriter( fileNameByNum( this.num ) ) );
      this.writeHead();
    }

    private String getFilename()
    {
      return fileNameByNum( this.num );
    }

    private String fileNameByNum( int n )
    {
      return n == 1 ? "index.html" : "index" + n + ".html";
    }

    public Index createSuccessor()
    {
      return new Index( this.num + 1, this.capacity );
    }

    public boolean isFull()
    {
      return this.remaining.get() < 1;
    }

    public void shutdown()
    {
      this.writeTail();
      this.out.close();
    }

    //-------------------------------------------------------------------------

    private void writeHead()
    {

      out
          .println( "<html><head><style type='text/css'>div.entry { border: 1px solid black; margin: 1em; padding: 2ex; } .entry img { background-color: pink; } div.meta { display: inline; } iframe.meta { width: 25%; height: 20em; }</style></head><body>" );
      out.println( "<h1>Index</h1><p>" );
      if ( this.num != 1 )
      {
        out.println( "<a href='" + fileNameByNum( this.num - 1 ) + "'>Previous</a> |" );
      }
      out.println( "<a href='" + fileNameByNum( this.num + 1 ) + "'>Next</a>" );
      out.println( "</p>" );
    }

    public void startEntry( int num )
    {
      this.remaining.decrementAndGet();
      out.println( "<div id='img" + num + "' class='entry'>" );
      out.println( "<h2>" + num + "</h2>" );
    }

    public void addContext( String key, String value )
    {
      out.println( "<p>" + key + " : " + value + "</p>" );
    }

    public void addImage( String rule, String href, String hrefMetadata )
    {
      out.println( "<div><h3>" + rule + "</h3>" );
      out.println( "<a href='" + href + "'><img class='" + rule + "' src='" + href + "' /></a>" );
      if ( hrefMetadata != null )
      {
        out.println( "<div class='meta'><iframe class='meta' src='" + hrefMetadata + "'></iframe><br/><a href='" + hrefMetadata + "'>Metadata</a></div>" );
      }
      out.println( "</div>" );
    }

    public void addException( String rule, String href )
    {
      out.println( "<h3>" + rule + "</h3>" );
      out.println( "<p><a href='" + href + "'>Exception</a></p>" );
    }

    public void endEntry()
    {
      out.println( "</div>" );
    }

    private void writeTail()
    {
      out.println( "</body></html>" );
    }

  } // inner-class

}
