package ch.ige.imagehandler.batch;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JToolBar;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableModel;

import org.apache.log4j.Logger;

import ch.ige.image.EncodedImage;
import ch.ige.image.Image;
import ch.ige.image.gui.ImageDisplay;
import ch.ige.imagehandler.batch.BatchConvert.AbortBatchException;

import com.sun.media.jai.widget.DisplayJAI;

/**
 * 
 *
 * <p>Copyright (c) 2008, Swiss Federal Institute of Intellectual Property</p>
 * @author   Michael Locher
 * @version  $Id: DisplayConsumer.java,v 1.6 2009/03/19 13:26:21 locher Exp $
 */
public class DisplayConsumer implements Consumer
{

  private static final Logger LOG = Logger.getLogger( DisplayConsumer.class );

  private final AtomicInteger count = new AtomicInteger( 0 );

  final BlockingQueue<WhatNext> decision = new SynchronousQueue<WhatNext>();

  private final Collection<WhatNext> processingOptions = new ArrayList<WhatNext>();

  private final ExecutorService decisionWorker = Executors.newSingleThreadExecutor();

  volatile boolean quit = false;

  JFrame frame;

  JTable contextTable;

  JPanel orig;

  JPanel converted;

  JSplitPane split;

  JButton defaultDecision;

  private JSlider defaultZoomSlider;

  JPanel nextReady;

  public DisplayConsumer( boolean optionToProceedWithoutProcessing, Map<String, Consumer> processings )
  {
    super();
    this.initFrame( optionToProceedWithoutProcessing, processings );
  }

  private void initFrame( boolean optionToProceedWithoutProcessing, Map<String, Consumer> processings )
  {
    this.frame = new JFrame( "Image Conversion" );
    this.frame.addWindowListener( new WindowAdapter()
    {
      @Override
      public void windowClosing( WindowEvent e )
      {
        super.windowClosing( e );
        DisplayConsumer.this.quit = true;
        try
        {
          DisplayConsumer.this.decision.offer( new QuitAction(), 20, TimeUnit.SECONDS );
        }
        catch ( InterruptedException ex )
        {
          LOG.debug( "proper window close failed", ex );
        }
      }
    } );

    Container cp = this.frame.getContentPane();
    cp.setLayout( new BorderLayout() );

    this.orig = new JPanel();
    this.converted = new JPanel();

    this.contextTable = new JTable();
    this.contextTable.setCellSelectionEnabled( true );

    JPanel left = new JPanel( new BorderLayout() );
    left.add( this.contextTable, BorderLayout.NORTH );
    left.add( new JScrollPane( this.orig ), BorderLayout.CENTER );

    this.split = new JSplitPane( JSplitPane.HORIZONTAL_SPLIT, left, new JScrollPane( this.converted ) );
    this.split.setOneTouchExpandable( true );
    this.split.setDividerLocation( 0.5 );

    cp.add( this.split, BorderLayout.CENTER );

    JToolBar tb = new JToolBar();

    tb.add( new JButton( new QuitAction() ) );
    tb.addSeparator();

    this.nextReady = new JPanel();
    Dimension s = new Dimension( 16, 16 );
    this.nextReady.setPreferredSize( s );
    this.nextReady.setMaximumSize( s );
    tb.add( this.nextReady );
    tb.addSeparator();

    if ( optionToProceedWithoutProcessing || processings.isEmpty() )
    {
      ProceedAction proceed = new ProceedAction();
      processingOptions.add( proceed );
      this.defaultDecision = new JButton( proceed );
      tb.add( this.defaultDecision );
    }

    int i = 1;
    for ( Map.Entry<String, Consumer> proc: processings.entrySet() )
    {
      ProcessAction a = new ProcessAction( i++, proc.getKey(), proc.getValue() );
      this.processingOptions.add( a );
      JButton pb = new JButton( a );
      tb.add( pb );
      if ( this.defaultDecision == null )
      {
        this.defaultDecision = pb;
      }
    }

    tb.addSeparator();
    this.defaultZoomSlider = new JSlider( 25, 600, 200 );
    this.defaultZoomSlider.setMajorTickSpacing( 25 );
    this.defaultZoomSlider.setPaintTicks( true );
    this.defaultZoomSlider.setPaintLabels( true );
    JLabel l = new JLabel( "default zoom: " );
    l.setLabelFor( this.defaultZoomSlider );
    tb.add( l );
    tb.add( this.defaultZoomSlider );

    this.setOptionsState( false );
    cp.add( tb, BorderLayout.NORTH );

    this.orig.setLayout( new BoxLayout( this.orig, BoxLayout.Y_AXIS ) );
    this.converted.setLayout( new BoxLayout( this.converted, BoxLayout.Y_AXIS ) );

    this.frame.setSize( new Dimension( 800, 600 ) );
    this.frame.setExtendedState( Frame.MAXIMIZED_BOTH );
    this.frame.setVisible( true );
  }

  // @see ch.ige.imagehandler.batch.Consumer#notifyNextReady()
  public void notifyNextReady()
  {
    SwingUtilities.invokeLater( new Runnable()
    {
      public void run()
      {
        DisplayConsumer.this.nextReady.setBackground( Color.GREEN );
      }
    } );
  }

  // @see ch.ige.imagehandler.batch.Consumer#shutdown()
  public void shutdown()
  {
    this.decisionWorker.shutdown();
    this.shutdownAllOptions();
    this.frame.setVisible( false );
    this.frame.dispose();
  }

  // @see ch.ige.imagehandler.batch.Consumer#wantMetadata()
  public boolean wantMetadata()
  {
    for ( WhatNext o: this.processingOptions )
    {
      if ( o.wantMetadata() )
      {
        return true;
      }
    }
    return false;
  }

  // @see ch.ige.imagehandler.batch.Consumer#handleConverted(ch.ige.imagehandler.batch.JobDescription, ch.ige.imagehandler.batch.ImageWithContext, java.util.Map)
  public synchronized void handleConverted( JobDescription job, final ImageWithContext<EncodedImage> img,
      final Map<String, Future<ImageWithContext<Image>>> results ) throws AbortBatchException
  {
    if ( this.quit )
    {
      BatchConvert.abortBatch();
    }
    try
    {
      SwingUtilities.invokeAndWait( new Runnable()
      {
        public void run()
        {
          DisplayConsumer.this.nextReady.setBackground( Color.RED );
          DisplayConsumer.this.orig.setVisible( false );
          DisplayConsumer.this.converted.setVisible( false );
          DisplayConsumer.this.orig.removeAll();
          DisplayConsumer.this.converted.removeAll();

          DisplayConsumer.this.frame.setTitle( Integer.toString( DisplayConsumer.this.count.incrementAndGet() ) );

          setContextInfo( img );

          EncodedImage origImage = img.getImage();
          addImage( false, "orig", DisplayConsumer.this.orig, origImage );
          try
          {
            for ( Map.Entry<String, Future<ImageWithContext<Image>>> conv: results.entrySet() )
            {
              String rule = conv.getKey();
              try
              {
                ImageWithContext<Image> ic = conv.getValue().get();
                Image i = ic.getImage();
                if ( i instanceof EncodedImage )
                {
                  addImage( true, rule, DisplayConsumer.this.converted, (EncodedImage) i );
                }
              }
              catch ( ExecutionException e )
              {
                addException( rule, DisplayConsumer.this.converted, e );
              }
            }
          }
          catch ( InterruptedException e )
          {
            throw new RuntimeException( e );
          }

          split.setDividerLocation( 0.5 );
          DisplayConsumer.this.orig.setVisible( true );
          DisplayConsumer.this.converted.setVisible( true );
          split.repaint();

          defaultDecision.requestFocus();
        }
      } );
      waitAndProcessDecision( job, img, results );
    }
    catch ( Exception e )
    {
      LOG.warn( "handle failed", e );
    }
  }

  protected void setContextInfo( ImageWithContext img )
  {
    final List<Entry<String, Object>> entries = new ArrayList<Entry<String, Object>>( img.getContextEntries() );
    if ( entries.isEmpty() )
    {
      this.contextTable.setModel( new DefaultTableModel() );
    }
    this.contextTable.setModel( new AbstractTableModel()
    {
      /** The serialization version. */
      private static final long serialVersionUID = 1L;

      public int getColumnCount()
      {
        return 2;
      }

      public int getRowCount()
      {
        return entries.size();
      }

      public Object getValueAt( int rowIndex, int columnIndex )
      {
        Entry<String, Object> entry = entries.get( rowIndex );
        if ( columnIndex == 0 )
        {
          return entry.getKey();
        }
        else if ( columnIndex == 1 )
        {
          return entry.getValue();
        }
        else
        {
          return "";
        }
      }
    } );
  }

  void addImage( boolean zoomable, String rule, Container cp, EncodedImage image )
  {
    try
    {
      if ( zoomable )
      {
        addZoomableImage( rule, cp, image );
      }
      else
      {
        addJAIImage( rule, cp, image );
      }
    }
    catch ( Exception e )
    {
      this.addException( rule, cp, e );
    }
  }

  private void addJAIImage( String rule, Container cp, EncodedImage image )
  {
    DisplayJAI d = new DisplayJAI( image.asRenderedImage() );
    JLabel l = new JLabel( rule );
    l.setLabelFor( d );
    cp.add( l );
    cp.add( d );
  }

  private void addZoomableImage( String rule, Container cp, EncodedImage image )
  {
    final ImageDisplay d = new ImageDisplay();
    JLabel l = new JLabel( rule );
    l.setLabelFor( d );
    cp.add( l );
    cp.add( d );
    int zoomValue = this.defaultZoomSlider.getValue();
    d.setZoomFactor( zoomValue / 100.0 );
    d.setImage( image );
    cp.add( createZoomSlider( d, zoomValue ) );
  }

  private JSlider createZoomSlider( final ImageDisplay d, int zoomValue )
  {
    final JSlider zoomSlider = new JSlider( 25, 600, zoomValue );
    zoomSlider.setMajorTickSpacing( 25 );
    zoomSlider.setPaintTicks( true );
    zoomSlider.setPaintLabels( true );
    zoomSlider.addChangeListener( new ChangeListener()
    {
      public void stateChanged( ChangeEvent e )
      {
        if ( !zoomSlider.getValueIsAdjusting() )
        {
          d.setZoomFactor( zoomSlider.getValue() / 100.0 );
        }
      }
    } );
    return zoomSlider;
  }

  void addException( String rule, Container cp, Exception e )
  {
    StringWriter msg = new StringWriter();
    e.printStackTrace( new PrintWriter( msg ) );
    JTextArea txt = new JTextArea( msg.toString() );
    txt.setEditable( false );
    JLabel l = new JLabel( rule );
    l.setLabelFor( txt );
    cp.add( l );
    cp.add( txt );
  }

  //--------------------------------------

  private void setOptionsState( boolean active )
  {
    for ( WhatNext o: this.processingOptions )
    {
      o.setEnabled( active );
    }
  }

  private void shutdownAllOptions()
  {
    for ( WhatNext o: this.processingOptions )
    {
      o.shutdown();
    }
  }

  void registerDecision( final WhatNext d )
  {
    this.decisionWorker.execute( new Runnable()
    {
      public void run()
      {
        try
        {
          decision.put( d );
        }
        catch ( InterruptedException ex )
        {
          throw new RuntimeException( ex );
        }
      }
    } );
  }

  private void waitAndProcessDecision( JobDescription job, ImageWithContext img, Map<String, Future<ImageWithContext<Image>>> results )
      throws InterruptedException, AbortBatchException
  {
    WhatNext d;
    try
    {
      this.setOptionsState( true );
      d = this.waitForDecicion();
    }
    finally
    {
      this.setOptionsState( false );
    }
    this.processDecision( d, job, img, results );
  }

  private WhatNext waitForDecicion() throws InterruptedException
  {
    return this.decision.take();
  }

  public void processDecision( final WhatNext d, final JobDescription job, final ImageWithContext<EncodedImage> img,
      final Map<String, Future<ImageWithContext<Image>>> results ) throws AbortBatchException
  {
    Future<Void> done = this.decisionWorker.submit( new Callable<Void>()
    {
      // @see java.util.concurrent.Callable#call()
      public Void call() throws Exception
      {
        d.handleConverted( job, img, results );
        return null;
      }
    } );
    try
    {
      done.get();
    }
    catch ( ExecutionException e )
    {
      Throwable cause = e.getCause();
      LOG.error( "processing failed", cause );
      if ( cause instanceof AbortBatchException )
      {
        throw (AbortBatchException) cause;
      }
    }
    catch ( Exception e )
    {
      LOG.error( "processing failed", e );
    }
  }

  //--------------------------------------

  public static interface WhatNext extends Action, Consumer
  {
  }

  private abstract class AbstractWhatNext extends AbstractAction implements WhatNext
  {
    /** The serialization version. */
    private static final long serialVersionUID = 1L;

    public AbstractWhatNext( String name )
    {
      super( name );
    }

    // @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
    public void actionPerformed( ActionEvent e )
    {
      registerDecision( this );
    }

    // @see ch.ige.imagehandler.batch.Consumer#notifyNextReady()
    public void notifyNextReady()
    {
      // ignore
    }

    // @see ch.ige.imagehandler.batch.Consumer#wantMetadata()
    public boolean wantMetadata()
    {
      return false;
    }

    // @see ch.ige.imagehandler.batch.Consumer#shutdown()
    public void shutdown()
    {
      // ignore
    }
  }

  private class ProceedAction extends AbstractWhatNext
  {
    /** The serialization version. */
    private static final long serialVersionUID = 1L;

    public ProceedAction()
    {
      super( "proceed" );
      this.putValue( MNEMONIC_KEY, KeyEvent.VK_P );
    }

    // @see ch.ige.imagehandler.batch.Consumer#handleConverted(ch.ige.imagehandler.batch.JobDescription, ch.ige.imagehandler.batch.ImageWithContext, java.util.Map)
    public void handleConverted( JobDescription job, ImageWithContext<EncodedImage> img, Map<String, Future<ImageWithContext<Image>>> results )
    {
      // ignore
    }

  } // inner-class

  private class QuitAction extends AbstractWhatNext
  {
    /** The serialization version. */
    private static final long serialVersionUID = 1L;

    public QuitAction()
    {
      super( "quit" );
      this.putValue( MNEMONIC_KEY, KeyEvent.VK_Q );
    }

    // @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)\
    @Override
    public void actionPerformed( ActionEvent e )
    {
      DisplayConsumer.this.quit = true;
      super.actionPerformed( e );
    }

    // @see ch.ige.imagehandler.batch.Consumer#handleConverted(ch.ige.imagehandler.batch.JobDescription, ch.ige.imagehandler.batch.ImageWithContext, java.util.Map)
    public void handleConverted( JobDescription job, ImageWithContext<EncodedImage> img, Map<String, Future<ImageWithContext<Image>>> results )
    {
      // ignore
    }

  } // inner-class

  private class ProcessAction extends AbstractWhatNext
  {
    /** The serialization version. */
    private static final long serialVersionUID = 1L;

    private Consumer processing;

    public ProcessAction( int num, String title, Consumer processing )
    {
      super( Integer.toString( num ) + ": " + title );
      if ( num < 10 )
      {
        this.putValue( MNEMONIC_KEY, KeyEvent.VK_0 + num );
      }
      this.processing = processing;
    }

    // @see ch.ige.imagehandler.batch.Consumer#wantMetadata()
    @Override
    public boolean wantMetadata()
    {
      return this.processing.wantMetadata();
    }

    // @see ch.ige.imagehandler.batch.Consumer#shutdown()
    @Override
    public void shutdown()
    {
      this.processing.shutdown();
    }

    // @see ch.ige.imagehandler.batch.Consumer#handleConverted(ch.ige.imagehandler.batch.JobDescription, ch.ige.imagehandler.batch.ImageWithContext, java.util.Map)
    public void handleConverted( JobDescription job, ImageWithContext<EncodedImage> img, Map<String, Future<ImageWithContext<Image>>> results )
        throws AbortBatchException
    {
      this.processing.handleConverted( job, img, results );
    }

  } // inner-class

}
