package ch.ige.imagehandler.rule;

import ch.ige.image.Image;
import ch.ige.image.ImageException;

/**
 * Combines rules with the "OR" operator.
 * <p>
 * <em>Example</em><br>
 * 
 * <pre>
 *   &lt;or&gt;
 *     &lt;rule class=&quot;ch.ige.imagehandler.rule.A&quot;/&gt;
 *     &lt;rule class=&quot;ch.ige.imagehandler.rule.B&quot;/&gt;
 *     &lt;rule class=&quot;ch.ige.imagehandler.rule.C&quot;/&gt;            
 *   &lt;/or&gt;
 * </pre>
 * 
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 * 
 * @author Michael Locher
 * @version $Id: DisjunctiveList.java,v 1.9 2009/04/14 11:38:03 locher Exp $
 */
public class DisjunctiveList<I extends Image> extends RuleList<I>
{

  @Override
  protected String getConjunction()
  {
    return "OR";
  }

  /**
   * If at least one contained Checker is successful or no checkers are
   * contained the check is successful.
   * 
   * @throws RuleException
   *           if not successful the first exception which was thrown by a
   *           Checker will be emitted
   * @see ch.ige.imagehandler.rule.Checker#checkImage(ch.ige.image.Image)
   */
  public void checkImage( I image ) throws RuleException
  {
    this.logMessage();
    RuleException firstFailed = null;

    for ( Rule rule: this.getRules() )
    {
      try
      {
        if ( rule instanceof Checker )
        {
          Checker<I> checker = (Checker<I>) rule;
          checker.checkImage( image );
          return; // exit if one was successful
        }
      }
      catch ( RuleException e )
      {
        if ( firstFailed == null )
        {
          firstFailed = e;
        }
      }
    }
    if ( firstFailed != null )
    {
      throw firstFailed;
    }
  }

  /**
   * Returns the source image. DisjunctiveLists are not well suited for
   * converters (which rule should be applied?).
   * 
   * @see ch.ige.imagehandler.rule.Converter#convert(ch.ige.image.Image)
   */
  public I convert( I source ) throws RuleException, ImageException
  {
    this.logMessage();
    return source;
  }

}
