package ch.ige.imagehandler.rule;

import ch.ige.image.EncodedImage;

/**
 * Checks the byte-size of an encoded image.
 * <p>
 * <em>Parameters</em> <table border="1">
 * <tr>
 * <td>Name</td>
 * <td>Value</td>
 * </tr>
 * <tr>
 * <td>max (required)</td>
 * <td>allowed size in kilobytes</td>
 * </tr>
 * <tr>
 * <td>unit (required)</td>
 * <td>must be "kb"</td>
 * </tr>
 * </table>
 * 
 * <p>
 * <em>Example</em>
 * <pre>
 *  &lt;rule class=&quot;ch.ige.imagehandler.rule.FileSizeChecker&quot;&gt;
 *    &lt;param name=&quot;max&quot;&gt;500&lt;/param&gt;
 *    &lt;param name=&quot;unit&quot;&gt;kb&lt;/param&gt;
 *  &lt;/rule&gt;
 * </pre>
 * 
 * <p>
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 * 
 * @see ch.ige.image.EncodedImage#getFilesize()
 * 
 * @author Michael Locher
 * @version $Id: FileSizeChecker.java,v 1.11 2009/04/14 11:38:03 locher Exp $
 */
public class FileSizeChecker<I extends EncodedImage> extends AbstractRule<I> implements Checker<I>
{

  private static final String CONFIG_MAX_FILESIZE = "max";

  /* @see ch.ige.imagehandler.rule.AbstractRule#getRuleName() */
  @Override
  protected String getRuleName()
  {
    return "File Size Checker";
  }

  /* @see ch.ige.imagehandler.rule.Rule#checkParameters() */
  @Override
  public void checkParameters() throws RuleException
  {
    this.assertParameterAvailable( CONFIG_UNIT );
    this.assertParameterAvailable( CONFIG_MAX_FILESIZE );
    this.assertParameterPredicate( "only 'kb' allowed as unit", "kb".equals( this.getParameterValue( CONFIG_UNIT ).toLowerCase() ) );
    this.getMaxSize();
  }

  private int getMaxSize()
  {
    int size = Integer.parseInt( this.getParameterValue( CONFIG_MAX_FILESIZE ) );
    this.assertParameterPredicate( "filesize must be a positive integer", size > 0 );
    return size;
  }

  // --------------------------------------------------------------------------------------------
  // Check  

  /* @see ch.ige.imagehandler.rule.Checker#checkImage(ch.ige.image.Image) */
  public void checkImage( EncodedImage image ) throws RuleException
  {
    this.logCheck();
    if ( image == null )
    {
      throw new RuleException( "Only encoded images have a filesize" );
    }

    if ( this.getMaxSize() < ( image.getFilesize() / 1024 ) )
    {
      throw new RuleException( "filesize too big" );
    }

  }

}
