package ch.ige.imagehandler.rule;

import org.apache.log4j.Logger;

import ch.ige.image.Image;
import ch.ige.image.ImageException;
import ch.ige.image.MetadataException;
import ch.ige.image.color.GrayScale;
import ch.ige.image.color.GrayScale.Algorithm;

/**
 * Colors are converted to grayscale.
 * <p>
 * <em>Parameters</em> <table border="1">
 * <tr>
 * <td>Name</td>
 * <td>Value</td>
 * </tr>
 * <tr>
 * <td>algorithm (optional)</td>
 * <td>LINEAR (=default), PHOTOCD, BT601, BT709</td>
 * </tr>
 * <tr>
 * <td>skipIfGray (optional)</td>
 * <td>true (=default), false</td>
 * </tr>
 * 
 * </table>
 * <p>
 * <em>Examples</em>
 * <pre>
 *    &lt;rule class=&quot;ch.ige.imagehandler.rule.GrayScaleConverter&quot;/&gt;
 * </pre>
 * <pre>
 *    &lt;rule class=&quot;ch.ige.imagehandler.rule.GrayScaleConverter&quot;&gt;
 *      &lt;param name=&quot;algorithm&quot;&gt;BT601&lt;/param&gt;
 *      &lt;param name=&quot;skipIfGray&quot;&gt;true&lt;/param&gt; *      
 *    &lt;/rule&gt;
 * </pre>
 * 
 * <p>
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 * 
 * @see ch.ige.image.format.Format#encode(Image, boolean)
 * @author Michael Locher
 * @version $Id: GrayScaleConverter.java,v 1.7 2009/04/14 11:38:02 locher Exp $
 */
public class GrayScaleConverter<I extends Image> extends AbstractRule<I> implements Converter<I, Image>
{

  private static final Logger LOG = Logger.getLogger( GrayScaleConverter.class );

  private static final String CONFIG_ALGORITHM = "algorithm";

  private static final String CONFIG_SKIP_IF_GRAY = "skipIfGray";

  /* @see ch.ige.imagehandler.rule.AbstractRule#getRuleName() */
  @Override
  protected String getRuleName()
  {
    return "GrayScale Converter";
  }

  /* @see ch.ige.imagehandler.rule.Rule#checkParameters() */
  @Override
  public void checkParameters() throws RuleException
  {
    getAlgorithm();
  }

  private Algorithm getAlgorithm() throws RuleException
  {
    if ( this.hasParameter( CONFIG_ALGORITHM ) )
    {
      try
      {
        return Algorithm.valueOf( this.getParameterValue( CONFIG_ALGORITHM ) );
      }
      catch ( Exception e )
      {
        throw new RuleException( "algorithm parameter not valid" );
      }
    }
    return GrayScale.DEFAULT_ALGORITHM;
  }

  private boolean skipIfGray()
  {
    if ( this.hasParameter( CONFIG_SKIP_IF_GRAY ) )
    {
      return Boolean.valueOf( this.getParameterValue( CONFIG_SKIP_IF_GRAY ) ).booleanValue();
    }
    return true;
  }

  // --------------------------------------------------------------------------------------------
  // Convert

  /* @see ch.ige.imagehandler.rule.Converter#convert(ch.ige.image.Image) */
  public Image convert( I source ) throws RuleException, ImageException
  {
    if ( this.skipIfGray() )
    {
      try
      {
        if ( "GRAY".equals( source.getMetadata().getColorspaceType() ) )
        {
          return source;
        }
      }
      catch ( MetadataException e )
      {
        LOG.info( "failed to extract the colorspace information for shortcutting the grayscale operation", e );
      }
    }
    return source.grayscale( this.getAlgorithm() );
  }

}
