package ch.ige.imagehandler.rule;

import java.awt.Dimension;
import java.io.IOException;

import ch.ige.image.EncodedImage;
import ch.ige.image.Image;
import ch.ige.image.ImageException;
import ch.ige.image.MetadataException;
import ch.ige.image.format.PNGFormat;

/**
 * Creates a 72DPI PNG Thumbnail.
 * Colorspaces are standardized with a <code>StandardizeColorspaceConverter</code>.
 * <p>
 * <em>Parameters</em> <table border="1">
 * <tr>
 * <td>Name</td>
 * <td>Value</td>
 * </tr>
 * <tr>
 * <td>width (required)</td>
 * <td>width, positive integer</td>
 * </tr>
 * <tr>
 * <td>height (required)</td>
 * <td>height, positive integer</td>
 * </tr>
 * <tr>
 * <td>keepAspectRatio (required)</td>
 * <td>"true" or "false"</td>
 * </tr>
 * <tr>
 * <td>unit (required)</td>
 * <td>"mm" or "px"</td>
 * </tr>
 * </table>
 *
 * <p>
 * <em>Example</em>
 * <pre>
 *  &lt;rule class=&quot;ch.ige.imagehandler.rule.ThumbnailConverter&quot;&gt;
 *    &lt;param name=&quot;width&quot;&gt;240&lt;/param&gt;
 *    &lt;param name=&quot;height&quot;&gt;48&lt;/param&gt;
 *    &lt;param name=&quot;unit&quot;&gt;px&lt;/param&gt;
 *    &lt;param name=&quot;keepAspectRatio&quot;&gt;true&lt;/param&gt;
 *  &lt;/rule&gt;
 * </pre>
 *
 * <p>
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 *
 * @see ch.ige.image.WritableMetadata#setDPI(int)
 * @see ch.ige.image.Image#resize(Dimension)
 * @see ch.ige.image.Image#resizeKeepAspect(Dimension)
 * @see ch.ige.image.format.PNGFormat
 *
 * @author Michael Locher
 * @version $Id: ThumbnailConverter.java,v 1.18 2009/04/23 13:54:28 galli Exp $
 */
public class ThumbnailConverter<I extends Image> extends AbstractRule<I> implements Converter<I, EncodedImage>
{
  private static final PNGFormat TARGET_FORMAT = new PNGFormat();

  private static final boolean SIGN_RESULT = true;

  private static final Converter<Image, Image> STANDARDIZE_COLORSPACES = new StandardizeColorspaceConverter<Image>();

  private static final double MM_PER_INCH = 25.4;

  private static final int THUMBNAIL_DPI = 72;

  static
  {
    TARGET_FORMAT.setProgressive( true );
  }

  /* @see ch.ige.imagehandler.rule.AbstractRule#getRuleName() */
  @Override
  protected String getRuleName()
  {
    return "Thumbnail Converter";
  }

  /* @see ch.ige.imagehandler.rule.Rule#checkParameters() */
  @Override
  public void checkParameters() throws RuleException
  {
    this.assertParameterAvailable( CONFIG_WIDTH );
    this.assertParameterAvailable( CONFIG_HEIGHT );
    this.assertParameterAvailable( CONFIG_KEEP_ASPECT_RATIO );
    this.assertAcceptedUnit();
    this.getSpecifiedSize();
    this.keepAspectRatio();
  }

  private boolean keepAspectRatio()
  {
    return Boolean.valueOf( this.getParameterValue( CONFIG_KEEP_ASPECT_RATIO ) ).booleanValue();
  }

  private Dimension getSpecifiedSize()
  {
    return new Dimension( this.getWidth(), this.getHeight() );
  }

  private int getWidth()
  {
    return this.parsePositiveInteger( this.getParameterValue( CONFIG_WIDTH ) );
  }

  private int getHeight()
  {
    return this.parsePositiveInteger( this.getParameterValue( CONFIG_HEIGHT ) );
  }

  // --------------------------------------------------------------------------------------------
  // Convert

  /* @see ch.ige.imagehandler.rule.Converter#convert(ch.ige.image.Image) */
  public EncodedImage convert( I source ) throws RuleException, ImageException
  {
    this.logConvert();
    Image scaled = this.scale( STANDARDIZE_COLORSPACES.convert( source ) );
    scaled.getWritableMetadata().setDPI( 72 );
    try
    {
      return TARGET_FORMAT.encode( scaled, SIGN_RESULT );
    }
    catch ( IOException e )
    {
      throw new RuleException( e.toString() );
    }
  }

  private Image scale( Image source ) throws MetadataException
  {
    Dimension targetSize = this.isUnitPixel() ? this.getSpecifiedSize() : this.getTargetSizeInPixel( this.getSpecifiedSize() );
    if ( this.keepAspectRatio() )
    {
      return source.resizeKeepAspect( targetSize );
    }
    return source.resize( targetSize );
  }

  private Dimension getTargetSizeInPixel( Dimension specifiedSize ) throws MetadataException
  {
    final double factor = THUMBNAIL_DPI / MM_PER_INCH;
    Dimension result = new Dimension();
    result.setSize( factor * specifiedSize.width, factor * specifiedSize.height );
    return result;
  }
}
