package ch.ige.imagehandler.rule;

import java.awt.Dimension;
import java.awt.RenderingHints;
import java.awt.Transparency;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.SampleModel;
import java.awt.image.renderable.ParameterBlock;
import java.io.IOException;
import java.nio.ByteBuffer;

import javax.media.jai.ImageLayout;
import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;

import org.apache.log4j.Logger;

import ch.ige.image.BufferedImageWrapper;
import ch.ige.image.EncodedImage;
import ch.ige.image.Image;
import ch.ige.image.ImageException;
import ch.ige.image.ImageFactoryImpl;
import ch.ige.image.JAIImage;
import ch.ige.image.Metadata;
import ch.ige.image.MetadataBean;
import ch.ige.image.WritableMetadata;
import ch.ige.image.IOUtil.WritableByteBuffer;

import com.sun.image.codec.jpeg.JPEGCodec;
import com.sun.image.codec.jpeg.JPEGDecodeParam;
import com.sun.image.codec.jpeg.JPEGEncodeParam;
import com.sun.image.codec.jpeg.JPEGImageEncoder;

/**
 * Converts images from eAnmeldung.
 * <p>
 * <em>Parameters</em> <table border="1">
 * <tr>
 * <td>Name</td>
 * <td>Value</td>
 * </tr>
 * <tr>
 * <td>colorClaim</td>
 * <td>true, false</td>
 * </tr>
 * <tr>
 * <td>encode</td>
 * <td>true, false</td>
 * </tr>
 *
 * </table>
 * <p>
 * <em>Examples</em>
 * <pre>
 *    &lt;rule class=&quot;ch.ige.imagehandler.rule.TrademarkConverter&quot;&gt;
 *      &lt;param name=&quot;colorClaim&quot;&gt;false&lt;/param&gt;
 *      &lt;param name=&quot;encode&quot;&gt;false&lt;/param&gt;
 *    &lt;/rule&gt;
 * </pre>
 *
 * <p>
 * Copyright (c) 2005, Swiss Federal Institute of Intellectual Property
 *
 * @author Michael Locher
 * @author Christian Heinzer
 * @version $Id: TrademarkConverter.java,v 1.12 2009/04/14 11:38:03 locher Exp $
 */
public class TrademarkConverter<I extends EncodedImage> extends AbstractTrademarkRule<I> implements Converter<I, Image>
{
  private static final Logger LOG = Logger.getLogger( TrademarkConverter.class );

  private static final String CONFIG_ENCODE = "encode";

  private static final String CONFIG_COLORCLAIM = "colorClaim";

  /* @see ch.ige.imagehandler.rule.AbstractRule#getRuleName() */
  @Override
  protected String getRuleName()
  {
    return "Trademark Converter";
  }

  /* @see ch.ige.imagehandler.rule.Rule#checkParameters() */
  @Override
  public void checkParameters() throws RuleException
  {
    hasColorClaim();
    shouldEncode();
  }

  private boolean hasColorClaim()
  {
    this.assertParameterAvailable( CONFIG_COLORCLAIM );
    return Boolean.valueOf( this.getParameterValue( CONFIG_COLORCLAIM ) ).booleanValue();
  }

  private boolean shouldEncode()
  {
    this.assertParameterAvailable( CONFIG_ENCODE );
    return Boolean.valueOf( this.getParameterValue( CONFIG_ENCODE ) ).booleanValue();
  }

  // --------------------------------------------------------------------------------------------
  // Convert

  private static final double CM_PER_INCH = 2.54;

  private static final int MAX_RESOLUTION_DPI = 600;

  private static final int WIDTH_MAX_CM = 8;

  private static final int HEIGHT_MAX_CM = 8;

  private static final float JPEG_COMPRESSION_QUALITY = 0.75f; //0..1 where 1 is best quality

  /* @see ch.ige.imagehandler.rule.Converter#convert(ch.ige.image.Image) */
  public Image convert( I source ) throws RuleException, ImageException
  {
    if ( !( source instanceof EncodedImage ) )
    {
      throw new RuleException( "image must be encoded for this rule" );
    }
    final JAIImage parsed;
    try
    {
      ByteBuffer data = ( (EncodedImage) source ).getDataBuffer();
      parsed = JAIImage.parse( data );
    }
    catch ( Exception e )
    {
      throw new ImageException( "could not parse image with JAI", e );
    }

    try
    {
      final Metadata metadata = source.getMetadata();
      double resolution = metadata.getDPI();
      final Dimension size = metadata.getSize();
      final double width_cm = size.getWidth() * CM_PER_INCH / resolution;
      final double height_cm = size.getHeight() * CM_PER_INCH / resolution;

      // If image is too wide/high: adapt resolution
      if ( width_cm > WIDTH_MAX_CM || height_cm > HEIGHT_MAX_CM )
      {
        if ( width_cm / WIDTH_MAX_CM > height_cm / HEIGHT_MAX_CM )
        { //width is relatively farer from limit - scale both dimensions by value determined by width
          resolution = resolution * width_cm / WIDTH_MAX_CM;
        }
        else
        { //scale by height
          resolution = resolution * height_cm / HEIGHT_MAX_CM;
        }
      }

      double xScale = 1;
      double yScale = 1;
      //scale down if resolution exceeds limit
      if ( resolution > MAX_RESOLUTION_DPI )
      {
        xScale = MAX_RESOLUTION_DPI / resolution;
        yScale = MAX_RESOLUTION_DPI / resolution;
        resolution = MAX_RESOLUTION_DPI;
      }

      // JAI pipeline
      PlanarImage pipe = jaiPipeline( parsed, xScale, yScale );
      if ( !this.hasColorClaim() )
      {
        //convert to grayscale when there is no color claim
        pipe = grayScale( pipe );
      }

      final BufferedImage result = pipe.getAsBufferedImage();
      final WritableMetadata resultMetadata = new MetadataBean( metadata );
      resultMetadata.setDPI( (int) resolution );

      if ( this.shouldEncode() )
      {
        return encode( result, resolution );
      }
      return new BufferedImageWrapper( result, resultMetadata );
    }
    finally
    {
      parsed.dispose();
    }
  }

  private PlanarImage grayScale( PlanarImage img )
  {
    if ( img.getSampleModel().getNumBands() == 3 )
    { //RGB
      double[][] matrix = { { 0.114D, 0.587D, 0.299D, 0.0D } }; // Grayscale Conversion from RGB, intensity according to the CCIR-601/ITU-R BT.601 Standard recommandation
      ParameterBlock pb = new ParameterBlock();
      pb.addSource( img );
      pb.add( matrix );
      return JAI.create( "bandcombine", pb, null );
    }
    else if ( img.getSampleModel().getNumBands() == 1 )
    { //already singleband - nothing to do
      return img;
    }
    else
    { // different Color-Encoding, try to use JAI ColorConvert
      LOG.debug( "unknown color encoding - trying to use ColorConvert call" );
      try
      {
        ParameterBlock pblock = new ParameterBlock();
        pblock.addSource( img );
        ColorModel cm = new ComponentColorModel( ColorSpace.getInstance( ColorSpace.CS_GRAY ), new int[] { 8 }, false, false, Transparency.OPAQUE,
            DataBuffer.TYPE_BYTE );
        pblock.add( cm );
        ImageLayout layout = new ImageLayout();
        layout.setColorModel( cm );
        SampleModel sampleModel = cm.createCompatibleSampleModel( img.getWidth(), img.getHeight() );
        layout.setSampleModel( sampleModel );
        RenderingHints hints = new RenderingHints( javax.media.jai.JAI.KEY_IMAGE_LAYOUT, layout );
        return JAI.create( "ColorConvert", pblock, hints );
      }
      catch ( Exception e )
      {
        LOG.error( "exception converting to grayscale", e );
        return img;
      }
    }
  }

  private EncodedImage encode( final BufferedImage img, double resolution ) throws ImageException
  {
    WritableByteBuffer buf = new WritableByteBuffer( false );
    JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder( buf );
    JPEGEncodeParam encParams = encoder.getDefaultJPEGEncodeParam( img );
    encParams.setDensityUnit( JPEGDecodeParam.DENSITY_UNIT_DOTS_INCH );
    encParams.setXDensity( (int) resolution );
    encParams.setYDensity( (int) resolution );
    encParams.setQuality( JPEG_COMPRESSION_QUALITY, true );
    encoder.setJPEGEncodeParam( encParams );
    try
    {
      encoder.encode( img );
      return new ImageFactoryImpl().loadBytes( buf.getBuffer() );
    }
    catch ( IOException e )
    {
      throw new ImageException( "jpeg encoding failed", e );
    }
  }

}
