package ch.ige.ipcloadvalidity.server;

import java.sql.SQLException;
import java.util.Date;

import javax.ejb.EJBException;

import ch.ige.ipcloadvalidity.server.dao.IpcDAO;
import ch.ige.ipcloadvalidity.vo.IpcVO;
import ch.ige.ipcloadvalidity.vo.TransferVO;


/**
 * Business Object für IpcLoadValidity
 * <p>
 * Copyright (c) 2005, Eidgenössisches Institut für Geistiges Eigentum
 * @author A. Rüegsegger
 * @version $Id: IpcValidityBO.java,v 1.19 2007/01/09 10:30:32 ruegsegger Exp $
 */
public class IpcValidityBO
{
  private IpcDAO ipcDAO = new IpcDAO();


  public IpcValidityBO()
  {
  }

  /**
   * Schliesst das übergebene Symbol ab.
   * Ein Update wird nur durchgeführt, sofern das gültig_bis Datum leer oder
   * gleich dem übergebenen Datum ist (reload).
   *
   * Falls das Symbol nicht existiert, wird es eingefügt.
   * Zurückgegeben wird die Anzahl betroffener Datensätze.
   *
   * @param ipcVO IpcVO
   * @throws Exception
   * @return Integer anzahl betroffene Datensätze, 0 falls das Symbol bereits abgeschlossen ist
   */
  public Integer setValidTo( IpcVO ipcVO )
  {
    try
    {
      int anz = ipcDAO.updateValidTo( ipcVO, ipcVO.gueltigBis );
      if ( anz < 1 )
      {
        boolean exist = ipcDAO.exists( ipcVO.ipcSymb, ipcVO.gueltigVon );
        if ( !exist )
        {
          anz = ipcDAO.insert( ipcVO );
        }
      }
      return new Integer( anz );
    }
    catch ( SQLException ex )
    {
      throw new EJBException( ex );
    }
  }


  /**
   * Fügt ein neues Symbol ein.
   * Falls das Symbol schon existiert, wird Zähler auktualisiert.
   * Ist das Symbol noch nicht vorhanden, wird überprüft ob noch eine Version gültig ist.
   * Ist noch eine Version gültig, wird diese mit dem übergebenen validTo Datum abgeschlossen.
   *
   * @param ipcVO IpcVO
   * @param validTo Date
   * @throws Exception
   * @return TransferVO
   */
  public TransferVO setIpc( IpcVO ipcVO, Date validTo )
  {
    TransferVO transferVO = new TransferVO();

    try
    {
      boolean exists = ipcDAO.exists( ipcVO.ipcSymb, ipcVO.gueltigVon );
      if ( !exists )
      {
        // überprüfen ob eine gültiger version existiert ( where gültig_bis is null )
        IpcVO currentVO = ipcDAO.selectCurrent( ipcVO.ipcSymb );
        if ( currentVO != null )
        {
          // aktuelles symbol terminieren, bevor eine neue Version geschrieben wird
          if ( ipcDAO.updateValidTo( currentVO, validTo ) == 1 )
          {
            transferVO.terminates++;
            currentVO.gueltigBis = validTo; // damit in Log ersichtlich
            transferVO.terminatedVO = currentVO;
          }
          else
          {
            throw new EJBException( "insertion failed for " + ipcVO.ipcSymb +
                                 " current version has not been terminated" );
          }
        }

        transferVO.inserts = ipcDAO.insert( ipcVO );
      }
      else
      {
        // bereits vorhanden
        transferVO.exists++;
      }
    }
    catch ( SQLException ex )
    {
      // SQLException is an application exception which effects no rollback
      // so she is unpacked into EJBException (systemexception)
      throw new EJBException( ex );
    }

    return transferVO;
  }
}
