package ch.ige.jipps;

import javax.ejb.CreateException;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;

/**
 * The factory for getting a JIPPS properties service.
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   Eric Lewis
 * @version  $Id: JIPPSPropertiesFactory.java,v 1.2 2007/04/30 15:08:18 lewis_lee Exp $
 */
public class JIPPSPropertiesFactory
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( JIPPSPropertiesFactory.class );

  /**
   * Shows whether this is running on a J2EE server
   */
  private static boolean runsInJ2EEMode = false;

  /**
   * Contains the JIPPS properties within a J2EE server
   */
  private static JIPPSPropertiesLocal j2eeJippsProperties = null;

  /**
   * Contains the local JIPPS properties for testing etc.
   */
  private static JIPPSPropertiesServices localJippsProperties = null;

  /**
   * Hide this from public access, use {@link #getInstance()}.
   *
   */
  private JIPPSPropertiesFactory()
  {
    // Nothing to do
  }

  /**
   * Gets the JIPPS properties, either as a bean (on the J2EE server)
   * or a local class (for local tests).
   * @return The JIPPS properties
   */
  public final static JIPPSPropertiesServices getInstance()
  {
    if ( runsInJ2EEMode )
    {
      LOG.debug( "JIPPS properties created for J2EE server" );
      if ( JIPPSPropertiesFactory.j2eeJippsProperties == null )
      {
        throw new IllegalStateException( "Factory not initalized" );
      }
      return JIPPSPropertiesFactory.j2eeJippsProperties;
    }
    else
    {
      LOG.debug( "JIPPS properties created for standalone use" );
      return getStandaloneJippsProperties();
    }
  }

  /**
   * Gets the JIPPS properties for local use.
   * @return The JIPPS properties
   */
  public final static JIPPSPropertiesServices getStandaloneJippsProperties()
  {
    if ( localJippsProperties == null )
    {
      localJippsProperties = new JIPPSPropertiesServicesImpl();
    }
    return localJippsProperties;
  }

  /**
   * Gets the JIPPS properties for use within a J2EE server.
   * This must only be called by other beans that use this, in <code>ejbCreate()</code>.
   * @return The JIPPS properties
   * @throws CreateException
   * @throws NamingException
   */
  public final static JIPPSPropertiesLocal getEJBJippsProperties() throws CreateException, NamingException
  {
    if ( JIPPSPropertiesFactory.j2eeJippsProperties == null )
    {
      runsInJ2EEMode = true;
      InitialContext context = new InitialContext();
      JIPPSPropertiesLocalHome jippsPropertiesLocalHome = (JIPPSPropertiesLocalHome) context.lookup( JIPPSPropertiesLocalHome.COMP_NAME );
      JIPPSPropertiesFactory.j2eeJippsProperties = jippsPropertiesLocalHome.create();
    }

    return JIPPSPropertiesFactory.j2eeJippsProperties;
  }
}
