package ch.ige.jipps;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.jipps.pdf.PdfMerger;
import ch.ige.jipps.text.ConvertText;
import ch.ige.jipps.vo.AnswerVO;
import ch.ige.jipps.vo.PrintJobVO;
import ch.ige.util.address.IpiAddress;
import ch.ige.util.constants.PrintStatusConstants;
import ch.ige.util.enums.printer.ReportPaper;

/**
 * Creates lists of search hits and all other reports where the XML document has been created by another program (mostly BAGIS).
 * Note that the lists already arrive as XML documents.
 * <p>
 * Copyright (c) 2003, Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * @author Marc Bouquet
 * @version $Id: ListenControl.java,v 1.56 2009/02/12 14:37:04 lewis_lee Exp $
 */
public class ListenControl extends ReportControl
{
  /**
   * The log.
   */
  private static final Logger LOG = Logger.getLogger( ListenControl.class );

  /**
   * The configured key for a list's upper limit.
   */
  public static final String UPPER_LIST_LIMIT = "upper_list_limit";

  /**
   * Gets the number of search hits from the list.
   */
  private static final Pattern SEARCH_HITS = Pattern.compile( "<anzahl>(\\d+)</anzahl>" );

  /**
   * The constructor
   */
  public ListenControl()
  {
    super();
  }

  /**
   * Creates, saves and prints PDFs for the given print jobs.
   * @param printJobs The print jobs
   * @return The answers containing the stati for the individual print jobs
   */
  @Override
  public List<AnswerVO> makePrint( List<PrintJobVO> printJobs )
  {
    // TODO Refactor and check for possible bug below
    PrintJobVO printJobVO = null;
    List<Integer> printJobIds = null;
    String status = "";
    // If there's a batch ID in the first print job, it's a batch run
    boolean isBatchRun = printJobs.get( 0 ).batchId > 0;

    Iterator<PrintJobVO> printJobIterator = printJobs.iterator();
    while ( printJobIterator.hasNext() )
    {
      printJobVO = printJobIterator.next();

      LOG.debug( "Parameter gelesen." );
      printJobIds = new ArrayList<Integer>();

      // Since the list is printed on plain white paper, the IPI header is always used
      printJobVO.printIgekopf = true;
      status = render( printJobVO, printJobVO.formTyp.toLowerCase() );

      printJobVO.printJobId = getNextPrintJobId();
      printJobVO.paperMode = ReportPaper.PAPER_PLAIN.toDB();
      printJobVO.anzKopien = PrintJobVO.NO_COPIES;
      storePdf( printJobVO );

      // Send to JMS when the print job is ready for printing (not in batch)
      if ( !isBatchRun && printJobVO.druckStatus == PrintStatusConstants.READY_FOR_PRINTING )
      {
        sendPrintJob( printJobVO.printJobId, null, printJobVO.druckerId );
        logJmsMessage( isBatchRun, printJobVO, "(isSave) = 0 , (reportName) = null , (path) = null" );
      }
      printJobIds.add( new Integer( printJobVO.printJobId ) );
    }
    // TODO This fills the answer only for the last print job - a bug?
    addAnswer( printJobIds, printJobVO.formTyp, null, status, printJobVO.formattedApplicationNumber, printJobVO.formattedIpNumber );

    if ( isBatchRun && printJobVO.druckStatus == PrintStatusConstants.READY_FOR_PRINTING )
    {
      sendBatchJob( printJobVO.batchId, null, printJobVO.druckerId, false, null, null );
      logJmsMessage( isBatchRun, printJobVO, null );
    }
    return this.answers;
  }

  /**
   * Logs the information sent to JMS.
   * @param isBatchRun Shows whether the print job is part of a batch run
   * @param printJob The print job
   * @param additionalMessage Any additional information
   */
  private void logJmsMessage( boolean isBatchRun, PrintJobVO printJob, String additionalMessage )
  {
    StringBuffer message = new StringBuffer();
    message.append( "*** JMS-Nachricht (" );
    if ( isBatchRun )
    {
      message.append( "Batch" );
    }
    else
    {
      message.append( "Original" );
    }
    message.append( ") wurde gesendet:" );
    message.append( "(isBatch) = " );
    message.append( isBatchRun );
    message.append( "(PrintJobId) = " );
    message.append( printJob.printJobId );
    message.append( "(DruckerId) = " );
    message.append( printJob.druckerId );
    message.append( "(BatchId) = " );
    message.append( printJob.batchId );
    if ( additionalMessage != null )
    {
      message.append( additionalMessage );
    }
    LOG.info( message.toString() );
  }

  /**
   * Creates a merged PDF stream for the given print jobs.
   * @param printJobs a <code>List</code> of {@link ch.ige.jipps.vo.PrintJobVO} objects
   * @return a merged PDF stream for the print jobs
   * @throws Exception if the PDFs can't be created
   */
  @Override
  public byte[] makePdf( List<PrintJobVO> printJobs ) throws Exception
  {
    List<byte[]> pdfFiles = new ArrayList<byte[]>();

    for ( PrintJobVO printJobVO: printJobs )
    {
      String status = render( printJobVO, printJobVO.formTyp.toLowerCase() );
      if ( !status.equals( AnswerVO.STATUS_OK ) )
      {
        LOG.error( "Error while rendering list with print job ID " + printJobVO.printJobId + ": " + status );
      }
      pdfFiles.add( printJobVO.printJobFile );
    }
    return PdfMerger.mergePdfs( pdfFiles );
  }

  /**
   * Renders the XML document into a PDF and saves it in the <code>printJob</code>.
   * @param printJob The print job containing XML in {@link PrintJobVO#listData}; its {@link PrintJobVO#printJobFile}
   * is filled with the PDF
   * @param xslName The name of the XSL to use for conversion
   * @return The print status, which is either {@link AnswerVO#STATUS_OK} or which contains the error message
   */
  private String render( PrintJobVO printJob, final String xslName )
  {
    checkMaximumNumberOfSearchHits( printJob );

    String status = AnswerVO.STATUS_OK;
    try
    {

      // Add the IPI header
      printJob.listData = printJob.listData.replaceFirst( "<header>", "<header><igekopf>" + createIpiHeader( printJob.printIgekopf ) + "</igekopf>" );
      // Escape everything that confuses
      printJob.listData = ConvertText.replaceListTags( printJob.listData );
      printJob.listData = StringUtils.replace( printJob.listData, "&", "&amp;" );
      LOG.debug( "ListData " + printJob.formTyp + ": " + printJob.listData );
      printJob.printJobFile = createPdf( printJob.listData, xslName );
    }
    catch ( Exception ex )
    {
      LOG.error( "Rendering of PDF failed", ex );
      status = "Exception in ListenConrol.render(): " + ex.getMessage();
    }
    return status;
  }

  /**
   * Checks that the number of search hits is not higher than the
   * maximum number defined by the configuration with the key {@link #UPPER_LIST_LIMIT}.
   * @param printJob The print job to check, specifically its XML document
   * @throws IllegalArgumentException if the number of search hits is too high
   * @throws IllegalStateException if the number of search hits can't be read
   */
  private void checkMaximumNumberOfSearchHits( PrintJobVO printJob ) throws IllegalArgumentException, IllegalStateException
  {
    if ( printJob.formTyp.equals( PrintJobVO.LISTEN ) )
    {
      // Get the number of search hits.
      Matcher searchHitsMatcher = SEARCH_HITS.matcher( printJob.listData );
      Integer searchHits;
      if ( searchHitsMatcher.find() )
      {
        searchHits = new Integer( searchHitsMatcher.group( 1 ) );
      }
      else
      {
        throw new IllegalStateException( "Can't find the number of search hits in the list" );
      }

      // Get the configured maximum number of search hits and check if the list is within range.
      String upperListLimit = getJippsProperties().getProperty( UPPER_LIST_LIMIT );
      Integer maximumSearchHits;
      if ( StringUtils.isEmpty( upperListLimit ) )
      {
        throw new IllegalStateException( "Found no configuration for maximum number of search hits" );
      }
      else
      {
        maximumSearchHits = new Integer( upperListLimit );
      }

      if ( searchHits.intValue() > maximumSearchHits.intValue() )
      {
        throw new IllegalArgumentException( "Maximale Anzahl Suchtreffer zum Ausdrucken übertroffen, Liste wird nicht gedruckt." );
      }
    }
  }

  /**
   * Creates the IPI header with the IPI's address.
   * @param printIpiHeader If this is <code>true</code>, the IPI header is printed, if it's <code>false</code>,
   * it's not printed
   * @return The IPI header as XML
   */
  private String createIpiHeader( boolean printIpiHeader )
  {
    StringBuffer ipiHeader = new StringBuffer();
    appendTag( ipiHeader, "igekopfDe", IpiAddress.getGermanName(), printIpiHeader );
    appendTag( ipiHeader, "igekopfFr", IpiAddress.getFrenchName(), printIpiHeader );
    appendTag( ipiHeader, "igekopfIt", IpiAddress.getItalianName(), printIpiHeader );
    appendTag( ipiHeader, "igekopfEn", IpiAddress.getEnglishName(), printIpiHeader );
    appendTag( ipiHeader, "igekopfAdr", IpiAddress.getFullGermanAddress(), printIpiHeader );
    appendTag( ipiHeader, "igekopfKto", "", printIpiHeader );
    return ipiHeader.toString();
  }

  /**
   * Appends a tag to <code>xml</code>, using the <code>value</code> depending on <code>addValue</code>.
   * @param xml The XML to append to
   * @param name The name of the tag, for instance <code>igekopfDe</code>
   * @param value The value to add, depending on <code>addValue</code>
   * @param addValue If this is <code>true</code>, the <code>value</code> is added within the tag
   */
  private void appendTag( StringBuffer xml, String name, String value, boolean addValue )
  {
    xml.append( "<" );
    xml.append( name );
    xml.append( ">" );
    if ( addValue )
    {
      xml.append( value );
    }
    xml.append( "</" );
    xml.append( name );
    xml.append( ">" );
  }
}
