package ch.ige.jipps;

import java.rmi.RemoteException;
import java.sql.Date;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.List;

import javax.ejb.CreateException;
import javax.naming.NamingException;

import org.apache.log4j.Logger;

import ch.ige.dbutils.dao.UnitOfWorkException;
import ch.ige.jipps.constants.XslConstants;
import ch.ige.jipps.dao.BfakturDAO;
import ch.ige.jipps.dao.FaktparDAO;
import ch.ige.jipps.dao.LoeschungDAO;
import ch.ige.jipps.dao.LoeschungErrorDAO;
import ch.ige.jipps.dao.LoeschungLaufDAO;
import ch.ige.jipps.vo.DruckPeriVO;
import ch.ige.jipps.vo.FaktparVO;
import ch.ige.jipps.vo.LetterVO;
import ch.ige.jipps.vo.LoeschungLaufVO;
import ch.ige.jipps.vo.LoeschungVO;
import ch.ige.jipps.vo.PrintJobVO;
import ch.ige.jipps.xml.LetterWriteXml;
import ch.ige.util.constants.CancellationConstants;
import ch.ige.util.constants.CancellationProcessingConstants;
import ch.ige.util.constants.IpTypeConstants;
import ch.ige.util.constants.PrintStatusConstants;
import ch.ige.util.date.DateFormatter;
import ch.ige.util.enums.printer.ReportPaper;

/**
 * LoeschungControl: Steuerungsklasse für das Aufbereiten der Löschungsanzeigen.
 * <p>
 * Copyright (c) 2002, Eidgenössisches Institut für Geistiges Eigentum
 * @author Marc Bouquet
 * @version $Id: LoeschungControl.java,v 1.99 2008/10/28 14:21:23 lang Exp $
 */
public class LoeschungControl extends ReportControl
{
  /**
   * 
   */
  private static final String LOESCHUNGEN_PRINTJOB_NAME = "Loeschungen";

  private BfakturDAO bfaktDAO = null;

  private FaktparDAO faktparDAO = null;

  private LoeschungDAO loeschungDAO = null;

  private LoeschungLaufDAO loeschungLaufDAO = null;

  private LoeschungErrorDAO loeschErrorDAO = null;

  private int anzBfaktGel = 0;

  private int countError = 0; // Counter für fehlerhafte Löschungen.

  private int anzErstPrintJobs = 0; // Anzahl erstellte Löschungen als PrintJobs

  // wird für Email verwendet.
  private String dateStart;

  private String dateEnde;

  // Attribute für IGELogging
  private static final Logger LOG = Logger.getLogger( LoeschungControl.class );

  /**
   * Mit dieser Methode werden die fehlerhaften Löschungsanzeigen vom Jipps-Client an den Server übergeben. Die Daten
   * werden noch einmal verarbeitet.
   * @param bfaktIdList Liste mit den fehlerhaften Bfakt_Id's
   * @param laufId Nummer des Löschungslaufes
   */
  public void doAufbereitung( List bfaktIdList, int laufId )
  {
    try
    {
      loadLoeschung( bfaktIdList, laufId );
    }
    catch ( Exception ex )
    {
      LOG.error( "Fehler in doAufbreiten (direkt)", ex );
    }
  }

  /**
   * Diese Methode wird von Message Driven Bean aufgerufen und startet den gesamten Prozess für die Aufbereitung der
   * Löschungsanzeigen.
   * @param loeschTyp dieser kann folgende Werte haben Loeschung oder Schutzablauf
   */
  public void doAufbereitung( String loeschTyp )
  {
    LoeschungLaufVO laufVO = new LoeschungLaufVO();

    try
    {
      LOG.info( "Aufbereitung der Löschungen wurde gestartet." );
      List arrlist = loadBfaktur( loeschTyp, laufVO );
      anzBfaktGel = arrlist.size();
      /** TODO: Inhalt der Tabelle LOESCHUNG_LAUF löschen. */

      insertLoeschungLauf( laufVO );
      LOG.debug( "Insert in Tabelle LoeschungLauf wurde ausgeführt (LoeschungsLaufId: " + laufVO.loeschungLaufId + ")" );

      laufVO.anzLoeschGen = loadLoeschung( arrlist, laufVO.loeschungLaufId );
      LOG.debug( "Loeschungsanzeigen sind aufbereitet." );

      updateLoeschungLauf( laufVO );
      LOG.debug( "Update in Tabelle LoeschungLauf wurde ausgeführt." );

      createMail( laufVO );
      LOG.info( "Aufbereitung der Löschungen ist beendet." );
    }
    catch ( NullPointerException nex )
    {
      LOG.error( "NullPointerException in der Methode doAufbereitung", nex );
    }
    catch ( CreateException cex )
    {
      LOG.error( "CreateException beim Lookup auf jippsProperties", cex );
    }
    catch ( NamingException naex )
    {
      LOG.error( "NamingException beim Lookup auf jippsProperties", naex );
    }
    catch ( Exception ex )
    {
      LOG.error( "Fehler in der Methode doAufbereitung", ex );
    }
  }

  /**
   * Macht einen Eintrag in der Tabelle Loeschungs_Lauf. In dieser Tabelle werden die wichtigesten Daten eingetragen,
   * damit sichergestellt werden kann, dass alle Löschungsanzeigen aufbereitet werden.
   * @param laufVO Value-Object mit den Start-Werten
   */
  private void insertLoeschungLauf( LoeschungLaufVO laufVO )
  {
    // wird nur für Mail benötigt!
    this.dateStart = DateFormatter.mediumFormat( new Date( System.currentTimeMillis() ) );

    laufVO.jippsStat = 0;
    laufVO.anzBfaktGel = anzBfaktGel;

    if ( loeschungLaufDAO == null )
    {
      loeschungLaufDAO = new LoeschungLaufDAO();
    }
    try
    {
      laufVO.loeschungLaufId = loeschungLaufDAO.insert( laufVO );
    }
    catch ( UnitOfWorkException e )
    {
      LOG.error( "SQL-Fehler beim Insert der Start-Werte des Löschungsanzeigenlaufs", e );
    }
  }

  /**
   * Diese Methode fügt in der Tabelle LOESCHUNG-LAUF Werte ein die den Lauf abschliessen.
   * @param laufVO Value-Object
   */
  private void updateLoeschungLauf( LoeschungLaufVO laufVO )
  {
    // wird nur für Mail benötigt!
    this.dateEnde = DateFormatter.mediumFormat( new Date( System.currentTimeMillis() ) );

    laufVO.jippsStat = 3;

    if ( loeschungLaufDAO == null )
    {
      loeschungLaufDAO = new LoeschungLaufDAO();
    }
    try
    {
      loeschungLaufDAO.update( laufVO );
    }
    catch ( UnitOfWorkException e )
    {
      LOG.error( "SQL-Fehler beim Insert der Start-Werte des Löschungsanzeigen", e );
    }
  }

  /**
   * Diese Methode werden die einzelnen Löschungsanzeigen in die Datenbank geschrieben.
   * @param jlVO gefülltes Value-Object mit dem Inhalt einer Loeschung
   */
  private void insertLoeschung( LoeschungVO jlVO )
  {
    /** TODO Im Fehlerfall wird hier nur eine LOG-Meldung geschrieben. */
    if ( loeschungDAO == null )
    {
      loeschungDAO = new LoeschungDAO();
    }
    try
    {
      loeschungDAO.insert( jlVO );
    }
    catch ( UnitOfWorkException e )
    {
      LOG.error( "Exception beim Insert der Loeschungsanzeigen: (" + jlVO.bfaktId + ") ", e );
    }
  }

  /**
   * Mit dieser Methode wird das Value-Object mit den Werten aus der Datenbank gefüllt. Es wird pro Bfakt_Id ein
   * Value-Object erzeugt. Dies wird in einem späteren Zeitpunkt mit den weiteren Daten ergänzt.
   * @param loeschTyp dieser kann folgende Werte haben Loeschung oder Schutzablauf
   * @param laufVO Value-Object mit den Lauf-Werten.
   * @return List List mit den gefüllten Value-Objects.
   */
  private List loadBfaktur( String loeschTyp, LoeschungLaufVO laufVO )
  {
    if ( bfaktDAO == null )
    {
      bfaktDAO = new BfakturDAO();
    }
    try
    {
      if ( loeschTyp.equalsIgnoreCase( "loeschung" ) )
      {
        laufVO.druckArt = CancellationConstants.IP_NOT_PAID_INTEGER;
        return bfaktDAO.selectBFaktIdLoeschung( "5", "3" );
      }
      else if ( loeschTyp.equalsIgnoreCase( "schutzablauf" ) )
      {
        laufVO.druckArt = CancellationConstants.IP_EXPIRED_INTEGER;
        return bfaktDAO.selectBFaktIdLoeschung( "6", "1" );
      }
      else
      {
        return null;
      }
    }
    catch ( SQLException sqlex )
    {
      LOG.error( "SQL-Fehler beim Laden des JippsLoeschungVO", sqlex );
      return null;
    }
  }

  /**
   * Mit dieser Methode wird die Tabelle FAKTPART in ein Value-Object gefüllt, welche für die Aufbereitung der
   * ADR_NR_VTR und SORT_TXT benötigt wird sowie weitere Angaben die Abgefüllt werden.
   * @param adrNr Primaray-Key (Adress-Nummer)
   * @return FaktparVO Gefülltes Value-Object
   */
  private FaktparVO selectFaktPart( int adrNr )
  {
    if ( faktparDAO == null )
    {
      faktparDAO = new FaktparDAO();
    }
    try
    {
      return faktparDAO.select( adrNr );
    }
    catch ( SQLException sqlex )
    {
      LOG.error( "SQL-Fehler beim Laden des FaktparVO", sqlex );
      return null;
    }
  }

  /**
   * Mit dieser Methode wird jede einzelne Löschungsanzeige erstellt und in die Tabelle LOESCHUNG gespeichert. Danach
   * wir ein Update in der Tabelle LOESCHUNG_LAUF gemacht.
   * @param laufId Anzahl der Durchläufe
   * @param loeschung Liste mit den Value-Objects und BfaktIds
   * @return int Anzahl der Durchläufe
   */
  private int loadLoeschung( List loeschung, int laufId )
  {
    int counter = 0; // Counter für die generierten Löschungen.
    LoeschungVO jlVO = null;
    FaktparVO faktparVO = null;
    LetterAssembler ass = new LetterAssembler();
    LetterWriteXml letterWriteXml = new LetterWriteXml();
    List<LetterVO> listLetter = null; // Liste aller LetterVO's inkl. PotZahler
    LetterVO letterVO = null;

    // Letzte Bfaktur_Id die Verarbeitet wurde.
    int bfaktNr = 0;

    Iterator it = loeschung.iterator();

    while ( it.hasNext() )
    {
      try
      {
        jlVO = (LoeschungVO) it.next();
        jlVO.loeschungLaufId = laufId;
        bfaktNr = jlVO.bfaktId;

        listLetter = ass.loadLetterVO( jlVO.bfaktId, null, null, true );

        if ( listLetter.size() > 0 )
        {
          for ( int i = 0; i < listLetter.size(); i++ )
          {
            letterVO = listLetter.get( i );
            jlVO.loeschungXml = letterWriteXml.toXml( letterVO );
            jlVO.adrNr = letterVO.versadrVO.getAdrNr();
            jlVO.schutitTyp = letterVO.jippsFakturVO.schutitTyp;
            jlVO.bfaktId = letterVO.jippsFakturVO.bfaktId;
            faktparVO = selectFaktPart( jlVO.adrNr );
            if ( faktparVO != null )
            {
              jlVO.druckPeri = faktparVO.druckperi;
              jlVO.fachNr = faktparVO.fachNr;
              jlVO.adrNrVtr = getAdrNrVtr( faktparVO, letterVO );
              jlVO.sortTxt = getSortTxt( faktparVO, letterVO );
            }
            else
            {
              jlVO.druckPeri = "0";
              jlVO.fachNr = 0;
              jlVO.adrNrVtr = 0;
              jlVO.sortTxt = letterVO.jippsFakturVO.ihrZeichen;
            }

            jlVO.verarbStat = CancellationProcessingConstants.NEW_INTEGER;
            LOG.debug( "InsertLöschung mit Bfakt-Id: " + jlVO.bfaktId + " Adr-Nr.: " + jlVO.adrNr );
            insertLoeschung( jlVO );
            counter++;
          }
        }
        else
        {
          LOG.warn( "Keine LetterVO's geladen zu Bfakt-Id: " + jlVO.bfaktId );
        }
      }
      catch ( SQLException sqlex )
      {
        LOG.error( "SQL-Exception in der Methode loadLoeschung", sqlex );
        insertLoeschungError( laufId, bfaktNr, sqlex.getMessage() );
        countError++;
      }
      catch ( Exception ex )
      {
        LOG.error( "Exception in der Methode loadLoeschung", ex );
        insertLoeschungError( laufId, bfaktNr, ex.getMessage() );
        countError++;
      }
    }
    return counter;
  }

  /**
   * Diese Methode setzt anhand des SORT_CD die Wert für den ADR_NR_VTR. SORT_CD: ADR_NR_VTR: 0 oder 3 = 0 1,2 oder 3 =
   * BFAKTUR.ADR_NR null = 0
   * @param faktparVO gefülltes ValueObject
   * @param letterVO gefülltes Value-Object
   * @return int Wert der ins Value-Object geschrieben wird
   */
  private int getAdrNrVtr( FaktparVO faktparVO, LetterVO letterVO )
  {
    if ( faktparVO.sortCd.equals( "0" ) || faktparVO.sortCd.equals( "3" ) )
    {
      return 0;
    }
    if ( faktparVO.sortCd.equals( "1" ) || faktparVO.sortCd.equals( "2" ) || faktparVO.sortCd.equals( "4" ) )
    {
      return letterVO.jippsFakturVO.adrNr;
    }
    else
    {
      return 0;
    }
  }

  /**
   * Diese Methode setzt anhand der SORT_CD die Texte für den SORT_TEXT. SORT_CD: SORT_TEXT: 0 oder 1 =
   * BFAKTUR.IHR-ZEICHEN 2 = BFAKTUR.INHABER 3 oder 4 = BFAKTUR.SCHUTIT_NR
   * @param faktparVO gefülltes Value-Object
   * @param letterVO gefülltes Value-Object
   * @return String gemappter String
   */
  private String getSortTxt( FaktparVO faktparVO, LetterVO letterVO )
  {
    if ( faktparVO.sortCd.equals( "0" ) || faktparVO.sortCd.equals( "1" ) )
    {
      return letterVO.jippsFakturVO.ihrZeichen;
    }
    else if ( faktparVO.sortCd.equals( "2" ) )
    {
      return letterVO.jippsFakturVO.inhaber;
    }
    else if ( faktparVO.sortCd.equals( "3" ) || faktparVO.sortCd.equals( "4" ) )
    {
      return letterVO.jippsFakturVO.schutitNr;
    }
    else
    {
      return letterVO.jippsFakturVO.ihrZeichen;
    }
  }

  /**
   * Diese Methode schreibt die fehlerhafte Bfakt_Id in die Tabelle LOESCHUNG_ERROR.
   * @param laufId Löschungslaufnummer
   * @param bfaktId Fehlerhafte Bfaktur-Nummer
   * @param errorMsg Fehlermeldung
   */
  private void insertLoeschungError( int laufId, int bfaktId, String errorMsg )
  {
    if ( loeschErrorDAO == null )
    {
      loeschErrorDAO = new LoeschungErrorDAO();
    }

    try
    {
      loeschErrorDAO.insert( laufId, bfaktId, errorMsg );
    }
    catch ( Exception ex )
    {
      LOG.error( "Fehler beim der Methode insertLoeschungError", ex );
    }
  }

  /**
   * Steuerung zum Sortieren, Rendern und Speichern der Loeschungen
   * @param selList List mit Druckperioden
   * @param drucker Id des JippsDruckers
   * @param paketSize Groesse der Druckpakete
   * @param userName Name des Auslösers, wird für allfällige Druckrückmeldungen benutzt
   * @param loeschungLaufId Nummer des Löschunglaufes
   * @return String, enthält eine Message über den Status der Verarbeitung
   */
  public String renderLoeschungen( List<DruckPeriVO> selList, int drucker, int paketSize, String userName, int loeschungLaufId )
  {
    String statusMessage = "";

    int anzAlleDruckPeri = 0;
    int anzProDruckPeri = 0;
    int anzVerarbProDruckPeri = 0;

    List listLoeschPaket = null;
    LoeschungDAO loeschDAO = new LoeschungDAO();
    PrintJobVO printJobVO = null;

    anzErstPrintJobs = 0; // Klassenvariable muss jeweils auf 0 zurückgesetzt werden.

    // Iterator über die übergebenen Druckerperioden
    for ( DruckPeriVO periVO: selList )
    {
      anzProDruckPeri = periVO.anzahl;
      anzAlleDruckPeri = anzAlleDruckPeri + anzProDruckPeri; // Anzahl pro Druckperiode aufsummieren.

      while ( anzVerarbProDruckPeri <= anzProDruckPeri )
      {
        try
        {
          listLoeschPaket = loeschDAO.selectSorted( periVO.druckPeri, paketSize, loeschungLaufId );
        }
        catch ( SQLException sqlex )
        {
          LOG.error( sqlex.getMessage(), sqlex );
        }

        LOG.debug( "*** listLoeschPaket: " + listLoeschPaket.size() + "(DruckPeri = " + periVO.druckPeri + ", Paket = " + paketSize + ")" );

        int batchId = getNextBatchId();
        LOG.debug( "*** BatchId: " + batchId );

        // für ein Paket die Printjobs erstellen und anzErstPrintJobs entsprechende erhöhen...
        printJobVO = processPaket( loeschDAO, listLoeschPaket, batchId, drucker, userName );

        LOG.debug( "*** Ein Paket wurde erstellt" );
        // Anzahl verarbeitete Löschungen um PaketSize erhöhen.
        anzVerarbProDruckPeri = anzVerarbProDruckPeri + paketSize;

        /**
         * TODO REFACTORING Wenn alle byte[] null oder < 1kB sind wird eine JMS-Nachricht versendet. Die Methode sollte
         * so umgeschrieben werden, damit bei diesem Fall keine JMS versendet wird.
         */

        if ( printJobVO.druckStatus == PrintStatusConstants.READY_FOR_PRINTING || printJobVO.druckStatus == PrintStatusConstants.PRINTED )
        {
          sendBatchJob( printJobVO.batchId, null, printJobVO.druckerId, false, LOESCHUNGEN_PRINTJOB_NAME, null );

          StringBuffer bufLog = new StringBuffer();
          bufLog.append( "*** JMS-Nachricht wurde gesendet: (PrintJobId) = 0" );
          bufLog.append( ", (DruckerId) = " );
          bufLog.append( Integer.toString( printJobVO.druckerId ) );
          bufLog.append( ", (BatchId) = " );
          bufLog.append( Integer.toString( printJobVO.batchId ) );
          bufLog.append( ", (isSave) = 0, (reportName) = null , (path) = null" );
          LOG.info( bufLog.toString() );
        }
      }
    }

    statusMessage = anzErstPrintJobs + " von " + anzAlleDruckPeri + " Loeschungsanzeigen wurden verarbeitet.";
    return statusMessage;
  }

  /**
   * Methode erstellt für ein Paket von Löschungen die PrintJobs.
   * @param loeschDAO LoeschungDAO
   * @param listLoeschPaket Liste mit den Löschungen
   * @param batchId Batch-Nr.
   * @param drucker Drucker-Nr.
   * @param userName Username
   * @return anzErstPrintJobs Anzahl erstellte PrintJobs
   */
  private PrintJobVO processPaket( LoeschungDAO loeschDAO, List listLoeschPaket, int batchId, int drucker, String userName )
  {
    LoeschungVO loeschVO = null;
    PrintJobVO printJobVO = null;

    Iterator itLoesch = listLoeschPaket.iterator();

    for ( int i = 0; i < listLoeschPaket.size(); i++ )
    {
      try
      {
        loeschVO = (LoeschungVO) itLoesch.next();
        printJobVO = fillPrintJobVO( batchId, drucker, userName, loeschVO );

        try
        {
          printJobVO.printJobFile = createPdf( loeschVO.loeschungXml, XslConstants.LETTER );
        }
        catch ( Exception ex )
        {
          LOG.error( "Fehler beim erstellen des XML", ex );
        }

        // prüfen, ob Pdf-Bytearray null oder korrupt ist!
        if ( isPrintJobVOValid( printJobVO.printJobFile ) )
        {
          storePdf( printJobVO );
          LOG.debug( "*** Pdf gespeichert: " + i );

          loeschDAO.update( loeschVO.loeschungLaufId, loeschVO.bfaktId, loeschVO.adrNr, CancellationProcessingConstants.TRANSMITTED_TO_SAP_INTEGER );
          LOG.debug( "*** Update auf Loeschung gemacht" );
          anzErstPrintJobs++;
        }
        else
        {
          LOG.error( "PrintJobFile korrupt: " + printJobVO.batchId + " , " + printJobVO.printJobId + " , " + printJobVO.reportId );
          String errMessage = "PrintJobFile korrupt oder Null";
          insertLoeschungError( loeschVO.loeschungLaufId, loeschVO.bfaktId, errMessage );
        }
      }
      catch ( UnitOfWorkException e )
      {
        String errMessage = "Fehler beim Erstellen des Pdf-Streams: " + e.getMessage();
        LOG.error( errMessage, e );
        insertLoeschungError( loeschVO.loeschungLaufId, loeschVO.bfaktId, errMessage );
      }
    }
    return printJobVO;
  }

  /**
   * Überprüft das PrintJob_File auf Fehler. Methode ist public wegen JUnit-Test.
   * @param pdfarray Byte-Array mit dem Inhalt des Schreibens.
   * @return true, wenn PrintJob_File ok ist, sonst false
   */
  public boolean isPrintJobVOValid( byte[] pdfarray )
  {
    if ( ( pdfarray == null ) || ( pdfarray.length < 500 ) )
    {
      return false;
    }
    else
    {
      return true;
    }
  }

  /**
   * Füllt die Parameterwerte für eine Loeschungsanzeige in ein PrintJobVO ein
   * @param batchId integer, enthält pro Paket die gleiche Id
   * @param drucker id des JippsDruckers
   * @param maName Name des Auslösenden Mitarbeiters
   * @param loeschVO LoeschungsVO
   * @return printVO mit den Parameterwerten
   */
  private PrintJobVO fillPrintJobVO( int batchId, int drucker, String maName, LoeschungVO loeschVO )
  {
    PrintJobVO printVO = new PrintJobVO();
    printVO.batchId = batchId;
    printVO.druckerId = drucker;
    printVO.druckStatus = PrintStatusConstants.READY_FOR_PRINTING;
    printVO.erstDatum = loeschVO.createDate;
    printVO.formTyp = PrintJobVO.LETTER;
    printVO.senderUser = maName;
    printVO.printJobId = getNextPrintJobId();
    printVO.reportId = loeschVO.bfaktId;
    printVO.paperMode = ReportPaper.PAPER_WITH_IPI_LOGO_INCLUDING_COPIES.toDB();
    printVO.schutitTyp = IpTypeConstants.PATENT_CH;

    return printVO;
  }

  /**
   * Mit dieser Methode wird ein Mail generiert, welches dem Empfänger einen Statusbericht der Aufbereitung der
   * löschungsanzeige zustellt.
   * @param laufVO Value-Object
   * @throws CreateException Unerwarteter Create-Fehler
   * @throws RemoteException Unerwarteter Remote-Fehler
   * @throws NamingException Unerwarteter Naming-Fehler
   * @throws Exception Unerwarteter Fehler
   */
  private void createMail( LoeschungLaufVO laufVO ) throws CreateException, RemoteException, NamingException, Exception
  {
    String param = "";

    if ( laufVO.druckArt.equals( CancellationConstants.IP_NOT_PAID_INTEGER ) )
    {
      param = "Nichtbezahlung";
    }
    else
    {
      param = "Schutzablauf";
    }

    StringBuffer bufBetreff = new StringBuffer();
    bufBetreff.append( "Löschungsanzeigen (" );
    bufBetreff.append( param );
    bufBetreff.append( ")" );

    JIPPSPropertiesServices jippsProperties = JIPPSPropertiesFactory.getInstance();
    if ( laufVO.anzBfaktGel > 0 )
    {
      StringBuffer bufMsg = new StringBuffer();
      bufMsg.append( "Statusbericht der Aufbereitung:\nLauf-Nr:\t\t\t\t\t" );
      bufMsg.append( laufVO.loeschungLaufId );
      bufMsg.append( "\nVerarbeitungs-Start:\t\t\t" );
      bufMsg.append( dateStart );
      bufMsg.append( "\nVerarbeitungs-Ende:\t\t\t" );
      bufMsg.append( dateEnde );
      bufMsg.append( "\nAnzahl gelesene Löschungen:\t\t" );
      bufMsg.append( laufVO.anzBfaktGel );
      bufMsg.append( "\nAnzahl generierter Löschungen:\t" );
      bufMsg.append( laufVO.anzLoeschGen );
      bufMsg.append( " (inkl. zusätzliche Empfänger)\nAnzahl der fehlerhaften Löschungen:\t" );
      bufMsg.append( countError );

      jippsProperties.doMail( bufBetreff.toString(), bufMsg.toString(), "loeschung" );
      LOG.debug( "*** Meldung wurde an JIPPSProperties gesendet: " + bufMsg.toString() );
    }
    else
    {
      jippsProperties.doMailNoData( bufBetreff.toString(), null, null, "loeschung", 0 );
    }
  }
}
