package ch.ige.jipps;

import java.sql.Date;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.ejb.EJBException;
import javax.ejb.MessageDrivenBean;
import javax.ejb.MessageDrivenContext;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.ipobjects.ServicesFactory;
import ch.ige.ipobjects.common.GenericIP;
import ch.ige.ipobjects.common.dao.JippsPublicationConfirmationMessageConverter;
import ch.ige.ipobjects.common.dao.XxDomainDAO;
import ch.ige.ipobjects.common.vo.SwissregPublicationVO;
import ch.ige.ipobjects.common.vo.XxDomainVO;
import ch.ige.ipobjects.event.Event;
import ch.ige.ipobjects.key.vo.IPKeyVO;
import ch.ige.ipobjects.time.vo.EreigIDat;
import ch.ige.jipps.vo.AnswerVO;

/**
 * Reads from the JMS queue for publication confirmation letters coming from the Swissreg update.
 * <p>
 * Copyright (c) 2007, Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * @author Eric Lewis
 * @version $Id: PublicationConfirmationMDBean.java,v 1.14 2008/05/22 13:54:13 lewis_lee Exp $ <!--
 * begin-xdoclet-definition -->
 * @ejb.bean name="PublicationConfirmationMDBean" destination-type="javax.jms.Queue" transaction-type="Container"
 * destination-jndi-name="jms/JIPPS.publConfirmQueue" connection-factory-jndi-name="jms/IGEQueueConnectionFactory"
 * @jboss.destination-jndi-name name="jms/JIPPS.publConfirmQueue" <!-- end-xdoclet-definition -->
 */
public class PublicationConfirmationMDBean implements MessageDrivenBean, MessageListener
{
  /**
   * The context for this
   */
  @SuppressWarnings( "unused" )
  private MessageDrivenContext messageDrivenContext;

  /**
   * Converts the JMS message for the publication confirmation
   */
  private JippsPublicationConfirmationMessageConverter converter;

  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( PublicationConfirmationMDBean.class );

  /**
   * Initializes the EJB.
   */
  public void ejbCreate()
  {
    this.converter = new JippsPublicationConfirmationMessageConverter();
    try
    {
      ServicesFactory.getStandaloneServices();
      getEventNumbersToPrint();
    }
    catch ( SQLException e )
    {
      LOG.fatal( "Can't get the event number configuration", e );
    }
    catch ( Exception e )
    {
      LOG.fatal( "Can't create ipobjects services", e );
    }
  }

  /**
   * Cleans up for EJB.
   */
  public void ejbRemove()
  {
    this.converter = null;
  }

  // @see javax.jms.MessageListener#onMessage(javax.jms.Message)
  public void onMessage( Message message )
  {
    LOG.debug( "Publication confirmation message received from queue: " + message );
    try
    {
      List<SwissregPublicationVO> publications = this.converter.fromMessage( message );
      if ( LOG.isDebugEnabled() )
      {
        LOG.debug( "Got the following publications to print:" );
        for ( SwissregPublicationVO publication: publications )
        {
          LOG.debug( publication.toString() );
        }
      }

      List<Integer> eventNumbersToPrint = getEventNumbersToPrint();
      if ( LOG.isDebugEnabled() )
      {
        LOG.debug( "The following events should be printed: " + StringUtils.join( eventNumbersToPrint.iterator(), ", " ) );
      }

      List<AnswerVO> answers = printPublicationConfirmations( message, eventNumbersToPrint, publications );
      for ( AnswerVO answer: answers )
      {
        if ( !answer.printStatus.equals( AnswerVO.STATUS_OK ) )
        {
          LOG.error( "Couldn't print publication confirmation for print jobs " + StringUtils.join( answer.listPrintId.iterator(), ", " ) );
        }
      }
    }
    catch ( JMSException e )
    {
      throw new EJBException( "Couldn't create publication confirmation", e );
    }
    catch ( SQLException e )
    {
      throw new EJBException( "Couldn't read events to print from XxDomain", e );
    }
    catch ( IllegalArgumentException e )
    {
      throw new EJBException( "Couldn't create publication confirmation (Illegal data in the JMS queue)", e );
    }
    catch ( IllegalStateException e )
    {
      throw new EJBException( "Couldn't create publication confirmation (Illegal configuration found)", e );
    }
    catch ( Exception e )
    {
      throw new EJBException( "Couldn't create publication confirmation", e );
    }
  }

  /**
   * Prints the publication confirmations.
   * @param message The JMS message
   * @param eventNumbersToPrint The event (article) numbers that must be printed
   * @param publications The publications confirmations to print
   * @return The answers containing the print status
   * @throws IllegalArgumentException if the publications are empty
   */
  public List<AnswerVO> printPublicationConfirmations( Message message, List<Integer> eventNumbersToPrint, List<SwissregPublicationVO> publications )
      throws IllegalArgumentException
  {
    if ( publications == null || publications.isEmpty() )
    {
      throw new IllegalArgumentException( "Found no publications from queue: " + message );
    }
    if ( eventNumbersToPrint == null || eventNumbersToPrint.isEmpty() )
    {
      throw new IllegalArgumentException( "Found no event numbers to print" );
    }

    Integer ipKey = publications.get( 0 ).getIPKey();
    GenericIP ip = ServicesFactory.getInstance().findIP( IPKeyVO.create( ipKey ) );
    if ( LOG.isDebugEnabled() )
    {
      LOG.debug( "Printing for IP " + ip.getIPKey() );
    }

    Map<EreigIDat, Date> validEventDates = getValidEventDates( ip, publications, eventNumbersToPrint );
    if ( LOG.isDebugEnabled() )
    {
      LOG.debug( "Got the following valid event dates: " + validEventDates );
    }

    LetterControl letterControl = new LetterControl();
    return letterControl.printPublicationConfirmation( ipKey, validEventDates );
  }

  /**
   * Gets the event numbers (article numbers) that should be printed.
   * @return The event numbers to print
   * @throws SQLException if the event numbers can't be read from XxDomain
   * @throws IllegalStateException if there are no event numbers to print
   */
  private List<Integer> getEventNumbersToPrint() throws SQLException, IllegalStateException
  {
    List<XxDomainVO> domainValues = new XxDomainDAO().findDomainValue( "sr_delayed_print", Locale.GERMAN );
    List<Integer> eventsNumbersToPrint = new ArrayList<Integer>();
    for ( XxDomainVO domainValue: domainValues )
    {
      eventsNumbersToPrint.add( Integer.valueOf( domainValue.getKBez() ) );
    }
    if ( eventsNumbersToPrint.isEmpty() )
    {
      throw new IllegalStateException( "Found no event numbers to print in XxDomain" );
    }
    return eventsNumbersToPrint;
  }

  /**
   * From all events received from the JMS queue, this filters the ones that have an event number in
   * <code>eventNumbersToPrint</code>.
   * @param ip The IP with all events
   * @param publications The publications from the JMS queue
   * @param eventNumbersToPrint The event numbers that must be printed
   * @return All event dates that must be printed (keys) and their corresponding publication date (value)
   */
  private Map<EreigIDat, Date> getValidEventDates( GenericIP ip, List<SwissregPublicationVO> publications, List<Integer> eventNumbersToPrint )
  {
    Map<EreigIDat, Date> eventDates = new HashMap<EreigIDat, Date>();
    for ( SwissregPublicationVO publication: publications )
    {
      Event event = ip.getEvents().get( publication.getEreig_idat() );
      if ( event == null )
      {
        throw new IllegalArgumentException( "Got published event time " + publication.getEreig_idat() + ", but no corresponding event exists for IP "
            + ip.getIPKey() );
      }
      if ( eventNumbersToPrint.contains( Integer.valueOf( event.getArt_nr() ) ) )
      {
        eventDates.put( publication.getEreigIDat(), publication.getPubl_dat() );
      }
    }
    return eventDates;
  }

  // @see javax.ejb.MessageDrivenBean#setMessageDrivenContext(javax.ejb.MessageDrivenContext)
  public void setMessageDrivenContext( MessageDrivenContext messageDrivenContext )
  {
    this.messageDrivenContext = messageDrivenContext;
  }
}
