package ch.ige.jipps.currentstate;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import ch.ige.ipobjects.Types;
import ch.ige.ipobjects.common.GenericIP;
import ch.ige.ipobjects.common.vo.FeeVO;
import ch.ige.ipobjects.patent.Patent;
import ch.ige.ipobjects.patent.PatentCH;
import ch.ige.ipobjects.patent.PatentEP;
import ch.ige.ipobjects.patent.vo.IPCVO;
import ch.ige.ipobjects.time.vo.EreigIDat;
import ch.ige.jipps.vo.StandDatVO;
import ch.ige.util.constants.ReportTextNumberHandler;
import ch.ige.util.enums.report.ReportType;
import ch.ige.util.format.IpcClassFormatter;
import ch.ige.util.format.PctApplicationNumberFormatter;
import ch.ige.util.format.PctPublicationNumberFormatter;

/**
 * Builds the objects for the current state in the patent report.
 * <p>
 * Copyright (c) 2006, Swiss Federal Institute of Intellectual Property
 * </p>
 * @author Eric Lewis
 * @version $Id: PatentBuilder.java,v 1.19 2008/06/23 14:46:52 lewis_lee Exp $
 */
class PatentBuilder extends BaseBuilder
{
  // See interface
  @Override
  public List<CurrentStateElement> getLatestFee( GenericIP ip, ReportType reportType, EreigIDat currentState, Locale language ) throws SQLException
  {
    Patent patent = (Patent) ip;
    List<CurrentStateElement> fees = new ArrayList<CurrentStateElement>();

    FeeVO fee = getLatestPayedFee( ip, currentState );
    if ( fee != null )
    {
      CurrentStateElement latestFee = new StandDatVO();
      latestFee.setReportTextNumber( ReportTextNumberHandler.getReportTextNumberForFee( patent.getBAGISType() ) );
      String feeText = getLatestPayedFeeText( fee, language );
      addSingleText( fees, latestFee, feeText );
    }

    return fees;
  }

  // See interface
  @Override
  public List<CurrentStateElement> getPct( GenericIP ip, ReportType reportType, EreigIDat currentState, Locale language )
  {
    Patent patent = (Patent) ip;
    List<CurrentStateElement> pctDates = new ArrayList<CurrentStateElement>();

    // PCT patents never have PCT dates
    if ( !Types.PATENT_PCT.isInstance( patent ) )
    {
      if ( StringUtils.isNotEmpty( patent.getPCTApplicationNumber() ) )
      {
        CurrentStateElement pctApplication = new StandDatVO();
        pctApplication.setReportTextNumber( ReportTextNumberHandler.getReportTextNumbersForPCTApplication( patent.getBAGISType() ) );
        String pctApplicationText = PctApplicationNumberFormatter.format( patent.getPCTApplicationNumber(), patent.getProtectionStart() );
        addSingleText( pctDates, pctApplication, pctApplicationText );
      }

      if ( StringUtils.isNotEmpty( patent.getPCTPublicationNumber() ) )
      {
        CurrentStateElement pctPublication = new StandDatVO();
        pctPublication.setReportTextNumber( ReportTextNumberHandler.getReportTextNumbersForPCTPublication( patent.getBAGISType() ) );
        String pctPublicationText = PctPublicationNumberFormatter.format( patent.getPCTPublicationNumber(), patent.getPCTPublicationDate() );
        addSingleText( pctDates, pctPublication, pctPublicationText );
      }
    }

    return pctDates;
  }

  // See interface
  @Override
  public List<CurrentStateElement> getClassification( GenericIP ip, ReportType reportType, EreigIDat currentState, Locale language )
  {
    Patent patent = (Patent) ip;
    List<CurrentStateElement> formattedIpcClasses = new ArrayList<CurrentStateElement>();

    Date ipcPublicationDate = getIpcPublicationDate( patent );
    List<IPCVO> ipcClasses = getRelevantIpcClasses( patent, currentState );

    if ( !ipcClasses.isEmpty() )
    {
      StringBuilder listOfClasses = new StringBuilder();
      for ( IPCVO ipcClass: ipcClasses )
      {
        if ( listOfClasses.length() > 0 )
        {
          listOfClasses.append( ", " );
        }
        String formattedClass = IpcClassFormatter.format( ipcClass.getPat_ipc_cd().toDB(), ipcClass.getIpc_gueltig_von(), ipcClass.getIpc_symb(),
            ipcPublicationDate );
        /**
         * TODO The dash in front on the bracket is only added because XSL-FO would wrap the line seeing a blank
         * in-between. This should not be necessary and should be solved in the XSL.
         */
        formattedClass = StringUtils.replace( formattedClass, " (", "-(" );
        listOfClasses.append( formattedClass );
      }

      CurrentStateElement formattedIpcClass = new StandDatVO();
      formattedIpcClass.setReportTextNumber( ReportTextNumberHandler.getReportTextNumberForClassification( patent.getBAGISType() ) );
      addSingleText( formattedIpcClasses, formattedIpcClass, listOfClasses.toString() );
    }

    return formattedIpcClasses;
  }

  /**
   * Gets the publication date which is relevant for IPC formatting.
   * @param patent The patent
   * @return The publication date for IPC formatting
   */
  private Date getIpcPublicationDate( Patent patent )
  {
    Date ipcPublicationDate = null;
    // Normally, the publication date is taken, but if that doesn't exist,
    // take the deposit date.
    switch( patent.getBAGISType() )
    {
      case PATENT_CH:
        PatentCH patentCh = (PatentCH) patent;
        if ( patentCh.getRegistrationEvent() != null )
        {
          ipcPublicationDate = patentCh.getRegistrationEvent().getEreig_idat().getDate();
        }
        break;
      case PATENT_EP:
        PatentEP patentEp = (PatentEP) patent;
        if ( patentEp.getPublicationDateB() != null )
        {
          ipcPublicationDate = patentEp.getPublicationDateB();
        }
        else
        {
          ipcPublicationDate = patentEp.getPublicationDateA();
        }
        break;
      case PATENT_PCT:
        if ( patent.getPCTPublicationDate() != null )
        {
          ipcPublicationDate = patent.getPCTPublicationDate();
        }
        break;
      default:
        throw new IllegalArgumentException( "Found illegal IP type <" + patent.getType() + ">" );
    }

    // As the last straw, use the deposit date
    if ( ipcPublicationDate == null && patent.getProtectionStart() != null )
    {
      ipcPublicationDate = patent.getProtectionStart();
    }

    return ipcPublicationDate;
  }

  /**
   * Gets the IPC classes which are relevant for the state that the patent is in. This filtering applies only to Swiss patents.
   * @param patent The patent
   * @param currentState The current report date
   * @return Only the IPC classes which are relevant for the patent
   */
  private List<IPCVO> getRelevantIpcClasses( Patent patent, EreigIDat currentState )
  {
    if ( Types.PATENT_CH.isInstance( patent ) )
    {
      boolean isPublished = patent.getFirstPublication( currentState ) != null;
      List<IPCVO> relevantIpcClasses = new ArrayList<IPCVO>();
      for ( IPCVO ipcClass: patent.getIPC() )
      {
        // If the patent is published (even though it may still be an application), all definitive IPC classes are taken.
        if ( isPublished && !ipcClass.isPreClassification() )
        {
          relevantIpcClasses.add( ipcClass );
        }
        // If it's not yet published, all pre-classes are taken 
        else if ( !isPublished && ipcClass.isPreClassification() )
        {
          relevantIpcClasses.add( ipcClass );
        }
      }

      // There can be the case that the patent is not published, and there are only definitive IPC classes.
      // Since that's better than nothing ;-) they are taken.
      if ( relevantIpcClasses.isEmpty() )
      {
        for ( IPCVO ipcClass: patent.getIPC() )
        {
          if ( !isPublished && !ipcClass.isPreClassification() )
          {
            relevantIpcClasses.add( ipcClass );
          }
        }
      }
      return relevantIpcClasses;
    }
    else
    {
      return patent.getIPC();
    }
  }
}
