package ch.ige.jipps.dao;

import java.sql.Date;
import java.sql.SQLException;
import java.util.List;

import ch.ige.dbutils.dao.AbstractJippsDAO;
import ch.ige.dbutils.runner.IGEQueryRunner;
import ch.ige.dbutils.sql.SQL;
import ch.ige.dbutils.sql.SQLStatement;
import ch.ige.jipps.vo.AmvLaufVO;
import ch.ige.util.validation.Check;

/**
 * Accesses the table <code>JIPPS_AMV_LAUF</code> that contains batch runs for trademark renewals. Note that
 * <code>AMV</code> means <code>Automatische Markenverlängerung</code>, it's a trademark renewal.
 * <p>
 * Copyright (c) 2004, Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * 
 * @author Marc Bouquet
 * @author Eric Lewis
 * @version $Id: AmvLaufDAO.java,v 1.25 2008/05/06 07:37:43 lewis_lee Exp $
 */
public class AmvLaufDAO extends AbstractJippsDAO
{
  /**
   * The type of value object to return
   */
  private static final Class<AmvLaufVO> VO_TYPE = AmvLaufVO.class;

  /**
   * The base select that selects all batch runs
   */
  private static final SQLStatement<AmvLaufVO> BASE_SELECT = SQL
      .select( "SELECT amv_lauf_id, lauf_nr, verarb_ende_dat, verarb_stat, anz_amv, typ_cd, batch_id, created_dat, changed_dat " + "FROM jipps_amv_lauf " );

  /**
   * Selects all batch runs of a certain type
   */
  private static final SQLStatement<AmvLaufVO> SELECT_BATCH_RUN_BY_TYPE = BASE_SELECT.append( "WHERE typ_cd = ? " + "ORDER BY lauf_nr DESC" );

  /**
   * Selects one batch run by ID
   */
  private static final SQLStatement<AmvLaufVO> SELECT_BATCH_RUN_BY_ID = BASE_SELECT.append( "WHERE amv_lauf_id = ?" );

  /**
   * Selects the highest batch ID of a certain type
   */
  private static final SQLStatement<Integer> SELECT_HIGHEST_BATCH_ID_BY_TYPE = SQL
      .select( "SELECT MAX(amv_lauf_id) AS amv_lauf_id FROM jipps_amv_lauf WHERE typ_cd = ?" );

  /**
   * Selects the batch ID of a certain type and sequence number
   */
  private static final SQLStatement<Integer> SELECT_BATCH_ID_BY_SEQUENCE_NUMBER_AND_TYPE = SQL
      .select( "SELECT amv_lauf_id FROM jipps_amv_lauf WHERE lauf_nr = ? AND typ_cd = ?" );

  private static final SQLStatement<AmvLaufVO> NUMBER_OF_ROWS = SQL.select( "SELECT COUNT(amv_lauf_id) AS anz_amv FROM jipps_amv_lauf" );

  private static final SQLStatement<Integer> INSERT = SQL.update( "INSERT INTO jipps_amv_lauf ( amv_lauf_id, lauf_nr, verarb_stat, typ_cd ) "
      + "VALUES (seq_jipps_amv_lauf_id.nextval, ?, ?, ?)" );

  private static final SQLStatement<Integer> UPDATE = SQL.update( "UPDATE jipps_amv_lauf "
      + "SET verarb_ende_dat = sysdate, verarb_stat = ?, anz_amv = ?, batch_id = ? , changed_dat = sysdate " + "WHERE amv_lauf_id = ? " );

  private static final SQLStatement<Integer> DELETE = SQL.update( "DELETE FROM jipps_amv_lauf WHERE amv_lauf_id = ?" );

  private static final SQLStatement<Integer> DELETE_BY_DATE = SQL.update( "DELETE FROM jipps_amv_lauf WHERE created_dat < ?" );

  /**
   * Gets all batch runs of a specified type.
   * 
   * @param batchRunType The batch run type, see {@link ch.ige.jipps.vo.AmvLaufVO#typCd} for possible values
   * @return The batch runs of a certain type
   * @throws SQLException if the query fails
   */
  public List<AmvLaufVO> select( String batchRunType ) throws SQLException
  {
    Check.notEmpty( batchRunType, "batch run type" );
    return selectVOList( 50, VO_TYPE, SELECT_BATCH_RUN_BY_TYPE, batchRunType );
  }

  /**
   * Gets one batch run.
   * 
   * @param amvBatchRunId The unique ID of the AMV batch run
   * @return The batch run or <code>null</code> if none is found
   * @throws SQLException if the query fails
   */
  public AmvLaufVO selectAMVLaufToId( int amvBatchRunId ) throws SQLException
  {
    return selectVO( VO_TYPE, SELECT_BATCH_RUN_BY_ID, new Integer( amvBatchRunId ) );
  }

  /**
   * Gets the latest batch run ID of a certain batch type.
   * 
   * @param batchRunType The batch run type, see {@link ch.ige.jipps.vo.AmvLaufVO#typCd} for possible values
   * @return The latest batch run ID of the specified type
   * @throws SQLException if the query fails
   */
  public int getAmvLaufId( String batchRunType ) throws SQLException
  {
    Check.notEmpty( batchRunType, "batch run type" );
    return selectInt( getQueryRunner(), SELECT_HIGHEST_BATCH_ID_BY_TYPE, 0, batchRunType );
  }

  /**
   * Shows whether there are batch runs.
   * 
   * @return <code>true</code> if there are batch runs, otherwise <code>false</code>
   * @throws SQLException if the query fails
   */
  public boolean hasRows() throws SQLException
  {
    AmvLaufVO batchRun = selectVO( VO_TYPE, NUMBER_OF_ROWS );
    return batchRun.anzAmv > 0;
  }

  /**
   * Gets the batch run ID for a certain sequence ID and batch run type.
   * 
   * @param sequenceNumber
   * @param batchRunType The batch run type, see {@link ch.ige.jipps.vo.AmvLaufVO#typCd} for possible values
   * @return The batch run ID for the given data or <code>0</code> if none is found
   * @throws SQLException if the query fails
   */
  public int selectAmvLaufId( String sequenceNumber, String batchRunType ) throws SQLException
  {
    Check.notEmpty( sequenceNumber, "sequence number" );
    Check.notEmpty( batchRunType, "batch run type" );
    return selectInt( getQueryRunner(), SELECT_BATCH_ID_BY_SEQUENCE_NUMBER_AND_TYPE, 0, sequenceNumber, batchRunType );
  }

  /**
   * Inserts a batch run, but only
   * <ul>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#amvLaufId} (is generated)</li>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#laufNr}</li>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#verarbStat}</li>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#typCd}</li>
   * </ul>
   * 
   * @param amvBatchRun The AMV batch run to insert
   * @return The AVM batch run ID of the inserted batch run
   */
  public Integer insert( final AmvLaufVO amvBatchRun )
  {
    Check.notNull( amvBatchRun, "AMV batch run" );

    return this.execute( new SingleDAOUnitOfWork<Integer>()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "amv_lauf_insert";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        int count = runner.update( INSERT, amvBatchRun.laufNr, amvBatchRun.verarbStat, amvBatchRun.typCd );
        verifyUpdateCount( count, 1, "update count must be 1" );
        this.setResult( AmvLaufDAO.this.selectCurrentSequenceValue( runner, "seq_jipps_amv_lauf_id" ) );
      }
    } );
  }

  /**
   * Updates a batch run, but only
   * <ul>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#verarbEndeDat} (is set to current date/time)</li>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#changedDat} (is set to current date/time)</li>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#verarbStat}</li>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#anzAmv}</li>
   * <li>{@link ch.ige.jipps.vo.AmvLaufVO#batchId}</li>
   * </ul>
   * Normally, this is used when the batch run has finished.
   * 
   * @param amvBatchRun The AMV batch run to change
   */
  @SuppressWarnings( "unchecked" )
  public void update( final AmvLaufVO amvBatchRun )
  {
    Check.notNull( amvBatchRun, "AMV batch run" );

    this.execute( new SingleDAOUnitOfWork()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "amv_lauf_update";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        int count = runner.update( UPDATE, amvBatchRun.verarbStat, amvBatchRun.anzAmv, amvBatchRun.batchId, amvBatchRun.amvLaufId );
        verifyUpdateCount( count, 1, "update count must be 1" );
      }
    } );
  }

  /**
   * Deletes a whole batch run and all of its trademark renewals and batch run errors.
   * 
   * @param amvBatchRunId The unique ID of the AMV batch run
   */
  @SuppressWarnings( "unchecked" )
  public void delete( final int amvBatchRunId )
  {
    this.execute( new SingleDAOUnitOfWork()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "amv_lauf_delete";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        int count = runner.update( DELETE, new Integer( amvBatchRunId ) );
        verifyUpdateCount( count, 1, "update count must be 1" );
      }
    } );
  }

  /**
   * Deletes all AMV runs older than the given date.
   * 
   * @param date The date that the AMV runs are compared to; runs older than this date are deleted.
   * @return The number of deleted runs
   */
  @SuppressWarnings( "unchecked" )
  public Integer deleteByDate( final Date date )
  {
    return this.execute( new SingleDAOUnitOfWork<Integer>()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "amv_lauf_delete";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        Integer count = runner.update( DELETE_BY_DATE, date );
        setResult( count );
      }
    } );
  }

}
