package ch.ige.jipps.dao;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import ch.ige.dbutils.handlers.TemporalScalarHandler;
import ch.ige.dbutils.handlers.TypedResultSetHandler;
import ch.ige.dbutils.sql.SQL;
import ch.ige.dbutils.sql.SQLStatement;
import ch.ige.ipobjects.dao.AbstractIpObjectsBagisDAO;
import ch.ige.jipps.vo.BfaktPosVO;

/**
 * Accesses the table <code>BFAKTPOS</code>.
 * 
 * <p>Copyright (c) 2002-2006, Eidgenössisches Institut für Geistiges Eigentum</p>
 * @author Marc Bouquet
 * @author Eric Lewis
 * @version $Id: BfaktPosDAO.java,v 1.48 2008/07/23 13:18:12 lewis_lee Exp $
 */
public class BfaktPosDAO extends AbstractIpObjectsBagisDAO
{

  // Note: This is an AbstractIpObjectsBagisDAO because integer columns are mapped to EreigIDat objects

  /**
   * The type of value object to return
   */
  private static final Class<BfaktPosVO> VO_TYPE = BfaktPosVO.class;

  /**
   * The statement for getting all IP type report types. Note that the columns have to be renamed because the corresponding VO has different
   * member variable names.
   */
  // TODO ereig_txt_cd 2 should be a constant
  private final static SQLStatement<BfaktPosVO> SELECT_ALL = SQL
      .select( "SELECT bp.bfakt_id AS invoice_id, bp.bfakt_posnr AS line_item_number, bp.ereig_idat AS event_date, NVL(bp.text_nr, 0) AS text_number, "
          + "bp.anzahl AS number_of_articles, bp.frist AS deadline, NVL(TRIM(bp.schutit_nr), '') AS ip_number, t.ereig_txt AS publication_text "
          + "FROM bfaktpos bp " + "LEFT JOIN ereig_txt t ON bp.schutit_key = t.schutit_key AND bp.ereig_idat = t.ereig_idat AND t.ereig_txt_cd = 2 "
          + "WHERE bp.bfakt_id = ? " + " ORDER BY bp.bfakt_posnr" );

  /**
   * The SQL statement used to select IP keys for the invoice IDs used during the registration process
   */
  //TODO erfass_proc 9 should be a constant
  private final static String SELECT_IP_KEYS_FOR_REGISTRATION = "SELECT b2.schutit_key FROM bfaktpos b1, bfaktur b2 WHERE b1.bfakt_id = b2.bfakt_id AND b1.bfakt_id IN ({0}) AND b1.art_nr IN (SELECT art_nr FROM bartikel WHERE erfass_proc = 9) "
      + "ORDER BY b2.schutit_nr";

  /**
   * Selects the publication date of the invoice's first line item.
   */
  private final static SQLStatement<Timestamp> SELECT_PUBLICATION_DATE = SQL.select( "SELECT e.publ_dat FROM bfaktpos b "
      + "JOIN ereignis e ON b.schutit_key = e.schutit_key AND b.ereig_idat = e.ereig_idat " + "WHERE b.bfakt_id = ? AND rownum = 1" );

  /**
   * Gets all invoice line items for an invoice.
   * @param invoiceId The invoice's identifier
   * @return The invoice lines for an invoice
   * @throws SQLException if the query fails
   */
  public List<BfaktPosVO> select( final int invoiceId ) throws SQLException
  {
    return selectVOList( 5, VO_TYPE, SELECT_ALL, Integer.valueOf( invoiceId ) );
  }

  /**
   * Gets the publication date of the first invoice line items for an invoice.
   * @param invoiceId The invoice's identifier
   * @return The first invoice line item's publication date or <code>null</code> if none exists
   * @throws SQLException if the query fails
   */
  public Timestamp getPublicationDate( final int invoiceId ) throws SQLException
  {
    return selectTemporal( getQueryRunner(), TemporalScalarHandler.TIMESTAMP, SELECT_PUBLICATION_DATE, Integer.valueOf( invoiceId ) );
  }

  /**
   * Gets the IP keys for the invoice IDs used during the registration process.
   * @param invoiceIds The <code>Integer</code> objects representing invoice IDs
   * @return The IP keys for all invoice IDs which contain a registration event; this is empty if <code>invoiceIds</code> is empty
   * @throws IllegalArgumentException if <code>invoiceIds</code> is null
   * @throws SQLException if the query fails
   */
  public List<Integer> getRegistrationIpKeys( final List<Integer> invoiceIds ) throws IllegalArgumentException, SQLException
  {
    if ( invoiceIds == null )
    {
      throw new IllegalArgumentException( "Got no invoice IDs" );
    }
    if ( invoiceIds.isEmpty() )
    {
      return Collections.emptyList();
    }

    List<Integer> returnList = new ArrayList<Integer>();

    for ( List<Integer> listInvoiceIds: getInvoiceIdLists( invoiceIds, 999 ) )
    {
      String invoiceIdsString = StringUtils.join( listInvoiceIds.iterator(), ',' );
      SQLStatement<List<Integer>> sqlQuery = SQL.select( MessageFormat.format( SELECT_IP_KEYS_FOR_REGISTRATION, new Object[] { invoiceIdsString } ) );
      returnList.addAll( getQueryRunner().query( sqlQuery, new IpKeyListHandler() ) );
    }

    return returnList;
  }

  /**
   * Splits a given <code>List</code> of <code>Integer</code> into one or more lists with a maximum size
   * of <code>maxSize</code>. If the given List has a size of
   * 22 and <code>maxSize</code> is 5, the method returns 4 lists with 5 entries and 1 list with 2 entries.
   * @param invoiceIds a List containing Integers
   * @param maxSize the maximum size of a List. if the size of invoiceIds exceeds this value, new Lists are generated.
   * @return a List containing Lists of Integer-objects with maxSize entries
   */
  public List<List<Integer>> getInvoiceIdLists( final List<Integer> invoiceIds, int maxSize )
  {
    List<List<Integer>> listListInvoiceIds = new ArrayList<List<Integer>>();
    List<Integer> listInvoiceIds = new ArrayList<Integer>( maxSize );
    for ( int i = 0; i < invoiceIds.size(); i++ )
    {
      if ( i == 0 || i == maxSize )
      {
        listInvoiceIds = new ArrayList<Integer>( maxSize );
        listListInvoiceIds.add( listInvoiceIds );
      }
      listInvoiceIds.add( invoiceIds.get( i ) );
    }
    return listListInvoiceIds;
  }

  /**
   * The <code>ResultSetHandler</code> which collects IP keys for registrations.
   */
  private static class IpKeyListHandler implements TypedResultSetHandler<List<Integer>>
  {
    public List<Integer> handle( ResultSet rs ) throws SQLException
    {
      List<Integer> list = new ArrayList<Integer>();
      while ( rs.next() )
      {
        list.add( Integer.valueOf( rs.getInt( "schutit_key" ) ) );
      }
      return list;
    }
  }
}
