package ch.ige.jipps.dao;

import java.sql.Date;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.MessageFormat;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ch.ige.dbutils.dao.AbstractBagisDAO;
import ch.ige.dbutils.handlers.TypedResultSetHandler;
import ch.ige.dbutils.runner.IGEQueryRunner;
import ch.ige.dbutils.sql.SQL;
import ch.ige.dbutils.sql.SQLStatement;
import ch.ige.ipobjects.time.vo.EreigIDat;
import ch.ige.jipps.vo.AmvLaufVO;
import ch.ige.jipps.vo.AmvVO;
import ch.ige.jipps.vo.BfakturVO;
import ch.ige.jipps.vo.LoeschungVO;
import ch.ige.util.validation.Check;

/**
 * Accesses the table <code>BFAKTUR</code>.
 * 
 * <p>Copyright (c) 2001-2006, Eidgenössisches Institut für Geistiges Eigentum</p>
 * 
 * @author Marc Bouquet
 * @author Eric Lewis
 * @version $Id: BfakturDAO.java,v 1.56 2008/05/19 14:50:19 lewis_lee Exp $
 */

public class BfakturDAO extends AbstractBagisDAO
{
  /**
   * The lowest invoice ID that can exist. If the queries are not limited,
   * a whole lot of invoices is fetched from the past.
   */
  private static final Integer LOWEST_INVOICE_ID = Integer.valueOf( BfaktIdLimiter.getInstance().getBfaktIdLowerLimit() );

  /**
   * The type of value object to return for cancellations
   */
  private static final Class<LoeschungVO> VO_TYPE_CANCELLATION = LoeschungVO.class;

  /**
   * The type of value object for invoices.
   */
  private static final Class<BfakturVO> VO_TYPE_INVOICE = BfakturVO.class;

  /**
   * The type of value object for trademark renewals.
   */
  private static final Class<AmvVO> VO_TYPE_TRADEMARK_RENEWAL = AmvVO.class;

  /**
   * Selects the IDs of the invoices used for cancellations.
   */
  private final static SQLStatement<LoeschungVO> SELECT_INVOICE_IDS_FOR_CANCELLATION = SQL.select( "SELECT bfakt_id" + " FROM bfaktur" + " WHERE bfakt_id > ?"
      + " AND trim(druck_art) = ? AND trim(status_druck) = ?" );

  /**
   * Selects the invoices for a certain form type.
   */
  private final static SQLStatement<BfakturVO> SELECT_INVOICES_FOR_FORM_TYPE = SQL
      .select( "SELECT DISTINCT b.bfakt_id, b.anz_kopie, b.ma_nr AS employee_number"
          + ", DECODE(e.publ_dat, null, DECODE(b.bfakt_dat, null, sysdate, b.bfakt_dat), e.publ_dat) AS publication_date"
          + ", b.schutit_key AS ip_key, b.schutit_nr" + " FROM bfaktur b" + " LEFT OUTER JOIN bfaktpos p ON b.bfakt_id = p.bfakt_id"
          + " LEFT OUTER JOIN ereignis e ON p.schutit_key = e.schutit_key AND p.ereig_idat = e.ereig_idat AND e.publ_dat IS NOT NULL" + " WHERE b.bfakt_id > ?"
          + " AND trim(b.formart_cd) = ? AND trim(b.status_druck) = ?" + " ORDER BY b.schutit_nr" );

  /**
   * Selects an invoice.
   */
  private final static SQLStatement<BfakturVO> SELECT_INVOICE = SQL
      .select( "SELECT bfakt_id, anz_kopie, ma_nr AS employee_number, trim(status_druck) AS print_status, bfakt_dat AS print_date, schutit_key AS ip_key FROM bfaktur WHERE bfakt_id = ?" );

  /**
   * Selects the print status for an invoice.
   */
  private final static SQLStatement<String> SELECT_PRINT_STATUS_FOR_INVOICE = SQL
      .select( "SELECT trim(status_druck) AS print_status FROM bfaktur WHERE bfakt_id = ?" );

  /**
   * Updates the print status
   */
  private final static SQLStatement<Integer> UPDATE_PRINT_STATUS = SQL.update( "UPDATE bfaktur " + "SET status_druck = ?, version = version + 1 "
      + "WHERE bfakt_id = ?" );

  /**
   * Updates the print date
   */
  private final static SQLStatement<Integer> UPDATE_PRINT_DATE = SQL.update( "UPDATE bfaktur " + "SET bfakt_dat = ?, version = version + 1 "
      + "WHERE bfakt_id = ?" );

  /**
   * Gets all invoices for a trademark renewal run
   */
  private final static SQLStatement<AmvVO> SELECT_INVOICES_FOR_TRADEMARK_RENEWAL = SQL
      .select( "SELECT adr_nr, bfakt_id, schutit_key, ma_nr AS employee_number, trim(schutit_nr) AS schutit_nr, schutit_key AS ip_key" + " FROM bfaktur"
          + " WHERE bfakt_id > ?" + " AND trim(bfakt_typ) = ? AND trim(status_druck) = ?" );

  /**
   * Gets all invoices for the given event dates
   */
  private final static String SELECT_INVOICES_FOR_EVENT_DATES = "SELECT f.bfakt_id, f.anz_kopie, f.ma_nr AS employee_number, TRIM(f.status_druck) AS print_status, "
      + "f.bfakt_dat AS print_date, f.schutit_key AS ip_key, p.bfakt_posnr, p.ereig_idat "
      + "FROM bfaktpos p "
      + "JOIN bfaktur f ON p.bfakt_id = f.bfakt_id "
      + "AND TRIM(formart_cd) = ''{0}'' AND status_druck = ''{1}'' "
      + "WHERE p.schutit_key = {2} AND p.ereig_idat IN ( {3} ) "
      + "ORDER BY p.bfakt_id, p.ereig_idat";

  /**
   * Gets the IDs of the invoices used for the cancellation letters.
   * 
   * @param printType Defines why the invoice is printed, see {@link ch.ige.util.constants.InvoiceConstants}
   * @param printStatus Defines how the invoice is printed, see {@link ch.ige.util.constants.PrintStatusConstants}
   * @return The cancellations containing only {@link ch.ige.jipps.vo.LoeschungVO#bfaktId}; they may also be empty.
   * @throws SQLException if the query fails
   */
  public List<LoeschungVO> selectBFaktIdLoeschung( final String printType, final String printStatus ) throws SQLException
  {
    Check.notEmpty( printType, "print type" );
    Check.notEmpty( printStatus, "print status" );

    return selectVOList( 20, VO_TYPE_CANCELLATION, SELECT_INVOICE_IDS_FOR_CANCELLATION, LOWEST_INVOICE_ID, printType.trim(), printStatus.trim() );
  }

  /**
   * Gets the invoices for a given form type.
   * This is used to print publication confirmation letters.
   * Note that the publication date is taken whenever possible from the event that was used
   * to publish the IP.
   * 
   * @param formType The form type, see {@link ch.ige.util.constants.InvoiceConstants}
   * @param printStatus Defines how the invoice is printed, see {@link ch.ige.util.constants.PrintStatusConstants}
   * @return The invoices, which may be empty
   * @throws SQLException if the query fails
   */
  public List<BfakturVO> selectBFakturByForm( final Integer formType, final String printStatus ) throws SQLException
  {
    Check.notNull( formType, "form type" );
    Check.notEmpty( printStatus, "print status" );

    return selectVOList( 20, VO_TYPE_INVOICE, SELECT_INVOICES_FOR_FORM_TYPE, LOWEST_INVOICE_ID, formType, printStatus.trim() );
  }

  /**
   * Gets the print status for an invoice.
   * 
   * @param invoiceId The invoice's identifier
   * @return The print status or <code>null</code> if the invoice is unknown or has no print status
   * @throws SQLException if the query fails
   */
  public String selectBFakturDruckstatus( final int invoiceId ) throws SQLException
  {
    return selectString( getQueryRunner(), SELECT_PRINT_STATUS_FOR_INVOICE, Integer.valueOf( invoiceId ) );
  }

  /**
   * Gets an invoice.
   * 
   * @param invoiceId The invoice's identifier
   * @return The invoice or <code>null</code> if none is found
   * @throws SQLException if the query fails
   */
  public BfakturVO selectInvoice( final int invoiceId ) throws SQLException
  {
    return selectVO( VO_TYPE_INVOICE, SELECT_INVOICE, Integer.valueOf( invoiceId ) );
  }

  /**
   * Get all invoices which belong to the given event dates of an IP and which are
   * ready to be printed.
   * 
   * @param formType The form type, see {@link ch.ige.util.constants.InvoiceConstants}
   * @param printStatus Defines how the invoice is printed, see {@link ch.ige.util.constants.PrintStatusConstants}
   * @param ipKey The IP's unique key
   * @param eventDates The event dates contained in the invoices
   * @return The invoices, each having the event dates that are contained in the invoice. Note that the sum of all
   *         these event dates can be identical with <code>eventDates</code> but may also be a smaller part of it.
   * @throws SQLException if the query fails
   */
  public Map<BfakturVO, Set<EreigIDat>> selectInvoices( final Integer formType, final String printStatus, final Integer ipKey, final List<EreigIDat> eventDates )
      throws SQLException
  {
    Check.notNull( formType, "form type" );
    Check.notEmpty( printStatus, "print status" );
    Check.notNegative( ipKey, "IP key" );
    Check.notNull( eventDates, "event dates" );

    if ( eventDates.isEmpty() )
    {
      return Collections.emptyMap();
    }
    else
    {
      // Can't use StringUtils.join(), since the toString() of EreigIDat returns
      // something like '2005-12-22#20'
      StringBuilder eventDatesList = new StringBuilder();
      for ( EreigIDat eventDate: eventDates )
      {
        if ( eventDatesList.length() > 0 )
        {
          eventDatesList.append( "," );
        }
        eventDatesList.append( eventDate.getCompareTime() );
      }
      SQLStatement<Map<BfakturVO, Set<EreigIDat>>> sqlStatement = SQL.select( MessageFormat.format( SELECT_INVOICES_FOR_EVENT_DATES, formType, printStatus,
          ipKey.toString(), eventDatesList.toString() ) );
      return getQueryRunner().query( sqlStatement, new InvoiceEventHandler() );
    }
  }

  /**
   * Implements the <code>ResultSetHandler</code> for {@link BfakturDAO#selectInvoices(Integer, String, Integer, List)}.
   */
  private static class InvoiceEventHandler implements TypedResultSetHandler<Map<BfakturVO, Set<EreigIDat>>>
  {
    // @see ch.ige.dbutils.handlers.TypedResultSetHandler#handle(java.sql.ResultSet)
    public Map<BfakturVO, Set<EreigIDat>> handle( ResultSet resultset ) throws SQLException
    {
      // This is a bit complicated, since BfakturVO doesn't implement equals() and hashCode(),
      // so a separated index of invoices is kept and everything is merged at the end.
      Map<Integer, BfakturVO> invoices = new HashMap<Integer, BfakturVO>();
      Map<Integer, Set<EreigIDat>> events = new HashMap<Integer, Set<EreigIDat>>();
      if ( resultset.next() )
      {
        do
        {
          Integer invoiceId = resultset.getInt( "bfakt_id" );
          if ( !invoices.containsKey( invoiceId ) )
          {
            BfakturVO invoice = new BfakturVO();
            invoice.bfaktId = invoiceId;
            invoice.anzKopie = resultset.getInt( "anz_kopie" );
            invoice.employeeNumber = resultset.getInt( "employee_number" );
            invoice.printStatus = resultset.getString( "print_status" );
            invoice.printDate = resultset.getDate( "print_date" );
            invoice.ipKey = resultset.getInt( "ip_key" );

            invoices.put( invoiceId, invoice );
          }
          if ( !events.containsKey( invoiceId ) )
          {
            events.put( invoiceId, new HashSet<EreigIDat>() );
          }
          Set<EreigIDat> eventDates = events.get( invoiceId );
          eventDates.add( EreigIDat.from( resultset.getInt( "ereig_idat" ) ) );

        } while ( resultset.next() );
      }

      Map<BfakturVO, Set<EreigIDat>> invoiceEvents = new HashMap<BfakturVO, Set<EreigIDat>>();
      for ( Integer invoiceId: invoices.keySet() )
      {
        if ( !events.containsKey( invoiceId ) )
        {
          throw new SQLException( "Found no events for invoice with ID " + invoiceId );
        }
        Set<EreigIDat> eventDates = events.remove( invoiceId );
        invoiceEvents.put( invoices.get( invoiceId ), eventDates );
      }
      // Anything left over?
      if ( !events.isEmpty() )
      {
        throw new SQLException( "Mismatch between event dates and invoices" );
      }

      return invoiceEvents;
    }
  }

  /**
   * Gets all invoices used for the automatic renewal of trademarks.
   * 
   * @param invoiceType The type of invoice
   * @param printStatus Defines how the invoice is printed, see {@link ch.ige.util.constants.InvoiceConstants}
   * @param trademarkRenewalRunId The ID of the current run of trademark renewals
   * @return A <code>List</code> of {@link ch.ige.jipps.vo.AmvVO} objects which may be empty
   * @throws SQLException if the query fails
   */
  public List<AmvVO> selectBFaktIdAMV( final int invoiceType, final int printStatus, final int trademarkRenewalRunId ) throws SQLException
  {
    List<AmvVO> trademarkRenewals = selectVOList( 20, VO_TYPE_TRADEMARK_RENEWAL, SELECT_INVOICES_FOR_TRADEMARK_RENEWAL, LOWEST_INVOICE_ID, Integer
        .valueOf( invoiceType ), Integer.valueOf( printStatus ) );

    // Force the trademark run ID and processing status
    for ( AmvVO trademarkRenewal: trademarkRenewals )
    {
      trademarkRenewal.amvLaufId = trademarkRenewalRunId;
      trademarkRenewal.verarbStat = AmvLaufVO.PROCESSING_STARTED;
    }
    return trademarkRenewals;
  }

  /**
   * Updates the print status of the invoice identified by <code>invoiceId</code>.
   * 
   * @param invoiceId The identifier of the invoice
   * @param printStatus The new print status to set
   */
  @SuppressWarnings( "unchecked" )
  public void updateBFakturDruckstatus( final int invoiceId, final String printStatus )
  {
    Check.notEmpty( printStatus, "print status" );

    this.execute( new SingleDAOUnitOfWork()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "bfaktur_update_druckstatus";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        int numberOfUpdates = runner.update( UPDATE_PRINT_STATUS, printStatus, Integer.valueOf( invoiceId ) );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't update invoice " + invoiceId + " with print status " + printStatus );
      }
    } );
  }

  /**
   * Updates the print date of the invoice identified by <code>invoiceId</code>.
   * 
   * @param invoiceId The identifier of the invoice
   * @param printDate The new print date to set
   */
  @SuppressWarnings( "unchecked" )
  public void updateBFakturBfaktDat( final int invoiceId, final Date printDate )
  {
    Check.notNull( printDate, "print date" );

    this.execute( new SingleDAOUnitOfWork()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "bfaktur_update_date";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        int numberOfUpdates = runner.update( UPDATE_PRINT_DATE, printDate, Integer.valueOf( invoiceId ) );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't update invoice " + invoiceId + " with print date " + printDate );
      }
    } );
  }
}
