package ch.ige.jipps.dao;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ch.ige.dbutils.dao.AbstractJippsDAO;
import ch.ige.dbutils.runner.IGEQueryRunner;
import ch.ige.dbutils.sql.SQL;
import ch.ige.dbutils.sql.SQLStatement;
import ch.ige.jipps.vo.DruckerVO;
import ch.ige.util.constants.PaperTrayConstants;
import ch.ige.util.validation.Check;

/**
 * Accesses the printers.
 * <p>Copyright (c) 2003, Eidgenössisches Institut für Geistiges Eigentum</p>
 * @author Marc Bouquet
 * @author Eric Lewis
 * @version $Id: DruckerDAO.java,v 1.44 2008/09/25 12:34:44 lang Exp $
 */
public class DruckerDAO extends AbstractJippsDAO
{
  //TODO Use a sequence for the printer ID

  /**
   * The type of value object to return
   */
  private static final Class<DruckerVO> VO_TYPE = DruckerVO.class;

  /**
   * The estimated number of printers
   */
  private static final int TOTAL_NUMBER_OF_PRINTERS = 25;

  /**
   * Always order by printer name.
   */
  private final static String ORDER_BY = " order by druckername";

  /**
   * The statement that gets the next free printer ID
   */
  private final static SQLStatement<Integer> SELECT_NEXT_PRINTER_ID = SQL.select( "SELECT max(drucker_id) + 1 AS max_id FROM drucker" );

  /**
   * The statement that gets all printers
   */
  private final static SQLStatement<DruckerVO> SELECT_ALL = SQL
      .select( "SELECT d.drucker_id, d.druckername AS drucker_name, d.controlserver AS control_server, " + "d.raum, d.farbe, d.optional_print_args, "
          + "NVL(l1.bin_refid, 0) AS ige_paper_id, NVL(l2.bin_refid, 0) AS white_paper_id, NVL(l3.bin_refid, 0) AS cover_paper_id, "
          + "l1.optional_print_args AS ige_print_args, l2.optional_print_args AS white_print_args, l3.optional_print_args AS cover_print_args "
          + "FROM drucker d " + "LEFT JOIN logidrucker l1 ON d.drucker_id = l1.drucker_id AND l1.schacht_id = ? "
          + "LEFT JOIN logidrucker l2 ON d.drucker_id = l2.drucker_id AND l2.schacht_id = ? "
          + "LEFT JOIN logidrucker l3 ON d.drucker_id = l3.drucker_id AND l3.schacht_id = ?" );

  /**
   * The statement that gets all printers
   */
  private final static SQLStatement<DruckerVO> SELECT_ALL_ORDERED = SELECT_ALL.append( ORDER_BY );

  /**
   * Selects non-color printers.
   */
  private final static SQLStatement<DruckerVO> SELECT_WHERE_NON_COLOR = SELECT_ALL.append( " WHERE d.farbe <> 1" ).append( ORDER_BY );

  /**
   * The statement that gets one printer
   */
  private final static SQLStatement<DruckerVO> SELECT_ONE = SELECT_ALL.append( " WHERE d.drucker_id = ?" );

  /**
   * Inserts one printer
   */
  private final static SQLStatement<Integer> INSERT_PRINTER = SQL
      .insert( "INSERT INTO drucker (drucker_id, druckername, controlserver, raum, farbe, optional_print_args) VALUES(?, ?, ?, ?, ?, ? )" );

  /**
   * Inserts one paper tray
   */
  private final static SQLStatement<Integer> INSERT_TRAY = SQL
      .insert( "INSERT INTO logidrucker (drucker_id, schacht_id, format, papiertyp, bin_refid, optional_print_args) VALUES(?, ?, ?, ?, ?, ?)" );

  /**
   * Updates one printer
   */
  private final static SQLStatement<Integer> UPDATE_PRINTER = SQL
      .update( "UPDATE drucker SET druckername = ?, controlserver = ?, raum = ?, farbe = ?, optional_print_args = ? WHERE drucker_id = ? " );

  /**
   * Inserts one paper tray
   */
  private final static SQLStatement<Integer> UPDATE_TRAY = SQL
      .update( "UPDATE logidrucker SET bin_refid = ?, optional_print_args = ? WHERE drucker_id = ? AND schacht_id = ?" );

  /**
   * Deletes one printer (and its paper trays by cascading delete).
   */
  private final static SQLStatement<Integer> DELETE_ONE = SQL.update( "DELETE FROM drucker WHERE drucker_id = ?" );

  /**
   * The statement that gets the control servers
   */
  private final static SQLStatement<DruckerVO> SELECT_CONTROL_SERVERS = SQL.select( "SELECT DISTINCT controlserver AS control_server FROM drucker" );

  /**
   * Gets all JIPPS printers, including their paper tray IDs.
   * @return All JIPPS printers
   * @throws SQLException if the query fails
   */
  public List<DruckerVO> getAllPrinters() throws SQLException
  {
    return selectVOList( TOTAL_NUMBER_OF_PRINTERS, VO_TYPE, SELECT_ALL_ORDERED, PaperTrayConstants.PAPER_WITH_IPI_LOGO, PaperTrayConstants.PAPER_PLAIN,
        PaperTrayConstants.PAPER_COVER );
  }

  /**
   * Gets all JIPPS printers, including their paper tray IDs.
   * @return All JIPPS printers
   * @throws SQLException if the query fails
   */
  public Map<Integer, String> getPrintersAsMap() throws SQLException
  {
    Map<Integer, String> printerMap = new HashMap<Integer, String>();
    List<DruckerVO> printerList = selectVOList( TOTAL_NUMBER_OF_PRINTERS, VO_TYPE, SELECT_ALL_ORDERED, PaperTrayConstants.PAPER_WITH_IPI_LOGO,
        PaperTrayConstants.PAPER_PLAIN, PaperTrayConstants.PAPER_COVER );
    for ( DruckerVO druckerVO: printerList )
    {
      printerMap.put( druckerVO.druckerId, druckerVO.druckerName + " / " + druckerVO.raum );
    }
    return printerMap;
  }

  /**
   * Gets one JIPPS printer, including its paper tray IDs.
   * @param printerId The printer's unique ID
   * @return The printer or <code>null</code> if it's unknown
   * @throws SQLException if the query fails
   */
  public DruckerVO select( int printerId ) throws SQLException
  {
    return selectVO( VO_TYPE, SELECT_ONE, PaperTrayConstants.PAPER_WITH_IPI_LOGO, PaperTrayConstants.PAPER_PLAIN, PaperTrayConstants.PAPER_COVER, printerId );
  }

  /**
   * Gets the next available printer ID.
   * This may be used if a new printer is inserted.
   * @return the next available printer ID
   * @throws SQLException if the query fails
   */
  public int selectNextId() throws SQLException
  {
    return selectInt( getQueryRunner(), SELECT_NEXT_PRINTER_ID, 0 );
  }

  /**
   * Gets the distinct control servers for printing.
   * @return The control servers
   * @throws SQLException if the query fails
   */
  public Set<String> selectControlServers() throws SQLException
  {
    List<DruckerVO> printers = selectVOList( 2, VO_TYPE, SELECT_CONTROL_SERVERS );
    Set<String> controlServers = new HashSet<String>( 2 );
    for ( DruckerVO printer: printers )
    {
      controlServers.add( printer.controlServer );
    }
    return controlServers;
  }

  /**
   * Inserts a new printer. Note that the printer's paper trays
   * are also inserted in a separate table <code>LOGIDRUCKER</code>.
   * @param printer The printer to insert
   */
  @SuppressWarnings( "unchecked" )
  public void insert( final DruckerVO printer )
  {
    Check.notNull( printer, "printer" );

    this.execute( new SingleDAOUnitOfWork()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "drucker_insert";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        int numberOfUpdates = runner.update( INSERT_PRINTER, printer.druckerId, printer.druckerName, printer.controlServer, printer.raum, printer.farbe,
            printer.optionalPrintArgs );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't insert printer " + printer );

        // IPI paper tray
        numberOfUpdates = runner.update( INSERT_TRAY, printer.druckerId, PaperTrayConstants.PAPER_WITH_IPI_LOGO, PaperTrayConstants.PAPER_FORMAT,
            PaperTrayConstants.PAPER_WITH_IPI_LOGO_NAME, printer.igePaperId, printer.igePrintArgs );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't update paper tray " + PaperTrayConstants.PAPER_WITH_IPI_LOGO + " for printer " + printer );

        // Plain paper tray
        numberOfUpdates = runner.update( INSERT_TRAY, printer.druckerId, PaperTrayConstants.PAPER_PLAIN, PaperTrayConstants.PAPER_FORMAT,
            PaperTrayConstants.PAPER_PLAIN_NAME, printer.whitePaperId, printer.whitePrintArgs );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't update paper tray " + PaperTrayConstants.PAPER_WITH_IPI_LOGO + " for printer " + printer );

        // Cover paper tray
        numberOfUpdates = runner.update( INSERT_TRAY, printer.druckerId, PaperTrayConstants.PAPER_COVER, PaperTrayConstants.PAPER_FORMAT,
            PaperTrayConstants.PAPER_COVER_NAME, printer.coverPaperId, printer.coverPrintArgs );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't update paper tray " + PaperTrayConstants.PAPER_WITH_IPI_LOGO + " for printer " + printer );
      }
    } );
  }

  /**
   * Updates the printer and its paper trays.
   * @param printer The printer to update
   */
  @SuppressWarnings( "unchecked" )
  public void update( final DruckerVO printer )
  {
    Check.notNull( printer, "printer" );

    this.execute( new SingleDAOUnitOfWork()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "drucker_update";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        int numberOfUpdates = runner.update( UPDATE_PRINTER, printer.druckerName, printer.controlServer, printer.raum, printer.farbe,
            printer.optionalPrintArgs, printer.druckerId );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't update printer " + printer );

        // IPI paper tray
        numberOfUpdates = runner.update( UPDATE_TRAY, printer.igePaperId, printer.igePrintArgs, printer.druckerId, PaperTrayConstants.PAPER_WITH_IPI_LOGO );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't insert paper tray " + PaperTrayConstants.PAPER_WITH_IPI_LOGO + " for printer " + printer );

        // Plain paper tray
        numberOfUpdates = runner.update( UPDATE_TRAY, printer.whitePaperId, printer.whitePrintArgs, printer.druckerId, PaperTrayConstants.PAPER_PLAIN );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't insert paper tray " + PaperTrayConstants.PAPER_PLAIN + " for printer " + printer );

        // Cover paper tray
        numberOfUpdates = runner.update( UPDATE_TRAY, printer.coverPaperId, printer.coverPrintArgs, printer.druckerId, PaperTrayConstants.PAPER_COVER );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't insert paper tray " + PaperTrayConstants.PAPER_COVER + " for printer " + printer );
      }
    } );
  }

  /**
   * Deletes the printer, including its paper trays.
   * @param printerId The unique ID of the printer to delete
   */
  @SuppressWarnings( "unchecked" )
  public void delete( final int printerId )
  {
    this.execute( new SingleDAOUnitOfWork()
    {
      // @see ch.ige.dbutils.dao.AbstractSingleDAOUnitOfWork#getTransactionName()
      @Override
      protected String getTransactionName()
      {
        return "drucker_delete";
      }

      // @see ch.ige.dbutils.dao.UnitOfWork#execute(ch.ige.dbutils.runner.IGEQueryRunner)
      public void execute( IGEQueryRunner runner ) throws SQLException
      {
        int numberOfUpdates = runner.update( DELETE_ONE, printerId );
        verifyUpdateCount( numberOfUpdates, 1, "Couldn't delete printer " + printerId );
      }
    } );
  }

  /**
   * @return a list containing all printers except the colour printers
   * @throws SQLException 
   */
  public List<DruckerVO> getBlackAndWhitePrinters() throws SQLException
  {
    return selectVOList( TOTAL_NUMBER_OF_PRINTERS, VO_TYPE, SELECT_WHERE_NON_COLOR, PaperTrayConstants.PAPER_WITH_IPI_LOGO, PaperTrayConstants.PAPER_PLAIN,
        PaperTrayConstants.PAPER_COVER );
  }
}
