package ch.ige.jipps.dao;

import java.sql.SQLException;
import java.text.MessageFormat;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import oracle.sql.BLOB;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.dbutils.dao.AbstractJippsDAO;
import ch.ige.dbutils.runner.IGEQueryRunner;
import ch.ige.dbutils.sql.SQL;
import ch.ige.dbutils.sql.SQLStatement;
import ch.ige.jipps.vo.PrintJobVO;
import ch.ige.util.validation.Check;

/**
 * Accesses the table <code>PRINTJOB</code>.
 * 
 * @author Johannes Lang
 * @author Eric Lewis
 * @version $Id: PrintjobDAO.java,v 1.38 2008/06/05 13:19:43 lang Exp $
 */
public class PrintjobDAO extends AbstractJippsDAO
{
  /**
   * 
   */
  private static final String PRINTJOB_IDS = "printjobIds";

  // TODO Refactor this
  // TODO Rename this to PrintJobDAO

  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( PrintjobDAO.class );

  /**
   * The type of value object to return
   */
  private static final Class<PrintJobVO> VO_TYPE = PrintJobVO.class;

  /**
   * The name of the batch ID sequence
   */
  private static final String BATCH_ID_SEQUENCE = "seq_batch_id";

  /**
   * The name of the print job ID sequence
   */
  private static final String PRINT_JOB_ID_SEQUENCE = "seq_printjob_id";

  /**
   * Used for the extra handling of letters due to the use of Oracle Bfile for letters.
   */
  private LetterPrintjobDAO letterPrintjobDAO;

  /**
   * The base selection
   */
  private static final SQLStatement<PrintJobVO> BASE_SELECT = SQL.select( "SELECT printjob_id AS print_job_id, batch_id, drucker_id, paper_mode, anz_kopien, "
      + "druck_status, senderapp AS sender_app, senderuser AS sender_user, erst_dat AS erst_datum, typ AS form_typ, report_id, "
      + "printjob_file AS print_job_file, cover_cd FROM printjob " );

  /**
   * Selects one print job
   */
  private static final SQLStatement<PrintJobVO> SELECT_PRINT_JOB = BASE_SELECT.append( "WHERE printjob_id = ?" );

  /**
   * Selects all print jobs in a batch run
   */
  private static final SQLStatement<PrintJobVO> SELECT_BATCH_PRINT_JOBS = BASE_SELECT.append( "WHERE batch_id = ? ORDER BY printjob_id" );

  /**
   * Selects the form type for one print job.
   */
  private static final SQLStatement<String> SELECT_FORM_TYPE_BY_PRINT_JOB_ID = SQL.select( "SELECT typ FROM printjob WHERE printjob_id = ?" );

  /**
   * Selects the form type for a batch run.
   */
  private static final SQLStatement<String> SELECT_FORM_TYPE_BY_BATCH_ID = SQL.select( "SELECT typ FROM printjob WHERE batch_id = ? AND rownum < 2" );

  /**
   * Inserts a print job
   */
  private static final SQLStatement<Integer> INSERT_PRINTJOB = SQL.insert( "INSERT INTO printjob "
      + "(printjob_id, batch_id, drucker_id, paper_mode, anz_kopien, "
      + "druck_status, senderapp, senderuser, erst_dat, typ, report_id, printjob_file, cover_cd)" + "VALUES (?,?,?,?,?,?,?,?,sysdate,?,?,?,?)" );

  /**
   * Updates a number of print jobs
   */
  private static final String UPDATE_PRINT_JOB = "UPDATE printjob SET druck_status = ?, drucker_id = ? WHERE printjob_id IN ( {0} )";

  /**
   * Deletes one print job using its ID
   */
  private static final SQLStatement<Integer> DELETE_PRINT_JOB_BY_PRINT_JOB_ID = SQL.update( "DELETE FROM printjob WHERE printjob_id = ?" );

  /**
   * Deletes one print job using a List of printjobIds
   */
  private static final SQLStatement<Integer> DELETE_PRINT_JOB_BY_PRINT_JOB_IDS = SQL.update( "DELETE FROM printjob WHERE printjob_id IN ( {0} ) " );

  /**
   * Deletes all print jobs of a batch run
   */
  private static final SQLStatement<Integer> DELETE_PRINT_JOBS_BY_BATCH_ID = SQL.update( "DELETE FROM printjob WHERE batch_id = ?" );

  /**
   * Deletes all print jobs within a batch run that match a report ID
   */
  private static final SQLStatement<Integer> DELETE_PRINT_JOB_BY_IDS = SQL.update( "DELETE FROM printjob WHERE report_id = ? AND batch_id = ?" );

  /**
   * Deletes all print jobs that are not letters and older than a certain date
   */
  private static final SQLStatement<Integer> DELETE_NON_LETTER_PRINT_JOBS_BY_DATE = SQL
      .update( "DELETE FROM printjob WHERE lower(typ) <> 'letter' and erst_dat < ?" );

  /**
   * The only constructor
   */
  public PrintjobDAO()
  {
    super();
  }

  /**
   * Reads a print job.
   * 
   * @param printJobId The unique identifier for a print job
   * @return The print job or <code>null</code> if it doesn't exist
   * @throws SQLException if the query fails
   */
  public PrintJobVO readPrintjob( final int printJobId ) throws SQLException
  {
    Check.notNegative( printJobId, "print job ID" );
    Integer printJobID = Integer.valueOf( printJobId );

    Boolean isLetter = isLetter( printJobID );
    // switch because of the bfilehandling for letters
    if ( isLetter == null )
    {
      // The print job wasn't found...
      return null;
    }
    else if ( isLetter )
    {
      return getLetterPrintjobDAO().readPrintjob( printJobID );
    }
    else
    {
      return selectVO( VO_TYPE, SELECT_PRINT_JOB, printJobID );
    }
  }

  /**
   * Gets all print jobs for a given batch run.
   * 
   * @param batchId The unique identifier of the batch run
   * @return A <code>List</code> of print jobs for the batch run
   * @throws SQLException if the print jobs can't be fetched
   */
  public List<PrintJobVO> readPrintJobs( final Integer batchId ) throws SQLException
  {
    Check.notNegative( batchId, "batch ID" );

    Boolean isLetterBatch = isLetterBatch( batchId );
    if ( isLetterBatch == null )
    {
      // The batch run wasn't found...
      return Collections.emptyList();
    }
    else if ( isLetterBatch )
    {
      return getLetterPrintjobDAO().readPrintjobs( batchId );
    }
    else
    {
      return selectVOList( 20, VO_TYPE, SELECT_BATCH_PRINT_JOBS, Integer.valueOf( batchId ) );
    }
  }

  /**
   * Gets all print jobs for a given invoice (<code>BFaktur</code> in German).
   * 
   * @param invoiceId The unique identifier of the invoice
   * @return The print job IDs for the invoice
   * @throws SQLException if the print job IDs can't be fetched
   */
  public List<Integer> readPrintJobsByBfaktId( final Integer invoiceId ) throws SQLException
  {
    return getLetterPrintjobDAO().readPrintJobsByBfaktId( invoiceId );
  }

  /**
   * Inserts a print job.
   * 
   * @param printJob The print job to insert
   * @throws SQLException
   */
  @SuppressWarnings( "unchecked" )
  public void writePrintjob( final PrintJobVO printJob ) throws SQLException
  {
    Check.notNull( printJob, "print job" );

    if ( printJob.formTyp.equalsIgnoreCase( PrintJobVO.LETTER ) )
    {
      getLetterPrintjobDAO().writePrintjob( printJob );
    }
    else
    {
      this.execute( new SingleDAOUnitOfWork()
      {
        public void execute( IGEQueryRunner runner ) throws SQLException
        {
          BLOB blob = null;
          try
          {
            blob = runner.createTemporaryOracleBLOB( printJob.printJobFile );
            int updateCount = runner.update( INSERT_PRINTJOB, new Object[] { Integer.valueOf( printJob.printJobId ), Integer.valueOf( printJob.batchId ),
                Integer.valueOf( printJob.druckerId ), Integer.valueOf( printJob.paperMode ), Integer.valueOf( printJob.anzKopien ),
                Integer.valueOf( printJob.druckStatus ), printJob.senderApp, printJob.senderUser, printJob.formTyp, Integer.valueOf( printJob.reportId ), blob,
                Boolean.valueOf( printJob.coverCd ) } );
            verifyUpdateCount( updateCount, 1, "update count must be 1" );
          }
          finally
          {
            if ( blob != null )
            {
              blob.freeTemporary();
            }
          }
        }
      } );
    }

  }

  /**
   * Updates print jobs with certain values.
   * 
   * @param printJobIds The unique identifiers for the print jobs
   * @param printStatus The new print status to set
   * @param printerId The new printer ID to set
   */
  @SuppressWarnings( "unchecked" )
  public void updatePrintjob( final List<Integer> printJobIds, final int printStatus, final int printerId )
  {
    Check.notNull( printJobIds, "print job IDs" );

    if ( !printJobIds.isEmpty() )
    {
      this.execute( new SingleDAOUnitOfWork()
      {
        public void execute( IGEQueryRunner queryRunner ) throws SQLException
        {
          String idList = StringUtils.join( printJobIds.iterator(), ", " );
          SQLStatement<Integer> update = SQL.update( MessageFormat.format( UPDATE_PRINT_JOB, new Object[] { idList } ) );
          int updatedRows = queryRunner.update( update, Integer.valueOf( printStatus ), Integer.valueOf( printerId ) );
          if ( updatedRows != printJobIds.size() )
          {
            LOG.warn( "Tried to update printjobs with IDs " + idList + ", but only " + updatedRows + " were updated" );
          }
        }
      } );
    }
  }

  /**
   * Delete one print identified by <code>printJobId</code>.
   * 
   * @param printJobId The unique identifier for a print job
   */
  @SuppressWarnings( "unchecked" )
  public void deletePrintjob( final int printJobId )
  {
    Check.notNegative( printJobId, "print job ID" );
    this.execute( new SingleDAOUnitOfWork()
    {
      public void execute( IGEQueryRunner queryRunner ) throws SQLException
      {
        int updatedRows = queryRunner.update( DELETE_PRINT_JOB_BY_PRINT_JOB_ID, Integer.valueOf( printJobId ) );
        verifyUpdateCount( updatedRows, 1, "update count must be 1 for deleting print job ID " + printJobId );
      }
    } );
  }

  /**
   * Delete one print identified by <code>printJobId</code>.
   * @param printJobIds The list of unique identifiers for a print job
   */
  @SuppressWarnings( "unchecked" )
  public void deletePrintjobs( final List<Integer> printJobIds )
  {
    Check.notNull( printJobIds, PRINTJOB_IDS );
    Check.notEmpty( printJobIds, PRINTJOB_IDS );
    this.execute( new SingleDAOUnitOfWork()
    {
      public void execute( IGEQueryRunner queryRunner ) throws SQLException
      {
        String idList = StringUtils.join( printJobIds.iterator(), ", " );
        SQLStatement<Integer> delete = SQL.update( MessageFormat.format( DELETE_PRINT_JOB_BY_PRINT_JOB_IDS.toString(), new Object[] { idList } ) );
        int updatedRows = queryRunner.update( delete );
        verifyUpdateCount( updatedRows, printJobIds.size(), "expected updatecount: " + printJobIds.size() );
      }
    } );
  }

  /**
   * Delete all print jobs within a batch run.
   * 
   * @param batchId The unique identifier of the batch run
   * @return The number of deleted print jobs
   */
  public Integer deletePrintjobs( final int batchId )
  {
    Check.notNegative( batchId, "batch ID" );
    return this.execute( new SingleDAOUnitOfWork<Integer>()
    {
      public void execute( IGEQueryRunner queryRunner ) throws SQLException
      {
        int updatedRows = queryRunner.update( DELETE_PRINT_JOBS_BY_BATCH_ID, Integer.valueOf( batchId ) );
        setResult( updatedRows );
      }
    } );
  }

  /**
   * Delete all print jobs within a batch run that belong to a certain report.
   * 
   * @param reportId The unique identifier of the report
   * @param batchId The unique identifier of the batch run
   * @return The number of deleted print jobs
   */
  public Integer deletePrintjob( final int reportId, final int batchId )
  {
    Check.notNegative( reportId, "report ID" );
    Check.notNegative( batchId, "batch ID" );
    return this.execute( new SingleDAOUnitOfWork<Integer>()
    {
      public void execute( IGEQueryRunner queryRunner ) throws SQLException
      {
        int updatedRows = queryRunner.update( DELETE_PRINT_JOB_BY_IDS, new Object[] { Integer.valueOf( reportId ), Integer.valueOf( batchId ) } );
        setResult( updatedRows );
      }
    } );
  }

  /**
   * Deletes all print jobs that are older than the given Date.
   * 
   * @param date Print jobs after this are deleted
   * @return The number of deleted print jobs
   */
  public Integer deletePrintjobs( final Date date )
  {
    Check.notNull( date, "date" );
    return this.execute( new SingleDAOUnitOfWork<Integer>()
    {
      public void execute( IGEQueryRunner queryRunner ) throws SQLException
      {
        int updatedRows = queryRunner.update( DELETE_NON_LETTER_PRINT_JOBS_BY_DATE, date );
        setResult( updatedRows );
      }
    } );
  }

  /**
   * Gets the next ID of the sequence for the print job batch ID.
   * 
   * @return The next batch ID
   * @throws SQLException if the query fails
   */
  public int getNextBatchId() throws SQLException
  {
    return selectNextSequenceValue( getQueryRunner(), BATCH_ID_SEQUENCE ).intValue();
  }

  /**
   * Gets the next ID of the sequence for the print job ID.
   * 
   * @return The next print job ID
   * @throws SQLException if the query fails
   */
  public int getNextPrintJobId() throws SQLException
  {
    return selectNextSequenceValue( getQueryRunner(), PRINT_JOB_ID_SEQUENCE ).intValue();
  }

  /**
   * Gets the <code>LetterPrintjobDAO</code> by lazy instantiation.
   * @return The <code>LetterPrintjobDAO</code>
   */
  private LetterPrintjobDAO getLetterPrintjobDAO()
  {
    if ( this.letterPrintjobDAO == null )
    {
      this.letterPrintjobDAO = new LetterPrintjobDAO();
    }
    return this.letterPrintjobDAO;
  }

  /**
   * Shows whether a print job was done for a letter.
   * @param printJobId The print job's ID
   * @return <code>true</code> if it's a letter print job, <code>false</code> if it's
   *         not a letter print job, <code>null</code> if the print job wasn't found
   * @throws SQLException if the query fails
   */
  private Boolean isLetter( Integer printJobId ) throws SQLException
  {
    Check.notNegative( printJobId, "print job ID" );
    String formType = selectString( this.getQueryRunner(), SELECT_FORM_TYPE_BY_PRINT_JOB_ID, printJobId );
    if ( formType == null )
    {
      return null;
    }
    return formType.equalsIgnoreCase( PrintJobVO.LETTER );
  }

  /**
   * Shows whether a batch run was done for a letter.
   * @param batchId The batch run's ID
   * @return <code>true</code> if it's a letter batch run, <code>false</code> if it's
   *         not a letter batch run, <code>null</code> if the batch run wasn't found
   * @throws SQLException if the query fails
   */
  private Boolean isLetterBatch( Integer batchId ) throws SQLException
  {
    Check.notNegative( batchId, "batch ID" );
    String formType = selectString( this.getQueryRunner(), SELECT_FORM_TYPE_BY_BATCH_ID, batchId );
    if ( formType == null )
    {
      return null;
    }
    return formType.equalsIgnoreCase( PrintJobVO.LETTER );
  }
}
