package ch.ige.jipps.dao.test;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import ch.ige.dbutils.dao.test.BaseDAOTestTemplate;
import ch.ige.jipps.dao.LoeschungErrorDAO;
import ch.ige.jipps.vo.LoeschungErrorVO;
import ch.ige.jipps.vo.LoeschungLaufVO;

/**
 * Tests {@link ch.ige.jipps.dao.LoeschungErrorDAO}.
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   Eric Lewis
 * @version  $Id: LoeschungErrorDAOTest.java,v 1.3 2008/05/22 09:24:11 lewis_lee Exp $
 */
public class LoeschungErrorDAOTest extends BaseDAOTestTemplate<LoeschungErrorDAO, LoeschungErrorVO>
{
  /**
   * An unknown cancellation error.
   */
  private static final int UNKNOWN_CANCELLATION_ERROR_ID = 15;

  /**
   * The error message to use.
   */
  private static final String ERROR_MESSAGE = "This is only a test from " + LoeschungDAOTest.class.toString();

  /**
   * If this exists, it's deleted during {@link #tearDown()}.
   */
  private Integer insertedCancellationErrorId = null;

  /**
   * Used for setting up a cancellation run.
   */
  private LoeschungLaufDAOTest loeschungLaufDAOTest;

  /**
   * The cancellation run set up for this test.
   */
  private LoeschungLaufVO cancellationRun = null;

  /**
   * The only constructor
   * @param name The test's name
   */
  public LoeschungErrorDAOTest( String name )
  {
    super( name );
  }

  @Override
  public void setUp() throws Exception
  {
    super.setUp();
    this.loeschungLaufDAOTest = new LoeschungLaufDAOTest( getName() );
    this.cancellationRun = this.loeschungLaufDAOTest.createAndInsert();
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOTestTemplate#getDAOClass()
  @Override
  protected Class<LoeschungErrorDAO> getDAOClass()
  {
    return LoeschungErrorDAO.class;
  }

  /**
   * Tests {@link ch.ige.jipps.dao.LoeschungErrorDAO#selectLoeschungError()}.
   * @throws Exception if the test fails
   */
  public void testSelectLoeschungError() throws Exception
  {
    insert();
    List<LoeschungErrorVO> errors = getDAO().selectLoeschungError();
    assertTrue( "Checking number of errors", errors.size() >= 1 );

    boolean foundInsertedError = false;
    for ( LoeschungErrorVO error: errors )
    {
      if ( error.loeschungLaufId == this.cancellationRun.loeschungLaufId )
      {
        foundInsertedError = true;
      }
      assertTrue( "Checking cancellation run ID", error.loeschungLaufId > 0 );
      assertTrue( "Checking invoice ID", error.bfaktId > 0 );
      assertTrue( "Checking error message", StringUtils.isNotEmpty( error.errorMessage ) );
    }
    assertTrue( "Checking for inserted cancellation run " + this.cancellationRun.loeschungLaufId, foundInsertedError );
  }

  /**
   * Tests {@link ch.ige.jipps.dao.LoeschungErrorDAO#selectLoeschungError(int)}.
   * @throws Exception if the test fails
   */
  public void testSelectLoeschungErrorForOneError() throws Exception
  {
    LoeschungErrorVO error = getDAO().selectLoeschungError( UNKNOWN_CANCELLATION_ERROR_ID );
    assertNull( "Checking unknown cancellation error", error );

    insert();
    error = getDAO().selectLoeschungError( this.insertedCancellationErrorId );
    assertEquals( "Checking cancellation error ID", this.insertedCancellationErrorId.intValue(), error.keyId );
    assertEquals( "Checking cancellation run ID", this.cancellationRun.getIdentifier().intValue(), error.loeschungLaufId );
    assertEquals( "Checking invoice ID", 99999, error.bfaktId );
    assertEquals( "Checking error message", ERROR_MESSAGE, error.errorMessage );
  }

  /**
   * Tests {@link ch.ige.jipps.dao.LoeschungErrorDAO#insert(int, int, java.lang.String)}.
   * @throws Exception if the test fails
   */
  public void testInsert() throws Exception
  {
    insert();
  }

  /**
   * Inserts a cancellation error.
   * @throws Exception if the insert fails
   */
  private void insert() throws Exception
  {
    LoeschungErrorVO error = createErrorToInsert();
    this.insertedCancellationErrorId = getDAO().insert( error.loeschungLaufId, error.bfaktId, error.errorMessage );
  }

  /**
   * Tests {@link ch.ige.jipps.dao.LoeschungErrorDAO#deleteLoeschungError(List)}.
   */
  public void testDeleteLoeschungError()
  {
    List<Integer> errorIds = new ArrayList<Integer>();
    errorIds.add( UNKNOWN_CANCELLATION_ERROR_ID );
    Integer numberOfDeletes = getDAO().deleteLoeschungError( errorIds );
    assertEquals( "Checking for unknown ID", 0, numberOfDeletes.intValue() );

    LoeschungErrorVO error = createErrorToInsert();
    this.insertedCancellationErrorId = getDAO().insert( error.loeschungLaufId, error.bfaktId, error.errorMessage );

    errorIds.clear();
    errorIds.add( this.insertedCancellationErrorId );
    numberOfDeletes = getDAO().deleteLoeschungError( errorIds );
    assertEquals( "Checking that the cancellation error was deleted", 1, numberOfDeletes.intValue() );
    this.insertedCancellationErrorId = null;
  }

  /**
   * Creates the cancellation error to insert.
   * @return The cancellation error to insert
   */
  private LoeschungErrorVO createErrorToInsert()
  {
    LoeschungErrorVO error = new LoeschungErrorVO();
    error.bfaktId = 99999;
    assertNotNull( "Checking that a cancellation run exists before inserting an cancellation error", this.cancellationRun );
    error.loeschungLaufId = this.cancellationRun.loeschungLaufId;
    error.errorMessage = ERROR_MESSAGE;
    return error;
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOTestTemplate#getSerializableVO()
  @Override
  protected LoeschungErrorVO getSerializableVO()
  {
    LoeschungErrorVO cancellationError = new LoeschungErrorVO();
    cancellationError.keyId = 33;
    cancellationError.loeschungLaufId = 44;
    cancellationError.bfaktId = 55;
    cancellationError.errorMessage = "Oh no! Not again!";
    return cancellationError;
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOTestTemplate#assertVOEqual(ch.ige.dbutils.vo.AbstractVO, ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected boolean assertVOEqual( LoeschungErrorVO originalVO, LoeschungErrorVO serializedVO )
  {
    return originalVO.keyId == serializedVO.keyId && originalVO.loeschungLaufId == serializedVO.loeschungLaufId && originalVO.bfaktId == serializedVO.bfaktId
        && originalVO.errorMessage.equals( serializedVO.errorMessage );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOTestTemplate#tearDown()
  @Override
  public void tearDown() throws Exception
  {
    if ( this.insertedCancellationErrorId != null )
    {
      List<Integer> cancellationErrorIds = new ArrayList<Integer>();
      cancellationErrorIds.add( this.insertedCancellationErrorId );
      getDAO().deleteLoeschungError( cancellationErrorIds );
    }
    if ( this.cancellationRun != null )
    {
      this.loeschungLaufDAOTest.delete( this.cancellationRun.getIdentifier() );
    }
    super.tearDown();
  }
}
