package ch.ige.jipps.header;

import java.sql.Date;
import java.sql.SQLException;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.ObjectUtils;

import ch.ige.ipobjects.common.GenericIP;
import ch.ige.ipobjects.report.dao.RegFormTextDAO;
import ch.ige.ipobjects.report.vo.RegAuszTypSTypVO;
import ch.ige.ipobjects.report.vo.RegFormTextVO;
import ch.ige.util.date.LanguageDateFormatter;
import ch.ige.util.enums.IPType;
import ch.ige.util.enums.report.ReportArea;
import ch.ige.util.enums.report.ReportType;

/**
 * Builds the object for the headers in the report.
 * <p>
 * Copyright (c) 2006, Swiss Federal Institute of Intellectual Property
 * </p>
 * @author Eric Lewis
 * @version $Id: Header.java,v 1.28 2008/09/22 10:05:14 lewis_lee Exp $
 */
public class Header
{
  /**
   * The default value
   */
  private static final String EMPTY = "";

  /**
   * Gets the print header for the report.
   * @param ipHeader The IP header, which contains the formatted IP number
   * @param reportType The report type
   * @param reportDate The report's date
   * @param language The language to use for the report
   * @return The print header to use in the report
   * @throws SQLException if the data can't be fetched
   */
  public static PrintHeader getPrintHeader( IpHeader ipHeader, ReportType reportType, Date reportDate, Locale language ) throws SQLException
  {
    LanguageDateFormatter dateFormatter = new LanguageDateFormatter( language );

    PrintHeaderVO printHeader = new PrintHeaderVO();
    printHeader.setPrintDate( formatDate( reportDate, dateFormatter ) );

    // The current state for designs is only used for designs in application state, so write the application number
    // instead of the IP number, which doesn't exist.
    List<RegFormTextVO> texts = null;
    int ipNumberLabelIndex = 0;
    if ( reportType.equals( ReportType.CURRENT_STATE ) && ipHeader.getIpType().equals( IPType.DESIGN ) )
    {
      texts = new RegFormTextDAO().findReportTexts( language, reportType, IPType.DESIGN, ReportArea.PAGE_HEADER );
      printHeader.setFormattedIpNumber( ipHeader.getFormattedApplicationNumber() );
      ipNumberLabelIndex = 2;
    }
    else
    {
      texts = new RegFormTextDAO().findReportTexts( language, ReportType.NO_TYPE, IPType.ALL, ReportArea.PAGE_HEADER );
      printHeader.setFormattedIpNumber( ipHeader.getFormattedIpNumber() );
      ipNumberLabelIndex = 1;
    }

    // Note: The texts are sorted by report text number, so getting them by index should work...
    // But for the design's current state, an alternative is required.
    RegFormTextVO formText = texts.get( 0 );
    printHeader.setPageLabel( formText.getText() + " " );

    formText = texts.get( ipNumberLabelIndex );
    printHeader.setIpNumberLabel( formText.getText() + " " );

    return printHeader;
  }

  /**
   * Gets the IP header for the report.
   * @param ip The IP containing all the data
   * @param reportDate The report's date
   * @param ipTypeReportType The IP type report type connection
   * @param language The language to use for the report
   * @return The IP header to use in the report
   */
  public static IpHeader getIpHeader( GenericIP ip, Date reportDate, RegAuszTypSTypVO ipTypeReportType, Locale language )
  {
    IpHeaderBuilder ipHeaderBuilder = getIpHeaderBuilder( ip );

    IpHeaderVO ipHeader = new IpHeaderVO();
    ipHeader.setReportType( ipTypeReportType.getReportType() );

    ipHeader.setReportLanguage( ipHeaderBuilder.getReportLanguage( ip, language ) );

    String reportTitle = ipTypeReportType.getReportTitle( ipHeader.getReportLanguage(), ip.getRegistrationState() );
    if ( reportTitle == null )
    {
      throw new IllegalStateException( "Found no report title for IP type " + ip.getBAGISType() + ", report type " + ipTypeReportType.getReportType()
          + " and registration state " + ip.getRegistrationState() );
    }
    ipHeader.setReportTitle( reportTitle );

    LanguageDateFormatter dateFormatter = new LanguageDateFormatter( ipHeader.getReportLanguage() );
    ipHeader.setReportDate( formatDate( reportDate, dateFormatter ) );

    ipHeader.setIpKey( ip.getIPKey() );
    ipHeader.setIpType( ip.getBAGISType() );
    ipHeader.setDescription( ipHeaderBuilder.getDescription( ip, reportDate ) );

    ipHeader.setIpNumber( ObjectUtils.toString( ip.getNumber() ) );
    ipHeader.setFormattedIpNumber( ipHeaderBuilder.getFormattedIpNumber( ip ) );
    ipHeader.setFormattedBaseIpNumber( ipHeaderBuilder.getFormattedBaseIpNumber( ip ) );

    ipHeader.setProtectionEnd( formatDate( ip.getProtectionEnd(), dateFormatter ) );
    ipHeader.setBaseProtectionStart( ipHeaderBuilder.getBaseProtectionStart( ip, dateFormatter ) );
    ipHeader.setCancellation( formatDate( ip.getCancellationDate(), dateFormatter ) );

    ipHeader.setRegistrationState( ip.getRegistrationState() );
    if ( ip.getRegistrationEvent() != null )
    {
      ipHeader.setRegistrationDate( formatDate( ip.getRegistrationEvent().getDate(), dateFormatter ) );
    }
    else
    {
      ipHeader.setRegistrationDate( EMPTY );
    }

    ipHeader.setApplicationDate( ipHeaderBuilder.getApplicationDate( ip, dateFormatter ) );
    ipHeader.setApplicationPublicationDate( ipHeaderBuilder.getApplicationPublicationDate( ip, dateFormatter ) );
    ipHeader.setFormattedApplicationNumber( ipHeaderBuilder.getFormattedApplicationNumber( ip ) );

    PublicationVO firstPublicationResults = ipHeaderBuilder.getPublication( ip, dateFormatter, reportDate, true );
    ipHeader.setFirstPublicationDate( firstPublicationResults.formattedPublicationDate );
    ipHeader.setFirstPublicationYear( firstPublicationResults.publicationYear );
    ipHeader.setFirstPublicationNumber( firstPublicationResults.publicationNumber );
    ipHeader.setEventOfFirstPublicationPublished( firstPublicationResults.publishPublicationEvent );

    ipHeader.setPublishedElectronically( firstPublicationResults.isPublishedElectronically );

    PublicationVO lastPublicationResults = ipHeaderBuilder.getPublication( ip, dateFormatter, reportDate, false );
    ipHeader.setLastPublicationDate( lastPublicationResults.formattedPublicationDate );
    ipHeader.setLastPublicationYear( lastPublicationResults.publicationYear );
    ipHeader.setLastPublicationNumber( lastPublicationResults.publicationNumber );
    ipHeader.setEventOfLastPublicationPublished( lastPublicationResults.publishPublicationEvent );

    ipHeader.setOppositionDeadline( ipHeaderBuilder.getOppositionDeadline( ip, reportDate, dateFormatter ) );

    return ipHeader;
  }

  /**
   * Gets the IP header builder for special cases.
   * @param ip The IP containing all data
   * @return The specific IP header builder for special cases
   */
  private static IpHeaderBuilder getIpHeaderBuilder( GenericIP ip )
  {
    switch( ip.getBAGISType() )
    {
      case DESIGN:
        return new DesignHeaderBuilder();
      case PATENT_CH:
      case PATENT_EP:
      case PATENT_PCT:
        return new PatentHeaderBuilder();
      case SPC:
        return new SpcHeaderBuilder();
      case TRADEMARK_CH:
        return new TrademarkHeaderBuilder();
      default:
        throw new IllegalArgumentException( "Unknown IP type " + ip.getBAGISType() );
    }
  }

  /**
   * Formats the <code>Date</code> or returns an empty <code>String</code>.
   * @param date The date to format
   * @param dateFormatter The date formatter
   * @return The formatted date or an empty <code>String</code> if <code>date</code> is <code>null</code>
   */
  static String formatDate( Date date, LanguageDateFormatter dateFormatter )
  {
    if ( date == null )
    {
      return EMPTY;
    }
    else
    {
      return dateFormatter.longFormat( date );
    }
  }
}
