package ch.ige.jipps.header;

import java.util.Locale;

import ch.ige.util.enums.IPType;
import ch.ige.util.enums.RegistrationState;
import ch.ige.util.enums.report.ReportType;

/**
 * Contains the data for the header in the report concerning the IP's base data.
 * <p>
 * Copyright (c) 2006, Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * 
 * @author Eric Lewis
 * @version $Id: IpHeader.java,v 1.19 2008/09/22 10:05:14 lewis_lee Exp $
 */
public interface IpHeader
{
  /**
   * Gets the type of report.
   * 
   * @return The report's type
   */
  public ReportType getReportType();

  /**
   * Gets the title of the report
   * 
   * @return The report's title
   */
  public String getReportTitle();

  /**
   * Gets the formatted date of the report.
   * 
   * @return The formatted date of the report
   */
  public String getReportDate();

  /**
   * Gets the report's language
   * 
   * @return The report's language
   */
  public Locale getReportLanguage();

  /**
   * Gets the unique key identifying the IP.
   * 
   * @return The unique key identifying the IP
   */
  public Integer getIpKey();

  /**
   * Gets the IP number, which is unique per IP type. This is a <code>String</code> as it's coming directly from the
   * database. For Euopean Patents, this is the EP patent number, for PCT patents, it's the WO publication number. For
   * some IP types, it includes a checksum.
   * 
   * @return The ip number
   */
  public String getIpNumber();

  /**
   * Gets the formatted version of {@link #getIpNumber()}, which loses the checksum and may be formatted with slashes
   * etc.
   * 
   * @return The formatted ip number
   */
  public String getFormattedIpNumber();

  /**
   * Gets the formatted IP number for an SPC's base patent.
   * 
   * @return The formatted IP number for an SPC's base patent
   */
  public String getFormattedBaseIpNumber();

  /**
   * Gets the IP type.
   * 
   * @return The IP type
   */
  public IPType getIpType();

  /**
   * Gets the IP's description.
   * 
   * @return The IP's description
   */
  public String getDescription();

  /**
   * Gets the image's filename.
   * 
   * @return The image's filename
   */
  public String getImageFilename();

  /**
   * Gets the formatted end of protection for an IP.
   * 
   * @return The formatted end of protection for an IP
   */
  public String getProtectionEnd();

  /**
   * For an SPC, this gets the formatted date when the original patent was registered.
   * 
   * @return The formatted start of protection for an SPC's base patent
   */
  public String getBaseProtectionStart();

  /**
   * Gets the formatted date of application.
   * 
   * @return The formatted date of application
   */
  public String getApplicationDate();

  /**
   * Gets the formatted date showing when the application was published.
   * 
   * @return The formatted date of application publication
   */
  public String getApplicationPublicationDate();

  /**
   * Gets the formatted application number.
   * 
   * @return The formatted application number
   */
  public String getFormattedApplicationNumber();

  /**
   * Gets the formatted date of cancellation.
   * 
   * @return The formatted date of cancellation
   */
  public String getCancellation();

  /**
   * Gets the IP's registration state.
   * 
   * @return The IP's registration state
   */
  public RegistrationState getRegistrationState();

  /**
   * Gets the formatted date when this IP was registered.
   * 
   * @return The formatted date when this IP was registered
   */
  public String getRegistrationDate();

  /**
   * Shows whether this is published by ePublication.
   * 
   * @return <code>true</code> if it's published electronically, otherwise <code>false</code>
   */
  public boolean isPublishedElectronically();

  /**
   * Gets the date of first publication.
   * 
   * @return The date of first publication
   */
  public String getFirstPublicationDate();

  /**
   * Gets the four-digit year of first publication (taken from the date of publication).
   * 
   * @return The four-digit year of first publication (taken from the date of publication)
   */
  public Integer getFirstPublicationYear();

  /**
   * Gets the publication number given when the IP was published. This is unique within a year.
   * 
   * @return The publication number given when the IP was published
   */
  public Integer getFirstPublicationNumber();

  /**
   * Indicates whether the event containing the date of publication must be published.
   * 
   * @return <code>true</code> if the event containing the date of publication must be published, otherwise
   *         <code>false</code>
   */
  public Boolean isEventOfFirstPublicationPublished();

  /**
   * Gets the date of last publication.
   * 
   * @return The date of last publication
   */
  public String getLastPublicationDate();

  /**
   * Gets the four-digit year of last publication (taken from the date of publication).
   * 
   * @return The four-digit year of last publication (taken from the date of publication)
   */
  public Integer getLastPublicationYear();

  /**
   * Gets the publication number given when the IP was published. This is unique within a year.
   * 
   * @return The publication number given when the IP was published
   */
  public Integer getLastPublicationNumber();

  /**
   * Indicates whether the event containing the date of publication must be published.
   * 
   * @return <code>true</code> if the event containing the date of publication must be published, otherwise
   *         <code>false</code>
   */
  public Boolean isEventOfLastPublicationPublished();

  /**
   * Gets the formatted last date for opposition to a trademark. After a trademark has been published, anyone can enter
   * opposition with the IPI.
   * 
   * @return The formatted last date for opposition to a trademark
   */
  public String getOppositionDeadline();
}
