package ch.ige.jipps.header;

import java.sql.Date;
import java.util.Calendar;

import org.apache.commons.lang.StringUtils;

import ch.ige.ipobjects.Types;
import ch.ige.ipobjects.common.GenericIP;
import ch.ige.ipobjects.event.Event;
import ch.ige.ipobjects.patent.Patent;
import ch.ige.ipobjects.patent.PatentEP;
import ch.ige.ipobjects.patent.PatentPCT;
import ch.ige.util.constants.NumericLiterals;
import ch.ige.util.date.LanguageDateFormatter;
import ch.ige.util.format.ApplicationNumberFormatter;

/**
 * Builds the special cases for the IP header in the report for patents.
 * <p>
 * Copyright (c) 2006, Swiss Federal Institute of Intellectual Property
 * </p>
 * @author Eric Lewis
 * @version $Id: PatentHeaderBuilder.java,v 1.14 2008/01/10 17:18:56 lewis_lee Exp $
 */
class PatentHeaderBuilder extends BaseIpHeaderBuilder
{
  // See interface
  @Override
  public String getFormattedApplicationNumber( GenericIP ip )
  {
    Patent patent = (Patent) ip;

    if ( patent.isPCTPatent() && StringUtils.isNotEmpty( patent.getApplicationNumber() ) )
    {
      // The first two characters contain the ISO country code, the rest is a number
      String countryCode = patent.getApplicationNumber().substring( 0, 2 );
      return ApplicationNumberFormatter.formatPct( patent.getApplicationNumber().substring( 2 ), patent.getBAGISType(), countryCode );
    }
    else
    {
      return super.getFormattedApplicationNumber( ip );
    }
  }

  // See interface
  @Override
  public String getApplicationPublicationDate( GenericIP ip, LanguageDateFormatter dateFormatter )
  {
    Patent patent = (Patent) ip;
    if ( patent.getPublicationOfApplication() != null )
    {
      return Header.formatDate( patent.getPublicationOfApplication().getDate(), dateFormatter );
    }
    else
    {
      return EMPTY;
    }
  }

  // See interface
  @Override
  public PublicationVO getPublication( GenericIP ip, LanguageDateFormatter dateFormatter, Date reportDate, boolean isFirstPublication )
  {
    // There's no last publication for patents
    if ( !isFirstPublication )
    {
      return super.getDefaultPublicationResults( ip, reportDate, isFirstPublication );
    }

    Patent patent = (Patent) ip;
    PublicationVO publication = getDefaultPublicationResults( ip, reportDate, isFirstPublication );

    // For Swiss patents, the publication year is required instead of the full publication date
    if ( Types.PATENT_CH.isInstance( patent ) )
    {
      Event publicationEvent = getPublicationEvent( patent, reportDate, isFirstPublication );
      if ( publicationEvent != null && publicationEvent.getDate() != null )
      {
        Calendar publicationDate = Calendar.getInstance();
        publicationDate.setTime( publicationEvent.getPubl_dat() );
        publication.formattedPublicationDate = EMPTY;
        publication.publicationYear = Integer.valueOf( publicationDate.get( Calendar.YEAR ) );
        publication.publicationNumber = publicationEvent.getPmmbl_seite();
      }
    }
    // For EP patents, it's the A publication date
    else if ( Types.PATENT_EP.isInstance( patent ) )
    {
      PatentEP patentEp = (PatentEP) patent;
      publication.formattedPublicationDate = Header.formatDate( patentEp.getPublicationDateA(), dateFormatter );
      // Publication year and number are irrelevant for EP patents
      publication.publicationYear = NumericLiterals.ZERO;
      publication.publicationNumber = null;
    }
    // For PCT patents, it's the PCT publication date (how surprising...)
    else if ( Types.PATENT_PCT.isInstance( patent ) )
    {
      PatentPCT patentPct = (PatentPCT) patent;
      publication.formattedPublicationDate = Header.formatDate( patentPct.getPCTPublicationDate(), dateFormatter );
      // Publication year and number are irrelevant for PCT patents
      publication.publicationYear = NumericLiterals.ZERO;
      publication.publicationNumber = null;
    }
    else
    {
      throw new IllegalArgumentException( "Found unknown patent type " + patent.getType() );
    }

    // Publication flag is irrelevant for patents
    publication.publishPublicationEvent = false;

    return publication;
  }
}
