package ch.ige.jipps.history.description;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.commons.lang.StringUtils;

import ch.ige.ipobjects.common.GenericIP;
import ch.ige.ipobjects.design.Design;
import ch.ige.ipobjects.event.Event;
import ch.ige.ipobjects.time.PointInTime;
import ch.ige.ipobjects.time.TimePeriod;
import ch.ige.jipps.history.HistoryItem;
import ch.ige.jipps.history.detail.BaseHistoryItemDetail;
import ch.ige.jipps.history.detail.design.DesignClassificationDetail;
import ch.ige.jipps.history.detail.design.DesignDescriptionDetail;
import ch.ige.jipps.history.detail.design.DesignItemDetail;
import ch.ige.jipps.history.detail.design.DesignPrioritiesDetail;
import ch.ige.jipps.history.detail.design.DesignTitleDetail;
import ch.ige.jipps.history.detail.design.NumberOfItemsDetail;
import ch.ige.jipps.vo.HistoryItemVO;
import ch.ige.util.enums.report.ReportType;

/**
 * Configures how the history part of the report is built for a design.
 * <p>
 * Copyright (c) 2006, Swiss Federal Institute of Intellectual Property
 * </p>
 * @author Esther Bandschapp
 * @version $Id: DesignHistoryDescription.java,v 1.21 2007/12/20 08:24:54 lewis_lee Exp $
 */
public class DesignHistoryDescription extends BaseHistoryDescription
{
  /**
   * Creates this with all required data.
   * @param ip The design
   * @param currentState The history is built for this date
   * @param language The language of the report
   * @param includeWaitListEvents If this is <code>true</code>, events that are on the wait list are also shown
   * @param startFromApplication If this is <code>true</code>, events starting from the application are shown,
   * otherwise from the registration
   * @param allowedReportGroups The report groups, a <code>Map</code> where the keys show whether an event must be
   * included on a report. The values can be ignored.
   * @param reportType The report type
   */
  public DesignHistoryDescription( GenericIP ip, PointInTime currentState, Locale language, boolean includeWaitListEvents, boolean startFromApplication,
      Map<Integer, String> allowedReportGroups, ReportType reportType )
  {
    super( ip, currentState, language, includeWaitListEvents, startFromApplication, allowedReportGroups, reportType );
  }

  // See interface
  @SuppressWarnings( "unchecked" )
  @Override
  public List<BaseHistoryItemDetail<? extends TimePeriod>> getIpSpecificDetailDescriptions()
  {
    Design design = (Design) this.ip;
    List<BaseHistoryItemDetail<? extends TimePeriod>> descriptions = new ArrayList<BaseHistoryItemDetail<? extends TimePeriod>>();
    descriptions.add( new DesignTitleDetail( design.getTitle() ) );
    descriptions.add( new DesignDescriptionDetail( design.getDesignDescription() ) );
    descriptions.add( new DesignClassificationDetail( design.getGroupedClassifications().getTimePeriods(), this.ip.getType() ) );
    descriptions.add( new NumberOfItemsDetail( design.getNumberOfItems() ) );
    descriptions.add( new DesignPrioritiesDetail( design.getGroupedPriorites().getTimePeriods(), design.getType(), this.ip ) );
    descriptions.add( new DesignItemDetail( design.getGroupedDesignItems().getTimePeriods() ) );
    return descriptions;
  }

  // See superclass
  @Override
  public List<HistoryItemVO> inspectAndFilterCreated( Event event, List<HistoryItemVO> createdItems )
  {
    List<HistoryItemVO> filteredItems = super.inspectAndFilterCreated( event, createdItems );
    // If the 'valid to...' dates are the same, only the last is kept.
    HistoryItem previousHistoryItem = null;
    for ( HistoryItemVO detail: filteredItems )
    {
      if ( StringUtils.isNotEmpty( detail.getEndDate() ) && detail.getCreator().shouldGroupEndDates() && previousHistoryItem != null
          && detail.getEndDate().equals( previousHistoryItem.getEndDate() ) )
      {
        previousHistoryItem.setEndDate( null );
      }
      previousHistoryItem = detail;
      detail.setCreator( null ); // Help the garbage collector to get rid of the creator
    }
    return filteredItems;
  }
}
