package ch.ige.jipps.history.description.test;

import java.sql.Date;
import java.util.Calendar;
import java.util.Collections;

import junit.framework.TestCase;
import ch.ige.dbutils.util.ClientInitializer;
import ch.ige.ipobjects.common.GenericIP;
import ch.ige.ipobjects.common.model.AbstractEventWrapper;
import ch.ige.ipobjects.common.model.GenericType;
import ch.ige.ipobjects.event.Event;
import ch.ige.ipobjects.time.BAGISTime;
import ch.ige.ipobjects.time.PointInTimeCollection;
import ch.ige.ipobjects.time.vo.EreigIDat;
import ch.ige.ipobjects.time.vo.GaplessPointInTimeCollection;
import ch.ige.util.enums.PublicationTarget;

/**
 * The base class for testing the creation of the publication text, using dummy IPs.
 * <p>
 * Copyright (c) 2006, Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * @author Eric Lewis
 * @version $Id: BasePublicationTextTestTemplate.java,v 1.10 2008/05/21 13:49:29 lewis_lee Exp $
 */
public abstract class BasePublicationTextTestTemplate extends TestCase
{
  static
  {
    ClientInitializer.initStandaloneClient();
  }

  /**
   * The only constructor
   * @param name The test's name
   */
  public BasePublicationTextTestTemplate( String name )
  {
    super( name );
  }

  /**
   * Tests an event which has definitely been published and whose publication is required.
   * @throws Exception if the test fails
   */
  public abstract void testDefinitiveAndRequiredPublication() throws Exception;

  /**
   * Tests an event which has definitely been published and whose publication is not required.
   * @throws Exception if the test fails
   */
  public abstract void testDefinitiveAndNonRequiredPublication() throws Exception;

  /**
   * Tests an event which has not yet been published and whose publication is required.
   * @throws Exception if the test fails
   */
  public abstract void testPreliminaryAndRequiredPublication() throws Exception;

  /**
   * Tests an event which has not yet been published and whose publication is not required.
   * @throws Exception if the test fails
   */
  public abstract void testPreliminaryAndNonRequiredPublication() throws Exception;

  /**
   * Since the publication changed around 2002, this checks for publication texts before and after the publication
   * change.
   * @throws Exception if the test fails
   */
  public abstract void testPublicationChange() throws Exception;

  /**
   * Sets the event in the IP.
   * @param ip The IP whose events are set with <code>event</code>
   * @param event The event to set
   */
  @SuppressWarnings( "unchecked" )
  protected void setEvent( GenericIP ip, Event event )
  {
    PointInTimeCollection<Event> events = new GaplessPointInTimeCollection<Event>( Collections.singletonList( event ), true );
    ip.set( GenericType.PROP_EVENTS, events );
  }

  /**
   * Implements an Event just for this test.
   */
  protected static class MockEvent extends AbstractEventWrapper
  {
    /**
     * The time to compare to
     */
    private int compareTime;

    /**
     * Shows whether this must be published
     */
    private boolean publ_pflicht;

    /**
     * Shows whether this has been definitely published
     */
    private boolean publ_definitiv;

    /**
     * The publication date
     */
    private Date publ_dat;

    /**
     * The page in the publication magazine
     */
    private Integer pmmbl_seite;

    /**
     * The issue of the publication magazine
     */
    private Integer publ_nr;

    /**
     * Shows where this is published
     */
    private PublicationTarget publicationTarget;

    /**
     * Creates this with all the data.
     * @param publ_pflicht Shows whether this must be published
     * @param publ_definitiv Shows whether this has been definitely published
     * @param publ_dat The publication date
     * @param pmmbl_seite The page in the publication magazine
     * @param publ_nr The issue of the publication magazine
     * @param publicationTarget Shows where this is published
     */
    public MockEvent( boolean publ_pflicht, boolean publ_definitiv, Date publ_dat, Integer pmmbl_seite, Integer publ_nr, PublicationTarget publicationTarget )
    {
      super( null );
      if ( publ_dat == null )
      {
        this.compareTime = 1999010100;
      }
      else
      {
        EreigIDat date = EreigIDat.min( publ_dat );
        this.compareTime = date.getCompareTime() - 200; // Two days before publication date
      }
      this.publ_pflicht = publ_pflicht;
      this.publ_definitiv = publ_definitiv;
      this.publ_dat = publ_dat;
      this.pmmbl_seite = pmmbl_seite;
      this.publ_nr = publ_nr;
      this.publicationTarget = publicationTarget;
    }

    // @see ch.ige.ipobjects.common.event.Event#isPubl_pflicht()
    @Override
    public boolean isPubl_pflicht()
    {
      return this.publ_pflicht;
    }

    // @see ch.ige.ipobjects.common.event.Event#isPubl_definitiv()
    @Override
    public boolean isPubl_definitiv()
    {
      return this.publ_definitiv;
    }

    // @see ch.ige.ipobjects.common.event.Event#getPubl_dat()
    @Override
    public Date getPubl_dat()
    {
      return this.publ_dat;
    }

    // @see ch.ige.ipobjects.common.event.Event#getPmmbl_jahr()
    @Override
    public Integer getPmmbl_jahr()
    {
      if ( this.publ_dat != null )
      {
        Calendar calendar = BAGISTime.createCalendar( this.publ_dat );
        return Integer.valueOf( calendar.get( Calendar.YEAR ) );
      }
      else
      {
        return null;
      }
    }

    // @see ch.ige.ipobjects.common.event.Event#getPmmbl_seite()
    @Override
    public Integer getPmmbl_seite()
    {
      return this.pmmbl_seite;
    }

    // @see ch.ige.ipobjects.common.event.Event#getPubl_nr()
    @Override
    public Integer getPubl_nr()
    {
      return this.publ_nr;
    }

    // @see ch.ige.ipobjects.common.event.Event#getCompareTime()
    @Override
    public int getCompareTime()
    {
      return this.compareTime;
    }

    // @see ch.ige.ipobjects.common.event.Event#getPublicationTarget()
    @Override
    public PublicationTarget getPublicationTarget()
    {
      return this.publicationTarget;
    }
  }
}
