package ch.ige.jipps.imageconv;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Map;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;

import ch.ige.jipps.constants.ImageTypeConstants;

import com.sun.media.jai.codec.ImageCodec;
import com.sun.media.jai.codec.MemoryCacheSeekableStream;

/**
 * Processes images: Converts them, saves them and extracts their metadata (height, width etc.).
 * 
 * <p>Copyright (c) 2004, Swiss Federal Institute of Intellectual Property</p>
 * @author Johannes Lang
 * @version $Id: ImageProcessor.java,v 1.28 2009/05/05 15:48:16 lewis_lee Exp $
 */
public class ImageProcessor
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( ImageProcessor.class );

  /**
   * This parameter controls whether images are processed or not.
   */
  public static final String PROCESSING_MODE = "ImageProcessor.processingMode";

  /**
   * Using {@link #PROCESSING_MODE}, this means that images are <strong>not</strong> processed.
   */
  public static final String NO_PROCESSING = "noProcessing";

  /**
   * Using {@link #PROCESSING_MODE}, this means that images <strong>must be</strong> processed.
   */
  public static final String PROCESS_IMAGES = "processImages";

  /**
   * Using {@link ch.ige.jipps.constants.PropertyConstants#IMAGE_CONVERTER_TOOL}, this means that JAI is used to
   * convert images.
   */
  public static String IMAGE_TOOL_JAI = "1";

  /**
   * Using {@link ch.ige.jipps.constants.PropertyConstants#IMAGE_CONVERTER_TOOL}, this means that ImageMagick is used
   * to convert images.
   */
  public static String IMAGE_TOOL_IMAGEMAGICK = "2";

  /**
   * In the <code>Map</code> created by {@link #writeImage(String, byte[], boolean)},
   * this is the key for the image type.
   * @see ImageTypeConstants#IMAGE_TYPE_JPG
   * @see ImageTypeConstants#IMAGE_TYPE_TIFF
   * @see ImageTypeConstants#IMAGE_TYPE_PNG
   */
  public static final String IMAGE_TYPE = "imageType";

  /**
   * In the <code>Map</code> created by {@link #writeImage(String, byte[], boolean)},
   * this is the key for the image width <strong>in centimeters</strong>.
   */
  public static final String IMAGE_WIDTH = "width";

  /**
   * In the <code>Map</code> created by {@link #writeImage(String, byte[], boolean)},
   * this is the key for the image height <strong>in centimeters</strong>.
   */
  public static final String IMAGE_HEIGHT = "height";

  /**
   * In the <code>Map</code> created by {@link #writeImage(String, byte[], boolean)},
   * this is the key for the image data (the <code>byte[]</code>).
   */
  public static final String IMAGE_DATA = "imageData";

  /**
   * The default image density in DPI.
   */
  protected static final int DEFAULT_DENSITY = 300;

  /**
   * This is shown whenever an image can't be processed.
   */
  private byte[] imageError;

  /**
   * The only instance of this.
   */
  private static ImageProcessor INSTANCE = new ImageProcessor();

  /**
   * This is private because this is a Singleton.
   */
  private ImageProcessor()
  {
    // Must never be called
  }

  /**
   * Gets the only instance of this.
   * @return The only instance
   */
  public static ImageProcessor getInstance()
  {
    return INSTANCE;
  }

  /**
   * Shows whether in general, images should be processed. For local tests for instance, one would not want the image
   * processing to work because of dependencies to either a local or a remote server installation.
   * 
   * @return <code>true</code> if images should be processed, otherwise <code>false</code>
   */
  public static boolean processImages()
  {
    boolean processImages = true;
    String processImageFlag = System.getProperty( PROCESSING_MODE );
    if ( processImageFlag != null )
    {
      if ( processImageFlag.equals( PROCESS_IMAGES ) )
      {
        processImages = true;
      }
      else if ( processImageFlag.equals( NO_PROCESSING ) )
      {
        processImages = false;
      }
    }
    return processImages;
  }

  /**
   * Writes an image into a file. 
   * @param pathWithoutFileType The path without the file type (like {@link ImageTypeConstants#IMAGE_TYPE_JPG})
   * @param image The image to write
   * @param replaceWithErrorImage If this is <code>true</code>, a dummy image containing the text
   *        <code><strong>IMAGE ERROR</strong></code> is taken whenever the real image can't be converted; if it's
   *        <code>false</code>, the corresponding Exception is thrown
   * @return The image properties for the image, see below for the keys
   * @see #IMAGE_TYPE
   * @see #IMAGE_WIDTH
   * @see #IMAGE_HEIGHT
   * @see #IMAGE_DATA
   * @throws IOException if the image can't be written
   */
  public Map<String, Object> writeImage( String pathWithoutFileType, byte[] image, boolean replaceWithErrorImage ) throws IOException
  {
    try
    {
      MemoryCacheSeekableStream memorySeekableStream = new MemoryCacheSeekableStream( new ByteArrayInputStream( image ) );
      String imageType = ImageCodec.getDecoderNames( memorySeekableStream )[0];
      if ( imageType.equals( ImageTypeConstants.IMAGE_TYPE_JPEG ) )
      {
        imageType = ImageTypeConstants.IMAGE_TYPE_JPG;
      }
      return writeImage( pathWithoutFileType, image, imageType );
    }
    catch ( RuntimeException e )
    {
      if ( replaceWithErrorImage )
      {
        return getErrorImage( pathWithoutFileType, e );
      }
      else
      {
        throw e;
      }
    }
    catch ( IOException e )
    {
      if ( replaceWithErrorImage )
      {
        return getErrorImage( pathWithoutFileType, e );
      }
      else
      {
        throw e;
      }
    }
  }

  /**
   * Writes an image depending on the image type.
   * 
   * @param pathWithoutFileType The path where the image is stored, excluding the file type
   * @param image The image's binary data
   * @param imageType The image's type, one of {@link ImageTypeConstants}; the default is {@link ImageTypeConstants#IMAGE_TYPE_TIFF}
   * @throws IOException if the image can't be written
   * @return The image's metadata, see {@link #writeImage(String, byte[], boolean)}
   */
  private Map<String, Object> writeImage( String pathWithoutFileType, byte[] image, String imageType ) throws IOException
  {
    ImageWriter imageWriter;

    if ( imageType.equalsIgnoreCase( ImageTypeConstants.IMAGE_TYPE_JPG ) )
    {
      imageWriter = new JPEGImageWriter();
    }
    else if ( imageType.equalsIgnoreCase( ImageTypeConstants.IMAGE_TYPE_PNG ) )
    {
      imageWriter = new PNGImageWriter();
    }
    else
    {
      if ( !imageType.equalsIgnoreCase( ImageTypeConstants.IMAGE_TYPE_TIFF ) )
      {
        LOG.warn( "Got unsupported image type '" + imageType + "' for file " + pathWithoutFileType + ", defaulting to TIFF" );
      }
      imageWriter = new TIFFImageWriter();
    }
    return imageWriter.processImage( pathWithoutFileType, image );
  }

  /**
   * Gets the image to be shown whenever a converting error occurs.
   * @param pathWithoutFileType The path without the file type (like {@link ImageTypeConstants#IMAGE_TYPE_JPG})
   * @return The 'error' image
   * @throws IOException if the file can't be read
   */
  private Map<String, Object> getErrorImage( String pathWithoutFileType, Exception e ) throws IOException
  {
    LOG.error( "Image '" + pathWithoutFileType + "' can't be created, defaulting to IMAGE ERROR", e );
    if ( this.imageError == null )
    {
      this.imageError = IOUtils.toByteArray( ImageProcessor.class.getResourceAsStream( "/res/image_error.jpg" ) );
    }
    return writeImage( pathWithoutFileType, this.imageError, ImageTypeConstants.IMAGE_TYPE_JPG );
  }

}
