package ch.ige.jipps.imageconv;

import java.awt.image.RenderedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Map;

import ch.ige.jipps.constants.FileConstants;
import ch.ige.jipps.constants.ImageTypeConstants;

import com.sun.media.jai.codec.ImageCodec;
import com.sun.media.jai.codec.ImageDecoderImpl;
import com.sun.media.jai.codec.MemoryCacheSeekableStream;
import com.sun.media.jai.codec.TIFFDecodeParam;
import com.sun.media.jai.codec.TIFFDirectory;
import com.sun.media.jai.codec.TIFFField;
import com.sun.media.jai.codecimpl.TIFFImageDecoder;

/**
 * Writes TIFF images.
 * @author Johannes Lang
 * @version $Id: TIFFImageWriter.java,v 1.19 2009/05/05 15:47:13 lewis_lee Exp $
 */
public class TIFFImageWriter extends ImageWriter
{
  /**
   * According to the TIFF specifiction, if {@link TIFFImageDecoder#TIFF_RESOLUTION_UNIT} returns this
   * value, the resolution is measured in inches. This is also the default.
   */
  private static final int TIFF_RESOLUTION_UNIT_INCH = 2;

  /**
   * According to the TIFF specifiction, if {@link TIFFImageDecoder#TIFF_RESOLUTION_UNIT} returns this
   * value, the resolution is measured in centimeters.
   */
  private static final int TIFF_RESOLUTION_UNIT_CM = 3;

  /**
   * Creates this.
   */
  public TIFFImageWriter()
  {
    super();
  }

  // @see ch.ige.jipps.imageconv.ImageWriter#processImage(java.lang.String, byte[])
  @Override
  public Map<String, Object> processImage( String pathWithoutFileType, byte[] image ) throws IOException
  {
    String destImageType = processImageData( image, pathWithoutFileType );
    Map<String, Object> map = getDimensions();
    map.put( ImageProcessor.IMAGE_TYPE, destImageType );
    return map;
  }

  /**
   * Processes the TIFF image.
   * @param image The image
   * @param pathWithoutFileType The path of the file without the ending (file extension)
   * @return The image type of the converted file, usually {@link ImageTypeConstants#IMAGE_TYPE_JPG}
   * @throws IOException if the image can't be processed
   */
  private String processImageData( byte[] image, String pathWithoutFileType ) throws IOException
  {
    MemoryCacheSeekableStream memSeekStream = new MemoryCacheSeekableStream( new ByteArrayInputStream( image ) );
    extractImageSize( memSeekStream );
    extractDensity( memSeekStream );
    memSeekStream = null;
    String sourceFilePath = pathWithoutFileType + FileConstants.EXTENSION_TIF;
    writeToFile( image, sourceFilePath );
    return callServlet( sourceFilePath, ImageTypeConstants.IMAGE_TYPE_TIFF );
  }

  /**
   * Extracts the image's height and width in pixels.
   * @param memSeekStream The memory cache seekable stream containing the image data
   * @throws IOException if the metadata can't be extracted
   */
  private void extractImageSize( MemoryCacheSeekableStream memSeekStream ) throws IOException
  {
    TIFFDecodeParam tiffparam = new TIFFDecodeParam();
    ImageDecoderImpl imDecoder = (ImageDecoderImpl) ImageCodec.createImageDecoder( ImageTypeConstants.IMAGE_TYPE_TIFF, memSeekStream, tiffparam );
    RenderedImage renderedImage = imDecoder.decodeAsRenderedImage();
    this.imageHeigthInPixel = renderedImage.getHeight();
    this.imageWidthInPixel = renderedImage.getWidth();
  }

  /**
   * Extracts the x and y resolution of the first TIFF image (index <code>0</code>) within the TIFF file.
   * @param memSeekStream The memory cache seekable stream containing the image data
   * @throws IOException if the metadata can't be extracted
   */
  private void extractDensity( MemoryCacheSeekableStream memSeekStream ) throws IOException
  {
    TIFFDirectory tiffDirectory = new TIFFDirectory( memSeekStream, 0 );

    TIFFField resolutionUnit = tiffDirectory.getField( TIFFImageDecoder.TIFF_RESOLUTION_UNIT );
    int encodedResolutionUnit = TIFF_RESOLUTION_UNIT_INCH;
    if ( resolutionUnit != null )
    {
      encodedResolutionUnit = resolutionUnit.getAsInt( 0 );
    }
    this.xDensity = getDensity( tiffDirectory.getField( TIFFImageDecoder.TIFF_X_RESOLUTION ), encodedResolutionUnit );
    this.yDensity = getDensity( tiffDirectory.getField( TIFFImageDecoder.TIFF_Y_RESOLUTION ), encodedResolutionUnit );
  }

  /**
   * Gets the density from the corresponding TIFF fiels.
   * @param densityField The TIFF field for the density (x or y)
   * @param resolutionUnit The resolution unit, either {@link #TIFF_RESOLUTION_UNIT_CM} or {@link #TIFF_RESOLUTION_UNIT_INCH}
   * @return The density in pixels per inch
   */
  private int getDensity( TIFFField densityField, int resolutionUnit )
  {
    try
    {
      long[] densityFields = densityField.getAsRational( 0 );
      int a = Long.valueOf( densityFields[0] ).intValue();
      int b = Long.valueOf( densityFields[1] ).intValue();
      if ( b != 0 )
      {
        int relation = ( a / b );
        switch( resolutionUnit )
        {
          case TIFF_RESOLUTION_UNIT_CM:
            // convert to pixels per inch
            return (int) ( relation * CONVERSION_DPCM_TO_DPI );
          case TIFF_RESOLUTION_UNIT_INCH:
            return relation;
          default:
            return relation;
        }
      }
      else
      {
        return 0;
      }
    }
    catch ( Exception ex )
    {
      return 0;
    }
  }

}
