package ch.ige.jipps.pdf;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.URIResolver;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.lang.StringUtils;
import org.apache.fop.apps.FOPException;
import org.apache.fop.apps.FOUserAgent;
import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FopFactory;
import org.apache.fop.apps.MimeConstants;
import org.apache.log4j.Logger;
import org.xml.sax.SAXException;

import ch.ige.jipps.configuration.PathConfiguration;
import ch.ige.jipps.constants.St36Constants;

/**
 * Wrapper-Class for FOP 0.93 Copyright (c) 2006, Swiss Federal Institute of Intellectual Property This class allows
 * generating a FOP area tree as well as a PDF.
 * @author Michael Künzli, Esther Bandschapp
 * @version $Id: PdfAreaTreeRenderer.java,v 1.11 2008/01/28 16:05:09 bandschapp Exp $
 */
public class PdfAreaTreeRenderer extends PdfRenderer
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( PdfAreaTreeRenderer.class );

  /**
   * The transformer factory
   */
  private TransformerFactory transformerFactory;

  /**
   * The config file directory
   */
  private String configFileUrl;

  /**
   * The hyphenation directory
   */
  private String hyphUrl;

  /**
   * The font directory
   */
  private String fontUrl;

  /**
   * The schema (base) directory
   */
  private String schemaDirectory;

  /**
   * The default constructor
   */
  public PdfAreaTreeRenderer()
  {
    super();
    try
    {
      this.transformerFactory = net.sf.saxon.TransformerFactoryImpl.class.newInstance();
    }
    catch ( InstantiationException e )
    {
      throw new RuntimeException( "Couldn't instatiate saxon TransformerFactory", e );
    }
    catch ( IllegalAccessException e )
    {
      throw new RuntimeException( "Couldn't instatiate saxon TransformerFactory", e );
    }
  }

  // @see ch.ige.jipps.pdf.PdfRenderer#setupFopFactory()
  @Override
  protected void setupFopFactory( FopFactory fopFactory )
  {
    super.setupFopFactory( fopFactory );
  }

  // @see ch.ige.jipps.pdf.PdfRenderer#setupFOUserAgent(org.apache.fop.apps.FOUserAgent)
  @Override
  public void setupFOUserAgent( FOUserAgent userAgent )
  {
    super.setupFOUserAgent( userAgent );

    String baseURL = createUrl( St36Constants.getSt36BaseDirectory() );
    userAgent.setBaseURL( baseURL );

    /**
     * set PDF/A mode to PDF/A-1b
     * @see http://www.pdf-tools.com/public/downloads/whitepapers/whitepaper-pdfa-de.pdf does not go in release 02/2007
     * yet.
     */
    // userAgent.getRendererOptions().put( "pdf-a-mode", "PDF/A-1b" );
  }

  /**
   * Updates the pdf renderer. This is needed because the config url etc. can vary between single patdocs which might be
   * using different schemas.
   * @param schemaId The schema ID
   */
  public void updatePDFRenderer( int schemaId )
  {
    this.setConfigFileUrl( St36Constants.getConfigFileUrl( schemaId ) );
    this.setFontUrl( St36Constants.getSt36FontDirectory( schemaId ) );
    this.setHyphUrl( St36Constants.getHyphUrl( schemaId ) );
    this.setSchemaDirectory( St36Constants.getSt36SchemaDirectory( schemaId ) );
    this.setupFopFactory( this.fopFactory );

    try
    {
      this.fopFactory.setUserConfig( new File( getConfigFileUrl() ) );
    }
    catch ( SAXException e )
    {
      throw new RuntimeException( "XML-related error while setting FOP user configuration", e );
    }
    catch ( IOException e )
    {
      throw new RuntimeException( "I/O-related error while setting FOP user configuration", e );
    }
    // set up hyhpenation base URL -> this is done via UriResolve
    LOG.debug( "setting hyph directory to " + getHyphUrl() );
    HyphUriResolver hyphUriResolver = new HyphUriResolver( getHyphUrl() );
    this.fopFactory.setURIResolver( hyphUriResolver ); // configure fopFactory

    // set up FOP font base dir
    LOG.debug( "setting font directory to " + getFontUrl() );
    this.fopFactory.setFontBaseURL( getFontUrl() );
  }

  /**
   * Renders a PDF by applying a XSL stylesheet to an XML document.
   * @param reportXml The XML document, all in one String
   * @param xslDirectory The directory where the XSL is stored
   * @param areaTreeUrl
   * @param xslUriName The URI name of the xsl
   * @param outputStream The stream of the rendered pdf
   * @throws TransformerException if the transformation fails
   * @throws FOPException If FOP cannot do the job
   * @throws SAXException If the content of the xml is faulty
   * @throws IOException if any file can't be accessed
   */
  public void renderPdf( String reportXml, String xslDirectory, String areaTreeUrl, String xslUriName, OutputStream outputStream ) throws TransformerException,
      FOPException, SAXException, IOException
  {
    generateFopOutput( new StringReader( reportXml ), xslDirectory, MimeConstants.MIME_PDF, areaTreeUrl, outputStream, xslUriName );
  }

  /**
   * Builds a FOP area tree by applying a XSL stylesheet to an XML document.
   * @param reportXml The XML document, as a character reader.
   * @param xslDirectory The directory where the XSL is stored
   * @param xslUriName The URI name of the xsl
   * @param outputStream The stream of the rendered pdf
   * @throws IOException if any file can't be accessed
   * @throws TransformerException if the transformation fails
   * @throws FOPException If FOP cannot do the job
   * @throws SAXException If the content of the xml is faulty
   */
  public void buildAreaTree( String reportXml, String xslDirectory, String xslUriName, OutputStream outputStream ) throws TransformerException, FOPException,
      IOException, SAXException
  {
    LOG.info( "start buildAreaTree" );
    generateFopOutput( new StringReader( reportXml ), xslDirectory, MimeConstants.MIME_FOP_AREA_TREE, null, outputStream, xslUriName );
  }

  /**
   * Generates generic FOP output.
   * @param reportXml The XML document, as a character reader.
   * @param xslDirectory The directory where the XSL is stored
   * @param mimeType Define the type of output to generate. <br>
   * MimeConstants.MIME_PDF: generate a PDF document<br>
   * MimeConstants.MIME_FOP_AREA_TREE: generate a area tree<br>
   * @param areaTreeUrl the URL for to the area tree (null if no area tree should be used)
   * @param outputStream The stream of the rendered pdf
   * @param xslUriName The URI name of the xsl
   * @throws TransformerException If the document cannot be transformed
   * @throws FOPException If FOP cannot do the job
   * @throws IOException If a file cannot be found
   * @throws SAXException If the content of the xml is faulty
   */
  private void generateFopOutput( Reader reportXml, String xslDirectory, String mimeType, String areaTreeUrl, OutputStream outputStream, String xslUriName )
      throws TransformerException, FOPException, IOException, SAXException
  {
    LOG.info( "start renderPdf" );

    FOUserAgent foUserAgent = this.fopFactory.newFOUserAgent();
    setupFOUserAgent( foUserAgent );

    try
    {
      // Construct fop with desired output format
      Fop fop = this.fopFactory.newFop( mimeType, foUserAgent, outputStream );

      // get transformer for passed XSLT
      String streamSourceUrl = createUrl( xslDirectory + xslUriName );
      Transformer transformer = this.transformerFactory.newTransformer( new StreamSource( streamSourceUrl ) );

      // Set the value of a <param> in the stylesheet
      transformer.setParameter( "bib_present", "yes" );
      // Set the directory where the standard images can be found, like Swiss flag etc.
      transformer.setParameter( "bibimg_path", createUrl( getSchemaDirectory() ) );
      // Set the directory where saved images from the document can be found
      transformer.setParameter( "imgpath", createUrl( St36Constants.getSt36BaseDirectory() ) );

      if ( mimeType.equals( MimeConstants.MIME_PDF ) )
      {
        transformer.setParameter( "onlybib_at_path", areaTreeUrl );
      }
      else if ( mimeType.equals( MimeConstants.MIME_FOP_AREA_TREE ) )
      {
        transformer.setParameter( "onlybibitem", "yes" );
      }
      else
      {
        throw new RuntimeException( "generateFopOutput(): Unsupported MimeType: " + mimeType );
      }
      LOG.debug( "set transformer parameter bib_present: " + transformer.getParameter( "bib_present" ) );
      LOG.debug( "set transformer parameter bibimg_path: " + transformer.getParameter( "bibimg_path" ) );
      LOG.debug( "set transformer parameter onlybib_at_path: " + transformer.getParameter( "onlybib_at_path" ) );
      LOG.debug( "set transformer parameter imgpath: " + transformer.getParameter( "imgpath" ) );
      LOG.debug( "set transformer parameter onlybibitem: " + transformer.getParameter( "onlybibitem" ) );

      // Setup input for XSLT transformation
      Source src = new StreamSource( reportXml );

      // Resulting SAX events (the generated FO) must be piped through to
      // FOP
      Result res = new SAXResult( fop.getDefaultHandler() );

      // Start XSLT transformation and FOP processing
      transformer.transform( src, res );
    }
    finally
    {
      outputStream.close();
    }
    if ( LOG.isDebugEnabled() )
    {
      LOG.debug( "renderPdf(): Successfully generated PDF." );
    }
  }

  /**
   * Gets the configFileUrl.
   * @return the configFileUrl
   */
  public String getConfigFileUrl()
  {
    return this.configFileUrl;
  }

  /**
   * Sets the configFileUrl.
   * @param configFileUrl The configFileUrl to set
   */
  public void setConfigFileUrl( String configFileUrl )
  {
    this.configFileUrl = configFileUrl;
  }

  /**
   * Gets the fontUrl.
   * @return the fontUrl
   */
  public String getFontUrl()
  {
    return this.fontUrl;
  }

  /**
   * Sets the fontUrl.
   * @param fontUrl The fontUrl to set
   */
  public void setFontUrl( String fontUrl )
  {
    this.fontUrl = createUrl( fontUrl );
  }

  /**
   * Gets the hyphUrl.
   * @return the hyphUrl
   */
  public String getHyphUrl()
  {
    return this.hyphUrl;
  }

  /**
   * Sets the hyphUrl.
   * @param hyphUrl The hyphUrl to set
   */
  public void setHyphUrl( String hyphUrl )
  {
    this.hyphUrl = createUrl( hyphUrl );
  }

  /**
   * Gets the schemaDirectory.
   * @return the schemaDirectory
   */
  public String getSchemaDirectory()
  {
    return this.schemaDirectory;
  }

  /**
   * Sets the schemaDirectory.
   * @param schemaDirectory The schemaDirectory to set
   */
  public void setSchemaDirectory( String schemaDirectory )
  {
    this.schemaDirectory = schemaDirectory;
  }

  /**
   * Creates an url string out of the given text.
   * @param file The path
   * @return The transformed path
   */
  public static String createUrl( String file )
  {
    return StringUtils.replace( ( PathConfiguration.URL_FILE_PREFIX + file ), "\\", "/" );
  }

  /**
   * Class is used to redirect hyphenation URLs.
   * <p>
   * Note:
   * <p>
   * This UriResolver assumes that hyphenation URLs start with "hyph:" this must be configured in the FOP configuration
   * file (fop_xconf.xml).
   * @author Michael Kuenzli
   * @version $Id: PdfAreaTreeRenderer.java,v 1.11 2008/01/28 16:05:09 bandschapp Exp $
   */
  public class HyphUriResolver implements URIResolver
  {
    /**
     * The hyphenation protocol the URL starts with
     */
    private static final String HYPH_URL_PROT = "hyph:";

    /**
     * The actual hyphenation base url
     */
    private String actualHyphBaseUrl = null;

    /**
     * The constructor
     * @param actualHyphBaseUrl The actual hyphenation base url
     */
    public HyphUriResolver( String actualHyphBaseUrl )
    {
      this.actualHyphBaseUrl = actualHyphBaseUrl;
    }

    // @see javax.xml.transform.URIResolver#resolve(java.lang.String, java.lang.String)
    @SuppressWarnings( "synthetic-access" )
    public Source resolve( String href, String base )
    {
      if ( LOG.isDebugEnabled() )
      {
        LOG.debug( "Base: " + base + " Href: " + href );
      }
      Source source = null;
      if ( base.equals( HYPH_URL_PROT ) )
      {
        LOG.info( "resolve(): " + base + href + " -> " + this.actualHyphBaseUrl + href );
        source = new StreamSource( this.actualHyphBaseUrl + href );
      }
      return source;
    }
  }
}
