package ch.ige.jipps.pdf;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import com.lowagie.text.Document;
import com.lowagie.text.pdf.PdfCopy;
import com.lowagie.text.pdf.PdfReader;
import com.lowagie.text.pdf.SimpleBookmark;

/**
 * Merges several PDF files into one.
 * @author Johannes Lang
 * @version $Id: PdfMerger.java,v 1.20 2008/01/21 13:52:56 lewis_lee Exp $
 */
public class PdfMerger
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( PdfMerger.class );

  /**
   * This must not be instatiated.
   */
  private PdfMerger()
  {
    // This must not be instatiated
  }

  /**
   * Merges several PDF files into one.
   * Taken from
   * <a href="http://itext.ugent.be/library/com/lowagie/examples/general/copystamp/Concatenate.java">Concatenate.java</a>.
   * @param pdfFiles A <code>List</code> PDF files (their binary content)
   * @return The merged PDF files or <code>null</code> if <code>pdfFiles</code> is empty
   */
  @SuppressWarnings( "unchecked" )
  public static byte[] mergePdfs( List<byte[]> pdfFiles )
  {
    if ( pdfFiles.isEmpty() )
    {
      return null;
    }
    if ( pdfFiles.size() == 1 )
    {
      return pdfFiles.get( 0 );
    }

    ByteArrayOutputStream output = new ByteArrayOutputStream();
    try
    {
      int pageOffset = 0;

      List master = new ArrayList();

      Document document = null;
      PdfCopy writer = null;

      for ( int fileIndex = 0; fileIndex < pdfFiles.size(); fileIndex++ )
      {
        // we create a reader for a certain document
        PdfReader reader = new PdfReader( pdfFiles.get( fileIndex ) );
        reader.consolidateNamedDestinations();
        // we retrieve the total number of pages
        int numberOfPages = reader.getNumberOfPages();

        List bookmarks = SimpleBookmark.getBookmark( reader );
        if ( bookmarks != null )
        {
          if ( pageOffset != 0 )
          {
            SimpleBookmark.shiftPageNumbers( bookmarks, pageOffset, null );
          }
          master.addAll( bookmarks );
        }

        pageOffset += numberOfPages;

        if ( fileIndex == 0 )
        {
          // step 1: creation of a document-object
          document = new Document( reader.getPageSizeWithRotation( 1 ) );
          // step 2: we create a writer that listens to the document
          writer = new PdfCopy( document, output );
          copyMetaData( reader, document );
          // step 3: we open the document
          document.open();
        }

        if ( writer != null )
        {
          // step 4: we add content
          for ( int pageNumber = 1; pageNumber <= numberOfPages; pageNumber++ )
          {
            writer.addPage( writer.getImportedPage( reader, pageNumber ) );
          }
          if ( reader.getAcroForm() != null )
          {
            writer.copyAcroForm( reader );
          }
        }
      }
      if ( writer != null && !master.isEmpty() )
      {
        writer.setOutlines( master );
      }
      // step 5: we close the document
      if ( document != null )
      {
        document.close();
      }
    }
    catch ( Exception e )
    {
      LOG.error( "Couldn't merge PDF documents", e );
      return null;
    }
    return output.toByteArray();
  }

  /**
   * Copies the PDF metadata (title, subject, creator) from <code>source</code> into <code>destination</code>
   * @param source The PDF reader getting the metadata
   * @param target The document receiving the metadata
   */
  private static void copyMetaData( PdfReader source, Document target )
  {
    LOG.debug( "Copying metadata for PDF" );
    String oldTitle = (String) source.getInfo().get( "Title" );
    String oldSubject = (String) source.getInfo().get( "Subject" );
    String oldCreator = (String) source.getInfo().get( "Creator" );
    target.addTitle( oldTitle != null ? oldTitle : "" );
    target.addSubject( oldSubject != null ? oldSubject : "" );
    target.addCreator( oldCreator != null ? oldCreator : "IGE PrintService" );
    target.addCreationDate();
  }
}
