package ch.ige.jipps.test;

import java.util.Calendar;

import junit.framework.TestCase;
import ch.ige.dbutils.dao.AbstractDAO;
import ch.ige.dbutils.dao.DAO;
import ch.ige.ipobjects.key.vo.IPKeyVO;
import ch.ige.jipps.AuszugAssembler;
import ch.ige.jipps.vo.AuszugVO;
import ch.ige.util.client.ClientLogInitalizer;
import ch.ige.util.date.DateConverter;
import ch.ige.util.date.DateFormatter;
import ch.ige.util.enums.BAGISLanguageDomain;
import ch.ige.util.enums.IPType;
import ch.ige.util.enums.report.ReportType;

/**
 * Tests the {@link ch.ige.jipps.AuszugAssembler} for certain special cases.
 * <p>
 * Copyright (c) 2006, Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * @author Eric Lewis
 * @version $Id: AuszugAssemblerTest.java,v 1.36 2008/04/15 08:18:45 bandschapp Exp $
 */
public class AuszugAssemblerTest extends TestCase
{
  /**
   * The class to test
   */
  private AuszugAssembler auszugAssembler;

  /**
   * The IP to test - a design.
   */
  private static final IPKeyVO IP_TO_TEST = IPKeyVO.create( 1100387284 );

  /**
   * The IP's application date
   */
  private static final String IP_APPLICATION_DATE = "08.01.1991";

  /**
   * Some user...
   */
  private static final int USER_HW = 3;

  static
  {
    ClientLogInitalizer.initializeLoggingForTests( true, true );
    AbstractDAO.setRunningInJ2EEServer( false );
    System.setProperty( DAO.STANDALONE_CONFIG_PROPERTY_NAME, "true" );
  }

  /**
   * @see junit.framework.TestCase#TestCase(String)
   * @param args
   */
  public AuszugAssemblerTest( String args )
  {
    super( args );
  }

  /**
   * Sets up this test.
   * @throws Exception if the setting up fails
   */
  @Override
  public void setUp() throws Exception
  {
    super.setUp();
    this.auszugAssembler = new AuszugAssembler();
  }

  /**
   * Tests what happens when the report date is around the IP's application date.
   * @throws Exception if the test fails
   */
  public void testReportDate() throws Exception
  {
    Calendar ipApplicationDate = DateConverter.stringToCalendar( IP_APPLICATION_DATE );

    // Same day as application
    AuszugVO report = this.auszugAssembler.loadReport( IP_TO_TEST, ReportType.REGISTER_REPORT_AFTER_REGISTRATION, BAGISLanguageDomain.DE, ipApplicationDate
        .getTime(), USER_HW, false );
    assertNotNull( "Checking report for date " + DateFormatter.shortFormat( ipApplicationDate.getTime() ), report );

    // One day after application
    ipApplicationDate.add( Calendar.DAY_OF_MONTH, 1 );
    report = this.auszugAssembler.loadReport( IP_TO_TEST, ReportType.REGISTER_REPORT_AFTER_REGISTRATION, BAGISLanguageDomain.DE, ipApplicationDate.getTime(),
        USER_HW, false );
    assertNotNull( "Checking report for date " + DateFormatter.shortFormat( ipApplicationDate.getTime() ), report );

    // One day before application
    ipApplicationDate.add( Calendar.DAY_OF_MONTH, -2 );
    try
    {
      report = this.auszugAssembler.loadReport( IP_TO_TEST, ReportType.REGISTER_REPORT_AFTER_REGISTRATION, BAGISLanguageDomain.DE, ipApplicationDate.getTime(),
          USER_HW, false );
      fail( "Checking report for date " + DateFormatter.shortFormat( ipApplicationDate.getTime() ) + " must fail" );
    }
    catch ( Exception ex )
    {
      checkApplicationDateAndReportDate( ex, IP_APPLICATION_DATE, DateFormatter.shortFormat( ipApplicationDate.getTime() ) );
    }
  }

  /**
   * Tests the combination of IP type and report type.
   * @throws Exception if the test fails
   */
  public void testIpTypeReportType() throws Exception
  {
    AuszugVO report = this.auszugAssembler.loadReport( IP_TO_TEST, ReportType.REGISTER_REPORT_AFTER_REGISTRATION, BAGISLanguageDomain.DE, DateConverter
        .stringToDate( IP_APPLICATION_DATE ), USER_HW, false );
    assertNotNull( "Checking report for legal IP type / report type combination", report );

    try
    {
      report = this.auszugAssembler.loadReport( IP_TO_TEST, ReportType.CURRENT_STATE, BAGISLanguageDomain.DE,
          DateConverter.stringToDate( IP_APPLICATION_DATE ), USER_HW, false );
      fail( "Illegal combination of IP type and report type must fail" );
    }
    catch ( Exception ex )
    {
      assertTrue( "Checking for IP type in Exception message: " + ex.getMessage(), ex.getCause().getMessage().indexOf( IPType.TRADEMARK_CH.toString() ) > -1 );
      assertTrue( "Checking for report type in Exception message: " + ex.getMessage(),
          ex.getCause().getMessage().indexOf( ReportType.CURRENT_STATE.toString() ) > -1 );
    }
  }

  /**
   * Tests getting a report in the past. This specifically tests for an <code>IndexOutOfBoundsException</code> that
   * happened when creating a report in the past.
   * @throws Exception if the test fails
   */
  public void testReportInThePast() throws Exception
  {
    IPKeyVO ip = IPKeyVO.create( 1200752134 );

    String date1 = "15.09.2002";
    AuszugVO report = this.auszugAssembler.loadReport( ip, ReportType.REGISTER_REPORT_AFTER_REGISTRATION, BAGISLanguageDomain.DE, DateConverter
        .stringToDate( date1 ), USER_HW, false );
    assertNotNull( "Checking for " + date1, report );

    String date2 = "15.09.2001";
    try
    {
      report = this.auszugAssembler.loadReport( ip, ReportType.REGISTER_REPORT_AFTER_REGISTRATION, BAGISLanguageDomain.DE, DateConverter.stringToDate( date2 ),
          USER_HW, false );
      fail( "Checking before application date must fail" );
    }
    catch ( Exception ex )
    {
      checkApplicationDateAndReportDate( ex, "19.07.2002", date2 );
    }
  }

  /**
   * Checks that the application and report dates are contained in the Exception's message.
   * @param ex The Exception to check for its message
   * @param applicationDate The formatted application date
   * @param reportDate The formatted report date
   */
  private void checkApplicationDateAndReportDate( Exception ex, String applicationDate, String reportDate )
  {
    assertTrue( "Checking for application date in Exception message: " + ex.getMessage(), ex.getCause().getMessage().indexOf( applicationDate ) > -1 );
    assertTrue( "Checking for report date in Exception message: " + ex.getMessage(), ex.getCause().getMessage().indexOf( reportDate ) > -1 );
  }
}
