package ch.ige.jipps.test;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;

import junit.framework.TestCase;

import org.apache.commons.dbutils.handlers.ColumnListHandler;
import org.apache.log4j.Logger;

import ch.ige.dbutils.dao.AbstractDAO;
import ch.ige.dbutils.dao.DAO;
import ch.ige.dbutils.sql.SQL;
import ch.ige.dbutils.sql.SQL.Statement;
import ch.ige.jipps.PatentSpecificationControl;
import ch.ige.jipps.configuration.PathConfiguration;
import ch.ige.jipps.constants.FileConstants;
import ch.ige.jipps.imageconv.ImageProcessor;
import ch.ige.jipps.vo.AnswerVO;
import ch.ige.jipps.vo.PrintJobVO;
import ch.ige.util.client.ClientLogInitalizer;
import ch.ige.util.enums.IPType;
import ch.ige.util.enums.infrastructure.DBRealm;

/**
 * Tests {@link ch.ige.jipps.PatentSpecificationControl}.
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   bandschapp
 * @version  $Id: PatentSpecificationControlTest.java,v 1.21 2008/01/29 16:36:25 lewis_lee Exp $
 */
public class PatentSpecificationControlTest extends TestCase
{

  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( PatentSpecificationControlTest.class );

  /**
   * The unit to test 
   */
  private PatentSpecificationControl patentSpecificationControl = new PatentSpecificationControl();

  /**
   * An employee number
   */
  private static final int MA_NO = 3;

  /**
   * The form type to test
   */
  private static final String FORM_TYPE = PrintJobVO.PATENT_SPECIFICATION;

  /**
   * The IP type to test
   */
  private static final int IP_TYPE = IPType.PATENT_CH.toDB();

  /**
   * The application which sends the the printjob 
   */
  private static final String SENDER_APPLICATION = "Eclipse";

  /**
   * The user who sends the printjob
   */
  private static final String SENDER_USER = "JUnitTest";

  /**
   * The language to test, as a code for the <code>PrintJobVO</code>
   */
  private static final int LANGUAGE = 1;

  /**
   * The printer where to send the <code>PrintJobVO</code> to.
   * Note: It won't actually be printed.
   */
  private static final int PRINTER = 16;

  /**
   * The patDoc to be tested (PATDOC_ID)
   */
  private List<BigDecimal> patDocIds = new ArrayList<BigDecimal>();

  /**
   * Contains patent documents that exist in the DB.
   * This overrides the selection from the DB below.
   */
  private static final int[] KNOWN_PAT_DOC_IDS = null;

  static
  {
    ClientLogInitalizer.initializeLoggingForTests( true, true );
    AbstractDAO.setRunningInJ2EEServer( false );
    System.setProperty( DAO.STANDALONE_CONFIG_PROPERTY_NAME, "true" );
    // By default, don't process images, since they require a call to the image processing program/servlet
    System.setProperty( ImageProcessor.PROCESSING_MODE, ImageProcessor.NO_PROCESSING );
  }

  /**
   * The constructor
   * @param name The name of the test
   */
  public PatentSpecificationControlTest( String name )
  {
    super( name );
  }

  // @see junit.framework.TestCase#setUp()
  @SuppressWarnings( "unchecked" )
  @Override
  protected void setUp() throws Exception
  {
    super.setUp();
    PathConfiguration.setPathsForLocalTest();
    if ( KNOWN_PAT_DOC_IDS != null )
    {
      for ( int patDocId: KNOWN_PAT_DOC_IDS )
      {
        this.patDocIds.add( BigDecimal.valueOf( patDocId ) );
      }
    }
    else
    {
      try
      {
        Statement<List<BigDecimal>> SELECT = SQL.<List<BigDecimal>>select( "SELECT patdoc_id FROM patdoc WHERE rownum <= 10" );
        this.patDocIds = (List<BigDecimal>) AbstractDAO.getDirectQueryRunner( DBRealm.BAGIS ).query( SELECT, new ColumnListHandler( 1 ) );
        this.patDocIds.remove( 0 );
      }
      catch ( SQLException e )
      {
        fail( "Could not fetch Patdoc-ID's from database" );
      }
    }
  }

  /**
   * Test method for {@link ch.ige.jipps.PatentSpecificationControl#makePrint(java.util.List)}.
   */
  public void testMakePrint()
  {
    List<AnswerVO> answers = this.patentSpecificationControl.makePrint( createPrintJobs() );

    for ( AnswerVO answer: answers )
    {
      if ( !answer.printStatus.equals( AnswerVO.STATUS_OK ) )
      {
        fail( "The print job failed to print: " + answer );
      }
    }
  }

  /**
   * Test method for {@link ch.ige.jipps.PatentSpecificationControl#makePdf(java.util.List)}.
   */
  public void testMakePdf()
  {
    File pdfFile;
    try
    {
      byte[] pdfContent = this.patentSpecificationControl.makePdf( createPrintJobs() );

      assertNotNull( "Checking a PDF was generated", pdfContent );
      assertTrue( "Checking that PatDoc PDF file size is not 0", pdfContent.length > 0 );
      pdfFile = File.createTempFile( "patdoc_test", FileConstants.EXTENSION_PDF );
      FileOutputStream out = new FileOutputStream( pdfFile.getPath() );
      out.write( pdfContent );
      out.flush();
      out.close();
      System.out.print( "PDF written to " + pdfFile );
    }
    catch ( FileNotFoundException e )
    {
      LOG.error( "Error FileNotFoundException: " + e.getMessage(), e );
      fail( e.getMessage() );
    }
    catch ( ParseException e )
    {
      LOG.error( "Error ParseException: " + e.getMessage(), e );
      fail( e.getMessage() );
    }
    catch ( IOException e )
    {
      LOG.error( "Error IOException: " + e.getMessage(), e );
      fail( e.getMessage() );
    }
    catch ( Exception e )
    {
      LOG.error( "Error Exception: " + e.getMessage(), e );
      fail( e.getMessage() );
    }
  }

  /**
   * Creates the print jobs to execute.
   * @return The print jobs
   */
  private List<PrintJobVO> createPrintJobs()
  {
    List<PrintJobVO> result = new ArrayList<PrintJobVO>();
    for ( BigDecimal patDocId: this.patDocIds )
    {
      result.add( createPrintJob( patDocId.intValue() ) );
    }

    if ( result.isEmpty() )
    {
      fail( "Print jobs are empty, which is not allowed!" );
    }

    return result;
  }

  /**
   * Creates one print job with the given patdoc ID
   * @param patdocId The patent document ID
   * @return The <code>PrintJobVO</code> for that patdoc ID
   */
  private PrintJobVO createPrintJob( int patdocId )
  {
    PrintJobVO printJob = new PrintJobVO();
    printJob.formTyp = FORM_TYPE;
    printJob.schutitTyp = IP_TYPE;
    printJob.reportId = patdocId;
    printJob.maNr = MA_NO;
    printJob.senderApp = SENDER_APPLICATION;
    printJob.senderUser = SENDER_USER;
    printJob.sprache = LANGUAGE;
    printJob.druckerId = PRINTER;
    return printJob;
  }
}
