package ch.ige.jipps.test;

import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import junit.framework.TestCase;
import ch.ige.dbutils.dao.AbstractDAO;
import ch.ige.ipobjects.ServicesFactory;
import ch.ige.ipobjects.time.vo.EreigIDat;
import ch.ige.jipps.LetterControl;
import ch.ige.jipps.configuration.PathConfiguration;
import ch.ige.jipps.vo.AnswerVO;
import ch.ige.jipps.vo.PrintJobVO;
import ch.ige.util.constants.InvoiceConstants;
import ch.ige.util.constants.PrintExecutionConstants;
import ch.ige.util.constants.PrintStatusConstants;
import ch.ige.util.constants.PublicationStateConstants;
import ch.ige.util.date.DateConverter;
import ch.ige.util.date.DateFormatter;
import ch.ige.util.enums.infrastructure.DBRealm;
import ch.ige.util.enums.printer.ReportPaper;

/**
 * Tests the printing of publication confirmations in {@link ch.ige.jipps.LetterControl}.
 * <p>
 * Copyright (c) 2007, Swiss Federal Institute of Intellectual Property
 * </p>
 * @author Eric Lewis
 * @version $Id: PublicationConfirmationTest.java,v 1.13 2009/01/12 10:20:07 lewis_lee Exp $
 */
public class PublicationConfirmationTest extends TestCase
{
  static
  {
    ServicesFactory.initStandaloneClient();
    PathConfiguration.setPathsForLocalTest();
  }

  /**
   * The unit under test
   */
  private TestLetterControl letterControl = new TestLetterControl();

  /**
   * The IDs of the trademarks to test with. Note: This is only a subset of what is set up in the SQL statements during
   * the test, because these are the only ones that were created by a user with a JIPPS printer.
   */
  protected static final int[] TRADEMARKS = { 1200092665, 1200092665 };

  /**
   * The dates of the events to publish.
   */
  protected static final int[] EVENT_DATES = { 2007011510, 2007020510 };

  /**
   * The publication dates of the events.
   */
  protected static final String[] PUBLICATION_DATES = { "16.01.2007", "06.02.2007" };

  /**
   * The IDs of the invoices to test with.
   */
  private static final int[] INVOICES = { 6256291, 6276014 };

  /**
   * The number of the user that really has a JIPPS printer (Eric Lewis).
   * The invoices are old and must be forced to belong to a user who has a valid JIPPS printer.
   */
  private static final int USER_NUMBER = 992;

  /**
   * The only constructor
   * @param name The test's name
   */
  public PublicationConfirmationTest( String name )
  {
    super( name );
  }

  /**
   * Sets up the test data in the database.
   * @throws Exception if the updates fail
   */
  protected void setupDatabase() throws Exception
  {
    Connection connection = AbstractDAO.getDirectConnection( DBRealm.BAGIS );
    connection.setAutoCommit( true );
    try
    {
      for ( int i = 0; i < TRADEMARKS.length; i++ )
      {
        PreparedStatement invoiceLineStatement = connection
            .prepareStatement( "UPDATE bfaktpos SET schutit_key = ?, ereig_idat = ?, version = version + 1 WHERE bfakt_id = ?" );
        invoiceLineStatement.setInt( 1, TRADEMARKS[i] );
        invoiceLineStatement.setInt( 2, EVENT_DATES[i] );
        invoiceLineStatement.setInt( 3, INVOICES[i] );
        int updatedRows = invoiceLineStatement.executeUpdate();
        assertEquals( "Checking that the invoice line update worked", 1, updatedRows );

        PreparedStatement invoiceStatement = connection
            .prepareStatement( "UPDATE bfaktur SET formart_cd = ?, status_druck = ?, druck_art = ?, ma_nr = ?, version = version + 1 WHERE bfakt_id = ?" );
        invoiceStatement.setString( 1, InvoiceConstants.FORM_SWISSREG_PUBLICATION_CONFIRMATION.toString() );
        invoiceStatement.setString( 2, PrintStatusConstants.READY_FOR_PRINTING_STRING );
        invoiceStatement.setString( 3, PrintExecutionConstants.IMMEDIATE_PRINTING );
        invoiceStatement.setInt( 4, USER_NUMBER );
        invoiceStatement.setInt( 5, INVOICES[i] );
        updatedRows = invoiceStatement.executeUpdate();
        assertEquals( "Checking that the invoice update worked", 1, updatedRows );

        PreparedStatement eventStatement = connection
            .prepareStatement( "UPDATE ereignis SET publ_key = null, publ_dat = null, publ_stat = ?, version = version + 1 "
                + "WHERE (Schutit_key, ereig_idat) IN (SELECT schutit_key, ereig_idat FROM bfaktpos WHERE bfakt_id = ?)" );
        eventStatement.setInt( 1, PublicationStateConstants.READY_FOR_SYNCHRONISATION );
        eventStatement.setInt( 2, INVOICES[i] );
        updatedRows = eventStatement.executeUpdate();
        assertEquals( "Checking that the event update worked", 1, updatedRows );

      }
    }
    catch ( SQLException e )
    {
      e.printStackTrace();
      fail( "Couldn't set up database: " + e.getMessage() );
    }
    finally
    {
      connection.close();
    }
  }

  /**
   * Test method for {@link ch.ige.jipps.LetterControl#printPublicationConfirmation(Integer, Map)}. Note that this
   * doesn't really test publication confirmations, since they are very hard to find in the DB (at least when this test
   * was written). It just checks for some other letter.
   * @throws Exception if the test fails
   */
  public void testPrintPublicationConfirmation() throws Exception
  {
    setupDatabase();

    Map<EreigIDat, Date> trademarkEventDates = createEventDates();
    for ( int i = 0; i < TRADEMARKS.length; i++ )
    {
      try
      {
        int trademarkKey = TRADEMARKS[i];
        Map<EreigIDat, Date> currentEventDates = new HashMap<EreigIDat, Date>();
        EreigIDat currentEventDate = EreigIDat.from( EVENT_DATES[i] );
        currentEventDates.put( currentEventDate, trademarkEventDates.get( currentEventDate ) );
        List<PrintJobVO> printJobs = this.letterControl.createPublicationConfirmationPrintJobs( trademarkKey, currentEventDates );

        assertEquals( "Checking number of print jobs (no print job for registration event)", 1, printJobs.size() );
        PrintJobVO printJob = printJobs.get( 0 );
        assertTrue( "Checking batch ID", printJob.batchId > 0 );
        assertEquals( "Checking form type", PrintJobVO.LETTER, printJob.formTyp );
        assertTrue( "Checking printer ID", printJob.druckerId > 0 );
        assertEquals( "Checking report ID", INVOICES[i], printJob.reportId );
        assertEquals( "Checking paper mode", ReportPaper.PAPER_WITH_IPI_LOGO_INCLUDING_COPIES.toDB().intValue(), printJob.paperMode );
        assertEquals( "Checking number of copies", 1, printJob.anzKopien );
        assertEquals( "Checking print status", PrintStatusConstants.READY_FOR_PRINTING, printJob.druckStatus );
        assertEquals( "Checking language", PrintJobVO.USE_IP_LANGUAGE, printJob.sprache );
        assertEquals( "Checking cover", false, printJob.coverCd );
        assertEquals( "Checking sender application", PrintJobVO.SYSTEM, printJob.senderApp );
        assertEquals( "Checking sender user", PrintJobVO.SYSTEM, printJob.senderUser );
        assertEquals( "Checking publication date", PUBLICATION_DATES[i], DateFormatter.shortFormat( printJob.erstDatum, Locale.GERMAN ) );
      }
      catch ( Exception e )
      {
        fail( "Failed for test case " + ( i + 1 ) + " with trademark " + TRADEMARKS[i] + ": " + e.getMessage() );
      }
    }
  }

  /**
   * Creates the event dates from {@link #EVENT_DATES} and {@link #PUBLICATION_DATES}.
   * @return The event dates to use
   * @throws ParseException if the date is invalid
   */
  private static Map<EreigIDat, Date> createEventDates() throws ParseException
  {
    Map<EreigIDat, Date> newEventDates = new HashMap<EreigIDat, Date>();
    for ( int i = 0; i < EVENT_DATES.length; i++ )
    {
      newEventDates.put( EreigIDat.from( EVENT_DATES[i] ), DateConverter.stringToSqlDate( PUBLICATION_DATES[i] ) );
    }
    return newEventDates;
  }

  /**
   * Test method for {@link ch.ige.jipps.LetterControl#printPublicationConfirmation(Integer, Map)}., checking faults.
   * @throws Exception if the test fails
   */
  public void testPrintPublicationConfirmationWithFaults() throws Exception
  {
    Map<EreigIDat, Date> eventDates = new HashMap<EreigIDat, Date>();
    eventDates.put( EreigIDat.from( 2007030322 ), DateConverter.stringToSqlDate( "04.03.2007" ) );
    try
    {
      this.letterControl.printPublicationConfirmation( null, eventDates );
      fail( "Null IP key must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      this.letterControl.printPublicationConfirmation( null, eventDates );
      fail( "IP key with null value must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      this.letterControl.printPublicationConfirmation( -1, eventDates );
      fail( "IP key with negative value must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      this.letterControl.printPublicationConfirmation( TRADEMARKS[0], null );
      fail( "IP key with negative value must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    // Use the implementation directly
    List<AnswerVO> answers = new LetterControl().printPublicationConfirmation( TRADEMARKS[0], new HashMap<EreigIDat, Date>() );
    assertEquals( "Checking number of answers", 0, answers.size() );
  }

  /**
   * Used only to have access to protected methods.
   */
  private static class TestLetterControl extends LetterControl
  {
    /**
     * Creates this for testing
     */
    public TestLetterControl()
    {
      super();
    }

    // @see ch.ige.jipps.LetterControl#createPublicationConfirmationPrintJobs(java.lang.Integer, java.util.Map)
    @Override
    public List<PrintJobVO> createPublicationConfirmationPrintJobs( Integer ipKey, Map<EreigIDat, Date> eventDates ) throws SQLException
    {
      return super.createPublicationConfirmationPrintJobs( ipKey, eventDates );
    }
  }
}
