package ch.ige.jipps.text;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import ch.ige.util.constants.HtmlConstants;
import ch.ige.util.server.replace.P2d2;

/**
 * Conversion utilities for transforming XML data.<br>
 * This is mainly used for escaping XML data for XSL transformation.<br>
 * Then there are different kinds of new line formatting in the texts. They may be changed during synchronization. For
 * instance, addresses may be compacted from several lines to one String, using <code>\n</code>. These newline
 * characters however can't be used by FOP, so they have to be transformed to <code>&lt;br/&gt;</code> tags.<br>
 * Another case is splitting one address back into several lines in the value object.<br>
 * Finally, special tags are recognized at formatting time, like <code>&lt;br/&gt;</code> or <code>&lt;p/&gt;</code>
 * for line control or <code></code> for printing text in bold characters.
 * <p>
 * Copyright (c) 2003, Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * 
 * @author Priska Steiner
 * @author Eric Lewis
 * @version $Id: ConvertText.java,v 1.16 2008/07/21 13:37:07 lewis_lee Exp $
 */
public class ConvertText
{
  /**
   * The 'less than' character.
   */
  private static final String LT = "<";

  /**
   * The 'greater than' character.
   */
  private static final String GT = ">";

  /**
   * The 'less than' character, escaped for XML/HTML.
   */
  public static final String ESCAPED_LT = "&lt;";

  /**
   * The 'less than' character, escaped twice for XML/HTML.
   */
  private static final String DOUBLE_ESCAPED_LT = "&amp;lt;";

  /**
   * The 'greater than' character, escaped for XML/HTML.
   */
  public static final String ESCAPED_GT = "&gt;";

  /**
   * The 'greater than' character, escaped twice for XML/HTML.
   */
  private static final String DOUBLE_ESCAPED_GT = "&amp;gt;";

  /**
   * The 'less than' character, escaped for content within XML/HTML.
   */
  private static final String ESCAPED_ESCAPED_LT = "%%\\$\\$";

  /**
   * The 'greater than' character, escaped for content within XML/HTML.
   */
  private static final String ESCAPED_ESCAPED_GT = "%%\\*\\$";

  /**
   * The ampersand character, escaped for XML/HTML.
   */
  private static final String ESCAPED_AMPERSAND = "&amp;";

  /**
   * The ampersand character, escaped twice for XML/HTML.
   */
  public static final String DOUBLE_ESCAPED_AMPERSAND = "&amp;amp;";

  /**
   * The corresponding end of {@link P2d2#TAG_PREFIX}.
   */
  public static final String P2D2_TAG_SUFFIX = "%%*$";

  /**
   * These are tags that don't have any variations.
   */
  private static final String[] STANDALONE_HTML_TAGS = { "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>", "<?xml version=\"1.0\" encoding=\"UTF-8\"?>", };

  /**
   * These are HTML tags used in the document. They always exist as opening, closing and empty tag.
   */
  private static final String[] HTML_FULL_TAGS = { HtmlConstants.BREAK, HtmlConstants.PARAGRAPH };

  /**
   * These are HTML tags used in the document. They exist as opening and closing tag.
   */
  private static final String[] HTML_OPENING_AND_CLOSING_TAGS = { HtmlConstants.BOLD, HtmlConstants.HEADER_1, HtmlConstants.STRONG };

  /**
   * The XML tags used in the document. They always exist as starting, ending and closing tag.
   */
  private static final String[] XML_DOCUMENT_TAGS = { "<spaltentitel>", "<kriterium>", "<kritname>", "<kritwert>", "<listeEl>", "<spalten>", "<groesse>",
      "<header>", "<anzahl>", "<spalte>", "<zeilen>", "<titel>", "<zelle>", "<zeile>", "<zeit>", "<text>", "<zusammenfassung>", "<stammblattEl>",
      "<header_title>", "<header_date>", "<header_art_nr>", "<header_searchitem>", "<header_description>", "<header_nr>", "<header_printsystem>", "<art>",
      "<art_title>", "<art_data>", "<art_key>", "<art_value>", "<artprice>", "<artprice_title>", "<artprice_data>", "<artprice_description>",
      "<artprice_date>", "<artprice_price>", "<arttext>", "<arttext_title>", "<arttext_data>", "<arttext_key>", "<arttext_language>", "<arttext_value>",
      "<artfollow>", "<artfollow_title>", "<artfollow_data>", "<artfollow_key>", "<artfollow_value>", "<karteikarteEl>", "<karteikarte_elements>",
      "<karteikarte_content>", "<karteikarte_content_long>", "<formularEl>", "<art_numbers>", "<art_no_title>", "<art_no>", "<formtxt>", "<formtxt_title>",
      "<formtxt_data>", "<formtxt_nr>", "<formtxt_art>", "<formtxt_detail>", "<formtxt_language>", "<formtxt_txt>", "<igekopf>", "<igekopfDe>", "<igekopfFr>",
      "<igekopfIt>", "<igekopfEn>", "<igekopfAdr>", "<igekopfKto>" };

  /**
   * Used for escaping HTML.
   */
  protected static final List<String> HTML_TAGS;

  /**
   * The same as {@link #HTML_TAGS}, but without {@link #STANDALONE_HTML_TAGS}.
   */
  protected static final List<String> PURE_HTML_TAGS;

  /**
   * Used for escaping XML tags in lists.
   */
  protected static final List<String> LIST_TAGS;

  static
  {
    HTML_TAGS = new ArrayList<String>();
    addTags( HTML_TAGS, STANDALONE_HTML_TAGS, false, false );
    addTags( HTML_TAGS, HTML_OPENING_AND_CLOSING_TAGS, true, false );
    addTags( HTML_TAGS, HTML_FULL_TAGS, true, true );

    PURE_HTML_TAGS = new ArrayList<String>();
    addTags( PURE_HTML_TAGS, HTML_OPENING_AND_CLOSING_TAGS, true, false );
    addTags( PURE_HTML_TAGS, HTML_FULL_TAGS, true, true );

    LIST_TAGS = new ArrayList<String>();
    addTags( LIST_TAGS, STANDALONE_HTML_TAGS, false, false );
    addTags( LIST_TAGS, HTML_OPENING_AND_CLOSING_TAGS, true, false );
    addTags( LIST_TAGS, HTML_FULL_TAGS, true, true );
    addTags( LIST_TAGS, XML_DOCUMENT_TAGS, true, true );
  }

  /**
   * The hidden constructor: This must only be used in a static way.
   */
  private ConvertText()
  {
    // This must only be used in a static way
  }

  /**
   * Gets all the tags that are not escaped by {link #replaceHtmlTags(String)}.
   * 
   * @return The HTML tags that are not escaped
   */
  public static List<String> getHtmlTags()
  {
    return HTML_TAGS;
  }

  /**
   * Gets all the tags that are not escaped by {link #replaceListTags(String)}.
   * 
   * @return The tags that are not escaped
   */
  public static List<String> getListTags()
  {
    return LIST_TAGS;
  }

  /**
   * Adds a number of tags to an existing list of tags.
   * 
   * @param tagList The existing list of tags
   * @param tags The tags to add, every String being a tag like <code>&lt;p&gt;</code>. These tags are always added.
   * @param addClosingTag If this is <code>true</code>, the closing tag like <code>&lt;/p&gt;</code> is also added
   * @param addEmptyTag If this is <code>true</code>, the empty tag like <code>&lt;p/&gt;</code> is also added
   */
  private static void addTags( List<String> tagList, String[] tags, boolean addClosingTag, boolean addEmptyTag )
  {
    for ( int i = 0; i < tags.length; i++ )
    {
      String tag = tags[i];
      tagList.add( tag );
      if ( addClosingTag )
      {
        tagList.add( "</" + tag.substring( 1 ) );
      }
      if ( addEmptyTag )
      {
        tagList.add( StringUtils.removeEnd( tag, ">" ) + "/>" );
      }
    }
  }

  /**
   * Replaces existing <code>&lt;</code> and <code>&gt;</code> with {@link P2d2#TAG_PREFIX} and {@link #P2D2_TAG_SUFFIX},
   * except for predefined HTML tags.
   * 
   * @param xml The XML document to change
   * @return The XML document with escaped tags or <code>null</code> if <code>xml</code> is <code>null</code>
   */
  public static String replaceHtmlTags( String xml )
  {
    if ( xml == null )
    {
      return null;
    }
    else
    {
      return replaceTags( xml, HTML_TAGS );
    }
  }

  /**
   * Escapes all HTML tags in the XML, for instance
   * <code>&lt;strong&gt;</code> to <code>&amp;lt;strong&amp;gt;</code>.
   * @param xml The XML document to change
   * @return The XML document with escaped HTML tags or <code>null</code> if <code>xml</code> is <code>null</code>
   */
  public static String escapeHtmlTags( String xml )
  {
    if ( xml == null )
    {
      return null;
    }
    else
    {
      String convertedXml = xml;
      for ( String htmlTag: PURE_HTML_TAGS )
      {
        String escapedHtmlTag = htmlTag.replace( LT, ESCAPED_LT ).replace( GT, ESCAPED_GT );
        convertedXml = convertedXml.replace( htmlTag, escapedHtmlTag );
      }
      return convertedXml;
    }
  }

  /**
   * Replaces existing <code>&lt;</code> and <code>&gt;</code> with {@link P2d2#TAG_PREFIX} and {@link #P2D2_TAG_SUFFIX},
   * except for predefined list tags.
   * 
   * @param xml The XML document to change
   * @return The XML document with escaped tags or <code>null</code> if <code>xml</code> is <code>null</code>
   */
  public static String replaceListTags( String xml )
  {
    if ( xml == null )
    {
      return null;
    }
    else
    {
      return replaceTags( xml, LIST_TAGS );
    }
  }

  /**
   * Escapes tags in the XML stream using <code>&amp;lt;</code> and <code>&amp;gt;</code>.
   * 
   * @param xml The XML stream which is escaped
   * @param tags The tags that aren't escaped
   * @return The XML stream with escaped tags
   */
  private static String replaceTags( final String xml, final List<String> tags )
  {
    final String LtReplacement = "=çç=";
    final String GtReplacement = "ç==ç";

    String escapedXml = xml;
    // Prevent all tags to be replaced by escaping them
    for ( String tag: tags )
    {
      String escapedTag = tag.replace( LT, LtReplacement ).replace( GT, GtReplacement );
      escapedXml = StringUtils.replace( escapedXml, tag, escapedTag );
    }

    // Replace all '<' and '>' by their P2d2 counterpart
    escapedXml = StringUtils.replace( escapedXml, LT, P2d2.TAG_PREFIX );
    escapedXml = StringUtils.replace( escapedXml, GT, P2D2_TAG_SUFFIX );

    // And unescape all tags
    for ( String tag: tags )
    {
      String escapedTag = tag.replace( LT, LtReplacement ).replace( GT, GtReplacement );
      escapedXml = StringUtils.replace( escapedXml, escapedTag, tag );
    }

    return escapedXml;
  }

  /**
   * Converts all newline characters in <code>text</code> to HTML breaks (<code>&lt;br&gt;</code>).
   * 
   * @param text The text to look at
   * @return The same text with replaced newline characters
   */
  public static String convertNewlineToBreak( String text )
  {
    return text.replaceAll( "\n", HtmlConstants.BREAK );
  }

  /**
   * Splits the text by newline character. This means that every line is an element of the resulting <code>List</code>.
   * 
   * @param text The text to split
   * @return The <code>List</code> of lines, which may be empty if <code>text</code> is empty
   */
  public static List<String> splitLines( String text )
  {
    if ( StringUtils.isEmpty( text ) )
    {
      return Collections.emptyList();
    }
    else
    {
      return Arrays.asList( StringUtils.split( text, '\n' ) );
    }
  }

  /**
   * Splits the text by comma. For instance <code>one,two, three</code> is split into a <code>List</code> containing
   * <ul>
   * <li><code>one</code></li>
   * <li><code>two</code></li>
   * <li><code>three</code></li>
   * </ul>
   * 
   * @param text The text to split
   * @return The <code>List</code> of items, which may be empty if <code>text</code> is empty
   */
  public static List<String> splitCommaSeparatedList( String text )
  {
    if ( StringUtils.isEmpty( text ) )
    {
      return Collections.emptyList();
    }
    else
    {
      return Arrays.asList( text.split( ",\\s*" ) );
    }
  }

  /**
   * Prepares the XML String for XSL formatting.
   * <ul>
   * <li>Replaces breaks by calling {@link #convertBreaksToParagraph(String)}</li>
   * <li>Converts <code>&amp;lt;</code> to <code>&lt;</code></li>
   * <li>Converts <code>&amp;gt;</code> to <code>&gt;</code></li>
   * <li>Converts <code>&lt;br&gt;</code> to <code>&lt;br/&gt;</code></li>
   * <li>Converts <code>%%$$</code> to <code>&amp;lt;</code></li>
   * <li>Converts <code>%%*$</code> to <code>&amp;gt;</code></li>
   * </ul>
   * 
   * @param xml The XML String coming from JAXB
   * @return The transformed XML for XSL
   */
  public static String prepareXmlForXsl( final String xml )
  {
    // In WriteXML, the characters '<' and '>' are escaped to '&lt;' and '&gt;',
    // but now they have to be reverted in order to be interpreted as tags.
    String preparedXml = xml.replaceAll( ESCAPED_LT, LT );
    preparedXml = preparedXml.replaceAll( ESCAPED_GT, GT );

    preparedXml = preparedXml.replaceAll( HtmlConstants.BREAK, HtmlConstants.BREAK_EMPTY );

    preparedXml = convertBreaksToParagraph( preparedXml );

    // These Strings aren't tags - there may be a '<' or a '>' in a IP's name (for instance
    // designs), so convert them to printable characters for FOP.
    preparedXml = preparedXml.replaceAll( ESCAPED_ESCAPED_LT, ESCAPED_LT );
    preparedXml = preparedXml.replaceAll( ESCAPED_ESCAPED_GT, ESCAPED_GT );

    return preparedXml;
  }

  /**
   * Prepares the XML String for XSL formatting.
   * <ul>
   * <li>Converts <code>&amp;amp;lt;</code> to <code>&amp;lt;</code></li>
   * <li>Converts <code>&amp;amp;gt;</code> to <code>&amp;gt;</code></li>
   * <li>Converts <code>&amp;lt;</code> to <code>&lt;</code></li>
   * <li>Converts <code>&amp;gt;</code> to <code>&gt;</code></li>
   * <li>Converts <code>&amp;amp;&amp;amp;</code> to <code>&amp;amp;</code></li>
   * </ul>
   * This is used whenever a piece of XML (like a snapshot), which already contains escaped pseudo-HTML tags is set into
   * a simple String. The escaped tags are escaped again, and this reverts the process, just before XSL formatting.
   * 
   * @param xml The XML String coming from JAXB
   * @return The transformed XML for XSL
   */
  public static String prepareDoubleEscapedXmlForXsl( final String xml )
  {
    // The double escaped stuff has to be 'parked' with a character sequence that's
    // not replaced afterwards.
    final String doubleEscapedLtReplacement = "=çç=";
    final String doubleEscapedGtReplacement = "ç==ç";
    String preparedXml = xml.replaceAll( DOUBLE_ESCAPED_LT, doubleEscapedLtReplacement );
    preparedXml = preparedXml.replaceAll( DOUBLE_ESCAPED_GT, doubleEscapedGtReplacement );
    preparedXml = preparedXml.replaceAll( ESCAPED_LT, LT );
    preparedXml = preparedXml.replaceAll( ESCAPED_GT, GT );
    preparedXml = preparedXml.replaceAll( doubleEscapedLtReplacement, ESCAPED_LT );
    preparedXml = preparedXml.replaceAll( doubleEscapedGtReplacement, ESCAPED_GT );
    preparedXml = preparedXml.replaceAll( DOUBLE_ESCAPED_AMPERSAND, ESCAPED_AMPERSAND );
    return preparedXml;
  }

  /**
   * Removes blanks between <code>&lt;br&gt;</code> tags and converts two <code>&lt;br&gt;</code> tags to one
   * <code>&lt;p/&gt;</code> tag.
   * 
   * @param text The text to convert
   * @return The text without blanks between the breaks and converted breaks
   */
  private static String convertBreaksToParagraph( String text )
  {
    String noBlanks = text.replaceAll( "\\s*" + HtmlConstants.BREAK + "\\s*", HtmlConstants.BREAK );
    noBlanks = noBlanks.replaceAll( "\\s*" + HtmlConstants.BREAK_EMPTY + "\\s*", HtmlConstants.BREAK_EMPTY );
    noBlanks = noBlanks.replaceAll( HtmlConstants.BREAK + HtmlConstants.BREAK, HtmlConstants.PARAGRAPH_EMPTY );
    noBlanks = noBlanks.replaceAll( HtmlConstants.BREAK_EMPTY + HtmlConstants.BREAK_EMPTY, HtmlConstants.PARAGRAPH_EMPTY );
    return noBlanks;
  }
}
