package ch.ige.jipps.text.test;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import junit.framework.TestCase;
import ch.ige.dbutils.util.ClientInitializer;
import ch.ige.jipps.text.EventTextContainer;
import ch.ige.jipps.text.EventTextLoader;
import ch.ige.jipps.vo.HistoryItemVO;

/**
 * Tests {@link ch.ige.jipps.text.EventTextLoader}.
 *
 * <p>Copyright (c) 2006, Swiss Federal Institute of Intellectual Property</p>
 * @author    Eric Lewis
 * @version   $Id: EventTextLoaderTest.java,v 1.10 2007/04/03 09:29:20 lewis_lee Exp $
 */
public class EventTextLoaderTest extends TestCase
{
  static
  {
    ClientInitializer.initStandaloneClient();
  }

  /**
   * The event numbers to check
   */
  private static final int[] EVENT_NUMBERS = new int[] { 110449, 110450, 110451, 110533 };

  /**
   * The German event texts to check
   */
  private static final String[] EVENT_TEXTS_GERMAN = new String[] { "Sitzverlegung", "Änderung von Vertreterdaten", "Adressänderung", "Übertragung" };

  /**
   * The French event texts to check
   */
  private static final String[] EVENT_TEXTS_FRENCH = new String[] { "Changement du siège social", "Modification des données du mandataire",
      "Changement d'adresse", "Cession" };

  /**
   * The only constructor
   * @param name The test's name
   */
  public EventTextLoaderTest( String name )
  {
    super( name );
  }

  /**
   * Tests {@link ch.ige.jipps.text.EventTextLoader#insertEventText(List, Locale, boolean)} with German texts.
   * @throws Exception if the test fails
   */
  public void testInsertEventTextForGerman() throws Exception
  {
    List<EventTextContainer> eventTextContainers = createEventTextContainers();
    EventTextLoader.insertEventText( eventTextContainers, Locale.GERMAN, true );
    checkEventTextContainers( eventTextContainers, Locale.GERMAN );
  }

  /**
   * Tests {@link ch.ige.jipps.text.EventTextLoader#insertEventText(List, Locale, boolean)} with French texts.
   * @throws Exception if the test fails
   */
  public void testInsertEventTextForFrench() throws Exception
  {
    List<EventTextContainer> eventTextContainers = createEventTextContainers();
    EventTextLoader.insertEventText( eventTextContainers, Locale.FRENCH, true );
    checkEventTextContainers( eventTextContainers, Locale.FRENCH );
  }

  /**
   * Tests {@link ch.ige.jipps.text.EventTextLoader#insertEventText(List, Locale, boolean)}, checking that the texts are unique.
   * @throws Exception if the test fails
   */
  public void testInsertEventTextForUniqueness() throws Exception
  {
    // Yes, I know that index and numbering aren't the same...  ;-)

    List<EventTextContainer> eventTextContainers = createEventTextContainers();
    EventTextContainer container2 = eventTextContainers.get( 2 );
    EventTextContainer container3 = eventTextContainers.get( 3 );
    container3.setEventNumber( container2.getEventNumber() );
    EventTextLoader.insertEventText( eventTextContainers, Locale.FRENCH, false );
    assertEquals( "Ignoring that the texts must be unique", EVENT_NUMBERS[2], container2.getEventNumber().intValue() );
    assertEquals( "Ignoring that the texts must be unique", EVENT_TEXTS_FRENCH[2], container2.getEventText() );
    assertEquals( "Ignoring that the texts must be unique", EVENT_NUMBERS[2], container3.getEventNumber().intValue() );
    assertEquals( "Ignoring that the texts must be unique", EVENT_TEXTS_FRENCH[2], container3.getEventText() );

    eventTextContainers = createEventTextContainers();
    container2 = eventTextContainers.get( 2 );
    container3 = eventTextContainers.get( 3 );
    container3.setEventNumber( container2.getEventNumber() );
    EventTextLoader.insertEventText( eventTextContainers, Locale.FRENCH, true );
    assertEquals( "Assuring that the texts must be unique", EVENT_NUMBERS[2], container2.getEventNumber().intValue() );
    assertEquals( "Assuring that the texts must be unique", EVENT_TEXTS_FRENCH[2], container2.getEventText() );
    assertEquals( "Assuring that the texts must be unique", EVENT_NUMBERS[2], container3.getEventNumber().intValue() );
    assertEquals( "Assuring that the texts must be unique", null, container3.getEventText() );
  }

  /**
   * Tests {@link ch.ige.jipps.text.EventTextLoader#insertOptionalEventText(List, Locale, boolean)}, checking that the texts are unique.
   * @throws Exception if the test fails
   */
  public void testInsertOptionalEventTextForUniqueness() throws Exception
  {
    // Yes, I know that index and numbering aren't the same...  ;-)

    List<EventTextContainer> eventTextContainers = createEventTextContainers();
    EventTextContainer container2 = eventTextContainers.get( 2 );
    EventTextContainer container3 = eventTextContainers.get( 3 );
    container3.setEventNumber( null );
    EventTextLoader.insertOptionalEventText( eventTextContainers, Locale.FRENCH, false );
    assertEquals( "Checking existing text number", EVENT_NUMBERS[2], container2.getEventNumber().intValue() );
    assertEquals( "Checking existing text", EVENT_TEXTS_FRENCH[2], container2.getEventText() );
    assertEquals( "Checking null text number", null, container3.getEventNumber() );
    assertEquals( "Checking null text", null, container3.getEventText() );
  }

  /**
   * Tests {@link ch.ige.jipps.text.EventTextLoader#insertEventText(List, Locale, boolean)} with faults.
   * @throws Exception if the test fails
   */
  public void testInsertEventTextWithFaults() throws Exception
  {
    List<EventTextContainer> eventTextContainers = createEventTextContainers();
    try
    {
      EventTextLoader.insertEventText( null, Locale.GERMAN, true );
      fail( "Null text containers must fail" );
    }
    catch ( IllegalArgumentException ex )
    {
      // This is expected
    }

    // Not a fault per se, just check that an empty list comes out empty
    List<EventTextContainer> empty = new ArrayList<EventTextContainer>();
    EventTextLoader.insertEventText( empty, Locale.GERMAN, true );
    assertTrue( "Checking empty list", empty.isEmpty() );

    try
    {
      EventTextLoader.insertEventText( eventTextContainers, null, true );
      fail( "Null language must fail" );
    }
    catch ( IllegalArgumentException ex )
    {
      // This is expected
    }

    try
    {
      EventTextContainer container2 = eventTextContainers.get( 1 );
      container2.setEventNumber( null );
      EventTextLoader.insertEventText( eventTextContainers, Locale.GERMAN, true );
      fail( "Null event number must fail" );
    }
    catch ( IllegalArgumentException ex )
    {
      // This is expected
    }

    try
    {
      EventTextContainer container2 = eventTextContainers.get( 1 );
      container2.setEventNumber( -33 );
      EventTextLoader.insertEventText( eventTextContainers, Locale.GERMAN, true );
      fail( "Unknown event number must fail" );
    }
    catch ( IllegalArgumentException ex )
    {
      // This is expected
    }
  }

  /**
   * Creates event text containers with event text numbers.
   * @return A <code>List</code> of event text containers
   */
  private List<EventTextContainer> createEventTextContainers()
  {
    List<EventTextContainer> eventTextContainers = new ArrayList<EventTextContainer>();
    for ( int i = 0; i < EVENT_NUMBERS.length; i++ )
    {
      EventTextContainer eventTextContainer = new HistoryItemVO();
      eventTextContainer.setEventNumber( EVENT_NUMBERS[i] );
      eventTextContainers.add( eventTextContainer );
    }
    return eventTextContainers;
  }

  /**
   * Checks the event text containers after filling the texts.
   * @param eventTextContainers The event text containers to check
   * @param language The language to check for
   */
  private void checkEventTextContainers( List<EventTextContainer> eventTextContainers, Locale language )
  {
    assertEquals( "Checking number of event text containers", EVENT_NUMBERS.length, eventTextContainers.size() );
    for ( int i = 0; i < eventTextContainers.size(); i++ )
    {
      EventTextContainer container = eventTextContainers.get( i );
      assertEquals( "Checking number", EVENT_NUMBERS[i], container.getEventNumber().intValue() );
      String expectedEventText = null;
      if ( language.equals( Locale.GERMAN ) )
      {
        expectedEventText = EVENT_TEXTS_GERMAN[i];
      }
      else
      {
        expectedEventText = EVENT_TEXTS_FRENCH[i];
      }
      assertEquals( "Checking text", expectedEventText, container.getEventText() );
    }
  }
}
