package ch.ige.jipps.vo;

import java.sql.Timestamp;
import java.util.Date;
import java.util.Locale;

import ch.ige.dbutils.vo.AbstractVO;
import ch.ige.dbutils.vo.IdentifiableVO;
import ch.ige.util.constants.PrintStatusConstants;
import ch.ige.util.enums.BAGISLanguageDomain;
import ch.ige.util.enums.printer.ReportPaper;

/**
 * Contains the data for printing a report or a letter or an invoice. Note that the numeric print job types are used in
 * the client and admin applications and to get the correct printer for a print job.
 * <p>
 * Copyright (c) 2002, Eidgenössisches Institut für Geistiges Eigentum
 * @author Priska Steiner
 * @author Johannes Lang
 * @author Marc Bouquet
 * @author Elmar Isler
 * @author Kurt Jegge
 * @author Eric Lewis
 * @version $Id: PrintJobVO.java,v 1.73 2008/10/23 13:34:24 lewis_lee Exp $
 */
//TODO Why does this implement Cloneable, but doesn't overwrite clone()?
public class PrintJobVO extends AbstractVO implements Cloneable, IdentifiableVO<Integer>
{
  /**
   * The code for an invalid print job type
   */
  public static final int INVALID = -1;

  /**
   * The value of {@link #formTyp} when this print job is a report.
   */
  public static final String AUSZUG = "Auszug";

  /**
   * @see ch.ige.jipps.vo.PrintJobVO#AUSZUG
   */
  public static final int AUSZUG_NUMERIC = 1;

  /**
   * The value of {@link #formTyp} when this print job is a letter.
   */
  public static final String LETTER = "Letter";

  /**
   * @see ch.ige.jipps.vo.PrintJobVO#LETTER
   */
  public static final int LETTER_NUMERIC = 2;

  /**
   * The value of {@link #formTyp} when this print job is a list of search hits.
   */
  public static final String LISTEN = "Listen";

  /**
   * @see ch.ige.jipps.vo.PrintJobVO#LISTEN
   */
  public static final int LISTEN_NUMERIC = 3;

  /**
   * The value of {@link #formTyp} when this print job is the base sheet for an article (event template).
   */
  public static final String STAMMBLATT = "Stammblatt";

  /**
   * @see ch.ige.jipps.vo.PrintJobVO#STAMMBLATT
   */
  public static final int STAMMBLATT_NUMERIC = 4;

  /**
   * The value of {@link #formTyp} when this print job is a KOSI card.
   */
  public static final String KARTEIKARTE = "Karteikarte";

  /**
   * @see ch.ige.jipps.vo.PrintJobVO#KARTEIKARTE
   */
  public static final int KARTEIKARTE_NUMERIC = 5;

  /**
   * The value of {@link #formTyp} when this print job is a form (letter template).
   */
  public static final String FORMULAR = "Formular";

  /**
   * @see ch.ige.jipps.vo.PrintJobVO#FORMULAR
   */
  public static final int FORMULAR_NUMERIC = 6;

  /**
   * The value of {@link #formTyp} when this print job is an image list.
   */
  public static final String DESIGN_BILDER = "DesignBilder";

  // TODO Check if the correspondent numeric code for DESIGN_BILDER is 7

  /**
   * The value of {@link #formTyp} when this print job is a patent specification.
   */
  public static final String PATENT_SPECIFICATION = "PatentSpecification";

  /**
   * @see ch.ige.jipps.vo.PrintJobVO#PATENT_SPECIFICATION
   */
  public static final int PATENT_SPECIFICATION_NUMERIC = 8;

  /**
   * The value of {@link #formTyp} when this print job is a patent search report.
   */
  public static final String PATENT_SEARCH_REPORT = "PatentSearchReport";

  /**
   * @see ch.ige.jipps.vo.PrintJobVO#PATENT_SEARCH_REPORT
   */
  public static final int PATENT_SEARCH_REPORT_NUMERIC = 9;

  /**
   * All the possible values for {@link #formTyp}.
   */
  public static final String[] FORM_TYPES = { AUSZUG, LETTER, LISTEN, STAMMBLATT, KARTEIKARTE, FORMULAR, DESIGN_BILDER, PATENT_SPECIFICATION,
      PATENT_SEARCH_REPORT };

  /**
   * The value of {@link #sprache} if no specific language should be used for printing, so just take the IP's language.
   */
  public static final int USE_IP_LANGUAGE = 0;

  /**
   * The value to set in {@link #senderApp} and {@link #senderUser} if the print job has been generated by the system,
   * for instance in batch runs.
   */
  public static final String SYSTEM = "ActiveBatch";

  /**
   * The value to use for {@link #batchId} if this isn't a batch print job.
   */
  public static final int NO_BATCH_ID = 0;

  /**
   * The value to use for {@link #anzKopien} if no copies should be printed.
   */
  public static final int NO_COPIES = 0;

  // @see ch.ige.dbutils.vo.IdentifiableVO#getIdentifier()
  public Integer getIdentifier()
  {
    return this.printJobId;
  }

  // @see ch.ige.dbutils.vo.IdentifiableVO#setIdentifier(java.io.Serializable)
  public void setIdentifier( Integer printJobId )
  {
    this.printJobId = printJobId;
  }

  /**
   * The unique identifier of this print job.
   */
  public int printJobId;

  /**
   * If several print job were created in a batch run, then this contains the batch ID which ties them together. If it's
   * not a batch run, use {@link #NO_BATCH_ID}.
   */
  public int batchId;

  /**
   * If this is a letter which needs an address to send to, then this contains the identifiying address number.
   */
  public int adrNr;

  /**
   * The unique ID of the printer where the print job should be printed.
   */
  public int druckerId;

  /**
   * Shows on what kind of paper the print job is printed, one of {@link ch.ige.util.constants.ReportPaperConstants}.
   */
  // TODO Use ReportPaper enum
  public int paperMode;

  /**
   * Number of copies to print. May be {@link #NO_COPIES}.
   */
  public int anzKopien;

  /**
   * The print status, which can be
   * <ul>
   * <li>{@link ch.ige.util.constants.PrintStatusConstants#PREVIEW_PRINT}</li>
   * <li>{@link ch.ige.util.constants.PrintStatusConstants#READY_FOR_PRINTING}</li>
   * <li>{@link ch.ige.util.constants.PrintStatusConstants#PRINTED}</li>
   * </ul>
   */
  public int druckStatus;

  /**
   * The application that created the print job. May also be {@link #SYSTEM}.
   */
  public String senderApp;

  /**
   * The user who created the print job. May also be {@link #SYSTEM}.
   */
  public String senderUser;

  /**
   * The creation date.
   */
  public Timestamp erstDatum;

  /**
   * The print job type; this combined with <code>Control</code> defines what controller class is used for the report.
   * For instance, if this is {@link #AUSZUG}, then the class {@link ch.ige.jipps.AuszugControl} is used. It can be
   * <ul>
   * <li>{@link #AUSZUG}</li>
   * <li>{@link #DESIGN_BILDER}</li>
   * <li>{@link #LETTER}</li>
   * <li>{@link #PATENT_SPECIFICATION}</li>
   * <li>{@link #PATENT_SEARCH_REPORT}</li>
   * <li>{@link #LISTEN}</li>
   * <li>{@link #STAMMBLATT}</li>
   * <li>{@link #KARTEIKARTE}</li>
   * <li>{@link #FORMULAR}</li>
   * </ul>
   */
  public String formTyp;

  /**
   * For reports, this contains the IP key of the intellectual property.
   * For letters, it contains the BAGIS invoice ID (<code>bfaktId</code>).
   * For patent documents, it contains the document ID of the XML document to read (patent specification ID,
   * patent search report ID etc.).
   */
  public int reportId;

  /**
   * The binary data of the print job, usually a PDF file.
   */
  public byte[] printJobFile;

  /**
   * If this is <code>true</code>, the address head of the IPI is printed. If it's <code>false</code>, it's not
   * printed, usually because the report is being printed on paper that already contains the address (pre-printed).
   */
  public boolean printIgekopf;

  /**
   * This contains any error to show during the conversion from print job to PDF file.
   */
  public String status;

  /**
   * Contains one of the following:
   * <ul>
   *  <li>When printing lists of search hits, this contains an XML document according to <code>listen.xsd</code> somewhere
   *      else in this project.</li>
   *  <li>When printing a patent specification, this contains the patent specification in ST.36 format.</li>
   *  <li>When printing a patent search report, this contains the patent search report in ST.36 format.</li>
   * </ul>
   */
  //TODO rename to 'xmlDocument'
  public String listData;

  /**
   * If this is <code>true</code>, a cover page is printed.
   */
  public boolean coverCd;

  // ******************************************************************************************
  // The following variables are only used for processing, they're not saved in the database.
  // ******************************************************************************************

  /**
   * For reports, the date for which the report is shown; for letters, an optional print issue date (the date on the top of the letter).
   */
  //TODO Rename to print date or whatever
  public Date standDatum;

  /**
   * Shows whether the report is created for IPs that are only in application state ({@link ch.ige.util.constants.IpRegisterStateConstants#APPLICATION_STATE})
   * or IPs that are already registered ({@link ch.ige.util.constants.IpRegisterStateConstants#REGISTRATION_STATE}).
   */
  // TODO Use RegistrationState enum
  public int registerKz;

  /**
   * The IP type, one of {@link ch.ige.util.constants.IpTypeConstants}.
   */
  // TODO Use IPType enum
  public int schutitTyp;

  /**
   * The report type, one of {@link ch.ige.util.constants.ReportTypeConstants}.
   */
  // TODO Use ReportType enum
  public int regauszTyp;

  /**
   * The report's language, one of {@link ch.ige.util.constants.LanguageConstants} or {@link #USE_IP_LANGUAGE} if the
   * IP's language is taken.
   */
  // TODO Use BAGISLanguage enum
  public int sprache;

  /**
   * Gets the BAGIS language corresponding to {@link #sprache}.
   * @return The BAGIS language or <code>null</code> if {@link #sprache} is {@link #USE_IP_LANGUAGE}
   */
  public BAGISLanguageDomain getBagisLanguage()
  {
    if ( this.sprache == USE_IP_LANGUAGE )
    {
      return null;
    }
    else
    {
      return BAGISLanguageDomain.parse( this.sprache );
    }
  }

  /**
   * Shows whether this is printed for the second time, only for letters.
   */
  public boolean archiveCopy;

  /**
   * The user number of the user who created the print job.
   */
  public int maNr;

  /**
   * If this is <code>true</code>, a letter or invoice has additional recipients. Note that <code>PotZahl</code>
   * stands for <code>Potentielle Zahler</code>.
   */
  public boolean hasPotZahl;

  /**
   * The text of a watermark which is printed on the page. This is <strong>additional</strong> to the watermark which
   * appears on the development and integration platform.
   */
  public String watermark;

  /**
   * Is only used for letters: Defines if an existing row with the same report ID should be overwritten.
   */
  public boolean overwrite;

  /**
   * Is only used for letters: Defines if a letter should only be previewed and not printed.
   */
  public boolean preview;

  /**
   * Is <code>true</code> if this print job is part of a group (nothing to do with a batch) of letters with
   * a common report ID, but with different adrIds (potzahler); otherwise <code>false</code>;
   * is only used for letters. 
   */
  public boolean hasSiblings;

  /**
   * The IP number of a registred title. Can contain formatting 
   * elements (which is why it is a <code>String</code>.
   */
  public String formattedIpNumber;

  /**
   * The application number of an IP title. Can contain formatting 
   * elements (which is why it is a <code>String</code>.
   */
  public String formattedApplicationNumber;

  /**
   * The constructor which initializes this.
   */
  public PrintJobVO()
  {
    this.printJobId = 0;
    this.batchId = PrintJobVO.NO_BATCH_ID;
    this.adrNr = 0;
    this.druckerId = 1;
    this.paperMode = ReportPaper.PAPER_WITH_IPI_LOGO_INCLUDING_COPIES.toDB();
    this.anzKopien = PrintJobVO.NO_COPIES;
    this.druckStatus = PrintStatusConstants.PREVIEW_PRINT;
    this.senderApp = "";
    this.senderUser = "";
    this.erstDatum = null;
    this.formTyp = "";
    this.reportId = 0;
    this.printJobFile = null;
    this.printIgekopf = false;

    this.listData = "";

    this.standDatum = null;
    this.regauszTyp = 0;
    this.sprache = BAGISLanguageDomain.parse( Locale.GERMAN ).toDB();
    this.archiveCopy = false;
    this.coverCd = false;
    this.hasPotZahl = true;
    this.overwrite = false;
    this.hasSiblings = false;
  }

  // @see ch.ige.dbutils.vo.AbstractVO#toString()
  @Override
  public String toString()
  {
    // Don't show the PDF file
    byte[] pdfFile = this.printJobFile;
    this.printJobFile = null;
    try
    {
      return super.toString();
    }
    finally
    {
      this.printJobFile = pdfFile;
    }
  }
}
