package ch.ige.jipps.xml;

import java.io.IOException;
import java.io.Serializable;
import java.math.BigInteger;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.dbutils.util.LOBUtils;
import ch.ige.ipobjects.common.vo.ImageVO;
import ch.ige.ipobjects.exceptions.ImageDiscardedException;
import ch.ige.jipps.configuration.PathConfiguration;
import ch.ige.jipps.constants.PropertyConstants;
import ch.ige.jipps.constants.XslConstants;
import ch.ige.jipps.currentstate.CurrentStateElement;
import ch.ige.jipps.header.IpHeader;
import ch.ige.jipps.imageconv.ImageProcessor;
import ch.ige.jipps.snapshot.SnapshotVO;
import ch.ige.jipps.text.ConvertText;
import ch.ige.jipps.vo.AuszugVO;
import ch.ige.jipps.vo.HistoryItemVO;
import ch.ige.jipps.xml.auszug.Auszug;
import ch.ige.jipps.xml.auszug.BildType;
import ch.ige.jipps.xml.auszug.Header;
import ch.ige.jipps.xml.auszug.HistAuszTxt;
import ch.ige.jipps.xml.auszug.HistData;
import ch.ige.jipps.xml.auszug.Igekopf;
import ch.ige.jipps.xml.auszug.Kopf;
import ch.ige.jipps.xml.auszug.ObjectFactory;
import ch.ige.jipps.xml.auszug.Signature;
import ch.ige.jipps.xml.auszug.SnapshotReferenceType;
import ch.ige.jipps.xml.auszug.SnapshotType;
import ch.ige.jipps.xml.auszug.StampType;
import ch.ige.jipps.xml.auszug.StandAuszTxt;
import ch.ige.jipps.xml.auszug.StandData;
import ch.ige.util.address.IpiAddress;
import ch.ige.util.enums.BAGISLanguageDomain;
import ch.ige.util.enums.RegistrationState;

/**
 * Creates the report XML document.
 * <p>
 * Copyright (c) 2003-2006 Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * @author Priska Steiner
 * @author Esther Bandschapp
 * @author Eric Lewis
 * @version $Id: AuszugWriteXml.java,v 1.14 2009/03/23 09:51:15 locher Exp $
 */

public class AuszugWriteXml implements Serializable
{
  /**
   * The JAXB object factory
   */
  private ObjectFactory objectFactory;

  /**
   * The JAXB context
   */
  private static JAXBContext jaxbContext;

  /**
   * Marshalls the objects to XML
   */
  private static Marshaller marshaller;

  /**
   * The final object for JAXB
   */
  private Auszug auszug;

  /**
   * The basic tag for the snapshot.
   */
  private static final String SNAPSHOT_TAG = "snapshotXml";

  /**
   * The starting snapshot tag.
   */
  private static final String SNAPSHOT_TAG_START = "<" + SNAPSHOT_TAG + ">";

  /**
   * The ending snapshot tag.
   */
  private static final String SNAPSHOT_TAG_END = "</" + SNAPSHOT_TAG + ">";

  /**
   * Used to go through all snapshot values that contain <code>&lt;</code> or <code>&gt;</code>.
   */
  private static final Pattern SNAPSHOT_VALUE = Pattern.compile( "(<value>)(.*(?:" + ConvertText.ESCAPED_LT + "|" + ConvertText.ESCAPED_GT + "|"
      + ConvertText.DOUBLE_ESCAPED_AMPERSAND + ").*)(</value>)" );

  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( AuszugWriteXml.class );

  /**
   * Creates an XML document from the report.
   * @param report Contains all the data for the report
   * @return A <code>String</code> containing the XML document
   * @throws JAXBException if the XML can't be created
   * @throws IOException if the image can't be written
   */
  @SuppressWarnings( "unchecked" )
  public String toXml( AuszugVO report ) throws JAXBException, IOException
  {
    if ( this.objectFactory == null )
    {
      jaxbContext = JAXBContext.newInstance( "ch.ige.jipps.xml.auszug" );
      this.objectFactory = new ObjectFactory();
      marshaller = jaxbContext.createMarshaller();
    }

    this.auszug = this.objectFactory.createAuszug();

    this.auszug.setIgekopf( createIpiHeader( report ) );
    LOG.debug( "IGE kopf" );

    this.auszug.setHeader( createPrintHeader( report ) );
    LOG.debug( "header" );

    this.auszug.setKopf( createIpHeader( report.getIpHeader(), report.getImage() ) );
    LOG.debug( "kopf" );

    this.auszug.getStandData().addAll( createCurrentState( report ) );
    LOG.debug( "stand" );

    this.auszug.setRegistration( createRegistrationSnapshot( report.getIpHeader().getIpKey(), report.getRegistrationSnapshot() ) );
    LOG.debug( "registration snapshot" );

    this.auszug.setTitelHist( report.getHistoryTitle() );
    LOG.debug( "titelHist" );

    this.auszug.getHistData().addAll( createHistory( report ) );
    LOG.debug( "hist" );

    this.auszug.setSignature( createSignature( report ) );
    LOG.debug( "signature" );

    String xmlDocument = XmlMarshaller.toString( jaxbContext, marshaller, "auszug.xsd", this.auszug );
    if ( this.auszug.getRegistration() != null )
    {
      // There is now double escaped stuff in the registration snapshot,
      // so replace it by simple escaped stuff - but ONLY the snapshot!
      String snapshot = StringUtils.substringBetween( xmlDocument, SNAPSHOT_TAG_START, SNAPSHOT_TAG_END );
      snapshot = ConvertText.prepareDoubleEscapedXmlForXsl( snapshot );
      Matcher valueMatcher = SNAPSHOT_VALUE.matcher( snapshot );
      StringBuffer escapedSnapshot = new StringBuffer();
      while ( valueMatcher.find() )
      {
        String valueStartTag = valueMatcher.group( 1 );

        String escapedValue = valueMatcher.group( 2 );
        escapedValue = ConvertText.prepareXmlForXsl( escapedValue );
        escapedValue = ConvertText.replaceHtmlTags( escapedValue );
        // This is required, because the $ would be interpreted as a group value in the regex.
        escapedValue = StringUtils.replace( escapedValue, "$", "\\$" );

        String valueEndTag = valueMatcher.group( 3 );
        valueMatcher.appendReplacement( escapedSnapshot, valueStartTag + escapedValue + valueEndTag );
      }
      valueMatcher.appendTail( escapedSnapshot );

      StringBuilder newXmlDocument = new StringBuilder();
      newXmlDocument.append( StringUtils.substringBefore( xmlDocument, SNAPSHOT_TAG_START ) );
      newXmlDocument.append( SNAPSHOT_TAG_START );
      newXmlDocument.append( escapedSnapshot.toString() );
      newXmlDocument.append( SNAPSHOT_TAG_END );
      newXmlDocument.append( StringUtils.substringAfter( xmlDocument, SNAPSHOT_TAG_END ) );
      return newXmlDocument.toString();
    }
    else
    {
      return xmlDocument;
    }
  }

  /**
   * Creates the IPI header for the XML.
   * @param report Contains all the data for the report
   * @return The JAXB object for the IPI header in the XML
   * @throws JAXBException
   */
  private Igekopf createIpiHeader( AuszugVO report ) throws JAXBException
  {
    Igekopf ipiHeader = this.objectFactory.createIgekopf();

    if ( !report.printIpiHeader() )
    {
      ipiHeader.setIgekopfDe( StringUtils.EMPTY );
      ipiHeader.setIgekopfFr( StringUtils.EMPTY );
      ipiHeader.setIgekopfIt( StringUtils.EMPTY );
      ipiHeader.setIgekopfEn( StringUtils.EMPTY );
      ipiHeader.setIgekopfAdr( StringUtils.EMPTY );
      ipiHeader.setIgekopfKto( StringUtils.EMPTY );
    }
    else
    {
      ipiHeader.setIgekopfDe( IpiAddress.getGermanName() );
      ipiHeader.setIgekopfFr( IpiAddress.getFrenchName() );
      ipiHeader.setIgekopfIt( IpiAddress.getItalianName() );
      ipiHeader.setIgekopfEn( IpiAddress.getEnglishName() );
      ipiHeader.setIgekopfAdr( IpiAddress.getFullGermanAddress() );
      ipiHeader.setIgekopfKto( StringUtils.EMPTY );
    }
    return ipiHeader;
  }

  /**
   * Creates the print header.
   * @param report Contains all the data for the report
   * @return The print header
   * @throws JAXBException
   */
  private Header createPrintHeader( AuszugVO report ) throws JAXBException
  {
    Header header = this.objectFactory.createHeader();
    if ( report.getPrintHeader() != null )
    {
      header.setPrintDat( report.getPrintHeader().getPrintDate() );
      header.setPage( report.getPrintHeader().getPageLabel() );
      header.setEintrNrLbl( report.getPrintHeader().getIpNumberLabel() );
      // For patents, if there's no registration number, the application number
      // must be shown in the print header
      if ( report.getIpHeader().getIpType().isPatent() && report.getPrintHeader().getFormattedIpNumber().matches( "^0+$" ) )
      {
        header.setEintrNr( report.getIpHeader().getFormattedApplicationNumber() );
      }
      else
      {
        header.setEintrNr( report.getPrintHeader().getFormattedIpNumber() );
      }
    }

    return header;
  }

  /**
   * Creates the IP header.
   * @param ipHeader The IP header containing the data
   * @param image The IP's image, which may be <code>null</code>
   * @return The IP header object for JAXB
   * @throws IOException
   * @throws JAXBException
   */
  private Kopf createIpHeader( IpHeader ipHeader, ImageVO image ) throws JAXBException, IOException
  {
    Kopf kopf = this.objectFactory.createKopf();

    if ( ipHeader != null )
    {
      // Note: If a node defaults to null, it won't be shown in the XML document as null value
      kopf.setAuszugsTyp( BigInteger.valueOf( ipHeader.getReportType().toDB().longValue() ) );
      kopf.setAuszugsTitel( ipHeader.getReportTitle() );
      kopf.setSchutitKey( ipHeader.getIpKey().toString() );
      if ( ipHeader.getIpNumber() != null )
      {
        kopf.setSchutitNr( ipHeader.getIpNumber() );
      }
      kopf.setEintrNr( ipHeader.getFormattedIpNumber() );
      kopf.setStammEintrNr( defaultToNull( ipHeader.getFormattedBaseIpNumber() ) );
      kopf.setSchutitTyp( BigInteger.valueOf( ipHeader.getIpType().toDB().longValue() ) );
      kopf.setStammBeginn( ipHeader.getBaseProtectionStart() );
      if ( ipHeader.getRegistrationState() == RegistrationState.APPLICATION )
      {
        kopf.setSchutitRegisterStatus( XslConstants.REPORT_FOR_APPLICATION );
      }
      else
      {
        kopf.setSchutitRegisterStatus( XslConstants.REPORT_FOR_REGISTRATION );
      }

      if ( image != null )
      {
        kopf.setBild( createImageFile( image, ipHeader.getIpKey(), false ) );
      }
      else
      {
        kopf.setSchutitel( ConvertText.replaceHtmlTags( ipHeader.getDescription() ) );
      }

      kopf.setSchutzEnde( ipHeader.getProtectionEnd() );
      kopf.setAnmeldDat( ipHeader.getApplicationDate() );
      kopf.setAnmeldPublDat( ipHeader.getApplicationPublicationDate() );

      kopf.setGesuchNr( defaultToNull( ipHeader.getFormattedApplicationNumber() ) );

      kopf.setLoeschDat( ipHeader.getCancellation() );
      kopf.setEintrDat( ipHeader.getRegistrationDate() );

      kopf.setEPublikation( ipHeader.isPublishedElectronically() );

      kopf.setErstPublDat( ipHeader.getFirstPublicationDate() );
      kopf.setErstPublJahr( ipHeader.getFirstPublicationYear().toString() );
      if ( ipHeader.getFirstPublicationNumber() != null )
      {
        kopf.setErstPublNr( ipHeader.getFirstPublicationNumber().toString() );
      }
      kopf.setErstPublPflicht( toPseudoBoolean( ipHeader.isEventOfFirstPublicationPublished() ) );

      kopf.setLetztPublDat( ipHeader.getLastPublicationDate() );
      kopf.setLetztPublJahr( ipHeader.getLastPublicationYear().toString() );
      if ( ipHeader.getLastPublicationNumber() != null )
      {
        kopf.setLetztPublNr( ipHeader.getLastPublicationNumber().toString() );
      }
      else
      {
        // This is done only for compatibility
        kopf.setLetztPublNr( "0" );
      }
      if ( ipHeader.isEventOfLastPublicationPublished().booleanValue() )
      {
        kopf.setLetztPublPflicht( toPseudoBoolean( ipHeader.isEventOfLastPublicationPublished() ) );
      }

      kopf.setStandDat( ipHeader.getReportDate() );
      kopf.setSprache( BAGISLanguageDomain.parse( ipHeader.getReportLanguage() ).toDB().toString() );

      kopf.setWidersprFrist( ipHeader.getOppositionDeadline() );
    }

    return kopf;
  }

  /**
   * Creates the current state elements.
   * @param report Contains all the data for the report
   * @return A <code>List</code> of {@link ch.ige.jipps.xml.auszug.StandData} objects
   * @throws JAXBException if the elements can't be created
   */
  @SuppressWarnings( "unchecked" )
  private List<StandData> createCurrentState( AuszugVO report ) throws JAXBException
  {
    List<StandData> currentStateElements = new ArrayList<StandData>();
    if ( report.getCurrentStateElements() != null )
    {
      for ( CurrentStateElement currentStateElement: report.getCurrentStateElements() )
      {
        StandData standData = this.objectFactory.createStandData();
        standData.setStandEreigTitel( currentStateElement.getTitle() );
        StandAuszTxt standText = this.objectFactory.createStandAuszTxt();

        for ( String textLine: currentStateElement.getTextLines() )
        {
          standText.getStandTxt().add( ConvertText.replaceHtmlTags( textLine ) );
        }
        standData.setStandAuszTxt( standText );
        currentStateElements.add( standData );
      }
    }
    return currentStateElements;
  }

  /**
   * Creates the registration snapshot. This checks the snapshot XML document in
   * {@link ch.ige.jipps.snapshot.SnapshotVO#getXml()} and replaces all image references with saved versions of
   * {@link ch.ige.jipps.snapshot.SnapshotVO#getImages()}, kind of like in the IP header.
   * @param ipKey The unique key of the IP
   * @param registrationSnapshot Contains all the data for the registration snapshot
   * @return A registration containing the modified XML document with the snapshot
   * @throws JAXBException if the registration snapshot can't be created
   * @throws IOException if the image can't be written
   */
  @SuppressWarnings( "unchecked" )
  private SnapshotType createRegistrationSnapshot( Integer ipKey, SnapshotVO registrationSnapshot ) throws JAXBException, IOException
  {
    SnapshotType registration = null;
    if ( registrationSnapshot != null && StringUtils.isNotEmpty( registrationSnapshot.getXml() ) )
    {
      registration = this.objectFactory.createSnapshotType();
      registration.setSnapshotTitle( registrationSnapshot.getTitle() );
      registration.setSnapshotXml( registrationSnapshot.getXml() );

      Collection<ImageVO> images = registrationSnapshot.getImages().values();
      if ( !images.isEmpty() )
      {
        // Because of the ImageProcessor configuration, a snapshot reference type
        // is only created if images are actually created.
        SnapshotReferenceType snapshotReference = null;
        // Save all images
        for ( ImageVO image: images )
        {
          BildType bild = createImageFile( image, ipKey, true );
          if ( bild != null )
          {
            if ( snapshotReference == null )
            {
              snapshotReference = this.objectFactory.createSnapshotReferenceType();
            }
            snapshotReference.getBild().add( bild );
          }
        }
        registration.setSnapshotReference( snapshotReference );
      }
    }

    return registration;
  }

  /**
   * Creates the history.
   * @param report Contains all the data for the report
   * @return The history data items
   * @throws JAXBException if the XML binding goes wrong
   * @throws IOException if the image can't be written
   */
  @SuppressWarnings( "unchecked" )
  private List<HistData> createHistory( AuszugVO report ) throws JAXBException, IOException
  {
    List<HistData> convertedHistoryItems = new ArrayList<HistData>();
    if ( report.getHistoryItems() != null )
    {
      for ( HistoryItemVO historyItem: report.getHistoryItems() )
      {
        HistData histData = this.objectFactory.createHistData();
        histData.setVonEreigDat( historyItem.getStartDate() );
        histData.setBisEreigDat( historyItem.getEndDate() );
        if ( historyItem.getEndDateLabel() == null )
        {
          histData.setGueltigTxt( historyItem.getEndDateLabel() );
        }
        else
        {
          histData.setGueltigTxt( historyItem.getEndDateLabel().trim() );
        }

        histData.setArtLbl( historyItem.getEventText() );
        histData.setRegTxtLbl( historyItem.getReportText() );

        if ( !historyItem.getTextLines().isEmpty() )
        {
          HistAuszTxt detailTextLines = this.objectFactory.createHistAuszTxt();
          for ( String textLine: historyItem.getTextLines() )
          {
            detailTextLines.getHistTxt().add( ConvertText.replaceHtmlTags( textLine ) );
          }
          histData.setHistAuszTxt( detailTextLines );
        }

        histData.setPublTxt( historyItem.getPublication() );
        histData.setEreigTxt( ConvertText.replaceHtmlTags( historyItem.getRemarks() ) );

        if ( historyItem.getImage() != null )
        {
          histData.setBild( createImageFile( historyItem.getImage(), report.getIpHeader().getIpKey(), true ) );
        }

        convertedHistoryItems.add( histData );
      }
    }
    return convertedHistoryItems;
  }

  /**
   * Creates the signature object for XML.
   * @param report Contains all the data for the report
   * @return The signature for XML
   * @throws JAXBException if the signature can't be created
   */
  @SuppressWarnings( "unchecked" )
  private Signature createSignature( AuszugVO report ) throws JAXBException
  {
    Signature signature = this.objectFactory.createSignature();
    if ( report.getSignatureLines() != null )
    {
      signature.getZeile().addAll( report.getSignatureLines() );
    }
    if ( null != report.getSignatureStamp() )
    {
      StampType stamp = this.objectFactory.createStampType();
      stamp.setSrc( report.getSignatureStamp().getImageScr() );
      stamp.setWidth( report.getSignatureStamp().getImageWidth() );
      stamp.setHeight( report.getSignatureStamp().getImageHeight() );
      signature.setStamp( stamp );
    }
    return signature;
  }

  /**
   * Writes the image into a file and returns its information.
   * @param image The image to write into a file
   * @param ipKey The IP's unique key
   * @param useImageNumber If this is <code>true</code>, the image number which is unique for every image together
   * with <code>ipKey</code> is appended to the image file name after <code>ipKey</code>.
   * @return The image for XML processing or <code>null</code> if images should not be processed
   * @throws IOException if the image file can't be written
   * @throws JAXBException if the image type can't be created
   */
  private BildType createImageFile( ImageVO image, Integer ipKey, boolean useImageNumber ) throws IOException, JAXBException
  {
    if ( ImageProcessor.processImages() )
    {
      StringBuilder pathWithoutFileType = new StringBuilder();
      pathWithoutFileType.append( PathConfiguration.getVerifiedPath( PropertyConstants.JIPPS_PATH_IMAGE, true ) );
      pathWithoutFileType.append( "bild_" );
      pathWithoutFileType.append( ipKey );
      if ( useImageNumber )
      {
        pathWithoutFileType.append( "_" );
        pathWithoutFileType.append( image.getImageNumber() );
      }

      try
      {
        byte[] imageData = LOBUtils.readBLOB( image.getImage() );
        Map<String, Object> imageProperties = ImageProcessor.getInstance().writeImage( pathWithoutFileType.toString(), imageData, true );

        BildType imageType = this.objectFactory.createBildType();
        imageType.setSrc( pathWithoutFileType + "." + imageProperties.get( ImageProcessor.IMAGE_TYPE ) );

        String[] sizes = ImageCropper.getCroppedAndFormattedSizes( imageProperties );
        imageType.setWidth( sizes[0] );
        imageType.setHeight( sizes[1] );
        if ( useImageNumber )
        {
          imageType.setImageNumber( BigInteger.valueOf( image.getImageNumber().longValue() ) );
        }

        return imageType;
      }
      catch ( SQLException e )
      {
        IOException exception = new IOException( "Can't convert image BLOB" );
        exception.initCause( e );
        throw exception;
      }
      catch ( ImageDiscardedException e )
      {
        throw e.throwAsSystemException();
      }
    }
    else
    {
      return null;
    }
  }

  /**
   * Defaults a <code>String</code> to <code>null</code> if it's empty. Such XML nodes are then not created.
   * @param string The <code>String</code> to check
   * @return <code>null</code> if <code>string</code> is empty, otherwise <code>string</code>
   */
  private String defaultToNull( String string )
  {
    return defaultTo( string, null );
  }

  /**
   * Defaults a <code>String</code> to <code>defaultValue</code> if it's empty.
   * @param string The <code>String</code> to check
   * @param defaultValue The default value if <code>string</code> is empty
   * @return <code>defaultValue</code> if <code>string</code> is empty, otherwise <code>string</code>
   */
  private String defaultTo( String string, String defaultValue )
  {
    if ( StringUtils.isNotEmpty( string ) )
    {
      return string;
    }
    else
    {
      return defaultValue;
    }
  }

  /**
   * Translates a boolean value to <code>1</code> for true and <code>0</code> for false.
   * @param booleanValue The boolean value
   * @return <code>1</code> for true and <code>0</code> for false
   */
  private String toPseudoBoolean( Boolean booleanValue )
  {
    if ( booleanValue.booleanValue() )
    {
      return "1";
    }
    else
    {
      return "0";
    }
  }
}