package ch.ige.jipps.xml;

import java.io.IOException;
import java.io.Serializable;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.ipobjects.report.dao.RegFormTextDAO;
import ch.ige.ipobjects.report.vo.RegFormTextVO;
import ch.ige.ipobjects.time.vo.EreigIDat;
import ch.ige.jipps.configuration.PathConfiguration;
import ch.ige.jipps.constants.PropertyConstants;
import ch.ige.jipps.imageconv.ImageProcessor;
import ch.ige.jipps.vo.DesignImageVO;
import ch.ige.jipps.vo.DesignImagesVO;
import ch.ige.jipps.xml.bilderliste.BildType;
import ch.ige.jipps.xml.bilderliste.Bilder;
import ch.ige.jipps.xml.bilderliste.BilderListe;
import ch.ige.jipps.xml.bilderliste.HistoryType;
import ch.ige.jipps.xml.bilderliste.Igekopf;
import ch.ige.jipps.xml.bilderliste.ListeFooter;
import ch.ige.jipps.xml.bilderliste.ListeHeader;
import ch.ige.jipps.xml.bilderliste.ObjectFactory;
import ch.ige.util.address.IpiAddress;
import ch.ige.util.constants.ReportTextConstants;
import ch.ige.util.date.DateFormatter;
import ch.ige.util.enums.IPType;
import ch.ige.util.enums.report.ReportArea;
import ch.ige.util.enums.report.ReportType;

/**
 * Creates the XML for the design image list.
 * <p>
 * Copyright (c) 2003-2006 Eidgenössisches Institut für Geistiges Eigentum
 * </p>
 * @author Priska Steiner
 * @author Esther Bandschapp
 * @version $Id: DesignBilderWriteXml.java,v 1.5 2008/03/10 08:58:01 lewis_lee Exp $
 */
public class DesignBilderWriteXml implements Serializable
{
  /**
   * The JAXB object factory
   */
  private ObjectFactory objectFactory;

  /**
   * The JAXB context
   */
  private static JAXBContext jaxbContext;

  /**
   * Marshalls the objects to XML
   */
  private static Marshaller marshaller;

  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( DesignBilderWriteXml.class );

  /**
   * Used for an empty IPI header
   */
  private static final String EMPTY_STRING = "";

  /**
   * The constructor
   */
  public DesignBilderWriteXml()
  {
    // empty constructor
  }

  /**
   * Creates an XML document from the design images, representing a design image list.
   * @param designImagesVO A value objects with the image list's data
   * @return A <code>String</code> containing the XML document
   * @throws JAXBException if the marshalling fails
   * @throws IOException if the XML can't be created
   */
  public String toXml( DesignImagesVO designImagesVO ) throws JAXBException, IOException
  {
    if ( this.objectFactory == null )
    {
      jaxbContext = JAXBContext.newInstance( "ch.ige.jipps.xml.bilderliste" );
      this.objectFactory = new ObjectFactory();
      marshaller = jaxbContext.createMarshaller();
    }

    BilderListe imageList = this.objectFactory.createBilderListe();

    imageList.setIgekopf( createIpiHeader( designImagesVO ) );
    imageList.setListeHeader( createIpHeader( designImagesVO ) );
    LOG.debug( "ListeHeader" );

    try
    {
      RegFormTextDAO regFormTextDAO = new RegFormTextDAO();

      RegFormTextVO validToLabel = regFormTextDAO.findReportText( ReportTextConstants.VALID_TO_INTEGER, designImagesVO.language.toLocale(),
          ReportArea.MISCELLANEOUS );

      imageList.getBilder().addAll( createImages( designImagesVO.images, designImagesVO.ipKey, false, validToLabel.getText() ) );
      LOG.debug( "Bilder" );

      imageList.setHistory( createHistoryImages( designImagesVO, validToLabel.getText() ) );
      LOG.debug( "History" );

      // Since there are images in the history, set the history title
      if ( !imageList.getHistory().getBilder().isEmpty() )
      {
        RegFormTextVO historyTitle = regFormTextDAO
            .findReportText( ReportTextConstants.HISTORY_INTEGER, designImagesVO.language.toLocale(), ReportArea.HISTORY );
        imageList.setTitelHist( historyTitle.getText() );
      }
    }
    catch ( SQLException e )
    {
      IOException exception = new IOException( "Couldn't convert images" );
      exception.initCause( e );
      throw exception;
    }

    try
    {
      imageList.setListeFooter( createFooter( designImagesVO ) );
      LOG.debug( "ListeFooter" );
    }
    catch ( SQLException e )
    {
      IOException exception = new IOException( "Couldn't create footer" );
      exception.initCause( e );
      throw exception;
    }

    return XmlMarshaller.toString( jaxbContext, marshaller, "bilderliste.xsd", imageList );
  }

  /**
   * Creates the IPI header.
   * @param designImagesVO A value object containing the data which are needed for the image list
   * @return The IPI header
   * @throws JAXBException if the IPI header can't be created by JAXB
   */
  private Igekopf createIpiHeader( DesignImagesVO designImagesVO ) throws JAXBException
  {
    Igekopf ipiHeader = this.objectFactory.createIgekopf();

    if ( !designImagesVO.printIpiHeader )
    {
      ipiHeader.setIgekopfDe( EMPTY_STRING );
      ipiHeader.setIgekopfFr( EMPTY_STRING );
      ipiHeader.setIgekopfIt( EMPTY_STRING );
      ipiHeader.setIgekopfEn( EMPTY_STRING );
      ipiHeader.setIgekopfAdr( EMPTY_STRING );
      ipiHeader.setIgekopfKto( EMPTY_STRING );
    }
    else
    {
      ipiHeader.setIgekopfDe( IpiAddress.getGermanName() );
      ipiHeader.setIgekopfFr( IpiAddress.getFrenchName() );
      ipiHeader.setIgekopfIt( IpiAddress.getItalianName() );
      ipiHeader.setIgekopfEn( IpiAddress.getEnglishName() );
      ipiHeader.setIgekopfAdr( IpiAddress.getFullGermanAddress() );
      ipiHeader.setIgekopfKto( EMPTY_STRING );
    }
    return ipiHeader;
  }

  /**
   * Creates the IP header.
   * @param designImagesVO A value object containing the data which are needed for the image list
   * @return The IP header
   * @throws JAXBException if the header can't be created by JAXB
   */
  private ListeHeader createIpHeader( DesignImagesVO designImagesVO ) throws JAXBException
  {
    ListeHeader header = this.objectFactory.createListeHeader();
    header.setEintrNr( designImagesVO.ipNumber );
    header.setAuszugsTitel( designImagesVO.reportTitle );
    header.setSprache( designImagesVO.language.toDB().toString() );
    header.setStandDat( designImagesVO.reportDate );
    return header;
  }

  /**
   * Creates the page footer.
   * @param designImagesVO A value object containing the data which are needed for the footer
   * @return The page footer
   * @throws JAXBException if the footer can't be created by JAXB
   * @throws SQLException if the <code>page</code> text can't be found
   */
  private ListeFooter createFooter( DesignImagesVO designImagesVO ) throws JAXBException, SQLException
  {
    ListeFooter footer = this.objectFactory.createListeFooter();
    List<RegFormTextVO> texts = new RegFormTextDAO().findReportTexts( designImagesVO.language.toLocale(), ReportType.CURRENT_STATE, IPType.DESIGN,
        ReportArea.PAGE_HEADER );
    String pageText = texts.get( 0 ).getText();
    // It's "Page:" in normal reports, but only "Page" for image lists
    footer.setPage( StringUtils.chomp( pageText, ":" ) );
    return footer;
  }

  /**
   * Creates the details of the images <code>imageList</code>, either for the current state or for the history.
   * @param designImages The design's images
   * @param ipKey The IP's unique key
   * @param isHistoryImage If this is <code>true</code>, all the additional information for an image in the image
   * history is set; otherwise that part is skipped
   * @param validToLabel The text to set for the <code>valid to...</code> text
   * @return The images for the XML document
   * @throws IOException if the images can't be written
   * @throws JAXBException if the images can't be created by JAXB
   * @throws SQLException if the image size can't be read
   */
  private List<Bilder> createImages( List<DesignImageVO> designImages, int ipKey, boolean isHistoryImage, String validToLabel ) throws IOException,
      JAXBException, SQLException
  {
    final long IMAGE_DATA_START_POSITION = 1;

    List<Bilder> allImages = new ArrayList<Bilder>();
    Integer currentOrderNumberId = new Integer( -9563 ); // Some non-existing value
    int previousEventDate = -1;
    for ( DesignImageVO designImageVO: designImages )
    {
      Bilder bilder = this.objectFactory.createBilder();

      // Start a new order number
      if ( designImageVO.getOrderNumberId() != null && !designImageVO.getOrderNumberId().equals( currentOrderNumberId ) )
      {
        bilder.setModOrdNr( designImageVO.getOrderNumber() );
        currentOrderNumberId = designImageVO.getOrderNumberId();

        if ( isHistoryImage )
        {
          // The event is only shown for the first order number
          if ( designImageVO.getPeriodStartTime() != previousEventDate )
          {
            // The start time and event text both come from the event that must be shown
            // as the event contining the image. This is normally not (!) the event that
            // contains it in the DB, but the registration event.
            bilder.setArtLbl( designImageVO.getEventText() );
            bilder.setVonEreigDat( DateFormatter.shortFormat( designImageVO.getEvent().getDate() ) );
          }
          // The ':' is shown in the XSL
          bilder.setGueltigTxt( StringUtils.chomp( validToLabel, ":" ) );
          bilder.setBisEreigDat( DateFormatter.shortFormat( EreigIDat.from( designImageVO.getPeriodStopTime() ).getDate() ) );

          previousEventDate = designImageVO.getPeriodStartTime();
        }
      }
      bilder.setBildNr( designImageVO.getImageNumberForOrderNumber().toString() );
      bilder.setBildNrVon( designImageVO.getNumberOfImagesForOrderNumber().toString() );

      String pathWithoutFileType = PathConfiguration.getVerifiedPath( PropertyConstants.JIPPS_PATH_IMAGE, true );
      pathWithoutFileType += "DesBild_" + ipKey + "_" + designImageVO.getImageId();

      int imageLength = new Long( designImageVO.getImage().length() ).intValue();
      byte[] imageData = designImageVO.getImage().getBytes( IMAGE_DATA_START_POSITION, imageLength );

      Map<String, Object> imageProperties = ImageProcessor.getInstance().writeImage( pathWithoutFileType, imageData, false );

      BildType bildType = this.objectFactory.createBildType();
      bildType.setSrc( pathWithoutFileType + "." + imageProperties.get( ImageProcessor.IMAGE_TYPE ) );

      String[] sizes = ImageCropper.getCroppedAndFormattedSizes( imageProperties );
      bildType.setWidth( sizes[0] );
      bildType.setHeight( sizes[1] );

      bilder.setBild( bildType );
      allImages.add( bilder );
    }
    return allImages;
  }

  /**
   * Writes the details of the images <code>imageList</code>.
   * @param designImagesVO A value object containing the data which are needed for the image list
   * @param validToLabel The text to set for the <code>valid to...</code> text
   * @return A <code>List</code> of {@link ch.ige.jipps.xml.bilderliste.Bilder} objects
   * @throws IOException if the images can't be written
   * @throws JAXBException if the images can't be created by JAXB
   * @throws SQLException if the image size can't be read
   */
  @SuppressWarnings( "unchecked" )
  private HistoryType createHistoryImages( DesignImagesVO designImagesVO, String validToLabel ) throws IOException, JAXBException, SQLException
  {
    HistoryType history = this.objectFactory.createHistoryType();
    history.getBilder().addAll( createImages( designImagesVO.historyImages, designImagesVO.ipKey, true, validToLabel ) );
    return history;
  }
}
