package ch.ige.pctedi.client.wpa;

import java.text.*;
import java.util.Date;
import org.apache.log4j.Logger;
import ch.ige.pctedi.client.XMLDoc;
import ch.ige.pctedi.vo.*;

/**
 *
 * <p>
 * Copyright (c) 2005, Eidgenössisches Institut für Geistiges Eigentum
 * @author A. Rueegsegger
 * @version $Id: ClassificationIpcr.java,v 1.10 2006/12/11 14:34:59 ruegsegger Exp $
 */

public class ClassificationIpcr
{
  private static final Logger LOG = Logger.getLogger( ClassificationIpcr.class );
  private SimpleDateFormat sdf = new SimpleDateFormat( "yyyyMMdd" );

  private XMLDoc xml;
  private WpaLogFile logFile = WpaLogFile.getInstance();

  public ClassificationIpcr( XMLDoc xml )
  {
    this.xml = xml;
    sdf.setLenient( false );
  }

  /**
   * Extrahiert die Werte aus dem XML und füllt sie in ein PatIpcVO ab
   * @param tag String XML-Element
   * @return PatIpcVO abgefülltes Object
   */
  public PatIpcVO extractPatIpcVO( Object element ) throws Exception
  {
    PatIpcVO patIpcVO = new PatIpcVO();

    IpcrClassificationVO ipcrClassificationVO = extractClassificationVO( element );

    patIpcVO.ipcSymb = getIpcSymbol( ipcrClassificationVO );
    patIpcVO.patIpcCd = getPatIpcCd( ipcrClassificationVO );
    patIpcVO.ipcGueltigVon = getIpcGueltigVon( ipcrClassificationVO );
    patIpcVO.classLevel = getClassLevel( ipcrClassificationVO );
    patIpcVO.classValue = getClassValue( ipcrClassificationVO );
    patIpcVO.erfassDat = getErfassDat( ipcrClassificationVO );
    patIpcVO.source = getSource( ipcrClassificationVO );
    patIpcVO.geschichte = getStatus( ipcrClassificationVO );
    patIpcVO.generatingOffice = getOffice( ipcrClassificationVO );

    return patIpcVO;
  }

  /**
   * <B510EP> contains <classification-ipcr>
   * which in turns contains all the new tags for each part of the classification
   * OR it can be stored in a fixed length field of 50 positions within the <text> data element.
   * @param tag String
   * @throws Exception
   * @return IpcrClassificationVO
   */
  private IpcrClassificationVO extractClassificationVO( Object element ) throws Exception
  {
    IpcrClassificationVO vo = new IpcrClassificationVO();

    vo.section = xml.getValueOfElement( element, "section" );
    vo.ipcclass = xml.getValueOfElement( element, "class" );
    vo.subclass = xml.getValueOfElement( element, "subclass" );
    vo.maingroup = xml.getValueOfElement( element, "main-group" );
    vo.subgroup = xml.getValueOfElement( element, "subgroup" );
    vo.ipcVersionIndicator = xml.getValueOfElement( element, new String[]
                                                   {"ipc-version-indicator", "date" } ); //YYYYMMDD date format
    vo.classificationLevel = xml.getValueOfElement( element, "classification-level" );
    vo.symbolPosition = xml.getValueOfElement( element, "symbol-position" );
    vo.classificationValue = xml.getValueOfElement( element, "classification-value" );
    vo.actionDate = xml.getValueOfElement( element, new String[]
                                           {"action-date", "date"} ); //YYYYMMDD date format
    vo.classificationStatus = xml.getValueOfElement( element, "classification-status" );
    vo.classificationDataSource = xml.getValueOfElement( element, "classification-data-source" );
    vo.generatingOffice = xml.getValueOfElement( element, new String[]
                                                 {"generating-office", "country"} );

    return vo;
  }

  /**
   * Extrahiert den Wert eines IPC-Symbols aus dem XML
   * @param tag String XML-Element
   * @return String IPC-Symbol wie es in Bagis gespeichert wird
   */
  private String getIpcSymbol( IpcrClassificationVO vo )
  {
    IpcSymbol symbol = new IpcSymbol( vo.section, vo.ipcclass, vo.subclass );

    try
    {
      symbol.setGroupPart( vo.maingroup, vo.subgroup );
    }
    catch ( ParseException ex )
    {
      logFile.appendMessage( WpaLogFile.WARN, "PCT " + xml.getSchutitGesuchNr() +
                             " Ungueltiges Format fuer Group Part IPC-Klasse: <" + vo.maingroup +
                             "/" + vo.subgroup + ">" );
      LOG.error( ex.getMessage() );
    }

    return symbol.getIpcSymbol();
  }

  /**
   * Extrahiert den Code eines Ipc-Symbols aus dem XML-Element symbol-position
   * Gültige Werte sind F=FirstClass und L=LaterClass.
   * Folgende Konvertierungen werden vorgenommen, bevor die Werte in Bagis gespeichert werden
   *  - F wird umgewandelt in eine 2 (Hauptklasse)
   *  - L wird umgewandelt in eine 3 (Nebenklasse)
   * @param tag String XML-Element
   * @return int Code des Ipc-Symbols wie er in Bagis gespeichert wird
   */
  private int getPatIpcCd( IpcrClassificationVO vo )
  {
    int patIpcCd = 0;

    if ( vo.symbolPosition != null && vo.symbolPosition.equals( "F" ) )
    {
      patIpcCd = PatIpcVO.IPC_CD_H;
    }
    else if ( vo.symbolPosition != null && vo.symbolPosition.equals( "L" ) )
    {
      patIpcCd = PatIpcVO.IPC_CD_N;
    }
    else
    {
      patIpcCd = PatIpcVO.IPC_CD_H; //Defaultwert

      String ipcSymbol = getIpcSymbol( vo );

      String line = "PCT " + xml.getSchutitGesuchNr() + " Bei IPC Klasse <" + ipcSymbol +
          "> keine oder ungültige Symbol-Position vorhanden. " + patIpcCd + " eingefügt";
      logFile.appendMessage( WpaLogFile.DEBUG, line );
    }

    return patIpcCd;
  }

  /**
   * Extrahiert Attribut generatingOffice aus dem XML-Element generating-office
   * Bezeichnet von welchem Land die Klassifizierung für ein Patent gemacht wurde
   * @param tag String XML-Element
   * @return String Attribut generatingOffice
   */
  private String getOffice( IpcrClassificationVO vo )
  {
    String generatingOffice = vo.generatingOffice;

    if ( generatingOffice == null || generatingOffice.trim().equals( "" ) )
    {
      generatingOffice = new PatIpcVO().generatingOffice; //default

      String ipcSymbol = getIpcSymbol( vo );

      String line = "PCT " + xml.getSchutitGesuchNr() + " Bei IPC Klasse <" + ipcSymbol
          + "> kein Generating-Office vorhanden. " + generatingOffice + " eingefügt";
      logFile.appendMessage( WpaLogFile.DEBUG, line );
    }
    return generatingOffice;
  }

  /**
   * Extrahiert Attribut geschichte aus dem XML-Element classification-status
   * Gültige Werte sind
   *  - B (Basic)
   *  - R (Reclassified data)
   *  - V (various incidental changes)
   *  - D (data to be deleted)
   * @param tag String XML-Element
   * @return String Attribut geschichte
   */
  private String getStatus( IpcrClassificationVO vo )
  {
    String status = vo.classificationStatus;

    if ( status == null || status.trim().equals( "" ) )
    {
      status = new PatIpcVO().geschichte; //default

      String ipcSymbol = getIpcSymbol( vo );

      String line = "PCT " + xml.getSchutitGesuchNr() + " Bei IPC Klasse <" + ipcSymbol
          + "> kein Classification-Status vorhanden. "
          + status + " eingefügt";
      logFile.appendMessage( WpaLogFile.DEBUG, line );
    }
    return status;
  }

  /**
   * Extrahiert Attribut source aus dem XML-Element classification-data-source
   * Gültige Werte sind
   *  - H (Human)
   *  - M (Machine)
   *  - G (Generated by software)
   * @param tag String XML-Element
   * @return String Attribut source
   */
  private String getSource( IpcrClassificationVO vo )
  {
    String source = vo.classificationDataSource;

    if ( source == null || source.trim().equals( "" ) )
    {
      source = new PatIpcVO().source; //default

      String ipcSymbol = getIpcSymbol( vo );

      String line = "PCT " + xml.getSchutitGesuchNr() + " Bei IPC Klasse <" + ipcSymbol +
          "> keine Classification-Data-Source vorhanden. " +
          source + " eingefügt";
      logFile.appendMessage( WpaLogFile.DEBUG, line );
    }
    return source;
  }


  /**
   * Extrahiert Attribut erfassDat aus dem XML-Element action-date
   * Falls kein Action-Date vorhanden, wird das Erteilungsdatum/Publikationsdatum oder das Anmeldedatum
   * @param tag String XML-Element
   * @return String Attribut erfassDat
   */
  private Date getIpcGueltigVon( IpcrClassificationVO vo ) throws Exception
  {
    Date ipcGueltigVon = parseDate( vo.ipcVersionIndicator );

    if( ipcGueltigVon == null )
    {
      String ipcSymbol = getIpcSymbol( vo );

      String line = "PCT " + xml.getSchutitGesuchNr() + " Bei IPC Klasse <" + ipcSymbol
          + "> kein IPC-Version-Indicator oder ungültiges Datum vorhanden <" + vo.ipcVersionIndicator + " >" +
          " Jüngstes Versions-Datum wird als Gültig von eingesetzt";
      logFile.appendMessage( WpaLogFile.DEBUG, line );
    }
    return ipcGueltigVon;
  }


  /**
   * Extrahiert Attribut classLevel aus dem XML-Element classification-level
   * Gültige Werte sind
   *  - C (Core)
   *  - A (Advanced)
   *  - S (Subclass)
   * @param tag String XML-Element
   * @return String Attribut classLevel
   */
  private String getClassLevel( IpcrClassificationVO vo )
  {
    String classLevel = vo.classificationLevel;

    if ( classLevel == null || classLevel.trim().equals( "" ) )
    {
      classLevel = new PatIpcVO().classLevel; //default

      String ipcSymbol = getIpcSymbol( vo );

      String line = "PCT " + xml.getSchutitGesuchNr() + " Bei IPC Klasse <" + ipcSymbol
          + "> kein Class-Level vorhanden. " + classLevel + " eingefügt";
      logFile.appendMessage( WpaLogFile.DEBUG, line );
    }
    return classLevel;
  }

  /**
   * Extrahiert Attribut classValue aus dem XML-Element classification-value
   * Gültige Werte sind
   *  - I (inventive)
   *  - N (non-inventive)
   * @param tag String XML-Element
   * @return String Attribut classValue
   */
  private String getClassValue( IpcrClassificationVO vo )
  {
    String classValue = vo.classificationValue;

    if ( classValue == null || classValue.trim().equals( "" ) ||
         ! ( classValue.equals( PatIpcVO.CLASS_VALUE_INVENTIVE ) ||
             classValue.equals( PatIpcVO.CLASS_VALUE_NON_INVENTIVE ) ) )
    {
      classValue = new PatIpcVO().classValue; //default

      String ipcSymbol = getIpcSymbol( vo );

      String line = "PCT " + xml.getSchutitGesuchNr() + " Bei IPC Klasse <" + ipcSymbol +
          "> kein oder ungültiger Classification-Value vorhanden. " + classValue + " eingefügt";
      logFile.appendMessage( WpaLogFile.DEBUG, line );
    }
    return classValue;
  }

  /**
   * Extrahiert Attribut erfassDat aus dem XML-Element action-date
   * Falls kein Action-Date vorhanden, wird das Erteilungsdatum/Publikationsdatum oder das Anmeldedatum
   * @param tag String XML-Element
   * @return String Attribut erfassDat
   */
  private Date getErfassDat( IpcrClassificationVO vo ) throws Exception
  {
    Date erfassDat = parseDate( vo.actionDate );

    // Falls kein oder ungültiges Action-Date wird das Anmeldedatum aus application-reference DATE eingesetzt
    if ( erfassDat == null )
    {
      String actionDate = xml.getValueOf( "application-reference", "document-id/date" );
      erfassDat = parseDate( actionDate );

      if ( erfassDat == null )
      {
        // Ist weder Action-Date noch Anmeldedatum vorhanden wird Tagesdatum abgefüllt,
        // da laut Datenbankdefinition das Feld nicht leer sein darf
        erfassDat = new Date();

        String ipcSymbol = getIpcSymbol( vo );

        String line = "PCT " + xml.getSchutitGesuchNr() + " Bei IPC Klasse <" + ipcSymbol +
            "> kein oder ungültiges Anmeldedatum vorhanden. Tagesdatum eingefügt";

        logFile.appendMessage( WpaLogFile.DEBUG, line );
      }
    }
    return erfassDat;
  }

  /**
   * Wandelt das action Date im Format yyyyMMdd in ein gültiges Datum um
   * @param actionDate String
   * @return Date oder null falls Datum nicht geparsed werden kann
   */
  private Date parseDate( String actionDate )
  {
    try
    {
      return sdf.parse( actionDate );
    }
    catch ( Exception ex )
    {
      return null;
    }
  }
}
