package ch.ige.pctedi.client.wpa;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import ch.ige.pctedi.client.MappingUnicode;
import ch.ige.pctedi.client.XMLDoc;
import ch.ige.pctedi.vo.SchutitVO;

/**
 *
 * <p>
 * Copyright (c) 2005, Eidgenössisches Institut für Geistiges Eigentum
 * @author A. Rueegsegger
 * @version $Id: Schutit.java,v 1.14 2008/03/11 14:11:01 ruegsegger Exp $
 */

public class Schutit
{
  private SimpleDateFormat sdf = new SimpleDateFormat( "yyyyMMdd" );

  private WpaLogFile logFile = WpaLogFile.getInstance();

  private XMLDoc xml;

  public Schutit()
  {
    sdf.setLenient( false );
  }

  public SchutitVO extract( XMLDoc xmlDoc ) throws Exception
  {
    this.xml = xmlDoc;
    SchutitVO schutitVO = new SchutitVO();

    schutitVO.sprache = getPublicationLanguage();
    schutitVO.gesuchNr = getGesuchNr();
    schutitVO.schutitNr = getSchutitNr();
    schutitVO.beschreibung = getBeschreibung();
    if ( schutitVO.beschreibung != null )
    {
      schutitVO.uppBeschreibung = schutitVO.beschreibung.toUpperCase();
    }

    Date filingDate = getFillingDate();
    schutitVO.schutzBegin = filingDate;
    schutitVO.erstEreigDat = filingDate;
    schutitVO.schutzEnde = computeSchutzEnde( filingDate );

    // Fix 2300 (rho 18.12.2001)
    // Für PCT werden keine Jahresgebühren fällig,
    // deshalb wird ein Datum genommen, welches weit in der Zukunft liegt.
    // Die Zahl 2300 hat keine tiefere Bedeutung.
    schutitVO.faelligJahr = 2300;

    return schutitVO;
  }

  /**
   * language of publication - use ISO639 language code
   * Veröffentlichungssprachen vom OPMI sind deutsch, französisch und englisch
   * von daher sollten keine anderen Sprachen daherkommen.
   * @return int
   */
  private int getPublicationLanguage() throws Exception
  {
    String lang = xml.getAttribute( "publication-reference", "document-id", "lang" );
    if ( lang.equalsIgnoreCase( "de" ) )
    {
      return 1;
    }
    else if ( lang.equalsIgnoreCase( "fr" ) )
    {
      return 2;
    }
    else if ( lang.equalsIgnoreCase( "en" ) )
    {
      return 4;
    }
    else
    {
      // gemäss RHO 22.1.2001 keine Message nötig
      return 1;
    }
  }

  /**
   * Die Länge der Gesuch-Nummer ist 12 Stellen.
   * 1-2:  zweistelliger Ländercode
   * 3-6:  vierstellige Jahreszahl (falls die Jahreszahl nur zweistellig ist sie auf vier
   *       Stellen zu ergänzen. als Grenzwert wird 70 genommen, da PCT’s erst seit ca. 1978 in Kraft sind)
   * 7-12: sechsstellige Laufnummer (ist mit führenden Nullen zu ergänzen)
   * Automatisch durchgeführte Korrekturen sind nicht zu protokollieren.
   *
   * <doc-number>GB2004/004445</doc-number>
   * @return String
   */
  private String getGesuchNr() throws Exception
  {
    String gesuchNr = xml.getValueOf( "application-reference", "document-id/doc-number" );
    gesuchNr = gesuchNr.replaceAll( "/", "" );

    if ( gesuchNr.length() != 12 )
    {
      logFile.appendMessage( WpaLogFile.ERROR, "PCT " + xml.getSchutitGesuchNr() + " Ungueltige Anmeldenummer <" + gesuchNr + ">" );
    }
    return gesuchNr;
  }

  /**
   * <doc-number>2005/062225</doc-number>
   * @return String
   */
  private String getSchutitNr() throws Exception
  {
    String schutitNr = xml.getSchutitNr();

    if ( schutitNr.length() != 10 )
    {
      logFile.appendMessage( WpaLogFile.ERROR, "PCT " + xml.getSchutitGesuchNr() + " Ungueltige Publikationsnummer <" + schutitNr + ">" );
    }
    return schutitNr;
  }

  /**
   * Französicher Titel wird in der Beschreibung gespeichert
   * Da das Datenbankattribute auf 340 beschränkt ist, wird der Titel auf 340 Zeichen gekürzt
   * Der Titel sollte immer in den Sprachen französisch und englisch vorhanden sein in manchen Fällen sogar auf deutsch
   * Da es sich bei französisch um eine Amtssprache handelt, wird der französische Titel gelesen
   * @throws Exception
   * @return String
   */
  private String getBeschreibung() throws Exception
  {
    String beschreibung = xml.getValueOf( "invention-title[@lang='fr']" ); // "" falls nicht vorhanden

    if ( beschreibung != null && !beschreibung.equals( "" ) )
    {
      beschreibung = MappingUnicode.getInstance().replaceUnicodeValues( beschreibung );

      if ( beschreibung.length() > 340 )
      {
        logFile.appendMessage( WpaLogFile.WARN, "PCT " + xml.getSchutitGesuchNr() + " Titel: " + beschreibung + " auf 340 Zeichen gekürzt" );
        beschreibung = beschreibung.substring( 0, 340 );
      }
    }
    else
    {
      logFile.appendMessage( WpaLogFile.WARN, "PCT " + xml.getSchutitGesuchNr() + " Titel ist leer" );
      beschreibung = null;
    }
    return beschreibung;
  }

  private Date computeSchutzEnde( Date filingDate )
  {
    // Date-Objekt an die Klasse Calendar übergeben und 20 Jahre min 1 Tag rechnen
    GregorianCalendar cal = new GregorianCalendar();
    cal.setTime( filingDate );
    cal.add( Calendar.YEAR, 20 );
    cal.add( Calendar.DATE, -1 );

    // mit getTime(), also umwandlung cal
    return cal.getTime();
  }

  private Date getFillingDate() throws Exception
  {
    String fillingDate = null;
    try
    {
      fillingDate = xml.getValueOf( "application-reference", "document-id/date" );
      return sdf.parse( fillingDate );
    }
    catch ( Exception ex )
    {
      logFile.appendMessage( WpaLogFile.WARN, "PCT " + xml.getSchutitGesuchNr() + " Kein oder ungueltiges Anmeldedatum <" + fillingDate + "> vorhanden" );
      return null;
    }
  }
}
