package ch.ige.rbbagisexport.server.xml;

import java.io.IOException;
import java.io.Serializable;
import java.util.GregorianCalendar;
import java.util.List;

import javax.xml.bind.JAXBException;

import org.apache.log4j.Logger;

import ch.ige.chsearchreport.ApplicantNameType;
import ch.ige.chsearchreport.ApplicationReference;
import ch.ige.chsearchreport.ClassificationIpcr;
import ch.ige.chsearchreport.ClassificationsIpcr;
import ch.ige.chsearchreport.DateOfEarliestPriority;
import ch.ige.chsearchreport.DateSearchCompleted;
import ch.ige.chsearchreport.DocumentId;
import ch.ige.chsearchreport.Examiners;
import ch.ige.chsearchreport.InventionTitle;
import ch.ige.chsearchreport.MinimumDocumentation;
import ch.ige.chsearchreport.ObjectFactory;
import ch.ige.chsearchreport.PrimaryExaminer;
import ch.ige.chsearchreport.SearchReportData;
import ch.ige.chsearchreport.SrepAbstract;
import ch.ige.chsearchreport.SrepAdmin;
import ch.ige.chsearchreport.SrepCitations;
import ch.ige.chsearchreport.SrepFieldsSearched;
import ch.ige.chsearchreport.SrepFigureToPublish;
import ch.ige.chsearchreport.SrepForPub;
import ch.ige.chsearchreport.SrepInfo;
import ch.ige.chsearchreport.SrepInfoAdmin;
import ch.ige.chsearchreport.SrepInventionTitle;
import ch.ige.chsearchreport.impl.ApplicantNameTypeImpl;
import ch.ige.rbbagisexport.server.vo.RBExportVO;
import ch.ige.xmlutil.bind.Bindings;
import ch.ige.xmlutil.doc.DocumentTypes;
import ch.ige.xmlutil.bind.Bindings;

/**
 * Helper Class to convert a RBExportVO into its XML representation (Schema ch-search-report). <p>Copyright (c) 2007,
 * Swiss Federal Institute of Intellectual Property</p>
 * @author kuenzli
 * @version $Id: ChSearchReportXMLConverter.java,v 1.12 2008/10/03 06:06:07 kuenzli Exp $
 */
/**
 * @author kuenzli
 *
 */
public class ChSearchReportXMLConverter implements Serializable
{
  //---------------------------------------------------------------------------- 
  // Constructors
  //---------------------------------------------------------------------------- 
  public ChSearchReportXMLConverter()
  {

  }

  //---------------------------------------------------------------------------- 
  // Privat Members
  //---------------------------------------------------------------------------- 
  // JAXB
  private transient ObjectFactory of = null;

  private static final Logger LOG = Logger.getLogger( ChSearchReportXMLConverter.class );

  private static final String APPLICANT_NAMES_DELIMITER = "|";

  private static final String SREP_OFFICE_COUNTRY = "CH";

  private static final String APPLICATION_COUNTRY = "CH";

  private static final String SCHEMA_NAME = DocumentTypes.CHSEARCHREPORT1_2_IGE1_1.getPhysicalSchemaLocation();

  private static final long serialVersionUID = 1L;

  //---------------------------------------------------------------------------- 
  // Public Methods
  //---------------------------------------------------------------------------- 
  /**
   * Converts a RBExportVO into its XML representation. The XML adheres to the
   * Schema ch-search-report.
   * @param vo The RBExportVO
   * @return The xml as a String
   * @throws JAXBException
   */
  public String convertToXml( RBExportVO vo ) throws JAXBException
  {
    // *** fill JAXB object
    // create Root Element <search-report-data>
    SearchReportData searchReportData = objFactory().createSearchReportData();
    // -------------
    // write vo data
    // -------------
    // <search-report-data lang="de"> for example
    writeLanguage( searchReportData, vo );
    // <search-report-data><srep-info>
    writeSrepInfo( searchReportData, vo );
    // <search-report-data><srep-for-pub>
    writeSrepForPub( searchReportData, vo );
    // --------------------
    // write mandatory data
    // --------------------
    // <search-report-data srep-office="CH"> for example
    searchReportData.setSrepOffice( SREP_OFFICE_COUNTRY );
    // <search-report-data date-produced="2007-09-10">  for example
    searchReportData.setDateProduced( new java.sql.Date( new GregorianCalendar().getTimeInMillis() ) );
    // *** convert JAXB object to xml
    String xml = transformToXml( searchReportData );
    return xml;
  }

  //---------------------------------------------------------------------------- 
  // Private Methods
  //---------------------------------------------------------------------------- 
  /**
   * Puts language into attribute lang of element <search-report-data>
   * @param searchReportData The Search Report Data JAXB root object.
   * @param vo The Search Report Value Object.
   */
  private void writeLanguage( SearchReportData searchReportData, RBExportVO vo )
  {
    if ( vo.getLanguage() == null )
    {
      return;
    }
    String lang = null;
    switch( vo.getLanguage() )
    {
      case 1:
        lang = "de";
        break;
      case 2:
        lang = "fr";
        break;
      case 3:
        lang = "it";
        break;
      case 4:
        lang = "en";
        break;
      default:
        throw new IllegalArgumentException( "Language code >" + vo.getLanguage().toString() + "< could not be converted." );
    }
    searchReportData.setLang( lang );
  }

  /**
   * Writes the element &lt;srep-info> into &lt;search-report-data>.
   * @param srData JAXB class representing &lt;search-report-data>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  private void writeSrepInfo( SearchReportData srData, RBExportVO vo ) throws JAXBException
  {
    SrepInfo srepInfo = objFactory().createSrepInfo();
    // -------------
    // write vo data
    // -------------
    // <srep-info><application-reference><document-id><date>
    writeApplicationReference( srepInfo, vo );
    // <srep-info><date-of-earliest-priority><date>
    writePriorityDate( srepInfo, vo );
    // <srep-info><applicant-name><name>
    writeApplicantName( srepInfo, vo );
    // <srep-info><srep-invention-title><invention-title>
    writeTitle( srepInfo, vo );
    // --------------------
    // write mandatory data
    // --------------------
    // <srep-info><application-reference><document-id><srep-abstract>
    SrepAbstract srepAbstract = objFactory().createSrepAbstract();
    srepInfo.setSrepAbstract( srepAbstract );
    // <srep-info><application-reference><document-id><srep-figure-to-publish>
    SrepFigureToPublish srepFigureToPublish = objFactory().createSrepFigureToPublish();
    srepFigureToPublish.setFiginfo( 1 ); // some arbitrary number 
    srepInfo.setSrepFigureToPublish( srepFigureToPublish );
    // <srep-info><application-reference><document-id><srep-info-admin/>
    SrepInfoAdmin srepInfoAdmin = objFactory().createSrepInfoAdmin();
    srepInfo.setSrepInfoAdmin( srepInfoAdmin );
    // *** append <srep-info> to <search-report-data>
    srData.setSrepInfo( srepInfo );
  }

  /**
   * Writes the element &lt;srep-for-pub> into &lt;search-report-data>.
   * @param srData JAXB class representing &lt;search-report-data>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  private void writeSrepForPub( SearchReportData srData, RBExportVO vo ) throws JAXBException
  {
    SrepForPub srepForPub = objFactory().createSrepForPub();
    // -------------
    // write vo data
    // -------------
    // <srep-for-pub><classifications-ipcr>
    writeIpcClassifications( srepForPub, vo );
    // <srep-for-pub><srep-admin>
    writeSrepAdmin( srepForPub, vo );
    // --------------------
    // write mandatory data
    // --------------------
    // <srep-for-pub><srep-fields-searched><minimum-documentation> 
    SrepFieldsSearched srepFieldsSearched = objFactory().createSrepFieldsSearched();
    MinimumDocumentation minDoc = objFactory().createMinimumDocumentation();
    srepFieldsSearched.setMinimumDocumentation( minDoc );
    srepForPub.setSrepFieldsSearched( srepFieldsSearched );
    // <srep-for-pub><srep-citations><text> 
    SrepCitations srepCitations = objFactory().createSrepCitations();
    srepCitations.setText( "" );
    srepForPub.setSrepCitations( srepCitations );
    // *** append <srep-for-pub> to <search-report-data>
    srData.setSrepForPub( srepForPub );
  }

  /**
   * Writes the element &lt;srep-admin> into &lt;srep-for-pub>.
   * @param srepForPub JAXB class representing &lt;srep-for-pub>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  private void writeSrepAdmin( SrepForPub srepForPub, RBExportVO vo ) throws JAXBException
  {
    if ( vo.getInvestigator() != null && vo.getInvestigator() != "" )
    {
      // -------------
      // write vo data
      // -------------
      // <srep-admin><examiners><primary-examiner>
      SrepAdmin srepAdmin = objFactory().createSrepAdmin();
      writePrimaryExaminer( srepAdmin, vo );
      // --------------------
      // write mandatory data
      // --------------------
      // <srep-admin><date-search-completed>
      DateSearchCompleted dateSearchCompleted = objFactory().createDateSearchCompleted();
      dateSearchCompleted.setDate( new java.sql.Date( ( new GregorianCalendar().getTime().getTime() ) ) );
      srepAdmin.setDateSearchCompleted( dateSearchCompleted );
      // *** append <srep-admin> to <srep-for-pub>
      srepForPub.setSrepAdmin( srepAdmin );
    }
  }

  /**
   * Writes the element &lt;application-reference> into &lt;srep-info>.
   * @param srepInfo JAXB class representing &lt;srep-info>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  @SuppressWarnings( "unchecked" )
  private void writeApplicationReference( SrepInfo srepInfo, RBExportVO vo ) throws JAXBException
  {
    // -------------
    // write vo data
    // -------------
    // <srep-info><application-reference><document-id><date>
    ApplicationReference applicationReference = objFactory().createApplicationReference();
    DocumentId documentId = objFactory().createDocumentId();
    documentId.setDate( new java.sql.Date( vo.getApplicationDate().getTime() ) );
    // <srep-info><application-reference><document-id><doc-number>
    documentId.setDocNumber( vo.getAppNr().substring( 0, vo.getAppNr().length() - 1 ) );
    // --------------------
    // write mandatory data
    // --------------------
    // <srep-info><application-reference><document-id><country>
    documentId.setCountry( APPLICATION_COUNTRY );
    applicationReference.setDocumentId( documentId );
    // *** append <application-reference> to <srep-info>
    // info: add is unchecked
    srepInfo.getApplicationReference().add( applicationReference );
  }

  /**
   * Writes the element &lt;applicant-name> into &lt;srep-info>.
   * @param srepInfo JAXB class representing &lt;srep-info>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  private void writeApplicantName( SrepInfo srepInfo, RBExportVO vo ) throws JAXBException
  {
    ApplicantNameType applicantName = new ApplicantNameTypeImpl();
    // -------------
    // write vo data
    // -------------
    // <srep-info><applicant-name>
    StringBuffer applicantNames = new StringBuffer();
    if ( vo.getApplicantNames() != null && !vo.getApplicantNames().isEmpty() )
    {
      int i = 0;
      for ( String applName: vo.getApplicantNames() )
      {
        i++;
        if ( i != 1 )
        {
          applicantNames = applicantNames.append( APPLICANT_NAMES_DELIMITER );
        }
        applicantNames = applicantNames.append( applName );
      }
    }

    applicantName.setName( applicantNames.toString() );
    // *** append <applicant-name> to <srep-info>
    srepInfo.setApplicantName( applicantName );
  }

  /**
   * Writes the element &lt;date-of-earliest-priority> into &lt;srep-info>.
   * @param srepInfo JAXB class representing &lt;srep-info>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  private void writePriorityDate( SrepInfo srepInfo, RBExportVO vo ) throws JAXBException
  {
    if ( vo.getPriorityDate() == null )
    {
      return;
    }
    DateOfEarliestPriority dateOfEarliestPriority = objFactory().createDateOfEarliestPriority();
    // -------------
    // write vo data
    // -------------
    // <srep-info><date-of-earliest-priority>
    dateOfEarliestPriority.setDate( new java.sql.Date( vo.getPriorityDate().getTime() ) );
    // *** append <date-of-earliest-priority> to <srep-info>
    srepInfo.setDateOfEarliestPriority( dateOfEarliestPriority );
  }

  /**
   * Writes the element &lt;invention-title> into &lt;srep-info>.
   * @param srepInfo JAXB class representing &lt;srep-info>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  @SuppressWarnings( "unchecked" )
  private void writeTitle( SrepInfo srepInfo, RBExportVO vo ) throws JAXBException
  {
    SrepInventionTitle srepInventionTitle = objFactory().createSrepInventionTitle();
    srepInfo.setSrepInventionTitle( srepInventionTitle );
    if ( vo.getTitle() == null || vo.getTitle() == "" )
    {
      return;
    }

    InventionTitle inventTitle = objFactory().createInventionTitle();
    List inventTitleList = inventTitle.getContent();
    inventTitleList.add( vo.getTitle() ); // info: add is unchecked!
    //  <srep-info><application-reference><document-id><invention-title>
    srepInventionTitle.setInventionTitle( inventTitle );
  }

  /**
   * Writes the element &lt;classifications-ipcr> into &lt;srep-for-pub>.
   * @param srepForPub JAXB class representing &lt;srep-for-pub>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  @SuppressWarnings( "unchecked" )
  private void writeIpcClassifications( SrepForPub srepForPub, RBExportVO vo ) throws JAXBException
  {
    if ( vo.getIpcClasses() == null || vo.getIpcClasses().isEmpty() )
    {
      return;
    }

    ClassificationsIpcr classificationsIpcr = objFactory().createClassificationsIpcr();
    for ( String ipcClass: vo.getIpcClasses() )
    {
      ClassificationIpcr classificationIpcr = objFactory().createClassificationIpcr();
      classificationIpcr.setText( ipcClass.trim() );
      classificationsIpcr.getClassificationIpcr().add( classificationIpcr );// info: add is unchecked!
    }

    srepForPub.setClassificationsIpcr( classificationsIpcr );
  }

  /**
   * Writes the element &lt;primary-examiner> into &lt;srep-admin>.
   * @param srepAdmin JAXB class representing &lt;srep-admin>.
   * @param vo The Search Report Value Object.
   * @throws JAXBException 
   */
  private void writePrimaryExaminer( SrepAdmin srepAdmin, RBExportVO vo ) throws JAXBException
  {
    PrimaryExaminer primExaminer = objFactory().createPrimaryExaminer();
    primExaminer.setName( vo.getInvestigator() );
    Examiners examiners = objFactory().createExaminers();
    examiners.setPrimaryExaminer( primExaminer );
    srepAdmin.setExaminers( examiners );
  }

  //---------------------------------------------------------------------------- 
  // Private helper classes
  //---------------------------------------------------------------------------- 
  /**
   * Transform passed JAXB object to an XML String.
   * @param srData The JAXB root object.
   * @return The XML String.
   * @throws JAXBException
   */
  private String transformToXml( SearchReportData srData ) throws JAXBException
  {
//    JAXBTransformer transformer = new JAXBTransformer( "ch.ige.chsearchreport", "SearchReportData", SCHEMA_NAME );
    JAXBTransformer transformer = new JAXBTransformer( "ch.ige.chsearchreport", "SearchReportData", Bindings.CHSEARCHREPORT.getPhysicalSchemaLocation() );
    

    String xml;
    try
    {
      xml = transformer.transformToXml( srData );
    }
    catch ( IOException e )
    {
      LOG.error( "transformToXml(): IOException" );
      throw new JAXBException( e );
    }
    catch ( ClassNotFoundException e )
    {
      LOG.error( "transformToXml(): Class not found!" );
      throw new JAXBException( e );
    }
    return xml;
  }

  /**
   * Get JAXB object factory for Schema CHSEARCHREPORT. 
   * Create a object factory if its not yet available (). 
   * @return
   */
  private ObjectFactory objFactory()
  {
    if ( this.of == null )
    {
      this.of = new ch.ige.chsearchreport.ObjectFactory();
    }
    return this.of;
  }
}
