package ch.ige.st36xmlgenerator;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import ch.ige.ipobjects.patent.PatentDocument;
import ch.ige.ipobjects.time.BAGISTime;
import ch.ige.xmlutil.doc.XMLDocument;

/**
 * A container class for the XML Data and all required metadata to store it as a file
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   fabrizioli
 * @version  $Id: XMLDocumentWithMetadata.java,v 1.8 2008/09/25 07:29:27 fabrizioli Exp $
 */
public class XMLDocumentWithMetadata implements Serializable
{
  private Long schutitNr;

  private String lang;

  private XMLDocument document;

  private String filename;

  private String kindCode;

  private int eventTime;

  private static final String DEFAULT_EXTENSION = ".xml";

  private String errorMsg = "";

  /**
   * Constructor for doc with metadata
   * @param document
   * @param schutitNr
   * @param lang
   * @param kindCode
   * @param eventTime
   */
  public XMLDocumentWithMetadata( XMLDocument document, Long schutitNr, String lang, String kindCode, int eventTime )
  {
    this.document = document;
    this.lang = lang;
    this.schutitNr = schutitNr;
    this.kindCode = kindCode;
    this.eventTime = eventTime;
  }

  /**
   * Gets the generated XMLDocument.
   */
  public XMLDocument getDocument()
  {
    return document;
  }

  /**
   * Get the error message, associated to the generation of the XMLDocument
   */
  public String getErrorMsg()
  {
    return this.errorMsg;
  }

  /**
   * Get the filename, with which the XMLDocuemnt need to be saved (EPO)
   */
  protected String getFilename()
  {
    if ( filename == null )
    {
      StringBuilder name = new StringBuilder( "CH" );
      name.append( StringUtils.leftPad( String.valueOf( schutitNr ), 8, '0' ) );
      name.append( "_" );
      name.append( StringUtils.remove( BAGISTime.formatIDatISO8601( eventTime ), "-" ) );
      name.append( "_" );
      name.append( kindCode );
      name.append( "_" );
      name.append( lang.charAt( 0 ) );
      name.append( "0" );
      name.append( DEFAULT_EXTENSION );
      filename = name.toString();
    }
    return filename;
  }

  /**
   * Get the language of the XMLDocuemnt
   */
  public String getLanguage()
  {
    return this.lang;
  }

  /**
   * Get the register number of the XMLDocuemnt
   */
  public Long getRegisterNumber()
  {
    return this.schutitNr;
  }
  
  /**
   * Set the error message associated to the generation of the XMLDocument
   */
  public void setErrorMsg( String msg )
  {
    this.errorMsg = msg;
  }

  public String toString()
  {
    return ToStringBuilder.reflectionToString( this );
  }

  // @see ch.ige.ipobjects.patent.PatentDocument#cloneWithNewContent(java.lang.String)
  public PatentDocument cloneWithNewContent( String xml )
  {
    throw new UnsupportedOperationException( "Clone with new content not implemented" );
  }

  // @see ch.ige.ipobjects.patent.PatentDocument#save(java.io.File)
  public void save( String folder ) throws IOException
  {
    File folderForXML = new File( folder, getFolderForNameForXML() );
    if ( !folderForXML.exists() )
    {
      if ( !folderForXML.mkdirs() )
      {
        throw new IOException( "The folder could not be created " + folderForXML );
      }
    }
    saveXML( folderForXML, getFilename() );
  }

  private void saveXML( File folder, String filename ) throws IOException
  {
    if ( document != null )
    {
      document.saveToUTF8File( new File( folder, filename ) );
    }
    else
    {
      throw new IllegalArgumentException( "Document is null and cannot be saved" );
    }
  }

  private String getFolderForNameForXML()
  {
    return this.getFilename().substring( 0, this.getFilename().indexOf( "_" ) );
  }

  public String getFilePath()
  {
    return FilenameUtils.concat( getFolderForNameForXML(), getFilename() );
  }

}
