package ch.ige.st36xmlgenerator.batch;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.rmi.RemoteException;
import java.sql.SQLException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;

import javax.ejb.CreateException;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.Options;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.filefilter.FileFilterUtils;
import org.apache.log4j.Logger;

import ch.ige.ipobjects.common.DetailTerminatedException;
import ch.ige.ipobjects.event.vo.EreignisVO;
import ch.ige.ipobjects.time.BAGISTime;
import ch.ige.ipobjects.time.vo.EreigIDat;
import ch.ige.st36xmlgenerator.BuildInfo;
import ch.ige.st36xmlgenerator.InvalidPatentDocumentException;
import ch.ige.st36xmlgenerator.ST36BiblGenerator;
import ch.ige.st36xmlgenerator.ST36BiblGeneratorFactory;
import ch.ige.st36xmlgenerator.ST36BiblGeneratorProperties;
import ch.ige.st36xmlgenerator.XMLDocumentWithMetadata;
import ch.ige.st36xmlgenerator.exception.IllegalBatchArgumentException;
import ch.ige.st36xmlgenerator.exception.InconsistentEventException;
import ch.ige.st36xmlgenerator.exception.ST36GenerationRuntimeException;
import ch.ige.st36xmlgenerator.exception.ZipCreationException;
import ch.ige.util.client.batch.BatchException;
import ch.ige.util.client.batch.BatchReport;
import ch.ige.util.client.batch.BatchReportMailWithAttachments;
import ch.ige.util.client.batch.BatchReturnCode;
import ch.ige.util.client.batch.NewAbstractBatch;
import ch.ige.util.client.batch.BatchReportMailWithAttachments.MailAttachmentProvider;
import ch.ige.util.exception.ServiceLocatorException;
import ch.ige.xmlutil.ISO8601Util;

/**
 * Main Class for ST36XMLGenerator
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   fabrizioli
 * @version  $Id: ST36XMLGeneratorBatch.java,v 1.18 2008/10/28 16:04:24 fabrizioli Exp $
 */
public class ST36XMLGeneratorBatch extends NewAbstractBatch
{

  // constants
  protected static final Logger LOG = Logger.getLogger( ST36XMLGeneratorBatch.class );

  private static final String JOB_NAME = "ST36XMLGenerator";

  private static final String HELP_STRING = "Usage:  " + ST36XMLGeneratorBatch.class.getName() + " <max publ_date> ";

  private static final String MAIL_SUBJECT = ch.ige.util.BuildInfo.BUILD_ENV_NAME + " - Generierung der Patentschriften-Titelseiten";

  private static final String OLD_ZIP_MSG = "Eine Zip-Datei aus einem früheren Lauf wurde in die Ablage kopiert ";

  private static final String MSG_NO_DATA_GENERATED = "Es wurden keine Patentschriften-Titelseiten generiert.";

  private static final String MSG_START_EREIG_IDAT = "Suche startete bei Ereignis: ";

  private static final String MSG_END_EREIG_IDAT = "Suche endete bei Ereignis: ";

  private static final String FOLDER_MARKUP = "st36";

  protected File zipFile;

  public static String TEST_PROPERTY = "test.mail.file";

  /**
   * Used when the batch run fails.
   */
  static final BatchReturnCode ST36XMLGENERATOR_FAILED = new BatchReturnCode( 1, "Die Patentschrift - Titelseiten konnten nicht generiert werden." );

  /**
   * Used when the event query is not ok.
   */
  static final BatchReturnCode ST36XMLGENERATOR_EVENTSQUERY_FAILED = new BatchReturnCode( 2,
      "Fehler bei der Generierung der Patentschrift - Titelseiten. Die Ereignisse konnten nicht in BAGIS gefunden werden" );

  /**
   * Used when the data cannot be written to batch server.
   */
  static final BatchReturnCode ST36XMLGENERATOR_ACCESS_BATCH_FAILURE = new BatchReturnCode( 3,
      "Fehler bei der Generierung der Patentschrift - Titelseiten. Die generierten Daten konnten nicht auf die Ablage geschrieben werden." );

  /**
   * Used when the evaluation of BAGIS data is erroneous.
   */
  static final BatchReturnCode ST36XMLGENERATOR_UNEXPECTED_GENERATION_FAILURE = new BatchReturnCode( 4,
      "Fehler bei der Generierung der Patentschrift - Titelseiten. Ein unerwarteter Fehler ist bei der Generierung aufgetreten." );

  /**
   * Used when an sql Exception occurred.
   */
  static final BatchReturnCode ST36XMLGENERATOR_DATABASEACCESS_FAILED = new BatchReturnCode( 5,
      "Fehler bei der Generierung der Patentschrift - Titelseiten. Ein Fehler ist beim Zugriff auf die Datenbank aufgetreten." );

  /**
   * Used when an xml file cannot be validated.
   */
  static final BatchReturnCode ST36XMLGENERATOR_VALIDATION_FAILED = new BatchReturnCode( 6,
      "Fehler bei der Generierung der Patentschrift - Titelseiten. Eine XML-Datei konnte nicht validiert werden." );

  /**
   * Used when the zip file cannot be generated.
   */
  static final BatchReturnCode ST36XMLGENERATOR_ZIPCREATION_FAILURE = new BatchReturnCode( 7,
      "Fehler bei der Generierung der Patentschrift - Titelseiten. Die Zip-Datei konnte nicht erzeugt werden." );

  /**
   * Used when the generation folder exists before starting.
   */
  static final BatchReturnCode ST36XMLGENERATOR_GENERATIONFOLDEREXISTS_FAILURE = new BatchReturnCode( 8,
      "Es bestehen Daten aus früheren Ablauf. Bitte informieren Sie den Userdesk" );

  /**
   * Used when the generation folder exists before starting.
   */
  static final BatchReturnCode ST36XMLGENERATOR_ZIPFILECANNOTBEMOVED_FAILURE = new BatchReturnCode( 9,
      "Es bestehen Daten aus früheren Ablauf, die nocht kopiert werden können. Bitte informieren Sie den Userdesk" );

  /**
   * Used when the initialization failed.
   */
  static final BatchReturnCode ST36XMLGENERATOR_INIT_FAILURE = new BatchReturnCode( 10,
      "Fehler bei der Initialisierung der Generierung der Patentschrift - Titelseiten." );

  /**
   * Used when the date is not in the correct range.
   */
  private static final BatchReturnCode ST36XMLGENERATOR_ILLEGALDATE_EXCEPTION = new BatchReturnCode( 11, "Das Datum der Eingabe ist nicht korrekt" );

//  private static boolean test = false;
//
//  static
//  {
//    if ( test )
//    {
//      String path = "C:\\Temp\\" + JOB_NAME + "\\test" + System.currentTimeMillis() + ".txt";
//      System.setProperty( TEST_PROPERTY, path );
//    }
//  }

  private static final String REPORT_DATE_FORMAT = "dd.MM.yyyy";

  private static final String ZIP_PREFIX = "ST36_BIBL_";

  private ST36BiblGenerator generator;

  private ST36BiblGeneratorProperties configuration;

  /**
   * constructor
   */
  public ST36XMLGeneratorBatch()
  {
    super( JOB_NAME, new Options().addOption( "h", false, HELP_STRING ), true );
    init();
  }

  /**
   * constructor for tests only
   */
  public ST36XMLGeneratorBatch( boolean exit )
  {
    super( JOB_NAME, new Options().addOption( "h", false, HELP_STRING ), exit );
    init();
  }

  private void init() throws BatchException
  {
    LOG.debug( "init - begin" );

    try
    {
      generator = createGenerator();
      // configuration from datatabse for the client
      configuration = generator.getConfiguration();
      LOG.debug( "init - done " + configuration );
    }
    catch ( Exception e )
    {
      LOG.error( ST36XMLGENERATOR_INIT_FAILURE, e );
      throw new BatchException( ST36XMLGENERATOR_INIT_FAILURE, e );
    }

  }

  private ST36BiblGenerator createGenerator() throws RemoteException, CreateException, ServiceLocatorException
  {
    return ST36BiblGeneratorFactory.getEJB();

  }

  private void execute( Date maxPublDate ) throws ST36GenerationRuntimeException, IOException
  {
    configuration.setEndEreigidat( EreigIDat.max( maxPublDate ).getCompareTime() );
    configuration.validate();
    LOG.debug( configuration );
    writeInitializationInformation();

    Map<Integer, List<EreignisVO>> eventsPerSchutitKey = null;
    Map<Integer, List<XMLDocumentWithMetadata>> documentsPerSchutitKey = null;

    eventsPerSchutitKey = generator.getEventsToPublicate( configuration );

    documentsPerSchutitKey = generator.generateBibliographicData( eventsPerSchutitKey, configuration );

    // reload,  now new start date is written to DB
    configuration = generator.getConfiguration();
    LOG.debug( configuration );

    String zipname = ZIP_PREFIX + BAGISTime.formatForDebug( configuration.getStartPoint() ) + "_generiert_"
        + ISO8601Util.toTimeString( new Date() ).replace( ":", "_" ) + ".zip";
    this.zipFile = new File( configuration.getDataStoreLocationZip(), zipname );

    if ( !documentsPerSchutitKey.isEmpty() )
    {
      if ( !configuration.getDataStoreLocation().contains( FOLDER_MARKUP ) )
      {
        LOG.error( configuration.getDataStoreLocation() + " does nont contain the keyword st36xml: no zip created " );
      }
      else
      {
        saveDocuments( documentsPerSchutitKey.values(), configuration.getDataStoreLocation(), FilenameUtils.concat( configuration.getDataStoreLocationZip(),
            zipname ) );
      }
      int count = 0;
      for ( List<XMLDocumentWithMetadata> doclists: documentsPerSchutitKey.values() )
      {
        count += doclists.size();
      }
      this.writeBatchReportLine( getBody( documentsPerSchutitKey.values(), FilenameUtils.concat( configuration.getDataStoreLocationProcessed(), zipname ) ) );
      // sendInformationEmail( documentsPerSchutitKey.values(), FilenameUtils.concat( configuration.getDataStoreLocationProcessed(), zipname ) );
    }
    if ( this.zipFile == null || !this.zipFile.exists() )
    {
      this.writeBatchReportLine( MSG_NO_DATA_GENERATED );
    }
  }

  private Throwable getOriginatingException( Throwable e )
  {
    assert ( e != null );
    Throwable t = e.getCause();
    if ( t == null )
    {
      LOG.warn( e.getMessage() );
      return e;
    }
    while ( !( t.getCause() == null || instanceOfKnownServerExceptions( t ) ) )
    {
      t = t.getCause();
    }
    LOG.debug( "[TRACE] Exception from server of type:" + t.getClass().getName() );
    return t;

  }

  protected boolean instanceOfKnownServerExceptions( Throwable t )
  {
    return ( t instanceof InconsistentEventException || t instanceof SQLException || t instanceof InvalidPatentDocumentException
        || t instanceof DetailTerminatedException || t instanceof ZipCreationException );
  }

  private void moveZipToProcessedFolder( File source ) throws IOException
  {
    if ( source == null || !source.exists() )
    {
      LOG.info( "No zip available to copy" );
      return;
    }
    FileUtils.copyFile( source, new File( configuration.getDataStoreLocationProcessed(), source.getName() ) );
    FileUtils.forceDelete( source );

  }

  private void saveDocuments( Collection<List<XMLDocumentWithMetadata>> documents, String dataStoreLocation, String zipLocation ) throws IOException,
      ST36GenerationRuntimeException
  {
    for ( List<XMLDocumentWithMetadata> doclist: documents )
    {
      for ( XMLDocumentWithMetadata doc: doclist )
      {
        if ( doc.getDocument() != null )
        {
          doc.save( dataStoreLocation );
        }
        else if ( doc.getErrorMsg() != null && doc.getErrorMsg().length() > 0 )
        {
          LOG.warn( "Error message exists for doc " + doc );
        }
      }
    }
    File folderToZip = new File( dataStoreLocation );
    if ( folderToZip == null || !folderToZip.exists() )
    {
      return;// throw new ST36GenerationException("No data were not created");
    }
    ZipCreator.createZip( zipLocation, folderToZip, null );
    FileUtils.deleteDirectory( folderToZip );
  }

  private String getBody( Collection<List<XMLDocumentWithMetadata>> documents, String dataStoreLocation )
  {
    return new StatistikGenerator().generateStatics( documents, dataStoreLocation );
  }

  /**
   * Hauptverarbeitung:
   *
   * @param args
   */
  public static void main( String[] args )
  {
    LOG.debug( "Starting st36xmlgenerator batch " );
    if ( args == null || args.length != 1 )
    {
      LOG.info( HELP_STRING );
      throw new IllegalArgumentException( HELP_STRING );
    }

    ST36XMLGeneratorBatch batch = new ST36XMLGeneratorBatch();
    NewAbstractBatch.start( batch, args );
    // not executed when System.exit is called

  }

  @Override
  public BatchReport[] getBatchReports() throws BatchException
  {
    try
    {
      configuration = createGenerator().getConfiguration();
    }
    catch ( Exception e )
    {
      throw new ST36GenerationRuntimeException( "Error in looking up the configuration " + e.getMessage(), e );
    }

    return new BatchReport[] { new BatchReportMailWithAttachments( configuration.getSMTPServer(), configuration.getMailSender(), configuration
        .getMailDestination(), configuration.getMailCC(), MAIL_SUBJECT, new ST36XMLGneratorAttachmentProvider() ) };
  }

  @Override
  protected BatchReturnCode[] getPossibleReturnCodes()
  {
    return BatchReturnCode.getBatchReturnCodes( ST36XMLGeneratorBatch.class );
  }

  @Override
  protected void checkCommandLine( CommandLine commandLine, Map<String, Object> parsedCommandLine ) throws IllegalArgumentException
  {
    String[] args = commandLine.getArgs();
    if ( args == null || args.length != 1 )
    {
      throw new IllegalArgumentException( "Missing parameter end date: " + HELP_STRING );
    }

  }

  @Override
  protected void run( CommandLine commandLine, Map<String, Object> parsedCommandLine ) throws BatchException
  {
    LOG.debug( "Checking for existing zip files " );

    this.checkForExsitingZipFiles();
    String[] args = commandLine.getArgs();
    String publDateString = args[0];
    Date publDate = null;
    try
    {
      publDate = new SimpleDateFormat( REPORT_DATE_FORMAT ).parse( publDateString );
    }
    catch ( ParseException e )
    {
      LOG.error( "Error in parsing the date " + publDateString, e );
      throw new BatchException( ST36XMLGENERATOR_INIT_FAILURE );
    }
    if ( publDate == null )
    {
      throw new IllegalArgumentException( "Argument is not a valid date " + publDateString );
    }

    LOG.info( "Executing st36xmlgenerator batch for " + publDateString );

    try
    {
      execute( publDate );
    }
    catch ( RemoteException e )
    {
      Throwable t = this.getOriginatingException( e );
      if ( t.getClass().equals( SQLException.class ) )
      {
        throw new BatchException( ST36XMLGENERATOR_DATABASEACCESS_FAILED, e );
      }
      if ( t.getClass().equals( InvalidPatentDocumentException.class ) )
      {
        throw new BatchException( ST36XMLGENERATOR_VALIDATION_FAILED, e );
      }
      throw new BatchException( ST36XMLGENERATOR_UNEXPECTED_GENERATION_FAILURE, e );
    }
    catch ( IOException e )
    {
      throw new BatchException( ST36XMLGENERATOR_ACCESS_BATCH_FAILURE, e );
    }
    catch ( ZipCreationException e )
    {
      // Zip Erzeugung fehlgeschlagen 
      throw new BatchException( ST36XMLGENERATOR_ZIPCREATION_FAILURE, e );
    }
    catch ( IllegalBatchArgumentException e )
    {
      throw new BatchException( ST36XMLGENERATOR_ILLEGALDATE_EXCEPTION, e );
    }
    catch ( Exception e )
    {
      LOG.error( ST36XMLGENERATOR_UNEXPECTED_GENERATION_FAILURE, e );
      throw new BatchException( ST36XMLGENERATOR_UNEXPECTED_GENERATION_FAILURE, e );
    }

  }

  // @see ch.ige.util.client.batch.NewAbstractBatch#getBuildInfo()
  @Override
  protected String getBuildInfo()
  {
    return BuildInfo.BUILD_ID;
  }

  // @see ch.ige.util.client.batch.NewAbstractBatch#getCvsTag()
  @Override
  protected String getCvsTag()
  {
    return BuildInfo.CVS_TAG;
  }

  private void writeInitializationInformation()
  {

    this.writeBatchReportLine( "\n" );
    this.writeBatchReportLine( MSG_START_EREIG_IDAT + BAGISTime.formatForDebug( configuration.getStartPoint() ) );
    this.writeBatchReportLine( MSG_END_EREIG_IDAT + BAGISTime.formatForDebug( configuration.getEndPoint() ) );
    this.writeBatchReportLine( "\n" );
//    
//    this.writeBatchReportLine( "\n" );
//    this.writeBatchReportLine( DATA_FOLDER_INFO_TITEL );
//    this.writeBatchReportLine( "Ablage für neue Titelseiten: " + this.configuration.getDataStoreLocation() );
//    this.writeBatchReportLine( "Ablage von versandten Titelseiten: " + this.configuration.getDataStoreLocationProcessed() );
//    this.writeBatchReportLine( "\n" );

  }

  private void checkForExsitingZipFiles()
  {
    File zipFolder = new File( configuration.getDataStoreLocationZip() );
    if ( zipFolder.exists() )
    {
      File[] zipFiles = zipFolder.listFiles( (FileFilter) FileFilterUtils.prefixFileFilter( ZIP_PREFIX ) );
      if ( zipFiles == null || zipFiles.length == 0 )
      {
        return;
      }
      this.writeBatchReportLine( "\n" );
      for ( File oldZipFile: zipFiles )
      {

        try
        {
          this.moveZipToProcessedFolder( oldZipFile );
        }
        catch ( IOException e )
        {
          LOG.error( ST36XMLGENERATOR_ZIPFILECANNOTBEMOVED_FAILURE + ": " + oldZipFile.getName(), e );
          throw new BatchException( ST36XMLGENERATOR_ZIPFILECANNOTBEMOVED_FAILURE );
        }
        //this.writeBatchReportLine( OLD_ZIP_MSG + zipFile.getName() );
        LOG.info( OLD_ZIP_MSG + oldZipFile.getName() );
      }
      this.writeBatchReportLine( "\n" );
    }
    File generationFolder = new File( configuration.getDataStoreLocation() );
    if ( generationFolder.exists() )
    {
      File[] bibliographicData = generationFolder.listFiles();
      if ( bibliographicData != null && bibliographicData.length > 0 )
      {
        throw new BatchException( ST36XMLGENERATOR_GENERATIONFOLDEREXISTS_FAILURE );
      }
    }

  }

  protected class ST36XMLGneratorAttachmentProvider implements MailAttachmentProvider
  {

    public List<File> getAttachments()
    {
      if ( ST36XMLGeneratorBatch.this.zipFile == null )
      {
        LOG.debug( "Zip file not set" );
        return Collections.emptyList();
      }
      else if ( !ST36XMLGeneratorBatch.this.zipFile.exists() )
      {
        LOG.debug( "Zip file does not exist " + ST36XMLGeneratorBatch.this.zipFile.getAbsolutePath() );
        return Collections.emptyList();
      }
      else
      {
        List<File> result = new ArrayList<File>( 2 );
        result.add( ST36XMLGeneratorBatch.this.zipFile );
        return result;
      }

    }
  }

} // end class
