package ch.ige.st36xmlgenerator.batch;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.filefilter.TrueFileFilter;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.st36xmlgenerator.exception.ZipCreationException;

/**
 * Helper class for creating a zip from data in a folder
 * @author fabrizioli
 *
 */
public class ZipCreator
{

  private static final Logger LOG = Logger.getLogger( ZipCreator.class );

  /**
   * Timeout for zip generation in seconds
   */
  private static final int ZIP_EXECUTION_TIMEOUT = 90;

  /**
   * Create a zip using the JAVA internal mechanism. The zip file is overwritten if it already exists.
   * @param zipPath the path for the zip file to create, may not be null 
   * @param folder the folder that shall be added to the zip, is ignored if it is a file 
   * @param additionalFilePaths paths to additional files, that shall be added to the zip
   * @throws ZipCreationException if the creation of the zip fails or exceeds the timeout {@link #ZIP_EXECUTION_TIMEOUT}
   * @throws IllegalArgumentException if the zipPath is null
   */
  @SuppressWarnings( "unchecked" )
  public static void createZip( String zipPath, File folder, String... additionalFilePaths ) throws ZipCreationException
  {
    if ( zipPath == null )
    {
      throw new IllegalArgumentException( "The path to the zip has to be provided" );
    }
    if ( new File( zipPath ).exists() )
    {
      LOG.warn( "Zip file will be overwritten " + zipPath );
    }

    if ( folder != null && folder.exists() && folder.isDirectory() )
    {
      createZipInternal( zipPath, FileUtils.iterateFiles( folder, TrueFileFilter.INSTANCE, TrueFileFilter.INSTANCE ), additionalFilePaths );
    }
    else
    {
      createZipInternal( zipPath, null, additionalFilePaths );
    }

  }

  private static void createZipInternal( final String zipPath, final Iterator<File> fileIterator, final String... filePaths ) throws ZipCreationException
  {

    ExecutorService executor = Executors.newFixedThreadPool( 1 );
    Future<Integer> future = executor.submit( new Callable<Integer>()
    {
      public Integer call() throws ZipCreationException
      {
        LOG.debug( "[TRACE] create zip " + zipPath );
        // Create a buffer for reading the files
        byte[] buf = new byte[1024];

        try
        {
          // Create the ZIP file

          ZipOutputStream out = new ZipOutputStream( new FileOutputStream( zipPath ) );

          // Compress the files
          if ( fileIterator != null )
          {
            while ( fileIterator.hasNext() )
            {
              File nextFile = fileIterator.next();
              FileInputStream in = new FileInputStream( nextFile );

              // Add ZIP entry to output stream.
              String path = nextFile.getAbsolutePath();
              path = defineZipEntryPath( zipPath, path );
              addZipEntry( path, in, out, buf );
            }
          }
          if ( filePaths != null )
          {
            // Compress the files
            for ( String path: filePaths )
            {
              if ( path != null )
              {
                FileInputStream in = new FileInputStream( path );
                path = defineZipEntryPath( zipPath, path );
                addZipEntry( path, in, out, buf );
              }
            }
          }

          // Complete the ZIP file
          out.close();
          LOG.debug( "Created zip " + zipPath );
          return 0;
        }
        catch ( IOException e )
        {
          logAndThrowException( zipPath, e );
          return -1;
        }

      }
    } );
    int exitVal = 0;
    try
    {
      exitVal = future.get( ZIP_EXECUTION_TIMEOUT, TimeUnit.SECONDS );
    }
    catch ( InterruptedException e )
    {
      logAndThrowException( zipPath, e );
    }
    catch ( ExecutionException e )
    {
      logAndThrowException( zipPath, e );
    }
    catch ( TimeoutException e )
    {
      logAndThrowException( zipPath, e );
    }
    if ( exitVal != 0 )
    {
      logAndThrowException( zipPath, null );
    }

  }

  protected static String defineZipEntryPath( String zipPath, String sourcePath )
  {// returns null if sourcepath null, no exception
    String partialPath = StringUtils.substringAfter( sourcePath, FilenameUtils.getPath( zipPath ) );
    if ( partialPath != null && partialPath.length() > 0 )
    {
      sourcePath = partialPath;
    }
    return sourcePath;
  }

  protected static void addZipEntry( String path, FileInputStream in, ZipOutputStream out, byte[] buf ) throws IOException
  {

    out.putNextEntry( new ZipEntry( path ) );

    // Transfer bytes from the file to the ZIP file
    int len;
    while ( ( len = in.read( buf ) ) > 0 )
    {
      out.write( buf, 0, len );
    }
    LOG.debug( "Added to zip entry " + path );
    // Complete the entry
    out.closeEntry();
    in.close();
  }

  protected static void logAndThrowException( String zipPath, Throwable e ) throws ZipCreationException
  {
    LOG.error( "Generation of zip failed [" + zipPath + "]" );
    throw new ZipCreationException( "Generation of zip failed [" + zipPath + "]", e );

  }
}
