package ch.ige.ses.client.batch;

import java.io.*;
import java.util.*;
import org.apache.log4j.Logger;
import ch.ige.util.client.batch.AbstractBatch;
import ch.ige.ses.BuildInfo;
import ch.ige.ses.server.batchctrl.*;

/**
 * Basisklasse für alle SES Batch Jobs.
 * <p>
 * Erstellt die Meldung ob der Batch Job erfolgreich war oder nicht und
 * generiert die Rückmeldung. Als Rückmeldung wird in der Regel eine Datei
 * erstellt mit der entsprechenden Meldung. Optional kann im weiteren
 * eine Benachrichtigung per EMail aktiviert werden.
 * <p>
 * Copyright (c) 2002, Eidgenössisches Institut für Geistiges Eigentum
 * @author S. Nydegger, T. Iten
 * @version $Id: BatchBasis.java,v 1.19 2008/03/04 07:10:47 iten Exp $
 */
public class BatchBasis
{
  //
  // const
  //
  public static final  String   MSG_TYPE_OK   = "OK";
  public static final  String   MSG_TYPE_NOK  = "ERROR";

  private static final Logger   LOG = Logger.getLogger(BatchBasis.class);

  public static final String    FILE_PATH     = AbstractBatch.FILE_PATH;
  private static final String   FILE_EXT_OK   = AbstractBatch.OK;
  private static final String   FILE_EXT_NOK  = AbstractBatch.NOK;

  private static final String   DEF_MSG_TYPE  = MSG_TYPE_OK;
  private static final String   DEF_MSG_NR    = AbstractBatch.FILE_CONTENT_OK;
  private static final String   DEF_MSG_BODY  = "BatchJob erfolgreich beendet";

  private static final String   DEF_SYS_TITLE = "System Informationen:";

  
  //
  // fields
  //
  private String    fJobName      = null;           // batch job name

  private boolean   fNotifyFile   = true;           // notify options
  private boolean   fNotifyEmail  = true;
  private boolean   fNotifyConsole= false;

  private String    fSubjectNose  = null;
  private String    fSubject      = null;           // email subject
                                                    // (default = job name)
  // notify fields
  // for file, email-body or console output
  private String    fMsgType      = DEF_MSG_TYPE;   // message fields
  private String    fMsgNr        = DEF_MSG_NR;
  private String    fMsgTitle     = null;
  private String    fMsgBody      = DEF_MSG_BODY;

  private String    fSysTitle     = DEF_SYS_TITLE;  // system info
  private String    fSysMsg       = null;           // optional system messages

  private String    fFooter       = null;           // footer

  //---------------------------------------------------------------------------
  // constructors
  //---------------------------------------------------------------------------

  /**
   * Erstellt BatchJob Basisklasse.
   * Der angegebene Job Name wird zur Erstellung des Dateinamens (in der
   * folgenden Form) verwendet: FILE_PATH + jobName + FILE_EXT
   *
   * @param jobName Name des aktuellen Batch Jobs
   */
  public BatchBasis(String jobName)
  {
    this.fJobName  = jobName;
    this.fSubject  = jobName;
    this.fMsgTitle = "Batch Job: " + jobName;
    
    if ( getProviderUrl().equals( ch.ige.util.BuildInfo.PROVIDER_URL ) )
    {
      fSubjectNose  = ch.ige.util.BuildInfo.BUILD_ENV_NAME;
    }
    else
    {
      fSubjectNose  = "Unerwartete Umgebung";
    }
    
  }

  //---------------------------------------------------------------------------
  // public methods
  //---------------------------------------------------------------------------

  /**
   * Get file notification mode
   * @return notification mode
   */
  public boolean getNotifyFile()
  {
    return fNotifyFile;
  }

  /**
   * Enable or disable file notification
   * @param state
   */
  public void setNotifyFile(boolean state)
  {
    fNotifyFile = state;
  }


  /**
   * Get email notification mode
   * @return notification mode
   */
  public boolean getNotifyEmail()
  {
    return fNotifyEmail;
  }

  /**
   * Enable or disable email notification
   * @param state
   */
  public void setNotifyEmail(boolean state)
  {
    fNotifyEmail = state;
  }

  /**
   * Get console notification mode
   * @return notification mode
   */
  public boolean getNotifyConsole()
  {
    return fNotifyConsole;
  }

  /**
   * Enable or disable console notification
   * @param state
   */
  public void setNotifyConsole(boolean state)
  {
    fNotifyConsole = state;
  }

  /**
   * Set who to notify
   * 
   * @param fileState
   * @param emailState
   * @param consoleState
   */
  public void setNotifyAll(boolean fileState, boolean emailState, boolean consoleState)
  {
    fNotifyFile     = fileState;
    fNotifyEmail    = emailState;
    fNotifyConsole  = consoleState;
  }

  /**
   * Get message
   * @return message
   */
  public String getMessage ()
  {
    return getContent();
  }

  /**
   * set message
   * @param msgType
   * @param msgNr
   * @param msgBody
   */
  public void setMessage (String msgType, String msgNr, String msgBody)
  {
    fMsgType  = msgType;
    fMsgNr    = msgNr;
    fMsgBody  = msgBody;
  }

  /**
   * Add a system message
   * 
   * @param sysMessage
   */
  public void addSystemMessage (String sysMessage)
  {
    if (fSysMsg == null)
    {
      fSysMsg =  "  " + sysMessage + "\n";
    }
    else
    {
      fSysMsg = fSysMsg + "  " + sysMessage + "\n";
    }
  }

  /**
   * Set message footer
   * 
   * @param footer
   */
  public void setMessageFooter (String footer)
  {
    fFooter = footer;
  }

  /**
   * Set subject
   * 
   * @param subject
   */
  public void setSubject (String subject)
  {
    fSubject = subject;
  }

  /**
   * Notify Message
   */
  public void notifyMessage()
  {
    LOG.debug("notifyMessage - start...");

    if ( fNotifyFile )
    {
      LOG.debug("notifyMessage - notify file");
      writeFile(getFileName(),getContent());
    }

    if ( fNotifyEmail )
    {
      try
      {
        LOG.debug("notifyMessage - notify mail");

        BatchDelegate.getInstance().batchNotifyEmail
          ( fJobName, fSubjectNose + "-" + fSubject + " " + fMsgType, getContent() );
      }
      catch (Exception ex)
      {
        LOG.error("notifyMessage - batchNotifyEmail failed with exception: " + ex.getMessage());
      }
    }

    if ( fNotifyConsole )
    {
      if ( fMsgType == MSG_TYPE_OK)
      {
        LOG.debug("notifyMessage - notify out-console");
        System.out.println(getContent());
      }
      else
      {
        LOG.debug("notifyMessage - notify err-console");
        System.err.println(getContent());
      }
    }
  }

  /**
   * Notify message
   * 
   * @param msgType
   * @param msgNr
   * @param msgBody
   */
  public void notifyMessage (String msgType, String msgNr, String msgBody)
  {
    setMessage (msgType,msgNr,msgBody);
    notifyMessage();
  }

  
  /**
   * Check if batch job is enabled
   * 
   * @return True if enabled, otherwise false
   * @throws Exception
   */
  public boolean enabled() throws Exception
  {
    return BatchDelegate.getInstance().batchJobEnabled(fJobName);
  }


  //---------------------------------------------------------------------------
  // private methods
  //---------------------------------------------------------------------------

  private String getFileName()
  {
    if ( fMsgType.equals(MSG_TYPE_OK) )
    {
      return FILE_PATH + fJobName + FILE_EXT_OK;
    }
    else
    {
      return FILE_PATH + fJobName + FILE_EXT_NOK;
    }
  }


  private String getContent()
  {
    StringBuffer buf  = new StringBuffer();
    Date         date = new Date();

    // batch job message
    buf.append(fMsgType  + " " + fMsgNr + "\n \n");
    buf.append(fMsgTitle + "\n");
    buf.append(fMsgBody  + "\n \n");

    // system info
    buf.append(fSysTitle + "\n");
    if (fSysMsg != null) 
    {
      buf.append(fSysMsg);
    }
    buf.append("  Datum = " + date.toString() + "\n");
    buf.append("  " + javax.naming.Context.PROVIDER_URL + " = " + getProviderUrl()+ "\n");
    buf.append("  CVS Tag = " + BuildInfo.CVS_TAG + "\n");
    buf.append("  Build Id = " + BuildInfo.BUILD_ID + "\n");
    buf.append("  Build Environment = " + BuildInfo.BUILD_ENV_NAME + "\n");

    if ( ! getProviderUrl().equals( ch.ige.util.BuildInfo.PROVIDER_URL ) )
    {
      buf.append("\n");
      buf.append("WARNUNG: \n");
      buf.append("  Dieses Batch Programm wurde in einer nicht dafuer vorgesehen Umgebung ausgefuehrt ! \n" );
      buf.append("  Provider URL geplant  = " + ch.ige.util.BuildInfo.PROVIDER_URL + "\n");
      buf.append("  Provider URL effektiv = " + getProviderUrl()  + "\n" );
    }
    
    
    // optional footer message(s)
    if (fFooter != null)
    {
      buf.append(fFooter + "\n");
    }
      
    // return result
    return buf.toString();
  }

  
  /**
   * Get current used provider url. 
   * 
   * The provider url will be used by the ch.ige.util.ServiceLocator. 
   * If a corresponding system property is set, this provider url will be used, 
   * othewise the provider url delivered by the BuildInfo class.
   *  
   * @return provider url
   */
  private String getProviderUrl()
  {
    String result = System.getProperty( javax.naming.Context.PROVIDER_URL );
    
    if ( result == null )
    {
      result = ch.ige.util.BuildInfo.PROVIDER_URL;
    }
    
    return result;
  }


  private boolean writeFile(String fileName, String content)
  {
    boolean ret = false;

    try
    {
      BufferedWriter  bufWriter = new BufferedWriter(new FileWriter(fileName));
      bufWriter.write(content);
      bufWriter.close();
      ret = true;
    }
    catch (IOException ex)
    {
      String err = "writeFile - Datei (" + fileName + ") konnte nicht angelegt werden, " +
                   "Exception: " + ex.getMessage();

      LOG.error(err);

      System.err.println(err);
      System.err.println(content);
    }

    return ret;
  }

} // end class


