package ch.ige.ses.client.sapinbound;

import java.util.*;
import javax.jms.*;
import org.apache.log4j.*;
import ch.ige.ses.client.saprequest.*;
import ch.ige.ses.client.saputil.*;
import ch.ige.ses.server.sapctrl.*;
import ch.ige.ses.util.*;
import ch.ige.ses.vo.bagis.*;
import ch.ige.ses.vo.ses.*;
import ch.ige.util.vo.*;

/**
 * SAP Inbound Client - Automatic Mode
 * Diese Klasse verarbeitet die empfangenen JMS Object Messages. Diese werden
 * via BAPI Call an SAP weitergegeben.
 * <p>
 * Copyright (c) 2002, Eidgenössisches Institut für Geistiges Eigentum
 * @author T. Iten
 * @version $
 */
public class AutoObjectMessage
{
  //
  // const
  //
  private static final Logger LOG = Logger.getLogger(AutoObjectMessage.class);

  //
  // fields
  //
  private   int           sapTrace; // trace options (see property file)

  private   AutoRFC       rfc;      // remote function call module
  private   JmsQueue      queue;    // jms queue, used for sending replies
  private   ObjectMessage om;       // object message to process

  private   boolean       rollback; // rollback indicator for the jms message

  //---------------------------------------------------------------------------
  // constructors
  //---------------------------------------------------------------------------

  public AutoObjectMessage()
  {
    // lazy bone
  }

  //---------------------------------------------------------------------------
  // public methods
  //---------------------------------------------------------------------------

  public boolean process ( ObjectMessage objectMessage,
                           AutoRFC autoRfc, JmsQueue jmsQueue ) throws Exception
  {
    int       msgType = Jms.MSG_TYPE_NULL;
    String    msgDesc = null;


    //
    // init
    //
    rfc   = autoRfc;
    queue = jmsQueue;
    om    = objectMessage;

    rollback = false;


    //
    // check message
    //
    if ( om.getJMSRedelivered() )
    {
      LOG.warn("process - received redelivered message !");
    }

    if ( om.propertyExists(Jms.JMS_PROP_NAME_MSG_TYPE) &&
         om.propertyExists(Jms.JMS_PROP_NAME_MSG_DESC) )
    {
      msgType = om.getIntProperty(Jms.JMS_PROP_NAME_MSG_TYPE);
      msgDesc = om.getStringProperty(Jms.JMS_PROP_NAME_MSG_DESC);
    }
    else
    {
      ErrorVO err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                                  "process - received & SKIPPED ObjectMessage because " +
                                  "property " + Jms.JMS_PROP_NAME_MSG_TYPE + " or " +
                                  "property " + Jms.JMS_PROP_NAME_MSG_DESC + " is missing !" );

      LOG.warn(err.description);
      saveError(err);
      return rollback;
    }


    //
    // process message
    //
    LOG.debug("process - " + msgDesc + " (msgType=" + msgType + ")" );

    switch ( msgType )
    {
      // stop application
      case Jms.MSG_TYPE_CTRL_STOP:
        processCtrlStop();
        break;

      // create or change billing request
      case Jms.MSG_TYPE_SAP_RECHNUNG:
        processSapRechnung((SAPRechnungVO)om.getObject());
        break;

      // reject billing request
      case Jms.MSG_TYPE_SAP_ANNULLATION:
        processSapAnnulation((SAPRechnungAnnullationVO)om.getObject());
        break;

      // reject billing request list
      case Jms.MSG_TYPE_SAP_ANNULLATION_LIST:
        processSapAnnulationList((ArrayList) om.getObject());
        break;

      // sap jgb job
      case Jms.MSG_TYPE_SAP_JGB_JOB:
        processSapJgbJob((SAPJGBJobVO)om.getObject());
        break;

      // sap jgb job force finish
      case Jms.MSG_TYPE_CTRL_JGB_FINISH:
        processCtrlJgbFinish((Integer)om.getObject());
        break;

      // sap amv job force finish
      case Jms.MSG_TYPE_CTRL_AMV_FINISH:
        processCtrlAmvFinish((AmvLaufVO)om.getObject());
        break;

      // adresse create
      case Jms.MSG_TYPE_SAP_CUST_CREATE:
        processSapCustCreate((VersadrVO)om.getObject());
        break;

      // adresse status
      case Jms.MSG_TYPE_SAP_CUST_STATUS:
        processSapCustStatus((SAPAdrStatusVO)om.getObject());
        break;

      // adresse check
      case Jms.MSG_TYPE_SAP_CUST_CHECK:
        processSapCustCheck((ArrayList)om.getObject());
        break;

      // sap material (request reply mode)
      case Jms.MSG_TYPE_SAP_MATERIAL :
        processSapMaterial((String)om.getObject());
        break;

      // sap balance (request reply mode)
      case Jms.MSG_TYPE_SAP_BALANCE :
        processSapBalance((Integer)om.getObject());
        break;

      // sap stako job
      case Jms.MSG_TYPE_SAP_STAKO_JOB:
        processSapStakoJob((SAPStakoJobVO)om.getObject());
        break;

      // insert stako
      case Jms.MSG_TYPE_SAP_STAKO:
        processSapStako((StakoVO)om.getObject());
        break;

      default:
        LOG.warn( "process - received & SKIPPED unhandeld ObjectMessage Type " +
                  om.getIntProperty(Jms.JMS_PROP_NAME_MSG_TYPE) + " !" );
        break;

    } // switch


    //
    // return rollback state
    //
    return rollback;

  } // end method


  public void setSapTrace(int value)
  {
    sapTrace = value;
  }


  public boolean saveError (ErrorVO err)
  {
    try
    {
      LOG.info("saveError - save vo to ses error db");
      SapDelegateInbound.getInstance().receiveError(err);
      return true;
    }
    catch (Exception ex)
    {
      LOG.fatal("saveError - save vo to ses error db failed: " + ex.getMessage() );

      // save error vo to local file
      new ErrorFile ( SapInbound.getProperties().getProperty("sapinbound.data.path"),
                      SapInbound.getProgramIdentifier() ).save(err);

      return false;
    }
  }

  //---------------------------------------------------------------------------
  // private methods - process messages
  //---------------------------------------------------------------------------

  private void processCtrlStop()
  {
    LOG.debug("processCtrlStop - begin");
    SapInbound.shutdown();
    LOG.debug("processCtrlStop - end");
  }


  private void processSapRechnung(SAPRechnungVO re) throws Exception
  {
    LOG.debug("processSapRechnung - begin: bfakt_id = " + re.bfakt_id);

    // process
    if (re.isRechnung())
    {
      processSapRechnungEinzel(re);
    }
    else if (re.isAMV())
    {
      processSapRechnungAmv(re);
    }

    LOG.debug("processSapRechnung - end");
  }


  private void processSapRechnungEinzel(SAPRechnungVO re) throws Exception
  {
    StringBuffer  retMsg  = new StringBuffer();
    int           retCode = RFC.RET_CODE_SUCCESS;
    ErrorVO       err     = null;

    // begin
    LOG.debug("processSapRechnungEinzel - begin");

    //  simulation enabled: modify rechnung data (for tests only !)
    if ( SapInbound.getProperties().getProperty("simulation.rechnung.modify").equals("1") )
    {
      int old_bfakt_id = re.bfakt_id;
      TestData.modifyRechnung(re,SapInbound.getProperties().getProperty("simulation.datafile"));
      LOG.debug("processObjectMsg - modify SAPRechnungVO, bfakt_id " +  old_bfakt_id + " replaced with " + re.bfakt_id);
    }

    // delete billing request (if required)
    // important: ignore return code since changes are also allowed for billing
    //            request's which couldn't be created in sap during the first call !
    if ( re.hasChanged.equals(re.BOOLEAN_TYPES[1]) )
    {
      LOG.debug("processSapRechnungEinzel - delete billing request");
      rfc.deleteBillingReq(re,retMsg);
    }

    // create billing request
    LOG.debug("processSapRechnungEinzel - create billing request");
    retCode =  rfc.createBillingReq(re,retMsg);


    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapRechnungEinzel - end (SAPRechnungVO successfully processed)");
      trace(true,Jms.MSG_DESC_SAP_RECHNUNG,re.toString());
    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processSapRechnungEinzel - process message SAPRechnungVO failed: \n" + retMsg.toString(),
                          ErrorVO.DATA_TYPE_CD_SAP_RECHNUNG_VO, re,
                          re.employeeId, ErrorVO.EMPL_NOTIFY_PENDENT, ErrorVO.EMPL_MSG_TYPE_RECHNUNG );
      saveError(err);

      LOG.error("processSapRechnungEinzel - end (SAPRechnungVO failed to process)");
      trace(false,Jms.MSG_DESC_SAP_RECHNUNG,err.toString());
    }
  }


  private void processSapRechnungAmv (SAPRechnungVO re) throws Exception
  {
    StringBuffer  retMsg      = new StringBuffer();
    int           retCode     = RFC.RET_CODE_SUCCESS;
    AmvRechnungVO amvRe       = new AmvRechnungVO();
    ErrorVO       err         = null;
    int           oldBfaktId  = 0;                    // simulation ONLY


    //
    // begin / init
    //
    if ( SapInbound.getProperties().getProperty("simulation.rechnung.modify").equals("1") )
    {
      oldBfaktId = re.bfakt_id;
      TestData.modifyRechnung(re,SapInbound.getProperties().getProperty("simulation.datafile"));
      LOG.debug("processSapRechnungAmv - modify SAPRechnungVO, bfakt_id " +  oldBfaktId + " replaced with " + re.bfakt_id);
    }

    if (LOG.isDebugEnabled())
    {
      LOG.debug("processSapRechnungAmv - begin: amvLaufNr=" + re.amvLaufNr +
                 " amvRechnungCd=" + re.amvRechnungCd + " bfakt_id=" +  re.bfakt_id );
    }

    amvRe.amvLaufId     = re.amvLaufId;
    amvRe.rechnungCd    = re.amvRechnungCd;
    amvRe.bfaktId       = re.bfakt_id;
    amvRe.anzRechEmpf   = 1 + re.zusatzRechEmpf.size();


    //
    // create/change billing request
    //
    if ( re.amvRechnungCd == AmvRechnungVO.RE_CD_ERINNERUNG_1)
    {
      LOG.debug("processSapRechnungAmv - create billing request (erinnerung 1)");
      retCode =  rfc.createBillingReq(re,retMsg);
    }
    else
    {
      LOG.debug("processSapRechnungAmv - change billing request (erinnerung 2)");
      retCode =  rfc.changeBillingReq(re,retMsg);
    }


    //
    // handle retCode
    //
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapRechnungAmv - create/change billing request successfully processed");
      trace(true,Jms.MSG_DESC_SAP_RECHNUNG,re.toString());

      // complete amv rechnung data
      amvRe.rechnungStat  = AmvRechnungVO.RE_STAT_SAP_OK;
      amvRe.errorMsg      = null;
    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processSapRechnungAmv - create/change billing request failed to process: \n" + retMsg.toString(),
                          ErrorVO.DATA_TYPE_CD_SAP_RECHNUNG_VO, re,
                          re.employeeId, ErrorVO.EMPL_NOTIFY_KEINE, ErrorVO.EMPL_MSG_TYPE_RECHNUNG );
      saveError(err);

      LOG.error("processSapRechnungAmv - create/change billing request failed to process)");
      trace(false,Jms.MSG_DESC_SAP_RECHNUNG,err.toString());

      // complete amv rechnung data
      amvRe.rechnungStat  = AmvRechnungVO.RE_STAT_SAP_ERROR;
      amvRe.errorMsg      = retMsg.toString();
    }


    //
    // update ses db, table amv_rechnung
    //
    try
    {
      //  simulation enabled: modify rechnung data (for tests only !)
      if ( SapInbound.getProperties().getProperty("simulation.rechnung.modify").equals("1") )
      {
        amvRe.bfaktId = oldBfaktId;
      }

      LOG.debug("processSapRechnungAmv - update amv rechnung");
      SapDelegateInbound.getInstance().requestAmvRechnungUpdate(amvRe);
      LOG.debug("processSapRechnungAmv - update amv rechnung successfully done");
    }
    catch (Exception ex)
    {
      LOG.fatal ( "processSapRechnungAmv - requestAmvRechnungUpdate failed ! \n" +
                  "data: " + amvRe.toString() + "\n" +
                  "exception: " + ex.getMessage() );
      //TODO maintenance (optional) - weitere behandlung: inkonsistenz zwischen ses db und sap system
    }

    //
    // end
    //
    LOG.debug("processSapRechnungAmv - end");
  }


  private void processSapAnnulation(SAPRechnungAnnullationVO an)
  {
    StringBuffer  retMsg  = new StringBuffer();
    int           retCode = RFC.RET_CODE_SUCCESS;
    ErrorVO       err     = null;

    // init
    LOG.debug("processSapAnnulation - begin: bfakt_id = " + an.bfakt_id);

    // reject billing request
    retCode = rfc.rejectBillingReq(an,retMsg);

    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapAnnulation - end (SAPRechnungAnnullationVO successfully processed)");
      trace (true,Jms.MSG_DESC_SAP_ANNULLATION,an.toString());

    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processSapAnnulation - process message SAPRechnungAnnullationVO failed: \n" + retMsg.toString(),
                          ErrorVO.DATA_TYPE_CD_SAP_RECHNUNG_ANNULATION_VO, an,
                          null,  ErrorVO.EMPL_NOTIFY_PENDENT, ErrorVO.EMPL_MSG_TYPE_RECHNUNG );
      saveError(err);
      LOG.error("processSapAnnulation - end (SAPRechnungAnnullationVO failed to process)");
      trace (false,Jms.MSG_DESC_SAP_ANNULLATION,err.toString());
    }
  }


  private void processSapAnnulationList(ArrayList list)
  {
    StringBuffer  retMsg  = new StringBuffer();
    int           retCode = RFC.RET_CODE_SUCCESS;
    ErrorVO       err     = null;

    Iterator                  it          = list.iterator();
    boolean                   errOccured  = false;
    SAPRechnungAnnullationVO  an          = null;
    StringBuffer              buf         = new StringBuffer();

    // init
    LOG.debug("processSapAnnulationList - begin");

    if ( ! it.hasNext() )
    {
      LOG.warn("processSapAnnulationList - MSG_TYPE_SAP_ANNULLATION_LIST is EMPTY");
      trace (false,Jms.MSG_DESC_SAP_ANNULLATION_LIST,"Received empty List");
      return;
    }

    buf.append ( Jms.MSG_DESC_SAP_ANNULLATION_LIST +
                 " : bfakt_id and process - state \n" );


    // handle single list entires
    while ( it.hasNext() )
    {
      // get single vo
      an = (SAPRechnungAnnullationVO) it.next();
      LOG.debug("processSapAnnulationList - process list entry with SAPRechnungAnnullationVO.bfakt_id = " + an.bfakt_id);

      // clear return message
      retMsg.delete(0,retMsg.length());

      // reject billing request
      retCode = rfc.rejectBillingReq(an,retMsg);

      // handle retCode
      if (retCode == RFC.RET_CODE_SUCCESS)
      {
        buf.append(an.bfakt_id + " ok \n");
        LOG.debug ("processSapAnnulationList - SAPRechnungAnnullationVO.bfakt_id = " + an.bfakt_id + " successfully processed");
      }
      else
      {
        buf.append(an.bfakt_id + " NOK \n");
        LOG.error ("processSapAnnulationList - SAPRechnungAnnullationVO.bfakt_id = " + an.bfakt_id + " failed to processed");

        errOccured = true;
        err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                            "processSapAnnulationList - process message MSG_TYPE_SAP_ANNULLATION_LIST failed: \n" + retMsg.toString(),
                            ErrorVO.DATA_TYPE_CD_SAP_RECHNUNG_ANNULATION_VO, an,
                            null,  ErrorVO.EMPL_NOTIFY_KEINE, ErrorVO.EMPL_MSG_TYPE_RECHNUNG );
        saveError(err);
      }
    } // end while

    // handle entire list state
    if ( ! errOccured)
    {
      LOG.debug("processSapAnnulationList - end (SAPRechnungAnnullationVO List successfully processed)");
      trace(true,Jms.MSG_DESC_SAP_ANNULLATION_LIST,buf.toString());
    }
    else
    {
      LOG.error("processSapAnnulationList - end (SAPRechnungAnnullationVO List failed to process)");
      trace(false,Jms.MSG_DESC_SAP_ANNULLATION_LIST,buf.toString());
    }
  }


  private void processSapJgbJob(SAPJGBJobVO job) throws Exception
  {
    ErrorVO err = null;
    JGB     jgb = new JGB(rfc);

    // init
    if ( LOG.isDebugEnabled() )
    {
      LOG.debug( "processSapJgbJob - begin: " +
                 "jgbLaufId="   + job.jgbLaufId    + ", " +
                 "bfaktIdFrom=" + job.bfaktIdFrom  + ", " +
                 "bfaktIdTo="   + job.bfaktIdTo    + ", " +
                 "lastJob="     + job.lastJob );
    }

    // process jgb job
    jgb.process(job);

    // handle status
    if ( ! jgb.isErrorOccured() )
    {
      LOG.debug("processSapJgbJob - end (SAPJGBJobVO successfully processed)");
      trace (true,Jms.MSG_DESC_SAP_JGB_JOB,job.toString());
    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processSapJgbJob - process SAPJGBJobVO failed: \n" + jgb.getErrorMessage(),
                          ErrorVO.DATA_TYPE_CD_SAP_JGB_JOB_VO, job,
                          null, ErrorVO.EMPL_NOTIFY_KEINE, ErrorVO.EMPL_MSG_TYPE_JGB );
      saveError(err);
      LOG.error("processSapJgbJob - end (failed to process)");
      trace (true,Jms.MSG_DESC_SAP_JGB_JOB,err.toString());
    }
  }


  private void processCtrlJgbFinish(Integer jgbLaudId) throws Exception
  {
    ErrorVO   err     = null;
    JGB       ctrlJgb = new JGB(rfc);

    // init
    LOG.debug("processCtrlJgbFinish - begin");

    // force jgb finish
    ctrlJgb.forceFinish(jgbLaudId.intValue());

    // handle state
    if ( ! ctrlJgb.isErrorOccured() )
    {
      LOG.debug("processCtrlJgbFinish - end (successfully processed)");
      trace(true,Jms.MSG_DESC_CTRL_JGB_FINISH,"jgbLaudId = " + jgbLaudId.toString());
    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processCtrlJgbFinish - process control message force jgb finish failed: \n" + ctrlJgb.getErrorMessage(),
                          ErrorVO.DATA_TYPE_CD_STRING, "jgbLaudId = " + jgbLaudId.intValue(),
                          null, ErrorVO.EMPL_NOTIFY_KEINE, ErrorVO.EMPL_MSG_TYPE_JGB );
      saveError(err);

      LOG.error("processCtrlJgbFinish - end (failed to process)");
      trace(false,Jms.MSG_DESC_CTRL_JGB_FINISH,err.toString());
    }
  }


  private void processCtrlAmvFinish(AmvLaufVO vo)
  {
    StringBuffer  retMsg    = new StringBuffer();
    int           retCode   = RFC.RET_CODE_SUCCESS;
    ErrorVO       err       = null;


    // init
    LOG.debug("processCtrlAmvFinish - begin: laufNr=" + vo.laufNr);

    // commit tm extension
    LOG.debug("processCtrlAmvFinish - commit tm extension");
    retCode =  rfc.commitTmExtension(vo.laufNr,retMsg);

    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      vo.laufStat = AmvLaufVO.LAUF_STAT_FINISHED;

      LOG.debug("processCtrlAmvFinish - commit tm extension successfully processed");
      trace(true,Jms.MSG_DESC_CTRL_AMV_FINISH,vo.toString());
    }
    else
    {
      vo.laufStat = AmvLaufVO.LAUF_STAT_STOPPED;

      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processCtrlAmvFinish - commit tm extension failed to process: \n" + retMsg.toString(),
                          ErrorVO.DATA_TYPE_CD_STRING, vo.toString() );

      saveError(err);

      LOG.error("processCtrlAmvFinish - commit tm extension failed to process)");
      trace(false,Jms.MSG_DESC_CTRL_AMV_FINISH,err.toString());
    }

    // update ses db, table amv_lauf
    try
    {
      LOG.debug("processCtrlAmvFinish - update amv lauf stat");
      SapDelegateInbound.getInstance().requestAmvLaufStatUpdate
        (vo.amvLaufId,vo.laufStat);
      LOG.debug("processCtrlAmvFinish - update amv lauf stat successfully done");
    }
    catch (Exception ex)
    {
      LOG.fatal ( "processCtrlAmvFinish - requestAmvLaufStatUpdate failed ! \n" +
                  "amvLaufId: " + vo.amvLaufId + "\n" +
                  "laufStat:  " + vo.laufStat + "\n" +
                  "exception: " + ex.getMessage() );
      //TODO maintenance (optional) - weitere behandlung: inkonsistenz zwischen ses db und sap system
    }

    // end
    LOG.debug("processCtrlAmvFinish - end");
  }


  private void processSapCustCreate(VersadrVO adr) throws Exception
  {
    StringBuffer  retMsg  = new StringBuffer();
    int           retCode = RFC.RET_CODE_SUCCESS;
    ErrorVO       err     = null;

    // init
    LOG.debug("processSapCustCreate - begin: adrNr = " + adr.getAdrNr());

    // customer create (new or modify)
    retCode = rfc.customerCreate(adr,retMsg);

    // try to create new customer if missing within sap
    if ( (retCode != RFC.RET_CODE_SUCCESS) &&
         (adr.isIsNew() == false) &&
         (rfc.customerMissing(adr.getAdrNr()) == true) )
    {
      LOG.debug("processSapCustCreate - VersadrVO failed to update, retry with create new");
      adr.setIsNew(true);
      retCode = rfc.customerCreate(adr,retMsg);
    }


    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapCustCreate - VersadrVO successfully processed");
      trace(true,Jms.MSG_DESC_SAP_CUST_CREATE,adr.toString());
    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processSapCustCreate - process message VersadrVO failed: \n" + retMsg.toString(),
                          ErrorVO.DATA_TYPE_CD_SAP_VERS_ADR_VO, adr,
                          null, ErrorVO.EMPL_NOTIFY_PENDENT, ErrorVO.EMPL_MSG_TYPE_ADR );

      if (! adr.getEmployeeIdOrEmptyStr().equals(""))       // notify employee
      {                                                     // if available
        err.employeeId = adr.getEmployeeId();
      }
      saveError(err);

      LOG.error("processSapCustCreate - VersadrVO failed to process");
      trace(false,Jms.MSG_DESC_SAP_CUST_CREATE,err.toString());
    }

    // reply (only in request reply mode)
    if (om.getJMSReplyTo() != null)
    {
      LOG.debug("processSapCustCreate - VersadrVO send REPLY");
      SapReturn sapReturn = new SapReturn(retCode,retMsg);
      queue.reply(om,sapReturn);
    }

    // end
    LOG.debug("processSapCustCreate - end");
  }


  private void processSapCustStatus(SAPAdrStatusVO adrStatus) throws Exception
  {
    StringBuffer  retMsg  = new StringBuffer();
    int           retCode = RFC.RET_CODE_SUCCESS;
    ErrorVO       err     = null;

    // init
    if ( LOG.isDebugEnabled() )
      LOG.debug( "processSapCustStatus - begin: " +
                 "adrNr = " + adrStatus.adrNr + ", " +
                 "modus = " + adrStatus.modus + ", " +
                 "action = " + adrStatus.action );

    // set status
    if ( adrStatus.modus )                            // customer delete/undelete
    {
      retCode = rfc.customerDelete(adrStatus,retMsg);
      if (retCode == RFC.RET_CODE_SUCCESS)            // for delete/undelete set
      {                                               // also block/unblock
        adrStatus.modus = false;
        retCode = rfc.customerBlock(adrStatus,retMsg);
      }
    }
    else                                              // customer block/unblock
    {
      retCode = rfc.customerBlock(adrStatus,retMsg);
    }

    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapCustStatus - SAPAdrStatusVO successfully processed");
      trace(true,Jms.MSG_DESC_SAP_CUST_STATUS,adrStatus.toString());
    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processSapCustStatus - process message SAPAdrStatusVO failed: \n" + retMsg.toString(),
                          ErrorVO.DATA_TYPE_CD_SAP_ADR_STATUS_VO, adrStatus,
                          null, ErrorVO.EMPL_NOTIFY_PENDENT, ErrorVO.EMPL_MSG_TYPE_ADR );

      if (! adrStatus.employeeId.equals(""))              // notify employee
      {                                                   // if available
        err.employeeId      = adrStatus.employeeId;
      }

      saveError(err);

      LOG.error("processSapCustStatus - SAPAdrStatusVO failed to process");
      trace(false,Jms.MSG_DESC_SAP_CUST_STATUS,err.toString());
    }

    // reply (only in request reply mode)
    if (om.getJMSReplyTo() != null)
    {
      LOG.debug("processSapCustStatus - SAPAdrStatusVO send REPLY");
      SapReturn sapReturn = new SapReturn(retCode,retMsg);
      queue.reply(om,sapReturn);
    }

    // end
    LOG.debug("processSapCustStatus - end");
  }


  private void processSapCustCheck(ArrayList list) throws Exception
  {
    StringBuffer  retMsg    = new StringBuffer();
    int           retCode   = RFC.RET_CODE_SUCCESS;
    String        adrNrList = "";


    // init
    LOG.debug("processSapCustCheck - begin");

    for (Iterator iter = list.iterator(); iter.hasNext(); )
    {
      VersadrCheckVO item = (VersadrCheckVO)iter.next();
      adrNrList = adrNrList + item.adrNr + " ";
    }

    if ( LOG.isDebugEnabled() )
      LOG.debug("processSapCustCheck - process message VersadrCheckVO List: " + adrNrList);


    // chech adress list
    retCode = rfc.customerCheck(list,retMsg);
    retCode = RFC.RET_CODE_SUCCESS;


    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapCustCheck - VersadrCheckVO List successfully processed");
      trace(true,Jms.MSG_DESC_SAP_CUST_CHECK,adrNrList);
    }
    else
    {
      // request reply mode - do not save error into database,
      // since user will be notified immediately
      LOG.error("processSapCustCheck - VersadrCheckVO List failed to process");
      trace(false,Jms.MSG_DESC_SAP_CUST_CHECK,adrNrList);
    }

    // reply (only in request reply mode)
    if (om.getJMSReplyTo() != null)
    {
      LOG.debug("processSapCustCheck - VersadrCheckVO List send REPLY");
      SapReturn sapReturn = new SapReturn(retCode,retMsg,list);
      queue.reply(om,sapReturn);
    }

    // end
    LOG.debug("processSapCustCheck - end");
  }


  private void processSapMaterial(String art_nr)  throws Exception
  {
    StringBuffer  retMsg  = new StringBuffer();
    int           retCode = RFC.RET_CODE_SUCCESS;

    // init
    LOG.debug( "processSapMaterial - begin: art_nr = " + art_nr);

    // check
    retCode = rfc.existenceCheck(art_nr, retMsg);

    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapMaterial - art_nr exists on SAP");
      trace(true,Jms.MSG_DESC_SAP_MATERIAL,"art_nr = " + art_nr + " vorhanden auf SAP");
    }
    else
    {
      /* Kein Error Logging
         Failure heisst in diesem Fall: Artikel nicht gefunden */
      LOG.debug("processSapMaterial - art_nr does NOT exist on SAP");
      trace(true,Jms.MSG_DESC_SAP_MATERIAL,"art_nr = " + art_nr + " NICHT vorhanden auf SAP !");
    }

    // reply
    if (om.getJMSReplyTo() != null)
    {
      LOG.debug("processSapMaterial - SAP_MATERIAL send REPLY");
      SapReturn sapReturn = new SapReturn(retCode, retMsg);
      queue.reply(om, sapReturn);
    }

    // end
    LOG.debug("processSapMaterial - end");
  }


  private void processSapBalance(Integer kontoNummer) throws Exception
  {
    StringBuffer  retMsg  = new StringBuffer();
    int           retCode = RFC.RET_CODE_SUCCESS;
    double        balance = 0;
    int           ktoNr   = kontoNummer.intValue();

    //init
    LOG.debug("processSapBalance - begin: ktoNr = " + ktoNr);

    // get balance
    retCode = rfc.customerBalance(ktoNr, retMsg);
    balance = rfc.getBalanceRetval();

    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapBalance - ktoNr successfully processed");
      trace(true,Jms.MSG_DESC_SAP_BALANCE,"ktoNr = " + ktoNr);
    }
    else
    {
      /* Kein Error Logging
         Fehler wird direkt von BAGIS angezeigt */
      LOG.error("processSapBalance - ktoNr failed to process");
      trace ( false,Jms.MSG_DESC_SAP_BALANCE,
              "ktoNr = " + ktoNr + "\n" + retMsg.toString() );
    }

    // reply
    if (om.getJMSReplyTo() != null)
    {
      Double  dBalance = new Double(balance);
      LOG.debug("processObjectMsg - CUST_BALANCE send REPLY");
      SapReturn sapReturn = new SapReturn(retCode, retMsg, dBalance);
      queue.reply(om,sapReturn);
    }

    //end
    LOG.debug("processSapBalance - end");
  }


  private void processSapStakoJob(SAPStakoJobVO stakoJob) throws Exception
  {
    ErrorVO err   = null;
    Stako   stako = new Stako(rfc);

    // init
    LOG.debug("processSapStakoJob - begin");

    // process stako job
    stako.process(stakoJob);

    // check status
    if ( ! stako.isErrorOccured() )
    {
      LOG.debug("processSapStakoJob - end (SAPStakoJobVO successfully processed)");
      trace (true,Jms.MSG_DESC_SAP_STAKO_JOB,stakoJob.toString());
    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processSapStakoJob - process SAPStakoJobVO failed: \n" + stako.getErrorMessage(),
                          ErrorVO.DATA_TYPE_CD_SAP_STAKO_JOB_VO, stakoJob,
                          null, ErrorVO.EMPL_NOTIFY_KEINE, ErrorVO.EMPL_MSG_TYPE_STAKO );
      saveError(err);

      LOG.error("processSapStakoJob - end (SAPStakoJobVO failed to process)");
      trace (false,Jms.MSG_DESC_SAP_STAKO_JOB,err.toString());
    }
  }


  private void processSapStako(StakoVO stakoVO) throws Exception
  {
    StringBuffer  retMsg  = new StringBuffer();
    int           retCode = RFC.RET_CODE_SUCCESS;
    ErrorVO       err     = null;

    // init
    LOG.debug("processSapStako - begin: artNr = " + stakoVO.artNr);

    // insert stako
    retCode = rfc.insertStako(stakoVO,retMsg);

    // handle retCode
    if (retCode == RFC.RET_CODE_SUCCESS)
    {
      LOG.debug("processSapStako - StakoVO successfully processed");
      trace(true,Jms.MSG_DESC_SAP_STAKO,stakoVO.toString());
    }
    else
    {
      err = new ErrorVO ( "SapInbound" + SapInbound.getProgramId(),
                          "processSapStako - process message StakoVO failed: \n" + retMsg.toString(),
                          ErrorVO.DATA_TYPE_CD_STAKO_VO, stakoVO,
                          null, ErrorVO.EMPL_NOTIFY_PENDENT, ErrorVO.EMPL_MSG_TYPE_STAKO );

      saveError(err);

      LOG.error("processSapStako - StakoVO failed to process");
      trace(false,Jms.MSG_DESC_SAP_STAKO,err.toString());
    }

    // reply (only in request reply mode)
    if (om.getJMSReplyTo() != null)
    {
      LOG.debug("processSapStako - StakoVO send REPLY");
      SapReturn sapReturn = new SapReturn(retCode,retMsg);
      queue.reply(om,sapReturn);
    }

    // end
    LOG.debug("processSapStako - end");
  }


  //---------------------------------------------------------------------------
  // private methods - misc
  //---------------------------------------------------------------------------

  private void trace (boolean actionOk, String actionDesc, String data)
  {
    TraceVO  vo;

    try
    {
      if (sapTrace == 0) return;                     // trace is switched off

      vo = new TraceVO();                             // assemble message
      vo.process    = SapInbound.getProgramName();
      vo.processNr  = SapInbound.getProgramId();

      if (actionOk)
        vo.action = vo.ACTION_OK;
      else
        vo.action = vo.ACTION_NOK;
      vo.action = vo.action + vo.ACTION_SEP + actionDesc;

      vo.data       = data;

      if (sapTrace == 1)                             // trace to stdout
      {
        System.out.println("### TRACE - START");
        System.out.println("TRACE-ACTION:");
        System.out.println(vo.action);
        System.out.println("TRACE-DATA:");
        System.out.println(vo.data);
        System.out.println("### TRACE - END");
      }
      else if (sapTrace == 2)                        // trace to ses db
      {
        LOG.debug("trace - write message to ses db");
        SapDelegateInbound.getInstance().receiveTrace(vo);
      }

    }
    catch (Exception ex)
    {
      LOG.error("trace - exception: " + ex.toString() );
    }
  }





} // end class

