package ch.ige.ses.client.sapinbound;

import java.util.*;
import java.sql.Date;
import org.apache.log4j.Logger;

import ch.ige.ses.vo.ses.*;
import ch.ige.ses.vo.bagis.*;
import ch.ige.ses.client.saputil.*;
import ch.ige.ses.server.sapctrl.SapDelegateInbound;

/**
 * SAP Inbound Client - Controlling Schnittstelle
 * Diese Klasse implementiert die Stako Verarbeitung von SES nach SAP
 * <p>
 * Copyright (c) 2003, Eidgenössisches Institut für Geistiges Eigentum
 * @author Stefan Nydegger
 * @version $Id: Stako.java,v 1.28 2006/12/11 12:43:32 iten Exp $
 */
public class Stako
{
  private static final Logger   LOG = Logger.getLogger(Stako.class);

  private SapDelegateInbound  fDelegate   = null;  // execute inbound requests
  private Date          bisDatum          = null;  // original bisDate from Date-Range
  private AutoRFC       fAutoRfc          = null;  // rfc handle auto mode
  private ManualRFC     fManualRfc        = null;  // rfc handle manual mode

  private int           fBlockSizeSesUpd  = 0;     // block-size (Number of days) for ses update
  private boolean       fErrOccured       = false; // state and message from calls
  private String        fErrMsg           = "";    // to process() & finish()
  private int           fCntError         = 0;
  private boolean       isEreignisMA      = true;  // modus: true  = select from tables ereignis/ma>
                                                   //        false = select from tables bfaktur/bfaktpos
  private String        fExclMa           = "";    //ma_nr of employee's
  /**
   * construct object for processing stako
   * @param rfc                 sap remote function call handle
   * @throws Exception          unexpected error
   */
  public Stako(RFC rfc) throws Exception
  {
    SapProperties sp = new SapProperties();

    // init delegate
    fDelegate = SapDelegateInbound.getInstance();

    // set rfc depending the program action
    if ( SapInbound.getProgramAction().equals("auto") )
    {
      fAutoRfc = (AutoRFC)rfc;
    }
    else
    {
      fManualRfc = (ManualRFC)rfc;
    }
    // get sap stako properties
    fBlockSizeSesUpd = sp.getInt(SapInbound.getProperties(),"sap.stako.block_size.ses_update",1);
    fExclMa = SapInbound.getProperties().getProperty("sap.stako.ma_nr.exclude");
  }

  /**
   * Methode process() starts either:<br>
   * - complete stako process including update of table artikel<br>
   * - restart of a failed stako process<br>
   * - only stako process without update of attribute standpreis on table artikel<br>
   * - only update the attribute standpreis on table artikel
   * @param sapStakoJobVO VO with modus
   */
  public void process(SAPStakoJobVO sapStakoJobVO)
  {
    LOG.debug ("Stako Method process - *** start Modus: " + sapStakoJobVO.modus + " ***");

    switch(sapStakoJobVO.modus)
    {
      // start new stako process
      case SAPStakoJobVO.STAKO_MODUS_START:
        startCompleteStakoProcess(sapStakoJobVO);
        break;
      // stako process without update of attribute standpreis on bagis table artikel
      case SAPStakoJobVO.STAKO_MODUS_NOUPDATE:
        startStakoProcess(sapStakoJobVO);
        break;
      // restart interrupted or failed stako process
      case SAPStakoJobVO.STAKO_MODUS_RESTART:
        restartStakoProcess(sapStakoJobVO);
        break;
      // update attribute standpreis on bagis table artikel
      case SAPStakoJobVO.STAKO_MODUS_UPDATE:
        updateArtikel();
        break;
    }
  } // end method

  /**
   * process stako controlling job
   * - updates attribute standpreis in table artikel<br>
   * - performs stako controlling process
   * @param sapStakoJobVO ValueObject with modus
   */
  private void startCompleteStakoProcess(SAPStakoJobVO sapStakoJobVO)
  {
    StakoKontrolleVO stakoKontrolleVO = null;
    // updates attribute standpreis in table artikel
    updateArtikel();
    // calculate vonDatum and bisDatum and insert row in stako_kontrolle
    stakoKontrolleVO = calculateDate(sapStakoJobVO);
    if(stakoKontrolleVO == null)
    {
      LOG.debug("Stako Method: startCompleteStakoProcess failed because stakoKontrolleVO is NULL!");
      return;
    }
    else
    {
      handleStakoProcess(stakoKontrolleVO);
    }
  }

  /**
   * handle stako process<br>
   *  - get list from BAGIS tables ereignis and ma with StakoVO<br>
   *  - divide list in block's ses update size<br>
   *  - process each block and update table stako_kontrolle<br>
   *  - get list from BAGIS tables bfaktur and bfaktpos with SAPRechnungVO to process<br>
   *  - divide list in block's ses update size<br>
   *  - process each block and update table stako_kontrolle<br>
   *  - finish stako_process<br>
   * @param stakoKontrolleVO Value Object
   * @throws Exception  unexpected error
   */
  private void handleStakoProcess(StakoKontrolleVO stakoKontrolleVO)
  {
    ArrayList         al               = null;  // list with all StakoKontrolleVO's

    try
    {
      bisDatum = stakoKontrolleVO.bisDat;     // store bisDatum
      // modus EreignisMA
      isEreignisMA = true;
      // read Ereignis/MA
      al = fDelegate.stakoSelectEreignisMA(stakoKontrolleVO.vonDat, stakoKontrolleVO.bisDat, fExclMa);
      LOG.debug("Stako Method handleStakoProcess Ereignis/MA " + al.size() + " VO selected");
      // update table stako_kontrolle with count of read records
      stakoKontrolleVO.anzahlEreignisMaGelesen += al.size();
      stakoKontrolleVO.herkunftStat = StakoKontrolleVO.STATUS_HERKUNFT_NOTHING_SAVED;
      fDelegate.stakoSetStakoKontrolle(stakoKontrolleVO);
      // divide list in blocks
      LOG.debug("Stako Method handleStakoProcess call processList");
      processList(al, stakoKontrolleVO);
      LOG.debug("Stako Method handleStakoProcess process List executed");
      // aktuelles stakoKontrolleVO lesen
      stakoKontrolleVO = fDelegate.stakoGetStakoKontrolle(stakoKontrolleVO.stakoKontrolleId);
      // restore bisDatum
      stakoKontrolleVO.bisDat = bisDatum;
      // modus Bfaktur/Bfaktpos
      isEreignisMA = false;
      // read bfaktur/bfaktpos
      al = fDelegate.stakoSelectBfakturBfaktPos(stakoKontrolleVO.vonDat, stakoKontrolleVO.bisDat, fExclMa);
      LOG.debug("process - processStako Bfaktur/Bfaktpos " + al.size() + " VO selected");
      // update table stako_kontrolle with count of read records
      stakoKontrolleVO.anzahlBfakturGelesen += al.size();
      fDelegate.stakoSetStakoKontrolle(stakoKontrolleVO);
      // divide list in blocks
      LOG.debug("Stako Method handleStakoProcess call processList with Bfaktur/Bfaktpos");
      processList(al, stakoKontrolleVO);
      // update stako controll and finish the stako process
      stakoKontrolleVO.bisDat = bisDatum;  // restore bisDatum
      stakoKontrolleVO.herkunftStat = StakoKontrolleVO.STATUS_HERKUNFT_NOTHING_SAVED;
      stakoKontrolleVO.stakoStat = fErrOccured ? StakoKontrolleVO.STATUS_VERARBEITUNG_FINISHED_WITH_ERRORS
                                              : StakoKontrolleVO.STATUS_VERARBEITUNG_FINISHED;
      fDelegate.stakoSetStakoKontrolle(stakoKontrolleVO);
    }
    catch (Exception ex)
    {
      fErrOccured = true;
      fErrMsg     = "Stako Method processStako failed, with exception in Method handleStakoProcess: " + ex.getMessage();
      LOG.error("process - " + fErrMsg);
    }
    // check for errors
    if ( fErrOccured )
    {
      fErrMsg = "finished with errors, processed records: " + getStatistic(stakoKontrolleVO);
      LOG.error("Stako Method handleStakoProcess " + fErrMsg);
      return;
    }
  } // end method

  /**
   * Divide list in blocks depending of ses update size or full list
   * @param al                ArrayList with StakoVO's
   * @param stakoKontrolleVO  Value Object
   */
  private void processList(ArrayList al, StakoKontrolleVO stakoKontrolleVO)
  {
    GregorianCalendar calBis       = new GregorianCalendar();
    GregorianCalendar calCount     = new GregorianCalendar();
    ArrayList         block        = null;
    double            count        = 0;
    int               numberOfRuns = 0;
    int               n             = 0;

    if (fBlockSizeSesUpd == 0)
    {
      processBlock(al, stakoKontrolleVO);
    }
    else
    {
      LOG.debug("Stako Methode processList: stakoKontrolleVO = \n" + stakoKontrolleVO.toString());
      calCount.setTime(stakoKontrolleVO.bisDat);
      //calculate first bisDatum: bisDatum = vonDatum + ses-size in days
      calBis.setTime(stakoKontrolleVO.vonDat);
      // calculate number of runs
      count = Math.abs(calCount.get(GregorianCalendar.DATE) - calBis.get(GregorianCalendar.DATE));
      count = count / fBlockSizeSesUpd;
      LOG.debug("Stako Method processList count = " + count);
      numberOfRuns = new Double(count).intValue() + 1;
      LOG.debug("numberOfRuns = " +numberOfRuns + " blockSize = " +fBlockSizeSesUpd);

      calBis.add(GregorianCalendar.DATE, fBlockSizeSesUpd);
      //new bisDatum has to be less then vo.bisDatum...
      if(calBis.getTime().getTime() > stakoKontrolleVO.bisDat.getTime())
      {
        calBis.setTime(stakoKontrolleVO.bisDat);
      }
      stakoKontrolleVO.bisDat = new java.sql.Date(calBis.getTime().getTime());

      for (int i = 0; i < numberOfRuns; i++)
      {
        block = new ArrayList();

        for (n = n; n < al.size() &&
             ((((StakoVO)al.get(n)).datum.getTime() < stakoKontrolleVO.bisDat.getTime())
             || i == numberOfRuns-1); n++)
        {
          block.add((StakoVO)al.get(n));
        }
        if(block.size() > 0)
        {
          // ArrayList mit StakoVO an SAP übermitteln
          processBlock(block, stakoKontrolleVO);
        }
        //calculate new end date
        calBis.add(GregorianCalendar.DATE, fBlockSizeSesUpd);
        //new bisDatum has to be less then VO.bisDatum...
        if(calBis.getTime().getTime() > calCount.getTime().getTime())
        {
          calBis.setTime(bisDatum);
        }
        stakoKontrolleVO.bisDat = new java.sql.Date(calBis.getTime().getTime());
      }
    }
  }

  /**
   * Calculates vonDatum and bisDatum and insert a new row in table stako_kontrolle
   * @param  sapStakoJobVO Value Object mit StakoJob Daten
   * @return StakoKontrolleVO ValueObject mit stako-kontroll daten
   */
  private StakoKontrolleVO calculateDate(SAPStakoJobVO sapStakoJobVO)
  {
    StakoKontrolleVO  newStakoVO       = new StakoKontrolleVO();
    StakoKontrolleVO  stakoKontrolleVO = null;
    GregorianCalendar cal              = new GregorianCalendar();

    try
    {
      LOG.debug("Stako Methode calculateDate - getLastStakoKontrolle");
      // get last row in table stako_kontrolle
      stakoKontrolleVO = fDelegate.stakoGetLastStakoKontrolle();
      if(stakoKontrolleVO == null)
      {
        fErrOccured = true;
        fErrMsg     = "Stako Method calculateDate, no initial row in table stako_kontrolle";
        LOG.error(fErrMsg);
        return null;
      }
      // calculate vonDatum
      cal.setTime(stakoKontrolleVO.bisDat);
      cal.add(GregorianCalendar.MONTH, 1);
      cal.set(GregorianCalendar.DATE, 1);
      newStakoVO.vonDat = new java.sql.Date(cal.getTime().getTime());
      // store new row
      fDelegate.stakoCreateStakoKontrolle(newStakoVO);
      // get created row
      stakoKontrolleVO = fDelegate.stakoGetLastStakoKontrolle();
      // calc bisDatum, will not be stored in table stako_kontrolle, is only used for select data
      if(sapStakoJobVO.bisDatum == null)
      {
        //bisDatum is last day of the month of vonDatum
        cal.add(GregorianCalendar.MONTH, 1);
        cal.add(GregorianCalendar.DATE, - 1);
      }
      else
      {
        cal.setTime(sapStakoJobVO.bisDatum);
      }
      stakoKontrolleVO.bisDat = new java.sql.Date(cal.getTime().getTime());
      LOG.debug("Stako Method calculate Date stakoKontrolleVO.bisDat = " + stakoKontrolleVO.bisDat);
      return stakoKontrolleVO;
    }
    catch (Exception ex)
    {
      fErrOccured = true;
      fErrMsg = "Stako Method processStakoKontrolle: Exception occured: " + ex.getMessage();
      return null;
    }
  }

  /** process block of StakoVO's
   *  - insert vo's into sap system
   *  - update table stako_kontrolle in ses db
   * @param al ArrayList of StakoVO
   * @param stakoKontrolleVO ValueObject
   */
  private void processBlock(ArrayList al, StakoKontrolleVO stakoKontrolleVO)
  {
    Iterator      it                   = al.iterator();
    StakoVO       vo                   = null;
    StringBuffer  retMsg               = new StringBuffer();     // sap call return message
    int           retCode              = RFC.RET_CODE_SUCCESS;   // sap call return code
    int           anzahlRecordsToSap   = 0;
    int           anzahlRecordsToSapOK = 0;

    // insert vo's into sap system
    while ( it.hasNext() )
    {
      vo    = (StakoVO) it.next();
      // assign kontroll_id
      vo.kontrollId = stakoKontrolleVO.stakoKontrolleId;
      if (retMsg.length() > 0)
        retMsg.delete(0,retMsg.length());

      // insert stako on sap (single commit)
      if ( fAutoRfc != null)
      {
        retCode = fAutoRfc.insertStako(vo ,retMsg);
      }
      else
      {
        retCode = fManualRfc.testInsertStako(vo);
      }
      // update record counts
        anzahlRecordsToSap++;
      // Errorhandling
      if (retCode == RFC.RET_CODE_SUCCESS)
      {
        anzahlRecordsToSapOK++;
      }
      else
      {
        fErrOccured = true;
        fCntError++;

        ErrorVO errorVO         = new ErrorVO();
        errorVO.employeeId      = null;
        errorVO.employeeMsgType = ErrorVO.EMPL_MSG_TYPE_STAKO;
        errorVO.employeeNotify  = ErrorVO.EMPL_NOTIFY_KEINE;
        errorVO.dataTypeCd      = ErrorVO.DATA_TYPE_CD_STAKO_VO;
        errorVO.data            = vo;
        errorVO.source          = "SapInbound" + SapInbound.getProgramId();
        errorVO.description     = "processBlock - stakoVO with art_nr " + vo.artNr +
                                  " failed: \n" + retMsg.toString();

        saveError(errorVO);
      }
    }

    // update table stako_kontrolle
    try
    {
      // setting count of read records depending on modus
      if(isEreignisMA)
      {
        stakoKontrolleVO.anzahlEreignisMaAnSap += anzahlRecordsToSap;
        stakoKontrolleVO.anzahlEreignisMaAnSapOK += anzahlRecordsToSapOK;
        stakoKontrolleVO.herkunftStat = StakoKontrolleVO.STATUS_HERKUNFT_MITARBEITER_SAVED;
      }
      else
      {
        stakoKontrolleVO.anzahlBfakturAnSap += anzahlRecordsToSap;
        stakoKontrolleVO.anzahlBfakturAnSapOK += anzahlRecordsToSapOK;
        stakoKontrolleVO.herkunftStat = StakoKontrolleVO.STATUS_HERKUNFT_ALL_SAVED;
      }
      fDelegate.stakoSetStakoKontrolle(stakoKontrolleVO);
      LOG.debug("Stako Method processBlock successfully processed.");
    }
    catch (Exception ex)
    {
      fErrOccured   = true;
      fErrMsg = "stako method processBlock - Exception occured: " + ex.getMessage();
      LOG.fatal(fErrMsg);
    }
  } // end method

  /**
   * restart an interrupted or failed stako process
   * @param sapStakoJobVO Value Object
   */
  private void restartStakoProcess(SAPStakoJobVO sapStakoJobVO)
  {
    StakoKontrolleVO  stakoKontrolleVO = null;
    GregorianCalendar cal              = new GregorianCalendar();
    GregorianCalendar calVon           = new GregorianCalendar();
    try
    {
      // get Data from cancelled Job
      stakoKontrolleVO = fDelegate.stakoGetLastStakoKontrolle();
      if(stakoKontrolleVO == null)
      {
        fErrOccured = true;
        fErrMsg =("Stako Method: restartStakoProcess failed because stakoKontrolleVO is NULL!");
        LOG.error(fErrMsg);
        return;
      }
      if(stakoKontrolleVO.stakoStat != StakoKontrolleVO.STATUS_VERARBEITUNG_CANCELLED)
      {
        fErrOccured = true;
        fErrMsg = "Stako Method restartStakoProcess: stako process cannot be restarted - " +
                  "restart is only allowed if STAKO_STAT == STATUS_VERARBEITUNG_CANCELLED";
        LOG.error(fErrMsg);
        return;
      }
      LOG.debug("Stako Method restartStakoProcess - calculate bisDatum");
      // new vonDatum  = (old bisDatum + 1)
      calVon.setTime(stakoKontrolleVO.bisDat);
      calVon.add(GregorianCalendar.DATE, 1);
      stakoKontrolleVO.vonDat = new java.sql.Date(calVon.getTime().getTime());
      // calculate bisDatum
      if(sapStakoJobVO.bisDatum == null)
      {
        //bisDatum is the last day of month from vonDatum
        cal.setTime(stakoKontrolleVO.vonDat);
        cal.set(GregorianCalendar.DATE, 1);
        cal.add(GregorianCalendar.MONTH, 1);
        cal.add(GregorianCalendar.DATE, - 1);
      }
      else
      {
        cal.setTime(sapStakoJobVO.bisDatum);
      }
      stakoKontrolleVO.bisDat = new java.sql.Date(cal.getTime().getTime());
      stakoKontrolleVO.stakoStat = StakoKontrolleVO.STATUS_VERARBEITUNG_STARTED;
      LOG.debug("Stako Method restartStakoKontrolle \n" + stakoKontrolleVO.toString());
      // process stako aufrufen
      handleStakoProcess(stakoKontrolleVO);
      LOG.debug("Stako Method: restartStakoProcess finished");
    }
    catch (Exception ex)
    {
      fErrOccured = true;
      fErrMsg = "Stako Method restartStakoProcess failed with exception: " + ex.getMessage();
      LOG.fatal(fErrMsg);
    }
  }

  /**
   *  start a stako process without updating attribute standpreis on table artikel
   *  @param sapStakoJobVO Value Object
   */
  private void startStakoProcess(SAPStakoJobVO sapStakoJobVO)
  {
    StakoKontrolleVO stakoKontrolleVO = null;

    // calculate vonDatum and bisDatum and insert row in stako_kontrolle
    stakoKontrolleVO = calculateDate(sapStakoJobVO);
    if(stakoKontrolleVO == null)
    {
      fErrOccured = true;
      fErrMsg = "Stako Method: startStakoProcess failed because stakoKontrolleVO is NULL!";
      LOG.debug(fErrMsg);
      return;
    }
    else
    {
      handleStakoProcess(stakoKontrolleVO);
    }
    LOG.debug("Stako Methode: startStakoProcess finished");
  }

  /**
   * select from SAP the stako relevantes articles,
   * reset the attribut standpreis of all records on the table artikel to 0 and
   * update the attribute standpreis of the selected articles to 1
   */
  private void updateArtikel()
  {
    int           retCode   = RFC.RET_CODE_SUCCESS;   // sap call return code
    ArrayList     al        = new ArrayList();
    StringBuffer  retMsg    = new StringBuffer();     // sap call return message

    if ( fAutoRfc != null)
    {
      //get stako relevant articles from SAP
      retCode = fAutoRfc.getArtikel(retMsg);
      // Fehlerhandling
      if (retCode != RFC.RET_CODE_SUCCESS)
      {
        ErrorVO errorVO         = new ErrorVO();
        errorVO.employeeId      = null;
        errorVO.employeeMsgType = ErrorVO.EMPL_MSG_TYPE_STAKO;
        errorVO.employeeNotify  = ErrorVO.EMPL_NOTIFY_PENDENT;
        errorVO.dataTypeCd      = ErrorVO.DATA_TYPE_CD_NO_DATA;
        errorVO.source          = "SapInbound" + SapInbound.getProgramId();
        errorVO.description     = "updateArtikel - failed with message: \n" + retMsg.toString();

        fErrOccured = true;
        fErrMsg     = errorVO.description;

        LOG.error(fErrMsg);
        saveError(errorVO);
      }

      al = fAutoRfc.getArtAL();
    }
    else
    {
      al = fManualRfc.testGetArtikel();
    }

    // update artikel state
    try
    {
      fDelegate.stakoUpdateArtikelStatus(al);
      LOG.debug("Stako Method updateArtikel : artikel updated");
    }
    catch (Exception ex)
    {
      fErrOccured = true;
      fErrMsg = "Stako Method updateArtikel failed with exception: " + ex.getMessage();
      LOG.fatal(fErrMsg);
    }
  }

  /**
   * override value from property file
   * note:
   * this method is only used in manual mode.
   * @param size block size for ses updates
   */
  public void setBlockSizeSesUpdate(int size)
  {
    fBlockSizeSesUpd = size;
  }

  /**
   * get error message as string
   * @return message
   */
  public String getErrorMessage()
  {
    LOG.error("fErrMsg = " + fErrMsg);
    return fErrMsg;
  }

  /**
   * indicate error state of process() and updateArtikel() method calls
   * @return state  true = error occured, false = no error occured
   */
  public boolean isErrorOccured()
  {
    return fErrOccured;
  }

  /**
  * gather statistic info about the processed records by calling the
  * method process().
  * @param stakoKontrolleVO Value Object
  * @return statistic info
  */
 private String getStatistic(StakoKontrolleVO stakoKontrolleVO)
 {
   int totalAnSap = stakoKontrolleVO.anzahlEreignisMaAnSap + stakoKontrolleVO.anzahlBfakturAnSap;
   int sapOK = stakoKontrolleVO.anzahlBfakturAnSapOK + stakoKontrolleVO.anzahlEreignisMaAnSapOK;
   int sapNotOK = totalAnSap - sapOK;

    return ( "processed records: " +
             "total="  + totalAnSap  + " : " +
             "sap ok=" + sapOK + ", sap err=" + sapNotOK + " : " );
  }


  private boolean saveError (ErrorVO err)
  {
    try
    {
      LOG.info("saveError - save vo to ses error db");
      SapDelegateInbound.getInstance().receiveError(err);
      return true;
    }
    catch (Exception ex)
    {
      LOG.fatal("saveError - save vo to ses error db failed: " + ex.getMessage() );

      // save error vo to local file
      new ErrorFile ( SapInbound.getProperties().getProperty("sapinbound.data.path"),
                      SapInbound.getProgramIdentifier() ).save(err);

      return false;
    }
  }


} // end class

