package ch.ige.ses.client.saputil;

import java.util.*;
import java.io.*;
import org.apache.log4j.*;
import ch.ige.util.crypto.VigenereCypher;

/**
 * Handling SAP Property - Files for SAP Inbound Client and SAP Outbound Server.
 * <p>
 * Copyright (c) 2002, Eidgenössisches Institut für Geistiges Eigentum
 * @author T. Iten
 * @version $Id: SapProperties.java,v 1.14 2006/12/29 09:20:36 iten Exp $
 */
public class SapProperties
{
  //
  // const
  //
  private static final Logger LOG = Logger.getLogger(SapProperties.class);
  private static final String VCK = "A-Z";        // key
  private static final String VCD = "#";          // delimiter

  public static final String  PLACEHOLDER_PID = "$$PID";


  //---------------------------------------------------------------------------
  // constructors
  //---------------------------------------------------------------------------
  public SapProperties()
  {
    // lazy bone
  }

  //---------------------------------------------------------------------------
  // public methods
  //---------------------------------------------------------------------------

  /**
   * get interger value from properties. returns the default value if the
   * property isn't available or the conversion fails.
   *
   * @param  prop   properties to read from
   * @param  name   name of property
   * @param  def    default value
   *
   * @return integer value
   */
  public int getInt(Properties prop, String name, int def)
  {
    int value = def;

    try
    {
      value = Integer.parseInt(prop.getProperty(name));
    }
    catch (Exception ex)
    {
      LOG.warn("getInt - reading property " + name + " failed. using default value !");
      value = def;
    }

    return value;

  } // end method


  /**
   * replace all placeholders with the given value within the properties.<br>
   * note:  the placeholders will only be replaced in the value section of the
   *        proprty files.
   *
   * @param  prop         properties to work on
   * @param  placeholder  name of placeholder
   * @param  value        value for the given placeholder
   *
   * @return Properies with replaced values
   */
  public Properties setPlaceholderValue(Properties prop, String placeholder, String value)
  {
    Enumeration e;
    String      key,val;
    int         idx;


    e = prop.propertyNames();

    while ( e.hasMoreElements() )
    {
      key = e.nextElement().toString();
      val = prop.getProperty(key);

      idx = val.indexOf(placeholder);
      while (idx != -1 && placeholder != value)
      {
        val = val.substring(0,idx) + value + val.substring(idx + placeholder.length());
        prop.setProperty(key,val);
        idx = val.indexOf(placeholder);
      }
    } // end while

    return prop;

  } // end method


  /**
   * set system properties (i.e. define java environment, in which the
   * application will run). this also defines against which app server the
   * client will run !
   * 
   * if the value is set to 'default' the system property will not be set.
   * in this case the default provider url delivered by the masterbuild
   * will be used!
   *
   * @param  prop Properties
   *
   * @exception Exception Properties konnten nicht eingelesen werden
   */
  public void setSystemProperties(Properties prop) throws Exception
  {
    String url  = prop.getProperty("java.naming.provider.url");
    
    if ( ! url.equalsIgnoreCase("default") )
    {
      System.setProperty("java.naming.provider.url",url);
    }

  } // end method


  /**
   * Einlesen der Programmeigenschaften aus den Property Datei.
   * Die Datei enthält neben den gemeinsamen Properties auch alle plattform-
   * abhängigen Einstellungen. Diese beginnen jeweils mit folgenden
   * Bezeichnungen:
   *              dev       Development (Entwicklung)
   *              test      Testsystem  (Integration)
   *              acc       Acceptance  (Acceptance)
   *              prod      Production  (Production)
   *
   * Beachte:  Beim Start des Programms wird als Startparameter angegeben, für
   *           welche Umgebung der Prozess gestartet wird. Das Programm liest
   *           anschliessend die entsprechenden Properties ein. Innerhalb der
   *           Applikation können die Properties anschliessend ohne die platt-
   *           formabhängige Bezeichnung gelesen werden.
   *
   *           Beispiel: Die Property dev.jco.clientpool.key wird innerhalb der
   *                     Applikation mit jco.clientpool.key referenziert.
   *
   *
   * @param  file     Pfad und Name der Propertydatei
   * @param  env      Umgebung ïn welcher das Programm läuft. Der mögliche
   *                  Wertebereich ist im Array PROGRAM_ENV_TYPES definiert.
   * @param  envTypes Array mit Environment Types
   *
   * @return Properties Eingelesene Properties
   *
   * @exception Exception Properties konnten nicht eingelesen werden
   */
  public Properties load(String file, String env,String[] envTypes) throws Exception
  {
    Properties  prop      = new Properties();
    Properties  propnew   = new Properties();
    InputStream is;

    Enumeration e;
    String      key;
    String      newkey;
    boolean     check;

    try
    {
      //
      // check param env
      //
      check = false;
      for (int i = 0; i < envTypes.length; i++)
      {
        if (envTypes[i].equals(env))
          check = true;
      }
      if (! check)
        throw new Exception("Parameter env (" + env + ") hat falschen Wert !");


      //
      // load all properties
      //
      is = new FileInputStream(file);

      if (is == null)
        throw new Exception("Property Datei (" + file + ") nicht gefunden !");

      prop.load(is);

      //
      // filter needed properties
      //
      e = prop.propertyNames();
      while ( e.hasMoreElements() )
      {
        key = e.nextElement().toString();

        if (key.indexOf(env + ".") == 0)
        {
          newkey = key.substring(env.length()+1);
        } else {
          newkey = key;
        }

        check = true;
        for (int i = 0; i < envTypes.length; i++)
        {
          if (newkey.indexOf(envTypes[i] + ".")  == 0)
            check = false;
        }
        if (check)
          propnew.setProperty(newkey,prop.getProperty(key));

      } // end while

      //
      // return properties
      //
      return propnew;

    }
    catch (Exception ex)
    {
      LOG.error("load - " + ex.toString());
      throw new Exception("load - " + ex.toString());
    }

  } // end method


  /**
   * Encode Property Text
   *
   * @param   plainText   input data
   * @return  String      encoded output data
   */
  public static String encodeProperty (String plainText)
  {
    String buf = new VigenereCypher(VCK).encode(plainText);
    return (VCD + buf + VCD);
  }


  /**
   * Decode Property Text
   *
   * @param   encodedText   input data
   * @return  String        decoded output data
   */
  public static String decodeProperty (String encodedText)
  {
    String buf = encodedText.substring ( VCD.length(),
                                         encodedText.length()-VCD.length() );
    return new VigenereCypher(VCK).decode(buf);
  }

} // end class


