package ch.ige.ses.server.bagisoutbound;

import java.util.ArrayList;

import javax.ejb.EJBException;
import javax.ejb.SessionBean;
import javax.ejb.SessionContext;
import javax.naming.Context;
import javax.naming.InitialContext;

import org.apache.log4j.Logger;

import ch.ige.ses.server.dao.bagis.BfakturDAO;
import ch.ige.ses.server.dao.bagis.JoinSAPRechnungDAO;
import ch.ige.ses.server.error.SESErrorWorkLocal;
import ch.ige.ses.server.error.SESErrorWorkLocalHome;
import ch.ige.ses.server.sapctrl.SapDelegateInbound;
import ch.ige.ses.server.sapctrl.SapInboundCtrlLookup;
import ch.ige.ses.util.ReplaceTexte;
import ch.ige.ses.vo.bagis.SAPRechnungVO;
import ch.ige.ses.vo.bagis.SAPRechnungsPositionVO;
import ch.ige.ses.vo.ses.AmvRechnungVO;
import ch.ige.ses.vo.ses.ErrorVO;

/**
 * Bean zur Verarbeitung von BAGIS-Rechnungen und Druckvorschau
 * <p>
 * Copyright (c) 2002, Eidgenössisches Institut für Geistiges Eigentum
 * @author   Pirmin Suter
 * @version  $Id: SESRechnungBean.java,v 1.76 2007/12/24 10:13:23 iten Exp $
 * 
 * @ejb.bean 
 *  name="SESRechnung"  
 *  display-name="SESRechnung"
 *  view-type="local"
 *  jndi-name="SESRechnung"
 *  local-jndi-name="SESRechnungLocal"
 *  type="Stateless"
 *  transaction-type="Container"
 * @ejb.transaction 
 *  type="Required"
 * @ejb.ejb-ref
 *   ejb-name="SESErrorWork"
 *   view-type="local"
 * @ejb.ejb-ref
 *   ejb-name="SapInboundCtrl"
 *   view-type="local"
 */
public class SESRechnungBean implements SessionBean
{
  
  //
  // const
  //
  private static final Logger LOG = Logger.getLogger(SESRechnungBean.class);

  //
  // fields
  //
  SessionContext          sessionContext;
  SapInboundCtrlLookup    sapInboundCtrlLookup = new SapInboundCtrlLookup();

  //---------------------------------------------------------------------------
  //  bean methods
  //---------------------------------------------------------------------------

  /**
   * create ejb
   */
  public void ejbCreate()
  {
    // lazy bone
  }

  public void ejbRemove()
  {
    // lazy bone
  }

  public void ejbActivate()
  {
    // lazy bone
  }

  public void ejbPassivate()
  {
    // lazy bone
  }

  public void setSessionContext(SessionContext sessionContext)
  {
    this.sessionContext = sessionContext;
  }

  //---------------------------------------------------------------------------
  //  public methods
  //---------------------------------------------------------------------------

  /**
   * Neue Rechnungsdaten auf BAGIS lesen und in SAP anlegen
   *
   * @param bfaktId Rechnungsnummer
   * 
   * @ejb.interface-method
   */
  public void createRechnung (int bfaktId)
  {
    try
    {
      SAPRechnungVO vo;

      // get data
      vo = getRechnung(bfaktId);
      vo.hasChanged = "0";                        // neue Rechnung anlegen

      // check max. empfänger
      checkMaxEmpf(vo);

      // send data to sap inbound service
      sapInboundCtrlLookup.getLocalInterface().sendRechnung(vo);
    }
    catch (Exception ex)
    {
      LOG.error("createRechnung - Exception: " + ex.getMessage());
      throw new EJBException(ex.getMessage());
    }
  }


  /**
   * Bestehende Rechnungsdaten auf BAGIS lesen und in SAP bestehende
   * Rechnung mutieren.
   *
   * @param bfaktId Rechnungsnummer
   * 
   * @ejb.interface-method
   */
  public void changeRechnung (int bfaktId)
  {
    try
    {
      SAPRechnungVO vo;

      // get data
      vo = getRechnung(bfaktId);
      vo.hasChanged = "1";                        // best. Rechnung mutieren

      // check max. empfänger
      checkMaxEmpf(vo);

      // send data to sap inbound service
      sapInboundCtrlLookup.getLocalInterface().sendRechnung(vo);
    }
    catch (Exception ex)
    {
      LOG.error("changeRechnung - Exception: " + ex.getMessage());
      throw new EJBException(ex.getMessage());
    }
  }


  /**
   * Rechnungsdaten auf BAGIS lesen und zurückgeben (für Druckvorschau).
   * Beachte: Die Überprüfung der max. Anzahl Empfänger wird nicht gemacht.
   *
   * @param   bfaktId         Rechnungsnummer
   * @return  SAPRechnungVO   Rechnungsdaten
   * 
   * @ejb.interface-method
   */
  public SAPRechnungVO previewRechnung (int bfaktId)
  {
    try
    {
      return getRechnung(bfaktId);
    }
    catch (Exception ex)
    {
      LOG.error("previewRechnung - Exception: " + ex.getMessage());
      throw new EJBException(ex.getMessage());
    }
  }


  /**
   * Liest alle für die automatische Markenverlängerung bereitstehenden
   * Rechnungsnummern auf BAGIS ein für den angegebenen Erinnerungstyp.
   *
   * @param   amvRechnungCd   Erinnerungstyp 1 oder 2
   * @return  ArrayList       Integer Liste mit Rechnungsnummern (bfakt_id's)
   * 
   * @ejb.interface-method
   */
  public ArrayList getRechnungNrListAMV(int amvRechnungCd)
  {
    try
    {
      String      statusDruck = null;
      ArrayList   list        = null;

      // init
      LOG.debug("getRechnungNrListAMV (amvRechnungCd=" + amvRechnungCd + ") - starting...");

      switch (amvRechnungCd)
      {
        case AmvRechnungVO.RE_CD_ERINNERUNG_1:
          statusDruck = "4";
          break;
        case AmvRechnungVO.RE_CD_ERINNERUNG_2:
          statusDruck = "5";
          break;
        default:
          throw new Exception( "Aufruf getRechnungNrListAMV() mit ungültigen Parameter. " +
                               "amvRechnungCd = " + amvRechnungCd + " ist nicht erlaubt." );
      }

      // get data
      list = new BfakturDAO().selectBfaktIdForAMV(statusDruck);
      LOG.debug("getRechnungNrListAMV - " + list.size() + " records found");

      // end
      LOG.debug("getRechnungNrListAMV - successfully done");
      return list;

    }
    catch (Exception ex)
    {
      LOG.error("getRechnungNrListAMV - Exception: " + ex.getMessage());
      throw new EJBException(ex.getMessage());
    }
  }


  /**
   * Neue Rechnungsdaten auf BAGIS lesen und in SAP anlegen für die
   * Erinnerungen 1 und 2 im Rahmen der automatischen Markenverlängerungen.
   *
   * @param amvLaufId       Lauf Id der Verarbeitung
   * @param amvLaufNr       Lauf Nr der Verarbeitung
   * @param amvRechnungCd   Erinnerungstyp 1 oder 2
   * @param bfaktId         Rechnungsnummer
   * 
   * @ejb.interface-method
   */
  public void createRechnungAMV ( int amvLaufId, String amvLaufNr,
                                  int amvRechnungCd, int bfaktId )
  {
    SAPRechnungVO vo;

    try
    {
      // init
      LOG.debug("createRechnungAMV (amvLaufId=" + amvLaufId +
        ",amvLaufNr=" + amvLaufNr + ",amvRechnungCd=" + amvRechnungCd +
        ",bfaktId=" + bfaktId + ") - starting...");

      // get data
      vo = getRechnung(bfaktId);
      vo.amvLaufId      = amvLaufId;
      vo.amvLaufNr      = amvLaufNr;
      vo.amvRechnungCd  = amvRechnungCd;

      // check max. empfänger
      checkMaxEmpf(vo);

      // send data to sap inbound service
      sapInboundCtrlLookup.getLocalInterface().sendRechnung(vo);

      // end
      LOG.debug("createRechnungAMV - successfully done");

    }
    catch (Exception ex)
    {
      LOG.error("createRechnungAMV - Exception: " + ex.getMessage());
      throw new EJBException(ex.getMessage());
    }
    finally
    {
      // enable faster clean up from garbage collector
      vo = null;
    }
  }


  //---------------------------------------------------------------------------
  //  private methods
  //---------------------------------------------------------------------------

  /**
   * Rechnungsdaten von BAGIS lesen.
   *
   * @param   bfaktId         Rechnungsnummer
   * @throws  Exception       Unerwarteter Fehler
   * @return  SAPRechnungVO   Rechnungsdaten
   */
  private SAPRechnungVO getRechnung(int bfaktId) throws Exception
  {
    ReplaceTexte        rt  = new ReplaceTexte();
    JoinSAPRechnungDAO  dao = new JoinSAPRechnungDAO();
    SAPRechnungVO       vo  = null;
    String              buf = null;
    String              logHead = "getRechnung(bfakt_id=" + bfaktId + ") - ";

    try
    {
      LOG.debug(logHead + "### started...");

      // get data from bagis db
      LOG.debug(logHead + "get data from bagis db");
      vo = dao.readRechnung(bfaktId);
      if (vo == null)
      {
        buf = "Ungültige Rechnungsnummer. Für die bfakt_id <" + bfaktId + "> " +
              "wurde kein Eintrag gefunden !";
        LOG.error("getRechnung - " + buf);
        throw new SESException(SESException.MASK,buf);
      }

      // check employeeId
      LOG.debug(logHead + "check employeeId");
      if(new Integer(vo.employeeId).intValue() <= 0)
      {
        buf = "Ungültige Employee_Id <" + vo.employeeId + "> " +
              "des Rechnung erstellenden Mitarbeiters festgestellt !";
        LOG.error("getRechnung - " + buf);
        throw new SESException(SESException.MASK,buf);
      }

      // handling kombi divers
      LOG.debug(logHead + "handling kombi divers");
      if (vo.posList.size() > 1)
      {
        vo = handlingKombiDivers(vo);
      }

      // replace placeholder within text fields
      LOG.debug(logHead + "replace placeholder within text fields");
      rt.PrepareTexte(vo);

      // return
      LOG.debug(logHead + "### successfully done");
      return vo;
    }
    catch (SESException se)
    {
      throw se;
    }
    catch (Exception ex)
    {
      buf = "Ein unerwarteter Fehler ist aufgetreten : " + ex.toString();
      LOG.error("getRechnung(bfakt_id=" + bfaktId + ") - " + buf);
      throw new SESException(SESException.MASK,buf);
    }
    finally
    {
      // enable faster clean up from garbage collector
      rt  = null;
      dao = null;
      vo  = null;
      buf = null;
    }
  }


  /**
   * Im Falle von Divers-Rechnung werden die einzelnen Position mit Mengen-
   * rabattpreisen zu einer Position zusammengefasst. Bei der Zusammenfassung
   * wird der Artikeltext und der Preis (BFAKTPOS.PREIS) der 2. Position
   * genommen.
   *
   * @param  vo              Rechnungsdaten
   * @return SAPRechnungVO   Rechnungsdaten verarbeitet
   */
  private SAPRechnungVO handlingKombiDivers(SAPRechnungVO vo)
  {
    SAPRechnungsPositionVO positionToCompare  = null;
    SAPRechnungsPositionVO positionActual     = null;
    SAPRechnungsPositionVO positionNew        = null;
    SAPRechnungsPositionVO positionDebug      = null;
    String KombiDiversTrue                    = "";


    // Herausfinden, ob es unterschiedliche SCHUTIT_KEY's hat
    for (int i = 0; i < vo.posList.size(); i++)
    {
      if(i == 0)
      {
        positionToCompare = (SAPRechnungsPositionVO)vo.posList.get(i);
      }
      else if (i > 0)
      {
        positionActual = (SAPRechnungsPositionVO)vo.posList.get(i);
      }

      if(i > 0 && positionToCompare.schutit_key != positionActual.schutit_key)
      {
        KombiDiversTrue = "true";
        break;
      }
    }

    // KombiDivers
    if (KombiDiversTrue.equals("true"))
    {
      LOG.debug("posList.size: " + vo.posList.size());
      for (int i = 0; i < vo.posList.size(); i++)
      {
        LOG.debug("i: " + i);
        if (i == 1)
        {
          positionNew = (SAPRechnungsPositionVO)vo.posList.get(i);
          positionNew.anzahl = vo.posList.size()-1;
          LOG.debug("positionNew.anzahl: " + positionNew.anzahl);
          vo.posList.set(i, positionNew);
        }
      }

      for (int i = vo.posList.size()-1; i > 1; i--)
      {
        LOG.debug("i: " + i);
        positionDebug = (SAPRechnungsPositionVO)vo.posList.get(i);
        LOG.debug("Positionsnummer: " + positionDebug.bfakt_posnr);
        vo.posList.remove(i);
      }
    }

    // end
    return vo;
  }


  /**
   * Falls die Rechnungsdaten die maximale Anzahl Rechnugsempfäner über-
   * schreitet werden die Rechnungsdaten mit einer entsprechenden Meldung
   * in der Tabelle Error gespeichert.
   *
   * Detailsbeschreibung:
   * Falls mehr als SAPRechnungVO.MAX_RECH_EMPF Zusatzempfänger existieren,
   * so werden diese im Feld zusatzRechEmpfNichtVerarbeitet gespeichert, da
   * sie beim Anlegen der Rechung in SAP nicht berücksichtigt werden.
   *
   * @param   vo  zu prüfende Rechnungsdaten
   */
  private void checkMaxEmpf (SAPRechnungVO vo)
  {
    if ( vo.zusatzRechEmpfNichtVerarbeitet != null &&
         vo.zusatzRechEmpfNichtVerarbeitet.size() > 0 )
    {
      LOG.debug ( "checkMaxEmpf - Nicht verarbeitete zusätzliche Rechnungsempfänger " +
                  "bei bfakt_id " + vo.bfakt_id );

      ErrorVO errorVO = new ErrorVO();

      errorVO.source          = "SESRechnungBean";
      errorVO.employeeId      = vo.employeeId;
      errorVO.employeeNotify  = ErrorVO.EMPL_NOTIFY_PENDENT;
      errorVO.employeeMsgType = ErrorVO.EMPL_MSG_TYPE_MAX_EMPF;
      errorVO.description     = "Nicht verarbeitete zusätzliche Rechnungsempfänger";
      errorVO.dataTypeCd      = ErrorVO.DATA_TYPE_CD_SAP_RECHNUNG_VO;
      errorVO.errorStat       = ErrorVO.ERROR_STAT_PENDENT;
      errorVO.data            = vo;

      // save error
      saveError(errorVO);

      // enable faster clean up from garbage collector
      errorVO = null;
    }
  }


  /**
   * save error vo into ses table error
   *
   * @param vo ErrorVO to store
   */
  private void saveError(ErrorVO vo)
  {
    try
    {
      SESErrorWorkLocalHome   sesErrorWorkLocalHome   = null;
      SESErrorWorkLocal       sesErrorWorkLocal       = null;

      Context ctx = new InitialContext();
      Object ref = ctx.lookup(sesErrorWorkLocalHome.COMP_NAME);

      sesErrorWorkLocalHome = (SESErrorWorkLocalHome)ref;
      sesErrorWorkLocal = sesErrorWorkLocalHome.create();

      sesErrorWorkLocal.sendError(vo);
    }
    catch (Exception ex)
    {
      LOG.error( "saveError - could NOT store ErrorVO into table error \n" +
                 " Exception: " + ex.getMessage() + " \n" +
                 " ErrorVO: " + vo.toString() );
    }
  }


} // end class






