package ch.ige.ses.server.dao.ses;

import ch.ige.ses.vo.ses.*;
import java.sql.*;
import java.sql.Date;
import java.util.*;

import ch.ige.util.server.DBHelper;

/**
 * Data Access Objekt für die Tabelle AMV_RECHNUNG
 * <p>
 * Copyright (c) 2004, Eidgenössisches Institut für Geistiges Eigentum</p>
 * @author T. Iten
 * @version $Id: AmvRechnungDAO.java,v 1.6 2006/12/13 14:54:56 iten Exp $
 */

public class AmvRechnungDAO
{
  //
  // const
  //
  public static final int FLD_LEN_ERROR_MSG   = 1000;


  //---------------------------------------------------------------------------
  // constructor
  //---------------------------------------------------------------------------
  public AmvRechnungDAO()
  {
    // lazy bone
  }

  //---------------------------------------------------------------------------
  // public methods
  //---------------------------------------------------------------------------

  /**
   * Fügt einen Datensatz in die Tabelle AMV_RECHNUNG ein. Falls das Feld
   * ErrorMsg ungleich null ist, so wird dieses auch eingefügt.
   *
   * @param       vo              Value Object
   * @exception   SQLException    Unerwarteter SQL Fehler
   */
  public void insert(AmvRechnungVO vo) throws SQLException
  {
    Connection        con = null;
    PreparedStatement ps  = null;

    try
    {
      // limit maximal field size
      if (vo.errorMsg != null)
      {
        if (vo.errorMsg.length() > FLD_LEN_ERROR_MSG)
          vo.errorMsg = vo.errorMsg.substring(0,FLD_LEN_ERROR_MSG-3) + "...";
      }

      // insert vo
      con = DBHelper.getConnection(DBHelper.SES);

      if (vo.errorMsg == null)
      {
        ps = con.prepareStatement (
              "insert into ses.amv_rechnung " +
              "(amv_lauf_id, rechnung_cd, bfakt_id, anz_rech_empf, rechnung_stat, created_dat, changed_dat) " +
              "values (?, ?, ?, ?, ?, sysdate, sysdate)" );
      }
      else
      {
        ps = con.prepareStatement (
              "insert into ses.amv_rechnung " +
              "(amv_lauf_id, rechnung_cd, bfakt_id, anz_rech_empf, rechnung_stat, error_msg, created_dat, changed_dat) " +
              "values (?, ?, ?, ?, ?, ?, sysdate, sysdate)" );
      }

      ps.setInt (1,vo.amvLaufId);
      ps.setInt (2,vo.rechnungCd);
      ps.setInt (3,vo.bfaktId);
      ps.setInt (4,vo.anzRechEmpf);
      ps.setInt (5,vo.rechnungStat);
      if (vo.errorMsg != null)
      {
        ps.setObject(6,vo.errorMsg);
      }

      ps.executeUpdate();
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }

  /**
   * Selektiert einen Datensatz der Tabelle AMV_RECHNUNG
   *
   * @param amvLaufId    PK AmvLauf
   * @param rechnungCd   PK Rechnungstyp
   * @param bfaktId      PK Rechnungsnummer
   * @return AmvRechnungVO mit Rechnungs-Informationen
   * @throws Exception Fehler
   */
  public AmvRechnungVO select( int amvLaufId, int rechnungCd, int bfaktId ) throws SQLException
  {
    Connection          con = null;
    PreparedStatement   ps  = null;
    ResultSet           rs  = null;
    AmvRechnungVO       amvRechnungVO = null;

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);
      ps  = con.prepareStatement
              ( getSelectStatement() +
                " where amv_lauf_id=? and rechnung_cd=? and bfakt_id=?" );

      ps.setInt(1,amvLaufId);
      ps.setInt(2,rechnungCd);
      ps.setInt(3,bfaktId);

      rs = ps.executeQuery();
      if ( rs.next() )
      {
        amvRechnungVO = fillAmvRechnungVO(rs);
      }

      return amvRechnungVO;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Selektiert eine Array Liste mit Datensätzen der Tabelle AMV_RECHNUNG
   * für die gegebene LaufId sortiert nach RechnungCd und BfaktId.
   *
   * Werden keine Daten gefunden wird eine leere Liste zurückgegeben.
   *
   * @param   amvLaufId     Lauf Id
   * @throws  SQLException  Unerwarteter SQL-Fehler
   * @return  ArrayList     Data Access Object oder null
   */
  public ArrayList select(int amvLaufId) throws SQLException
  {
    Connection          con = null;
    PreparedStatement   ps  = null;
    ResultSet           rs  = null;
    ArrayList           al  = new ArrayList();

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);
      ps  = con.prepareStatement
              ( getSelectStatement() +
                " where amv_lauf_id=?" +
                " order by rechnung_cd,bfakt_id" );

      ps.setInt(1,amvLaufId);

      rs = ps.executeQuery();
      while ( rs.next() )
      {
        al.add( fillAmvRechnungVO(rs) );
      }

      return al;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Selektiert eine Array Liste mit Datensätzen der Tabelle AMV_RECHNUNG
   * für die gegebene LaufId und RechnungCd sortiert nach BfaktId.
   *
   * Werden keine Daten gefunden wird eine leere Liste zurückgegeben.
   *
   * @param   amvLaufId     Lauf Id
   * @param   rechnungCd    Erinnerungstyp
   * @throws  SQLException  Unerwarteter SQL-Fehler
   * @return  ArrayList     Data Access Object oder null
   */
  public ArrayList selectByRechnungCd(int amvLaufId, int rechnungCd) throws SQLException
  {
    Connection          con = null;
    PreparedStatement   ps  = null;
    ResultSet           rs  = null;
    ArrayList           al  = new ArrayList();

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);
      ps  = con.prepareStatement
              ( getSelectStatement() +
                " where amv_lauf_id=? and rechnung_cd=?" +
                " order by bfakt_id" );

      ps.setInt(1,amvLaufId);
      ps.setInt(2,rechnungCd);

      rs = ps.executeQuery();
      while ( rs.next() )
      {
        al.add( fillAmvRechnungVO(rs) );
      }

      return al;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Selektiert eine Array Liste mit Datensätzen der Tabelle AMV_RECHNUNG für
   * die gegebene LaufId und RechnungStat sortiert nach rechnungCd und BfaktId.
   *
   * Werden keine Daten gefunden wird eine leere Liste zurückgegeben.
   *
   * @param   amvLaufId     Lauf Id
   * @param   rechnungStat  Status
   * @throws  SQLException  Unerwarteter SQL-Fehler
   * @return  ArrayList     Data Access Object oder null
   */
  public ArrayList selectByRechnungStat(int amvLaufId, int rechnungStat) throws SQLException
  {
    Connection          con = null;
    PreparedStatement   ps  = null;
    ResultSet           rs  = null;
    ArrayList           al  = new ArrayList();

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);
      ps  = con.prepareStatement
              ( getSelectStatement() +
                " where amv_lauf_id=? and rechnung_stat=?" +
                " order by rechnung_cd,bfakt_id" );

      ps.setInt(1,amvLaufId);
      ps.setInt(2,rechnungStat);

      rs = ps.executeQuery();
      while ( rs.next() )
      {
        al.add( fillAmvRechnungVO(rs) );
      }

      return al;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Selektiert eine Array Liste mit Datensätzen der Tabelle AMV_RECHNUNG für
   * die gegebene LaufId und RechnungStat sortiert nach rechnungCd und BfaktId.
   *
   * Werden keine Daten gefunden wird eine leere Liste zurückgegeben.
   *
   * @param   amvLaufId     Lauf Id
   * @param   rechnungStat  Status
   * @throws  SQLException  Unerwarteter SQL-Fehler
   * @return  ArrayList     Data Access Object oder null
   */
  public ArrayList selectIdByCdAndStat ( int amvLaufId,
                                         int rechnungCd, int rechnungStat) throws SQLException
  {
    Connection          con = null;
    PreparedStatement   ps  = null;
    ResultSet           rs  = null;
    ArrayList           al  = new ArrayList();

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);
      ps  = con.prepareStatement
              ( "select bfakt_id " +
                "from   ses.amv_rechnung " +
                "where  amv_lauf_id=? and rechnung_cd=? and rechnung_stat=? " +
                "order by bfakt_id" );

      ps.setInt(1,amvLaufId);
      ps.setInt(2,rechnungCd);
      ps.setInt(3,rechnungStat);

      rs = ps.executeQuery();
      while ( rs.next() )
      {
        al.add( new Integer(rs.getInt("bfakt_id")) );
      }

      return al;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Selektiert pro Verarbeitungsstatus die summierte Anzahl Rechnungsempfänger
   * der Tabelle AMV_RECHNUNG für die gegebene LaufId und RechnungCd.
   * Die Summe pro Verarbeitungsstatus wird in einer Map zurückgegeben.
   * Werden keine Daten gefunden wird eine leere Map zurückgegeben.
   *
   * @param   amvLaufId     Lauf Id
   * @param   rechnungCd    Erinnerungstyp
   * @throws  SQLException  Unerwarteter SQL-Fehler
   * @return  Map           beinhaltet summierten Rechnungsempfänger pro Verarbeitungsstatus
   */
  public Map selectGroupByStat(int amvLaufId, int rechnungCd) throws SQLException
  {
    Connection          con = null;
    PreparedStatement   ps  = null;
    ResultSet           rs  = null;
    Map                 map = new HashMap();

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);
      ps  = con.prepareStatement
              ( "select rechnung_stat, sum(anz_rech_empf) as summe" +
                " from ses.amv_rechnung" +
                " where amv_lauf_id=? and rechnung_cd=?" +
                " group by rechnung_stat" );

      ps.setInt(1,amvLaufId);
      ps.setInt(2,rechnungCd);

      rs = ps.executeQuery();
      while ( rs.next() )
      {
        map.put(new Integer(rs.getInt("rechnung_stat")), new Integer(rs.getInt("summe")) );
      }

      return map;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Selektiert die Anzahl Datensätze eines Laufes mit dem angegebenen Status.
   *
   * @param   amvLaufId     Gewünschter Lauf
   * @param   rechnungStat  Gewünschter Status
   * @return                Anzahl Datensätze
   * @throws  SQLException  Unerwarteter Fehler
   */
  public int selectCntByRechnungStat(int amvLaufId, int rechnungStat)
      throws SQLException
  {
    Connection          con = null;
    PreparedStatement   ps  = null;
    ResultSet           rs  = null;
    int                 cnt = 0;

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);
      ps  = con.prepareStatement
              ( "select count(*) as cnt  " +
                "from   ses.amv_rechnung " +
                "where  amv_lauf_id=?    " +
                "and    rechnung_stat=?  " );

      ps.setInt(1,amvLaufId);
      ps.setInt(2,rechnungStat);

      rs = ps.executeQuery();
      if ( rs.next() )
      {
        cnt = rs.getInt("cnt");
      }

      return cnt;
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Selektiert folgende Attribute der Tabelle ERROR
   *  error_id, source, employee_id, employee_notify,
   *  data_type_cd, error_stat, created_dat, changed_dat
   *
   * @param   errorSelKritVO  enthält die Selektionskriterien
   * @return  ArrayList       von ErrorVO die den Selektionskriteren entsprechen
   * @throws  Exception       unerwartete Fehlersituation
   */
  public ArrayList select(AmvRechnungSelKritVO amvRechnungSelKritVO) throws Exception
  {
    Connection con = null;
    Statement stmt = null;
    ResultSet rs = null;
    ArrayList al  = new ArrayList();

    StringBuffer select = new StringBuffer();
    StringBuffer where  = new StringBuffer();
    StringBuffer order  = new StringBuffer();
    String and = "";

    select.append( getSelectStatement() );
    where.append("WHERE ");
    order.append(" ORDER BY amv_lauf_id DESC, bfakt_id ");

    if (!amvRechnungSelKritVO.amvLaufId.equals(""))
    {
      where.append("amv_lauf_id = " + amvRechnungSelKritVO.amvLaufId);
      and = " AND ";
    }

    if (!amvRechnungSelKritVO.rechnungCd.equals(""))
    {
      where.append(and);
      where.append("rechnung_cd = " + amvRechnungSelKritVO.rechnungCd);
      and = " AND ";
    }

    if (!amvRechnungSelKritVO.bfaktId.equals(""))
    {
      where.append(and);
      where.append("bfakt_id = " + amvRechnungSelKritVO.bfaktId);
      and = " AND ";
    }

    if (!amvRechnungSelKritVO.rechnungStat.equals(""))
    {
      where.append(and);
      where.append("rechnung_stat = " + amvRechnungSelKritVO.rechnungStat);
      and = " AND ";
    }

    if (!amvRechnungSelKritVO.errorMsg.equals(""))
    {
      if (amvRechnungSelKritVO.errorMsg.equals(AmvRechnungSelKritVO.ERROR))
      {
        where.append(and);
        where.append("error_msg is not null");
        and = " AND ";
      }
      else if (amvRechnungSelKritVO.errorMsg.equals(AmvRechnungSelKritVO.NOERROR))
      {
        where.append(and);
        where.append("error_msg is null");
        and = " AND ";
      }
    }

    String sql =  select.toString() +
                  ( where.toString().equals("WHERE ") ? "" : where.toString() ) +
                  order.toString();

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);
      stmt = con.createStatement();
      rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      {
        AmvRechnungVO amvRechnungVO = this.fillAmvRechnungVO(rs);
        al.add(amvRechnungVO);
      }

      return al;
    }
    finally
    {
      DBHelper.close(stmt, con);
    }
  }


  /**
   * Aktualisiert die folgenden Felder für die angegebenen Schlüsselwert Felder
   * (Primary Keys) des Value Objects:
   *  - ANZ_RECH_EMPF
   *  - RECHNUNG_STAT
   *  - ERROR_MSG (optional, falls ungleich NULL)
   *  - CHANGED_DAT (automatisch, Wert vom Value Object wird nicht verwendet)
   *
   * Beachte:
   * Eine Aktualisierung wird nur durchgeführt, sofern der Datensatz noch nicht
   * in einem "Endzustand" (2, 3 oder 4) ist. Die Aktualisierung ist also nur
   * möglich, falls der aktuelle Status 0 oder 1 ist.
   *
   * Hintergrund:
   * Damit wird verhindert, dass der Update vom SapInbound Service auf den
   * Status 2 (erfolgreich in Sap angelegt) vom SesAmvStart Client Update
   * auf den Staus 1 (erfolgreich an den SapInbound Service gesendet) über-
   * schrieben wird. Dieses Scenario ist zwar sehr unwahrscheinlich, ist aber
   * in zirka 1 Promille aller Fälle aufgetreten.
   *
   * @param   vo            Value Object einem zu aktualisierenden Datensatz
   * @return  int           Anzahl aktualisierter Datensätze
   * @throws  SQLException  Unerwarteter SQL Fehler
   */
  public int update(AmvRechnungVO vo) throws SQLException
  {
    Connection        con = null;
    PreparedStatement ps  = null;

    try
    {
      // limit maximal field size
      if (vo.errorMsg != null)
      {
        if (vo.errorMsg.length() > FLD_LEN_ERROR_MSG)
          vo.errorMsg = vo.errorMsg.substring(0,FLD_LEN_ERROR_MSG-3) + "...";
      }

      // update data
      con = DBHelper.getConnection(DBHelper.SES);

      if (vo.errorMsg == null)
      {
        ps  = con.prepareStatement (
                "update ses.amv_rechnung " +
                "set    anz_rech_empf=?, rechnung_stat=?, changed_dat=sysdate " +
                "where  amv_lauf_id=? and rechnung_cd=? and bfakt_id=? " +
                "and    rechnung_stat in (0,1) " );

        ps.setInt(1,vo.anzRechEmpf);
        ps.setInt(2,vo.rechnungStat);
        ps.setInt(3,vo.amvLaufId);
        ps.setInt(4,vo.rechnungCd);
        ps.setInt(5,vo.bfaktId);
      }
      else
      {
        ps  = con.prepareStatement (
                "update ses.amv_rechnung " +
                "set    anz_rech_empf=?, rechnung_stat=?, error_msg=?, changed_dat=sysdate " +
                "where  amv_lauf_id=? and rechnung_cd=? and bfakt_id=? " +
                "and    rechnung_stat in (0,1) " );

        ps.setInt(1,vo.anzRechEmpf);
        ps.setInt(2,vo.rechnungStat);
        ps.setString(3,vo.errorMsg);
        ps.setInt(4,vo.amvLaufId);
        ps.setInt(5,vo.rechnungCd);
        ps.setInt(6,vo.bfaktId);
      }


      return ps.executeUpdate();
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Aktualisiert das Feld RECHNUNG_STAT und CHANGED_DAT (automatisch) für die
   * angegebenen Schlüsselwert Felder (Primary Keys) des Value Objects.
   *
   * @param   vo            Value Object einem zu aktualisierenden Datensatz
   * @return  int           Anzahl aktualisierter Datensätze
   * @throws  SQLException  Unerwarteter SQL Fehler
   */
  public int updateRechnungStat(AmvRechnungVO vo) throws SQLException
  {
    Connection        con = null;
    PreparedStatement ps  = null;

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);

      ps  = con.prepareStatement (
              "update ses.amv_rechnung " +
              "set    rechnung_stat=?, changed_dat=sysdate " +
              "where  amv_lauf_id=? and rechnung_cd=? and bfakt_id=? " );

      ps.setInt(1,vo.rechnungStat);
      ps.setInt(2,vo.amvLaufId);
      ps.setInt(3,vo.rechnungCd);
      ps.setInt(4,vo.bfaktId);

      return ps.executeUpdate();
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Reset Status Feld aller fehlerhaften Datensätze. Die Fehlermeldung wird
   * auf 'reset' gesetzt, damit ersichtlich bleibt, dass der Datensatz bereits
   * zurückgesetzt wurde.
   *
   * @param   amvLaufId     Lauf Identifikation
   * @return  int           Anzahl aktualisierter Datensätze
   * @throws  SQLException  Unerwarteter SQL Fehler
   */
  public int resetRechnungStatError(int amvLaufId) throws SQLException
  {
    Connection        con = null;
    PreparedStatement ps  = null;

    try
    {
      con = DBHelper.getConnection(DBHelper.SES);

      ps  = con.prepareStatement (
              "update ses.amv_rechnung " +
              "set    rechnung_stat=?, error_msg=?, changed_dat=sysdate " +
              "where  amv_lauf_id=? and rechnung_stat=? " );

      ps.setInt(1,AmvRechnungVO.RE_STAT_SELECTED);
      ps.setString(2,"reset");
      ps.setInt(3,amvLaufId);
      ps.setInt(4,AmvRechnungVO.RE_STAT_SAP_ERROR);

      return ps.executeUpdate();
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }


  /**
   * Löscht einen Datensatz aus der Tabelle AMV_RECHNUNG
   *
   * @param     vo            Value Object mit Primary Key Daten
   * @return                  Anzahl gelöschter Datensätze
   * @exception SQLException  Unerwarteter SQL Fehler
   */
  public int delete(AmvRechnungVO vo) throws SQLException
  {
    Connection        con = null;
    PreparedStatement ps  = null;

    try
    {
      con  = DBHelper.getConnection(DBHelper.SES);
      ps = con.prepareStatement
            ( "delete from ses.amv_rechnung " +
              "where  amv_lauf_id=? and rechnung_cd=? and bfakt_id=? " );

      ps.setInt(1,vo.amvLaufId);
      ps.setInt(2,vo.rechnungCd);
      ps.setInt(3,vo.bfaktId);

      return ps.executeUpdate();
    }
    finally
    {
      DBHelper.close(ps, con);
    }
  }



  //---------------------------------------------------------------------------
  // private methods
  //---------------------------------------------------------------------------

  /**
   * delivers sql statement for the select methods WITHOUT the where part
   *
   * note: the field must correspondent with the fields used in the
   *       method fillAmvRechnungVO !
   *
   * @return  select statement
   */
  private String getSelectStatement()
  {
    return
      ( "select amv_lauf_id, rechnung_cd, bfakt_id, " +
        "       anz_rech_empf, rechnung_stat, error_msg, " +
        "       created_dat, changed_dat " +
        "from   ses.amv_rechnung " );
  }


  /**
    * fill AmvLaufVO with data from (a valid) result set.
    *
    * @param      rs            valid result set
    * @return     AmvLaufVO     value object
    * @exception  SQLException  unexpected sql error
    */
  private AmvRechnungVO fillAmvRechnungVO (ResultSet rs) throws SQLException
  {
    AmvRechnungVO vo = new AmvRechnungVO();

    vo.amvLaufId    = rs.getInt("amv_lauf_id");
    vo.rechnungCd   = rs.getInt("rechnung_cd");
    vo.bfaktId      = rs.getInt("bfakt_id");
    vo.anzRechEmpf  = rs.getInt("anz_rech_empf");
    vo.rechnungStat = rs.getInt("rechnung_stat");
    vo.errorMsg     = rs.getObject("error_msg") == null ?
                        null : rs.getString("error_msg");
    vo.createdDat   = rs.getTimestamp("created_dat") == null ?
                        null : new Date(rs.getTimestamp("created_dat").getTime());
    vo.changedDat   = rs.getTimestamp("changed_dat") == null ?
                        null : new Date(rs.getTimestamp("changed_dat").getTime());

    return vo;
  } // end method


} // end class

