package ch.ige.ses.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Hashtable;

import javax.jms.BytesMessage;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.ObjectMessage;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueReceiver;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Session;
import javax.jms.TemporaryQueue;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import ch.ige.util.BuildInfo;
import ch.ige.util.ServiceLocator;

/**
 * Java Message Service - Queue Handling
 *
 * simple helper class for basic jms initializations. the class supports
 * the following jms queue modes:
 *  - Sender
 *  - Receiver
 *  - Sender & Receiver
 *  - Requestor
 *  - Receiver & Replier
 *
 * important notes:
 *  - you can use only one mode at the same time
 *  - j2ee beans should only use the Sender mode. for receiver tasks you can
 *    use message driven beans. the requestor pattern isn't suitable for j2ee.
 *
 * Copyright (c) 2002, Eidgenössisches Institut für Geistiges Eigentum
 * @author T. Iten
 * @version $Id: JmsQueue.java,v 1.17 2007/12/24 10:13:23 iten Exp $
 */
public class JmsQueue
{
  //
  // const
  //
  private static final boolean  DEFAULT_SESSION_TRANSACTED       = false;
  private static final int      DEFAULT_SESSION_ACKNOWLEDGE_MODE = Session.AUTO_ACKNOWLEDGE;

  /** run on a client application */
  public static final int       RUN_ON_CLIENT = 0;

  /** run this class within an application server */
  public static final int       RUN_ON_APPSRV = 1;
  
  
  //
  // fields
  //
  
  // jms basic params (set by constructor)
  private int       fRunOn                    = RUN_ON_APPSRV;
  private String    fJndiFactoryName          = null;
  private String    fJmsQueueName             = null;
  private boolean   fSessionTransacted        = DEFAULT_SESSION_TRANSACTED;
  private int       fSessionAcknowledgeMode   = DEFAULT_SESSION_ACKNOWLEDGE_MODE;

  // connection params (set by connect())
  private QueueConnectionFactory  fFactory    = null;
  private QueueConnection         fConnection = null;
  private QueueSession            fSession    = null;
  private Queue                   fQueue      = null;

  // sender & receiver fields, used depending on the mode
  //    Mode                fSender    fReceiver
  //    -----------------------------------------
  //    Sender                  x
  //    Receiver                           x
  //    Sender & Receiver       x          x
  //    Requestor               x          x
  //    Receiver & Replier      x          x
  private QueueSender     fSender   = null;
  private QueueReceiver   fReceiver = null;

  // misc
  private boolean  fConnectedFlag = false;  // connection state (if used as Sender & Receiver)
  private int      fTimeToLive    = 0;      // time to live [ms] used for requestor mode

  //---------------------------------------------------------------------------
  // constructors
  //---------------------------------------------------------------------------

  private JmsQueue() 
  {
    // do not use this constructor
  }
  
  /**
   * Create a JmsQueue helper class.
   * 
   * Important:
   * Depending on which in environment the instance will be executed, the
   * runOn param has to be set. This is very important because of the JNDI
   * lookup, which is diffrent on a client than on the application server!
   * 
   * @param runOn           runnign environment 
   * @param jndiFactoryName factory name
   * @param jmsQueueName    jms queue name
   */
  public JmsQueue(int runOn, String jndiFactoryName, String jmsQueueName)
  {
    this.fRunOn           = runOn;
    this.fJndiFactoryName = jndiFactoryName;
    this.fJmsQueueName    = jmsQueueName;
  }


  /**
   * Create a JmsQueue helper class.
   * 
   * Important:
   * Depending on which in environment the instance will be executed, the
   * runOn param has to be set. This is very important because of the JNDI
   * lookup, which is diffrent on a client than on the application server!
   * 
   * @param runOn           runnign environment 
   * @param jndiFactoryName factory name
   * @param jmsQueueName    jms queue name
   * @param transacted      transcation setting
   * @param acknowledgeMode acknowledge mode
   */
  public JmsQueue ( int runOn, String  jndiFactoryName, String jmsQueueName,
                    boolean transacted, int acknowledgeMode )
  {
    this(runOn, jndiFactoryName,jmsQueueName);

    this.fSessionTransacted      = transacted;
    this.fSessionAcknowledgeMode = acknowledgeMode;
  }


  //---------------------------------------------------------------------------
  // sender methods
  //---------------------------------------------------------------------------

  public QueueSender connectSender()
    throws NamingException, JMSException
  {
    this.connect();
    fSender = fSession.createSender(fQueue);

    return fSender;
  }


  public QueueSender connectSender(int priority)
    throws NamingException, JMSException
  {
    this.connect();
    fSender = fSession.createSender(fQueue);
    fSender.setPriority(priority);

    return fSender;
  }


  public QueueSender getSender()
  {
    return this.fSender;
  }

  //---------------------------------------------------------------------------
  // receiver methods
  //---------------------------------------------------------------------------

  public QueueReceiver connectReceiver()
    throws NamingException, JMSException
  {
    this.connect();
    fReceiver   = fSession.createReceiver(fQueue);

    return fReceiver;
  }


  public QueueReceiver connectReliableReceiver(String clientId)
    throws NamingException, JMSException
  {
    this.connect(true,clientId);
    fReceiver = fSession.createReceiver(fQueue);

    return fReceiver;
  }


  public QueueReceiver getReceiver()
  {
    return this.fReceiver;
  }

  //---------------------------------------------------------------------------
  // request & reply methods
  //---------------------------------------------------------------------------

  public void connectRequestor(int timeToLive) throws NamingException, JMSException
  {
    fTimeToLive = timeToLive * 1000;                  // timeToLive in [sec] will be
                                                      // saved internally in [ms]
    connectSender();

    if (fTimeToLive > 0)                              // if not set, the message
      fSender.setTimeToLive(fTimeToLive);             // will never expire

    fConnection.start();
  }


  public Object request(ObjectMessage objectMessage) throws Exception
  {
    //TODO maintenance (optional) - request art einstellbar via message property
    return requestBytesMessage(objectMessage);
  }


  public int getTimeToLive()
  {
    return (this.fTimeToLive / 1000);
  }


  public void reply(Message receivedMessage, Serializable replyObject)
    throws JMSException, IOException
  {
    //TODO maintenance (optional) - request art einstellbar via message property */
    replyBytesMessage(receivedMessage,replyObject);
  }


  //---------------------------------------------------------------------------
  // common methods
  //---------------------------------------------------------------------------

  public void disconnect()
  {
    // close connection
    // (closes also all related sessions, queues, sender and receiver)
    try
    {
      if ( fConnection != null)
      {
        fConnection.close();
      }
    }
    catch (Exception ex)
    {
      // ignore
    }
    finally
    {
      fSender   = null;               // important: set to null because of
      fReceiver = null;               // initialization part within reply method

      fConnectedFlag = false;
    }
  } // end method


  public QueueConnection getConnection()
  {
    return this.fConnection;
  }

  public QueueSession getSession()
  {
    return this.fSession;
  }

  public Queue getQueue()
  {
    return this.fQueue;
  }

  //---------------------------------------------------------------------------
  // private methods
  //---------------------------------------------------------------------------

  private void connect() throws NamingException, JMSException
  {
    this.connect(false,"null");
  }


  private void connect(boolean reliable, String clientId)
    throws NamingException, JMSException
  {
    if ( ! this.fConnectedFlag)
    {
      //
      // get initial contect depending on the running environment
      //
      Context  ctx = null;
      if ( this.fRunOn == RUN_ON_APPSRV )
      {
        ctx = new InitialContext();
      }
      else
      {
        Hashtable env = ServiceLocator.getRemoteJNDIEnvironment();
        ctx = new InitialContext(env);
      }

      //
      // get connection
      //
      fFactory    = (QueueConnectionFactory) ctx.lookup(fJndiFactoryName);
      fConnection = fFactory.createQueueConnection();

      if (reliable)
      {
        fConnection.setClientID(clientId);
      }

      fSession = fConnection.createQueueSession(fSessionTransacted,fSessionAcknowledgeMode);
      fQueue   = fSession.createQueue(fJmsQueueName);

      this.fConnectedFlag = true;
    }

  } // end method


  private Object requestBytesMessage(ObjectMessage objectMessage) throws Exception
  {
    TemporaryQueue      tempQueue;            // queue for reply message
    javax.jms.Message   message;              // reply message
    Object              object   = null;      // return object or null if timeout expired


    tempQueue = fSession.createTemporaryQueue();
    fReceiver = fSession.createReceiver(tempQueue);
    objectMessage.setJMSReplyTo(tempQueue);

    fSender.send(objectMessage);

    if (fTimeToLive == 0)
    {
      message = fReceiver.receive();
    }
    else
    {
      message = fReceiver.receive(fTimeToLive);
    }

    if (message != null)
    {
      // BytesMessage mode
      // extract byte[] of Message
      BytesMessage bm = (BytesMessage)message;
      byte[] barr = new byte[100000];
      bm.readBytes(barr);

      ByteArrayInputStream bais = new ByteArrayInputStream(barr);
      ObjectInputStream ois = new ObjectInputStream(bais);
      object = ois.readObject();
    }

    return object;
  } // end method


  private Object requestObjectMessage(ObjectMessage objectMessage) throws Exception
  {
    TemporaryQueue      tempQueue;            // queue for reply message
    javax.jms.Message   message;              // reply message
    Object              object   = null;      // return object or null if timeout expired


    tempQueue = fSession.createTemporaryQueue();
    fReceiver = fSession.createReceiver(tempQueue);
    objectMessage.setJMSReplyTo(tempQueue);

    fSender.send(objectMessage);

    if (fTimeToLive == 0)
    {
      message = fReceiver.receive();
    }
    else
    {
      message = fReceiver.receive(fTimeToLive);
    }

    if (message != null)
    {
      object = ((javax.jms.ObjectMessage)message).getObject();
    }

    return object;
  } // end method


  private void replyBytesMessage(Message receivedMessage, Serializable replyObject)
    throws JMSException, IOException
  {
    javax.jms.Queue q = (javax.jms.Queue) receivedMessage.getJMSReplyTo();

    if (q != null)
    {
      if (fSender == null)                          // init at first reply, queue
      {                                             // will be set for each reply
        fSender = fSession.createSender(null);
      }

      // BytesMessage mode
      // es wird eine BytesMessage verschickt, wegen Classloader - Problematik auf AppServer
      byte[] barr = null;
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      ObjectOutputStream os = new ObjectOutputStream(baos);
      os.writeObject(replyObject);
      barr = baos.toByteArray();

      BytesMessage bm = fSession.createBytesMessage();
      bm.writeBytes(barr);
      bm.setJMSCorrelationID(receivedMessage.getJMSCorrelationID());

      fSender.send (q,bm);
    }
  } // end method


  private void replyObjectMessage(Message receivedMessage, Serializable replyObject)
    throws JMSException, IOException
  {
    javax.jms.Queue         q  = (javax.jms.Queue) receivedMessage.getJMSReplyTo();
    javax.jms.ObjectMessage om = null;

    if (q != null)
    {
      if (fSender == null)                          // init at first reply, queue
      {                                             // will be set for each reply
        fSender = fSession.createSender(null);
      }

      om = fSession.createObjectMessage(replyObject);
      om.setJMSCorrelationID(receivedMessage.getJMSMessageID());

      fSender.send (q,om);
    }
  } // end method

} // end class



