package ch.ige.wdltool.analyzer;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.hibernate.validator.InvalidValue;

import ch.ige.dbutils.dao.DAOPool;
import ch.ige.ipobjects.common.dao.RegistrationFinderDAO;
import ch.ige.ipobjects.key.vo.IPKeyList;
import ch.ige.ipobjects.key.vo.IPKeyVO;
import ch.ige.ipobjects.time.PointInTime;
import ch.ige.ipobjects.time.vo.EreigIDat;
import ch.ige.ipobjects.trademark.Trademark;
import ch.ige.util.date.DateFormatter;
import ch.ige.util.enums.IPType;
import ch.ige.wdltool.checker.WdlChecker;
import ch.ige.wdltool.dao.AnalyzedDAO;
import ch.ige.wdltool.ejb.WdlServices;
import ch.ige.wdltool.ejb.WdlServicesFactory;
import ch.ige.wdltool.ejb.impl.TrademarkImpl;
import ch.ige.wdltool.exception.ExceptionThrower;
import ch.ige.wdltool.exception.WdlValidationException;
import ch.ige.wdltool.util.BatchUtils;
import ch.ige.wdltool.vo.AbstractTermVO;
import ch.ige.wdltool.vo.AnalyzedVO;
import ch.ige.wdltool.vo.IpiTermCandidateVO;
import ch.ige.wdltool.vo.IpiTermVO;
import ch.ige.wdlutil.exception.WdlSyntaxException;
import ch.ige.wdlutil.text.LanguageGuesser;
import ch.ige.wdlutil.vo.ClassBlockVO;
import ch.ige.wdlutil.vo.ClassTextGroupVO;
import ch.ige.wdlutil.vo.WdlVO;

/**
 * Analyzes WDLs and saves term candidates in the database.
 * <p>
 * Copyright (c) 2007, Swiss Federal Institute of Intellectual Property
 * </p>
 * @author Christian Heinzer
 * @version $Id: WdlAnalyzer.java,v 1.47 2008/12/08 15:53:56 heinzer Exp $
 */
public class WdlAnalyzer
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( WdlAnalyzer.class );

  /**
   * Supported LOG size  
   * When analyzing a range of trademarks and the current log size exeeds this limit the logging on trademark level gets suspended.
   * This limit is respected approximately for the full / final log (minimal extra space for summary will still be needed)
   */
  private static final int MAXIMUM_LOG_SIZE = 1024 * 1024 * 10;

  /**
   * Used to get other services
   */
  private WdlServices wdlServices;

  /**
   * Public constructor.
   * @param wdlServices The WDL services
   */
  public WdlAnalyzer( WdlServices wdlServices )
  {
    this.wdlServices = wdlServices;
  }

  /**
   * Analyzes the WDLs of trademarks entered in the date range from...to.
   * @param from Start of range to analyze - only the day is relevant, time is ignored
   * @param to End of range to analyze - only the day is relevant, time is ignored
   * @return Logging information with details for all analyzed WDLs. This will be clipped when {@link #MAXIMUM_LOG_SIZE} is exeeded.
   */
  public String analyzeWdls( Date from, Date to )
  {
    /**
     * Syntax error count.
     */
    int syntaxErrorCount = 0;

    /**
     * The list of IP keys of trademarks with syntactically incorrect WDL
     */
    List<Integer> syntaxErrorIpKeys = new ArrayList<Integer>();

    /**
     * New terms count.
     */
    int newTermsCount = 0;

    /**
     * Existing terms count (active or candidate)
     */
    int existingTermsCount = 0;

    /**
     * Existing term candidates count.
     */
    int existingTermCandidatesCount = 0;

    /**
     * Term insertion failure count.
     */
    int insertionErrorCount = 0;

    // get the keys in Range from BAGIS
    EreigIDat fromDateIDAT = EreigIDat.min( from );
    EreigIDat toDateIDAT = EreigIDat.max( to );
    IPKeyList ipKeys = new IPKeyList();

    try
    {
      ipKeys = new IPKeyList( new RegistrationFinderDAO().findRegistrationsBetween( IPType.TRADEMARK_CH, fromDateIDAT, toDateIDAT ) );
    }
    catch ( SQLException e )
    {
      LOG.error( "Error selecting keys in range : " + e.getMessage(), e );
      ExceptionThrower.throwSQLException( "Error selecting keys in range ", e );
    }

    StringBuilder fullLog = new StringBuilder();
    BatchUtils.addLine( fullLog, BatchUtils.TRADEMARK_SEPARATOR );
    boolean clipLogging = false; //gets true when the log size would exceed the limit -> only the summary will be added anyway

    for ( IPKeyVO key: ipKeys )
    {
      AnalysisResultVO analysisResult = new AnalysisResultVO( key.getIPKey() );

      try
      {
        analysisResult = WdlServicesFactory.getLocalInstance( true ).analyzeWdl( key.getIPKey() );
      }
      catch ( Exception e )
      {
        analysisResult.setGeneralError( e );
      }
      if ( analysisResult.hasSyntaxError() )
      {
        syntaxErrorCount++;
        syntaxErrorIpKeys.add( analysisResult.getIPKey() );
      }
      else
      {
        newTermsCount += analysisResult.getCounterNewTerms();
        existingTermsCount += analysisResult.getCounterExistingTerms();
        existingTermCandidatesCount += analysisResult.getCounterExistingCandidateTerms();
        insertionErrorCount += analysisResult.getCounterInsertionError();
      }

      if ( clipLogging || fullLog.length() + analysisResult.getAnalyzerLog().length() > MAXIMUM_LOG_SIZE )
      { //overflow
        if ( clipLogging == false )
        { //first overflow
          fullLog.append( "Maximale Dateigrösse überschritten: Detailliertes Logging auf Titelebene deaktiviert" );
          clipLogging = true;
        }
      }
      else
      {
        fullLog.append( analysisResult.getAnalyzerLog() );
      }
    }

    int syntacticallyCorrectWdls = ipKeys.size() - syntaxErrorCount;
    int completeCount = existingTermsCount + newTermsCount;
    int existingAndActive = existingTermsCount - existingTermCandidatesCount;

    List<String> summary = new ArrayList<String>();
    summary.add( BatchUtils.TRADEMARK_SEPARATOR );
    summary.add( "ZUSAMMENFASSUNG FÜR MARKEN VON " + DateFormatter.shortFormat( from ) + " BIS " + DateFormatter.shortFormat( to ) );
    summary.add( BatchUtils.DETAILS_SEPARATOR );
    summary.add( "Analysierte Marken: " + ipKeys.size() );
    summary
        .add( "Davon mit syntaktisch korrekter WDL: " + syntacticallyCorrectWdls + BatchUtils.getPercentage( syntacticallyCorrectWdls, ipKeys.size(), false ) );

    if ( syntaxErrorCount > 0 )
    {
      summary.add( "Syntaktisch fehlerhafte WDLs (Schutit-Key): " + StringUtils.join( syntaxErrorIpKeys.iterator(), ", " ) );
    }

    summary.add( "Extrahierbare Begriffe: " + completeCount );

    if ( completeCount > 0 )
    {
      summary.add( BatchUtils.getIndentedLine( 1, "Davon neu: " + newTermsCount + BatchUtils.getPercentage( newTermsCount, completeCount, false ) ) );
      summary.add( BatchUtils.getIndentedLine( 2, "Davon neu, aber nicht als Kandidat erfassbar: " + insertionErrorCount
          + BatchUtils.getPercentage( insertionErrorCount, completeCount, true ) ) );
      summary.add( BatchUtils.getIndentedLine( 1, "Davon total bereits vorhanden: " + existingTermsCount
          + BatchUtils.getPercentage( existingTermsCount, completeCount, false ) ) );
      summary.add( BatchUtils.getIndentedLine( 2, "Davon bereits vorhanden und aktiv: " + existingAndActive
          + BatchUtils.getPercentage( existingAndActive, completeCount, true ) ) );
      summary.add( BatchUtils.getIndentedLine( 2, "Davon bereits vorhanden und inaktiv: " + existingTermCandidatesCount
          + BatchUtils.getPercentage( existingTermCandidatesCount, completeCount, true ) ) );
    }
    BatchUtils.insertLine( fullLog, summary );

    return fullLog.toString();
  }

  /**
   * Analyzes the WDL of a trademark with known IP key. This produces term candidates for new terms or updates existing
   * term candidates
   * @param ipKey The IP key of the trademark to analyze
   * @return Logging information procuded by analyze process
   */
  public AnalysisResultVO analyzeWdl( int ipKey )
  {
    AnalysisResultVO analysisResult = new AnalysisResultVO( ipKey );
    StringBuilder log = new StringBuilder();

    try
    {
      if ( DAOPool.get( AnalyzedDAO.class ).contains( ipKey ) )
      {
        BatchUtils.addLine( log, "Marke mit Schutit-Key " + ipKey + " wurde bereits analysiert und wird deshalb übersprungen." );
        return analysisResult;
      }
      else
      {
        Trademark trademark = TrademarkImpl.getTrademarkForKey( ipKey );
        PointInTime now = TrademarkImpl.getCurrentIDat();
        String wdlText = TrademarkImpl.getWdlText( trademark, now );
        Locale trademarkLanguage = trademark.getLocale();
        Locale wdlLanguage = LanguageGuesser.getWdlLocale( wdlText );
        if ( wdlLanguage == null )
        {
          wdlLanguage = trademarkLanguage;
        }

        BatchUtils.addBaseInformation( log, trademark );

        boolean useWdlLanguage = false;
        Locale language = trademarkLanguage;
        if ( !trademarkLanguage.equals( wdlLanguage ) )
        {
          BatchUtils.addLine( log, "ACHTUNG: Nehme WDL-Sprache " + wdlLanguage.getDisplayLanguage( Locale.GERMAN ) + " statt Markensprache "
              + trademarkLanguage.getDisplayName( Locale.GERMAN ) );
          LOG.info( "Taking WDL language '" + wdlLanguage.getLanguage() + "' instead of trademark language '" + trademarkLanguage.getLanguage()
              + "' for IP key " + trademark.getIPKey() );
          language = wdlLanguage;
          useWdlLanguage = true;
        }
        BatchUtils.addLine( log, BatchUtils.DETAILS_SEPARATOR );

        List<Integer> classNumbers = TrademarkImpl.getWdlClasses( trademark, now );

        WdlVO wdlVO = WdlChecker.parseAndInitialize( wdlText, language, classNumbers, false, false, false, null );
        LOG.debug( "Analyzing trademark " + trademark.getDescription() + " with key = " + trademark.getIPKey() );
        for ( ClassBlockVO blockVO: wdlVO.getClassBlocks() )
        {
          // note: possible limitation is ignored

          List<ClassTextGroupVO> groups = blockVO.getClassTextGroups();
          for ( ClassTextGroupVO groupVO: groups )
          {
            String classTextGroup = groupVO.getClassTextGroup();

            // search if it already exists
            LOG.debug( "searching in active: lang = " + language + " term = " + classTextGroup );
            if ( StringUtils.isEmpty( classTextGroup ) )
            {
              LOG.debug( "Empty classTextGroup was found. Was it null? " + classTextGroup == null );
              continue;
            }
            classTextGroup = classTextGroup.trim();

            if ( blockVO.getClassNumber() == null )
            {
              LOG.error( "ignoring class text group without class number " + classTextGroup );
              BatchUtils.addLine( log, 1, "Der Begriff '" + classTextGroup + "' konnte keiner Klasse zugeordnet werden und wird ignoriert." );
              continue;
            }

            List<AbstractTermVO> searchResults = new ArrayList<AbstractTermVO>();
            try
            {
              searchResults = this.wdlServices.findTermWithinActiveTerms( language, blockVO.getClassNumber(), classTextGroup, null );
            }
            catch ( Exception e )
            {
              LOG.debug( "Error searching on Server: " + e.getMessage(), e );
              BatchUtils.addLine( log, 1, "Suche nach aktivem Begriff '" + classTextGroup + "' schlug fehl - wird als nicht gefunden behandelt" );
            }

            if ( searchResults.size() == 0 ) // search also in inactive terms
            {
              List<IpiTermVO> inactiveSearchResults = new ArrayList<IpiTermVO>();

              try
              {
                inactiveSearchResults = this.wdlServices.findTermWithinInactiveTerms( language, blockVO.getClassNumber(), classTextGroup );
              }
              catch ( Exception e )
              {
                LOG.debug( "Error searching on Server: " + e.getMessage(), e );
                BatchUtils.addLine( log, 1, "Suche nach inaktivem Begriff '" + classTextGroup + "' schlug fehl - wird als nicht gefunden behandelt" );
              }

              if ( inactiveSearchResults.size() > 0 ) // convert into candidate
              {
                IpiTermVO ipiTermVO = inactiveSearchResults.get( 0 );
                LOG.debug( "Converting inactive term into a candidate. term = " + ipiTermVO.getTerm() + " lang = " + ipiTermVO.getLanguage() + " class = "
                    + ipiTermVO.getClassNumber() );
                BatchUtils.addLine( log, "Inaktiver Begriff wurde in Kandidat mit Zähler 1 konvertiert. Begriff = " + ipiTermVO.getTerm() + ", Sprache = "
                    + ipiTermVO.getLanguage() + ", Klassennummer = " + ipiTermVO.getClassNumber() );
                // do conversion
                IpiTermCandidateVO ipiTermCandidateVO = new IpiTermCandidateVO( ipiTermVO );
                this.wdlServices.convertInactiveTermToIpiTermCandidate( ipiTermCandidateVO );
                // candidate is created, counter will already be 1 initially so we can go on with the next term
                continue;
              }
            }
            if ( searchResults.size() == 0 ) // current term is not active
            {

              IpiTermCandidateVO termCandidate = null;
              try
              {
                termCandidate = this.wdlServices.findIpiTermCandidate( language, blockVO.getClassNumber(), classTextGroup );
              }
              catch ( Exception e )
              {
                LOG.debug( "Error searching on Server: " + e.getMessage(), e );
                BatchUtils.addLine( log, 1, "Suche nach Begriffskandidat '" + classTextGroup + "' schlug fehl - wird als nicht gefunden behandelt" );
              }

              if ( termCandidate != null ) // update count + date
              {
                analysisResult.incCounterExistingTerms();
                analysisResult.incCounterExistingCandidateTerms();
                termCandidate.setUsageCount( termCandidate.getUsageCount() + 1 );
                if ( termCandidate.getLatestOccurrence().before( new Timestamp( trademark.getRegistrationEvent().getDate().getTime() ) ) )
                {
                  termCandidate.setLatestOccurrence( new Timestamp( trademark.getRegistrationEvent().getDate().getTime() ) );
                }
                try
                {
                  this.wdlServices.updateIpiTermCandidate( termCandidate );
                }
                catch ( Exception e )
                {
                  BatchUtils.addLine( log, 1, "Zähler für Begriffskandidat '" + classTextGroup + "' konnte nicht erhöht werden." );
                  LOG.error( "Error updating IpiTermCandidate " + e.getMessage(), e );
                  if ( e instanceof WdlValidationException )
                  {
                    WdlValidationException validationException = (WdlValidationException) e;
                    StringBuilder valueErrorMessage = new StringBuilder();
                    for ( InvalidValue invalidValue: validationException.getInvalidValues() )
                    {
                      if ( valueErrorMessage.length() > 0 )
                      {
                        valueErrorMessage.append( ", " );
                      }
                      valueErrorMessage.append( invalidValue.getMessage() );
                    }
                    LOG.error( "Cause was validation: " + valueErrorMessage );
                  }
                }
              }
              else
              // does not exist, add as candidate
              {
                analysisResult.incCounterNewTerms();
                termCandidate = new IpiTermCandidateVO();
                termCandidate.setCreationUser( "Analyzer-Batch" );
                termCandidate.setChangeUser( termCandidate.getCreationUser() );
                termCandidate.setClassNumber( blockVO.getClassNumber() );
                termCandidate.setLanguage( language );
                termCandidate.setLatestOccurrence( new Timestamp( trademark.getRegistrationEvent().getDate().getTime() ) );
                termCandidate.setTerm( classTextGroup );
                try
                {
                  this.wdlServices.insertIpiTermCandidate( termCandidate );
                  if ( useWdlLanguage )
                  {
                    BatchUtils.addLine( log, 1, "Begriffskandidat '" + termCandidate.getTerm() + "' wurde mit WDL-Sprache "
                        + wdlLanguage.getDisplayLanguage( Locale.GERMAN ) + " statt Markensprache " + trademarkLanguage.getDisplayName( Locale.GERMAN )
                        + " eingefügt." );
                  }
                  LOG.debug( "Successfully inserted a new term candidate. Details: ID = " + termCandidate.getIdentifier() + " term = "
                      + termCandidate.getTerm() + " lang = " + termCandidate.getLanguage() + " class = " + termCandidate.getClassNumber() );
                  BatchUtils.addLine( log, "Neuer Begriffskandidat eingefügt: ID " + termCandidate.getIdentifier() + ", Begriff '" + termCandidate.getTerm()
                      + "'" + ", Sprache = " + termCandidate.getLanguage() + ", Klassennummer = " + termCandidate.getClassNumber() );
                }
                catch ( Exception e )
                {
                  analysisResult.incCounterInsertionError();

                  BatchUtils.addLine( log, "Begriffskandidat konnte nicht eingefügt werden: " + classTextGroup );
                  if ( e instanceof WdlValidationException )
                  {
                    LOG.debug( "Could not insert candidate: " + classTextGroup + " ;" + ( (WdlValidationException) e ).getInvalidValues() );
                    BatchUtils.addLine( log, "Grund: " + ( (WdlValidationException) e ).getInvalidValues() );
                  }
                  else
                  {
                    LOG.debug( "Could not insert candidate: " + classTextGroup + " ;" + e.getMessage(), e );
                  }
                }
              }
            }
            else
            {
              analysisResult.incCounterExistingTerms();
            }
          }
        }

        int completeCount = analysisResult.getCounterExistingTerms() + analysisResult.getCounterNewTerms();

        BatchUtils.addLine( log, "Anzahl extrahierter Begriffe: " + completeCount );
        if ( completeCount > 0 )
        {
          BatchUtils.addLine( log, 1, "Davon neu: " + analysisResult.getCounterNewTerms()
              + BatchUtils.getPercentage( analysisResult.getCounterNewTerms(), completeCount, false ) );
          BatchUtils.addLine( log, 1, "Davon bereits vorhanden (Kandidat oder aktiv): " + analysisResult.getCounterExistingTerms()
              + BatchUtils.getPercentage( analysisResult.getCounterExistingTerms(), completeCount, false ) );
        }
        DAOPool.get( AnalyzedDAO.class ).insert( new AnalyzedVO( ipKey ) );
        return analysisResult;
      }
    }
    catch ( WdlSyntaxException e )
    {
      LOG.debug( "Syntax incorrect: " + e.getMessage(), e );
      BatchUtils.addLine( log, 1, "WDL syntaktisch inkorrekt: " + e.getMessage() );
      analysisResult.setSyntaxError();
      return analysisResult;

    }
    catch ( Throwable t )
    {
      LOG.debug( "Generic exception for key: " + ipKey + ", " + t.getMessage(), t );
      BatchUtils.addLine( log, 1, "Genereller Fehler für Marke mit Schutit-Key " + ipKey + ": " + t.getMessage() );
      analysisResult.setGeneralError( t );
      return analysisResult;
    }
    finally
    {
      BatchUtils.addLine( log, BatchUtils.TRADEMARK_SEPARATOR );
      analysisResult.setAnalyzerLog( log.toString() );
    }
  }
}