package ch.ige.wdltool.checker;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.SortedSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.dbutils.dao.DAOPool;
import ch.ige.ipobjects.trademark.Trademark;
import ch.ige.wdltool.dao.GeneralTermDAO;
import ch.ige.wdltool.dao.StopwordDAO;
import ch.ige.wdltool.ejb.impl.TrademarkImpl;
import ch.ige.wdltool.text.UpperCaseCharacterConverter;
import ch.ige.wdltool.text.WdlCharacterReplacer;
import ch.ige.wdltool.vo.AbstractTermVO;
import ch.ige.wdltool.vo.CheckedClassTextGroupVO;
import ch.ige.wdltool.vo.CheckedClassTextPart;
import ch.ige.wdltool.vo.ClassTextPartVO;
import ch.ige.wdltool.vo.StopwordVO;
import ch.ige.wdltool.vo.WdlCheckResultVO;
import ch.ige.wdlutil.analyze.WdlParser;
import ch.ige.wdlutil.analyze.WdlSyntaxConfiguration;
import ch.ige.wdlutil.exception.WdlSyntaxException;
import ch.ige.wdlutil.text.CharacterChecker;
import ch.ige.wdlutil.text.RegexUtils;
import ch.ige.wdlutil.text.SearchText;
import ch.ige.wdlutil.vo.ClassBlockVO;
import ch.ige.wdlutil.vo.ClassTextGroupVO;
import ch.ige.wdlutil.vo.WdlVO;

/**
 * Checks a WDL for correctness and matches it with terms from the WDL DB.
 * <p>
 * Copyright (c) 2007, Swiss Federal Institute of Intellectual Property
 * </p>
 * 
 * @author Eric Lewis
 * @version $Id: WdlChecker.java,v 1.88 2008/05/08 10:42:47 lewis_lee Exp $
 */
public class WdlChecker
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( WdlChecker.class );

  /**
   * Searches for multiple blanks
   */
  private static final Pattern MULTIPLE_BLANKS = Pattern.compile( " {2,}" );

  /**
   * All brackets
   */
  private static final char[] BRACKETS = new char[] { '(', ')', '[', ']' };

  /**
   * Checks for blanks before punctuation marks.
   */
  private static final Pattern BLANK_BEFORE_PUNCTUATION_MARK = Pattern.compile( " [,;\\.]" );

  /**
   * The stopword DAO to use
   */
  private StopwordDAO stopwordDao = DAOPool.get( StopwordDAO.class );

  /**
   * The term DAO to use for fetching all positive terms
   */
  private GeneralTermDAO generalTermDao = DAOPool.get( GeneralTermDAO.class );

  /**
   * Configures how this reacts to WDL syntax errors.
   */
  private Map<WdlSyntaxConfiguration, Boolean> wdlSyntaxConfiguration;

  /**
   * Creates this without any configuration.
   */
  public WdlChecker()
  {
    this( null );
  }

  /**
   * Creates this with a WDL syntax configuration.
   * @param wdlSyntaxConfiguration Configures how this reacts to WDL syntax errors
   */
  public WdlChecker( Map<WdlSyntaxConfiguration, Boolean> wdlSyntaxConfiguration )
  {
    super();
    this.wdlSyntaxConfiguration = wdlSyntaxConfiguration;
  }

  /**
   * Loads a trademark and its WDL and checks it.
   * 
   * @param trademarkIpKey The trademark to check
   * @param doStrictSyntaxCheck If this is <code>true</code>, the WDL text must stick to the given syntax exactly,
   *          otherwise the parsing is more tolerant
   * @return The check result
   * @throws WdlSyntaxException if the WDL syntax is invalid
   * @throws SQLException if the query for stopwords or terms fails
   */
  public WdlCheckResultVO checkWdl( Integer trademarkIpKey, boolean doStrictSyntaxCheck ) throws WdlSyntaxException, SQLException
  {
    Trademark trademark = TrademarkImpl.getTrademarkForKey( trademarkIpKey );
    return checkWdl( trademark, doStrictSyntaxCheck );
  }

  /**
   * Checks a trademark's WDL.
   * 
   * @param trademark The trademark to check
   * @param doStrictSyntaxCheck If this is <code>true</code>, the WDL text must stick to the given syntax exactly,
   *          otherwise the parsing is more tolerant
   * @return The check result
   * @throws WdlSyntaxException if the WDL syntax is invalid
   * @throws SQLException if the query for stopwords or terms fails
   */
  WdlCheckResultVO checkWdl( Trademark trademark, boolean doStrictSyntaxCheck ) throws WdlSyntaxException, SQLException
  {
    String wdlText = TrademarkImpl.getCurrentWdlText( trademark );
    List<Integer> wdlClasses = TrademarkImpl.getCurrentWdlClasses( trademark );
    Locale language = trademark.getLocale();
    return doWdlCheck( trademark.getDescription(), trademark.getIPKey(), trademark.getIPNumber(), trademark.getApplicationNumber(), trademark
        .getProlongationCount(), wdlText, language, wdlClasses, doStrictSyntaxCheck );
  }

  /**
   * Checks the pure WDL without loading a trademark.
   * 
   * @param wdlText The WDL text
   * @param language The WDL's language
   * @param doStrictSyntaxCheck If this is <code>true</code>, the WDL text must stick to the given syntax exactly,
   *          otherwise the parsing is more tolerant
   * @return The check result
   * @throws WdlSyntaxException if the WDL syntax is invalid
   * @throws SQLException if the query for stopwords or terms fails
   */
  public WdlCheckResultVO checkWdl( String wdlText, Locale language, boolean doStrictSyntaxCheck ) throws WdlSyntaxException, SQLException
  {
    return doWdlCheck( null, null, null, null, null, wdlText, language, null, doStrictSyntaxCheck );
  }

  /**
   * Checks the WDL.
   * 
   * @param trademarkName The trademark's name
   * @param ipKey The IP key
   * @param ipNumber The IP number
   * @param applicationNumber The trademark application number
   * @param prolongationCount The number of trademark prolongations
   * @param wdlText The real WDL text
   * @param language The language of the WDL
   * @param classNumbers The WDL class numbers
   * @param doStrictSyntaxCheck If this is <code>true</code>, the WDL text must stick to the given syntax exactly,
   *          otherwise the parsing is more tolerant
   * @return The check result
   * @throws WdlSyntaxException if the WDL syntax is invalid
   * @throws SQLException if the query for stopwords or terms fails
   */
  private WdlCheckResultVO doWdlCheck( final String trademarkName, final Integer ipKey, final Long ipNumber, final String applicationNumber,
      final Integer prolongationCount, final String wdlText, final Locale language, final List<Integer> classNumbers, boolean doStrictSyntaxCheck )
      throws WdlSyntaxException, SQLException
  {
    String cleanWdlText = WdlCharacterReplacer.replaceSpecialCharacters( wdlText );

    WdlVO wdlVO = parseAndInitialize( cleanWdlText, language, classNumbers, false, doStrictSyntaxCheck, true, this.wdlSyntaxConfiguration );
    wdlVO.setTrademarkName( trademarkName );
    wdlVO = checkWdl( wdlVO, getStopwords( language ), getPositiveTerms( language, getMaximumClassTextGroupLength( wdlVO ) ) );

    WdlCheckResultVO result = new WdlCheckResultVO( ipKey, prolongationCount, ipNumber, applicationNumber, wdlVO );
    return result;
  }

  /**
   * Gets the maximum length of all class text groups that exist in the WDL text.
   * 
   * @param wdlVO The WDL object containing the parsed WDL text
   * @return The maximum class text group length
   */
  private int getMaximumClassTextGroupLength( WdlVO wdlVO )
  {
    int maximumLength = 0;
    for ( ClassBlockVO classBlock: wdlVO.getClassBlocks() )
    {
      for ( ClassTextGroupVO classTextGroup: classBlock.getClassTextGroups() )
      {
        int currentLength = classTextGroup.getClassTextGroup().trim().length();
        if ( currentLength > maximumLength )
        {
          maximumLength = currentLength;
        }
      }
    }
    return maximumLength;
  }

  /**
   * Checks the class text group for multiple blanks and marks them as {@link CheckedClassTextPart#INVALID}.
   * 
   * @param checkedClassTextGroup The class text group to check
   */
  static void checkCharacters( CheckedClassTextGroupVO checkedClassTextGroup )
  {
    String classTextGroup = checkedClassTextGroup.getClassTextGroup();
    SortedSet<Integer> invalidCharacterPositions = CharacterChecker.checkWdl( classTextGroup );
    for ( Integer invalidCharacterPosition: invalidCharacterPositions )
    {
      int realPosition = invalidCharacterPosition - 1;
      checkedClassTextGroup.splitClassTextParts( realPosition, classTextGroup.substring( realPosition, realPosition + 1 ), CheckedClassTextPart.INVALID );
    }
  }

  /**
   * Checks a parsed WDL.
   * 
   * @param wdlVO The WDL object containing the full and parsed WDL text
   * @param stopwords The stopwords to use
   * @param positiveTerms The positive terms to use
   * @return The WDL with all class part texts checked
   */
  public static WdlVO checkWdl( WdlVO wdlVO, List<StopwordVO> stopwords, Map<Integer, List<AbstractTermVO>> positiveTerms )
  {
    for ( ClassBlockVO blockVO: wdlVO.getClassBlocks() )
    {
      List<CheckedClassTextGroupVO> checkedClassTextGroups = new ArrayList<CheckedClassTextGroupVO>( blockVO.getClassTextGroups().size() );

      for ( int i = 0; i < blockVO.getClassTextGroups().size(); i++ )
      {
        boolean isFirstClassTextGroup = ( i == 0 );
        CheckedClassTextGroupVO checkedClassTextGroup = (CheckedClassTextGroupVO) blockVO.getClassTextGroups().get( i );
        checkedClassTextGroups.add( checkedClassTextGroup );

        checkMultipleBlanks( checkedClassTextGroup );
        checkCharacters( checkedClassTextGroup );

        if ( blockVO.getClassNumber() != null )
        {
          checkTerms( positiveTerms.get( blockVO.getClassNumber() ), checkedClassTextGroup, isFirstClassTextGroup, wdlVO.getLanguage() );
        }
        else
        {
          LOG.warn( "WDL has no class number! " + wdlVO.getTrademarkName() + " with trademark number " + wdlVO.getTrademarkNumber() );
        }
        checkStopwords( stopwords, checkedClassTextGroup );
        changeInlinePunctuationMarks( checkedClassTextGroup );
        changeBrackets( checkedClassTextGroup );
      }
      checkDuplicateEntries( checkedClassTextGroups, wdlVO.getLanguage() );
      checkPunctuationMarks( checkedClassTextGroups );

    }
    return wdlVO;
  }

  /**
   * Checks whether the <code>CheckedClassTextGroupVO</code> has duplicate entries. If so, the class text parts are
   * now marked as being CheckedClassTextPart.DUPLICATED
   * 
   * @param checkedClassTextGroups The groups to check
   * @param language The language of the WDL
   */
  static void checkDuplicateEntries( List<CheckedClassTextGroupVO> checkedClassTextGroups, Locale language )
  {
    // This eats about 25% of the performance... Perhaps it could be done faster.
    Map<String, Integer> entries = new HashMap<String, Integer>( checkedClassTextGroups.size() );

    for ( int i = 0; i < checkedClassTextGroups.size(); i++ )
    {
      CheckedClassTextGroupVO classTextGroup = checkedClassTextGroups.get( i );
      String cleanClassTextGroup = classTextGroup.getClassTextGroup().trim().toLowerCase( language );
      if ( language.equals( Locale.FRENCH ) )
      {
        cleanClassTextGroup = SearchText.toLowerCaseAscii( cleanClassTextGroup, language );
      }
      Integer existingClassTextGroupIndex = entries.get( cleanClassTextGroup );
      if ( existingClassTextGroupIndex != null )
      {
        setAllClassTextPartsToDuplicate( classTextGroup );
        setAllClassTextPartsToDuplicate( checkedClassTextGroups.get( existingClassTextGroupIndex ) );
      }
      else
      {
        entries.put( cleanClassTextGroup, Integer.valueOf( i ) );
      }
    }
  }

  /**
   * Sets all class text parts of a class text group to {@link CheckedClassTextPart#DUPLICATE}.
   * 
   * @param classTextGroup The class text group that's changed
   */
  private static void setAllClassTextPartsToDuplicate( CheckedClassTextGroupVO classTextGroup )
  {
    for ( ClassTextPartVO classTextPart: classTextGroup.getClassTextParts() )
    {
      classTextPart.setCheckFlag( CheckedClassTextPart.DUPLICATE );
    }
  }

  /**
   * Parses a full WDL text and initializes all checked class text groups.
   * 
   * @param wdlText The full WDL text to parse
   * @param language The WDL's language
   * @param classNumbers The Nice class numbers of the trademark
   * @param lookForClassNumber If this is <code>true</code> and <code>wdlText</code> contains no class number, then
   *          it's fetched from <code>classNumbers</code>
   * @param doStrictSyntaxCheck If this is <code>true</code>, the WDL text must stick to the given syntax exactly,
   *          otherwise the parsing is more tolerant
   * @param keepBlanks If this is true, all blanks within class group texts are kept
   * @param wdlSyntaxConfiguration Configures how this reacts to WDL syntax errors; ignored if <code>null</code>
   * @return The corresponding WDL object
   * @throws WdlSyntaxException if the WDL text's syntax is invalid
   */
  public static WdlVO parseAndInitialize( String wdlText, Locale language, List<Integer> classNumbers, boolean lookForClassNumber, boolean doStrictSyntaxCheck,
      boolean keepBlanks, Map<WdlSyntaxConfiguration, Boolean> wdlSyntaxConfiguration ) throws WdlSyntaxException
  {
    WdlVO wdlVO;
    if ( doStrictSyntaxCheck )
    {
      wdlVO = WdlParser.parse( wdlText, keepBlanks, wdlSyntaxConfiguration ); // Implicitly does strict check
    }
    else
    {
      wdlVO = WdlParser.parse( wdlText, doStrictSyntaxCheck, keepBlanks );
    }
    wdlVO.setLanguage( language );

    for ( int i = 0; i < wdlVO.getClassBlocks().size(); i++ )
    {
      ClassBlockVO classBlock = wdlVO.getClassBlocks().get( i );

      if ( lookForClassNumber && classBlock.getClassNumber() == null )
      {
        LOG.info( "Got no class number in WDL '" + wdlText + "'" );
        if ( i < classNumbers.size() )
        {
          Integer defaultClassNumber = classNumbers.get( i );
          classBlock.setClassNumber( defaultClassNumber );
          LOG.info( "Set to class number " + defaultClassNumber + " from the DB" );
        }
        else
        {
          LOG.warn( "Can't set default class number, since index " + i + " is required from " + StringUtils.join( classNumbers.iterator(), ", " ) );
        }
      }

      List<ClassTextGroupVO> classTextGroups = classBlock.getClassTextGroups();
      List<ClassTextGroupVO> checkedClassTextGroups = new ArrayList<ClassTextGroupVO>( classTextGroups.size() );
      for ( ClassTextGroupVO classTextGroup: classTextGroups )
      {
        CheckedClassTextGroupVO checkedClassTextGroup = new CheckedClassTextGroupVO( classTextGroup.getClassTextGroup(), classTextGroup.getPunctuationMark() );
        checkedClassTextGroups.add( checkedClassTextGroup );
      }

      classBlock.setClassTextGroups( checkedClassTextGroups );
    }

    return wdlVO;
  }

  /**
   * Checks for stopwords and marks them as stopword.
   * 
   * @param stopwords The stopwords to use
   * @param checkedClassTextGroup The class text group whose text is marked
   */
  static void checkStopwords( List<StopwordVO> stopwords, CheckedClassTextGroupVO checkedClassTextGroup )
  {
    for ( StopwordVO stopwordVO: stopwords )
    {
      Pattern searchPattern;
      char lastCharacter = stopwordVO.getWord().charAt( stopwordVO.getWord().length() - 1 );
      if ( Character.isLetterOrDigit( lastCharacter ) )
      {
        searchPattern = RegexUtils.createUnicodeWordPattern( stopwordVO.getWord() );
      }
      else if ( lastCharacter == '\'' )
      {
        // A French stopword ending with a ' requires a following character - end of line is NOT valid!
        // A blank is tolerated, even if it's actually wrong (see below).
        searchPattern = RegexUtils.createUnicodeWordPattern( stopwordVO.getWord(), null, "(\\w|" + RegexUtils.LOWERCASE_CHARACTER + "| )" );
      }
      else
      {
        throw new IllegalArgumentException( "Stopword ending with '" + lastCharacter + "' is not supported" );
      }

      String classTextGroup = checkedClassTextGroup.getClassTextGroup();
      Matcher matcher = searchPattern.matcher( classTextGroup );
      while ( matcher.find() )
      {
        int position = matcher.start( 1 );
        String foundStopword = classTextGroup.substring( matcher.start( 1 ), matcher.end( 1 ) );
        ClassTextPartVO originalStopwordPart = checkedClassTextGroup.getPartAt( checkedClassTextGroup.new ModifiableInteger( position ) );
        CheckedClassTextPart checkFlag = originalStopwordPart.getCheckFlag();

        boolean stopwordHasBeenProcessed = false;
        // The stopword was found, but its leading blank is part of a multiple blank INVALID part
        if ( checkFlag.equals( CheckedClassTextPart.INVALID ) )
        {
          int positionAfterStopword = classTextGroup.indexOf( foundStopword.substring( 1 ), position );
          checkedClassTextGroup.splitClassTextParts( positionAfterStopword, foundStopword.substring( 1 ), CheckedClassTextPart.STOPWORD );
          stopwordHasBeenProcessed = true;
        }
        // Don't check if it already has been checked
        else if ( !( checkFlag.equals( CheckedClassTextPart.STOPWORD ) || checkFlag.equals( CheckedClassTextPart.CORRECT ) || checkFlag
            .equals( CheckedClassTextPart.CORRECT_WRONG_CASE ) ) )
        {
          checkedClassTextGroup.splitClassTextParts( position, foundStopword, CheckedClassTextPart.STOPWORD );
          stopwordHasBeenProcessed = true;
        }

        // Remember that a stopword ending with ' must be followed by a character.
        if ( stopwordHasBeenProcessed && matcher.groupCount() == 2 && matcher.group( 2 ).equals( " " ) )
        {
          int blankPosition = matcher.start( 2 );
          checkedClassTextGroup.splitClassTextParts( blankPosition, " ", CheckedClassTextPart.INVALID );
        }
      }
    }
  }

  /**
   * Checks the class text group for multiple blanks and marks them as {@link CheckedClassTextPart#INVALID}.
   * 
   * @param checkedClassTextGroup The class text group to check
   */
  static void checkMultipleBlanks( CheckedClassTextGroupVO checkedClassTextGroup )
  {
    String classTextGroup = checkedClassTextGroup.getClassTextGroup();
    Matcher matcher = MULTIPLE_BLANKS.matcher( classTextGroup );
    while ( matcher.find() )
    {
      String foundMultipleBlanks = classTextGroup.substring( matcher.start(), matcher.end() );
      checkedClassTextGroup.splitClassTextParts( matcher.start(), foundMultipleBlanks, CheckedClassTextPart.INVALID );
    }
  }

  /**
   * Checks whether the punctuation marks are correct over the whole class text:
   * <ul>
   * <li> The last class text group must contain a <code>.</code>, otherwise the previous character is
   * {@link CheckedClassTextPart#INVALID}. </li>
   * <li> Dots (<code>.</code>) within the class text are considered invalid if they're not part of an acronym.
   * </li>
   * <li> Blanks before punctuation marks like <code>,</code> or <code>;</code> or <code>.</code> are considered
   * {@link CheckedClassTextPart#INVALID}. </li>
   * <li> The character in the first class text group must be uppercase. The same is checked for every start of a
   * sentence, which means that a class text group ending with a dot (<code>.</code>) forces the next class text
   * group to start with an uppercase character. </li>
   * </ul>
   * 
   * @param checkedClassTextGroups The class text groups to check
   */
  static void checkPunctuationMarks( List<CheckedClassTextGroupVO> checkedClassTextGroups )
  {
    boolean previousClassTextGroupEndsWithDot = false;
    for ( int i = 0; i < checkedClassTextGroups.size(); i++ )
    {
      CheckedClassTextGroupVO checkedClassTextGroup = checkedClassTextGroups.get( i );
      String fullClassTextGroup = checkedClassTextGroup.getClassTextGroup();
      if ( checkedClassTextGroup.getPunctuationMark() != null )
      {
        fullClassTextGroup += checkedClassTextGroup.getPunctuationMark().toString();
      }

      // The whole class text must start with an uppercase character.
      if ( i == 0 && !startsWithUppercaseCharacter( fullClassTextGroup ) )
      {
        checkedClassTextGroup.splitClassTextParts( 0, fullClassTextGroup.substring( 0, 1 ), CheckedClassTextPart.INVALID );
      }
      // If the previous class text group ended with a dot, this must start with an uppercase character.
      if ( previousClassTextGroupEndsWithDot && !startsWithUppercaseCharacter( fullClassTextGroup ) )
      {
        int firstCharacterPosition = 0;
        // The leading spaces are also considered INVALID
        for ( int j = 0; j < fullClassTextGroup.length(); j++ )
        {
          if ( !Character.isWhitespace( fullClassTextGroup.charAt( j ) ) )
          {
            firstCharacterPosition = j;
            break;
          }
        }
        checkedClassTextGroup.splitClassTextParts( 0, fullClassTextGroup.substring( 0, firstCharacterPosition + 1 ), CheckedClassTextPart.INVALID );
      }

      List<Integer> acronymPositions = getAcronymPositions( fullClassTextGroup );
      int lastDotPosition = 0;
      int dotPosition = -1;
      do
      {
        dotPosition = fullClassTextGroup.indexOf( '.', lastDotPosition );
        if ( dotPosition > -1 )
        {
          // All dots except the last are invalid, unless they are part of an acronym
          boolean isLastDot = i == checkedClassTextGroups.size() - 1 && dotPosition == checkedClassTextGroup.getClassTextPartsLength() - 1;
          if ( !isLastDot && !acronymPositions.contains( dotPosition ) )
          {
            checkedClassTextGroup.splitClassTextParts( dotPosition, ".", CheckedClassTextPart.INVALID );
          }
        }
        lastDotPosition = dotPosition + 1;
      } while ( dotPosition > -1 );

      Matcher matcher = BLANK_BEFORE_PUNCTUATION_MARK.matcher( fullClassTextGroup );
      while ( matcher.find() )
      {
        checkedClassTextGroup.splitClassTextParts( matcher.start(), " ", CheckedClassTextPart.INVALID );
      }

      previousClassTextGroupEndsWithDot = fullClassTextGroup.endsWith( "." );
    }

    if ( checkedClassTextGroups.size() > 0 )
    {
      CheckedClassTextGroupVO lastClassTextGroup = checkedClassTextGroups.get( checkedClassTextGroups.size() - 1 );
      List<ClassTextPartVO> classTextPartsOfLastClassTextGroup = lastClassTextGroup.getClassTextParts();
      ClassTextPartVO lastClassTextPart = classTextPartsOfLastClassTextGroup.get( classTextPartsOfLastClassTextGroup.size() - 1 );
      String lastCharacter = StringUtils.right( lastClassTextPart.getClassTextPart(), 1 );
      // The last character of a class text must be a dot (which must be correct).
      // Otherwise it's invalid.
      if ( lastCharacter.equals( "." ) )
      {
        int position = lastClassTextGroup.getClassTextPartsLength() - 1;
        lastClassTextGroup.splitClassTextParts( position, lastCharacter, CheckedClassTextPart.CORRECT );
      }
      else
      {
        int position = lastClassTextGroup.getClassTextPartsLength() - 1;
        lastClassTextGroup.splitClassTextParts( position, lastCharacter, CheckedClassTextPart.INVALID );
      }
    }

  }

  /**
   * Gets the positions of dots within a text that are part of an acronym.
   * 
   * @param text The text to check
   * @return The positions of dots within acronyms
   */
  static List<Integer> getAcronymPositions( String text )
  {
    List<Integer> positions = new ArrayList<Integer>();
    if ( StringUtils.isEmpty( text ) )
    {
      return positions;
    }

    Pattern pattern = WdlParser.getAbbreviationsPattern();
    Matcher matcher = pattern.matcher( text );
    while ( matcher.find() )
    {
      String acronym = text.substring( matcher.start(), matcher.end() );
      String[] acronymNonDots = StringUtils.splitPreserveAllTokens( acronym, '.' );
      int index = matcher.start() - 1;
      for ( String nonDot: acronymNonDots )
      {
        if ( nonDot.length() > 0 )
        {
          index += nonDot.length() + 1; // Count the dot as well!
          positions.add( index );
        }
      }
    }
    return positions;
  }

  /**
   * Checks for terms and marks them as {@link CheckedClassTextPart#CORRECT} if it finds them.
   * 
   * @param positiveTerms The terms to use
   * @param checkedClassTextGroup The class text group whose text is marked
   * @param isFirstClassTextGroup If this is <code>true</code>, this checks the first class text group; in that
   *        case, the class text must start with an upper case letter; if this is <code>false</code>, it's handled
   *        normally
   * @param language The language 
   */
  static void checkTerms( List<AbstractTermVO> positiveTerms, CheckedClassTextGroupVO checkedClassTextGroup, boolean isFirstClassTextGroup, Locale language )
  {
    if ( positiveTerms != null )
    {
      UpperCaseCharacterConverter upperCaseCharacterConverter = new UpperCaseCharacterConverter( language );

      int previousLength = Integer.MAX_VALUE;
      for ( AbstractTermVO term: positiveTerms )
      {
        if ( term.getTerm().length() > previousLength )
        {
          throw new IllegalArgumentException( "Terms must be sorted by descending length" );
        }
        String classTextGroup = checkedClassTextGroup.getClassTextGroup();

        // If the search term is longer than the class text group, it makes no sense to look for it
        if ( term.getTerm().length() > classTextGroup.length() )
        {
          continue;
        }

        Pattern searchPattern = createTermPattern( term.getTerm(), language );
        Matcher matcher = searchPattern.matcher( classTextGroup );
        while ( matcher.find() )
        {
          int termPosition = matcher.start( 1 );
          String foundTerm = classTextGroup.substring( matcher.start( 1 ), matcher.end( 1 ) );

          // Check if the found term really is contained in the original text part. There were cases where one part
          // of the found term was in one original text part and another part of the found term was in a
          // second original text part - and the Checker blew up...
          // Note that the check for a 'blanks only' class text part comes from the special case below with a
          // multiple blank INVALID part.
          ClassTextPartVO originalTextPart = checkedClassTextGroup.getPartAt( checkedClassTextGroup.new ModifiableInteger( termPosition ) );
          boolean originalTextPartContainsOnlyBlanks = ( originalTextPart.getClassTextPart().trim().length() == 0 );
          if ( !originalTextPartContainsOnlyBlanks && !originalTextPart.getClassTextPart().contains( foundTerm ) )
          {
            continue;
          }

          if ( originalTextPart.getCheckFlag().equals( CheckedClassTextPart.INVALID ) )
          {
            // The term was found, but its leading blank is part of a multiple blank INVALID part
            if ( originalTextPartContainsOnlyBlanks )
            {
              int positionAfterTerm = classTextGroup.indexOf( foundTerm.substring( 1 ), termPosition );
              checkedClassTextGroup.splitClassTextParts( positionAfterTerm, foundTerm.substring( 1 ), CheckedClassTextPart.CORRECT );
            }
            else
            {
              // Leave all other INVALID parts alone
            }
          }
          else if ( !originalTextPart.getCheckFlag().equals( CheckedClassTextPart.CORRECT ) )
          {
            checkedClassTextGroup.splitClassTextParts( termPosition, foundTerm, CheckedClassTextPart.CORRECT );
          }

          // The term was found in a case-independet way. Now check for characters where the case differs.
          // Note: If the character is in another case at the beginning, then that's ok.
          // For instance if the term is 'computers', and it appears in the WDL at the beginning as 'Computers.',
          // then that's of course ok. So start at the first point where they could differ.
          String originalTerm = term.getTerm();
          if ( foundTerm.startsWith( " " ) )
          {
            originalTerm = " " + originalTerm;
          }

          if ( !foundTerm.equals( originalTerm ) )
          {
            boolean alreadyChecked = false;
            if ( isFirstClassTextGroup )
            {
              // In French, there's a catch: The special characters like 'é' are written as 'E' in upper case.
              // Since the Pattern searches case independent, a term 'epices' would also be found with 'Epices etc.',
              // even though the correct term is 'épices'. So check that case as well and make the found term UNKNOWN.
              if ( language.equals( Locale.FRENCH ) && Character.isUpperCase( foundTerm.charAt( 0 ) ) )
              {
                int checkPosition = checkFirstClassTextGroupInFrench( checkedClassTextGroup, termPosition, originalTerm, foundTerm, upperCaseCharacterConverter );
                checkCharacterCase( checkedClassTextGroup, termPosition, originalTerm, foundTerm, checkPosition );
                alreadyChecked = true;
              }
            }

            if ( !alreadyChecked )
            {
              if ( foundTerm.equalsIgnoreCase( originalTerm ) )
              {
                checkCharacterCase( checkedClassTextGroup, termPosition, originalTerm, foundTerm, ( isFirstClassTextGroup ? 1 : 0 ) );
              }
              else
              {
                // The term seems to be matching, but due to special character case changes, some false
                // positives come through: They are in fact invalid.
                ClassTextPartVO textPart = checkedClassTextGroup.getPartAt( checkedClassTextGroup.new ModifiableInteger( termPosition ) );
                textPart.setCheckFlag( CheckedClassTextPart.UNKNOWN );
              }
            }
          }
        }
        previousLength = term.getTerm().length();
      }
    }
  }

  /**
   * Checks the character case for a term that was found, but not exactly in the same case (upper case,
   * lower case or mixed case).
   * @param checkedClassTextGroup The class text group that's being checked
   * @param termPosition The position of the found term
   * @param originalTerm The term from the DB
   * @param foundTerm The term found in the WDL
   * @param characterComparisonStart The index where the comparison should start, either <code>0</code> or <code>1</code> 
   */
  private static void checkCharacterCase( CheckedClassTextGroupVO checkedClassTextGroup, int termPosition, String originalTerm, String foundTerm,
      int characterComparisonStart )
  {
    if ( characterComparisonStart != 0 && characterComparisonStart != 1 )
    {
      throw new IllegalArgumentException( "Character comparison can only start at 0 or 1" );
    }

    char[] foundCharacters = foundTerm.toCharArray();
    char[] originalCharacters = originalTerm.toCharArray();
    for ( int i = characterComparisonStart; i < foundCharacters.length; i++ )
    {
      char foundCharacter = foundCharacters[i];
      if ( i < originalCharacters.length )
      {
        char originalCharacter = originalCharacters[i];
        // As mentioned, if the case differs at the first position, it's ok.
        // The WdlChecker checks later that all sentences start with an uppercase character.
        if ( foundCharacter != originalCharacter )
        {
          if ( LOG.isDebugEnabled() )
          {
            LOG.debug( "Comparing '" + originalTerm + "' and '" + foundTerm + "': Character " + i + " is different: " + originalCharacter + " instead of "
                + foundCharacter );
          }
          int currentPosition = termPosition + i;
          ClassTextPartVO originalTextPart = checkedClassTextGroup.getPartAt( checkedClassTextGroup.new ModifiableInteger( currentPosition ) );
          if ( originalTextPart.getCheckFlag().equals( CheckedClassTextPart.CORRECT ) )
          {
            checkedClassTextGroup.splitClassTextParts( currentPosition, String.valueOf( foundCharacter ), CheckedClassTextPart.CORRECT_WRONG_CASE );
          }
        }
      }
    }
  }

  /**
   * French terms are special: A special character like <code>é</code> loses its accent when switched to
   * upper case (like <code>E</code>); so this checks the first class text group for a French class text.
   * @param checkedClassTextGroup The class text group that's being checked
   * @param termPosition The position of the found term
   * @param originalTerm The term from the DB
   * @param foundTerm The term found in the WDL
   * @param upperCaseCharacterConverter Used for checking whether a character is a special character
   * @return The character comparison start, for {@link #checkCharacterCase(CheckedClassTextGroupVO, int, String, String, int)}
   */
  private static int checkFirstClassTextGroupInFrench( CheckedClassTextGroupVO checkedClassTextGroup, int termPosition, String originalTerm, String foundTerm,
      UpperCaseCharacterConverter upperCaseCharacterConverter )
  {
    char originalFirstCharacter = originalTerm.charAt( 0 );
    List<Character> alternativeFirstLowerCaseCharacters = upperCaseCharacterConverter.firstCharacterConverter( foundTerm, true );
    if ( alternativeFirstLowerCaseCharacters == null )
    {
      // Don't check the first character; if it's lower case, another check
      // will take care of that error.
      return 1;
    }
    else
    {
      boolean firstCharacterIsValid = false;
      for ( Character alternativeCharacter: alternativeFirstLowerCaseCharacters )
      {
        if ( originalFirstCharacter == alternativeCharacter.charValue() )
        {
          firstCharacterIsValid = true;
          break;
        }
      }
      if ( !firstCharacterIsValid )
      {
        ClassTextPartVO textPart = checkedClassTextGroup.getPartAt( checkedClassTextGroup.new ModifiableInteger( termPosition ) );
        textPart.setCheckFlag( CheckedClassTextPart.UNKNOWN );
      }
      return 1;
    }
  }

  /**
   * Creates the Pattern for searching a term.
   * @param term The term to search
   * @param language The language of the term
   * @return The Pattern to use for searching the term in the DB results
   */
  public static Pattern createTermPattern( String term, Locale language )
  {
    String fullPattern = null;
    // The problem is that in French certain characters are written in different ways for lower case
    // and upper case. For instance, an 'é' can be written as upper case 'E' or 'É'.
    UpperCaseCharacterConverter upperCaseCharacterConverter = new UpperCaseCharacterConverter( language );
    List<Character> alternativeFirstCharacters = upperCaseCharacterConverter.firstCharacterConverter( term, false );
    if ( alternativeFirstCharacters != null )
    {
      alternativeFirstCharacters.add( Character.valueOf( term.charAt( 0 ) ) );
      StringBuilder termPatternStart = new StringBuilder( "[" );
      for ( Character alternativeFirstCharacter: alternativeFirstCharacters )
      {
        termPatternStart.append( alternativeFirstCharacter );
      }
      termPatternStart.append( "]" );

      // This avoids a JDK bug, see http://bugs.sun.com/view_bug.do?bug_id=6233084
      //TODO Remove this hack as soon as the JDK bug is fixed (starting with Java 7)
      if ( termPatternStart.toString().equals( "[Yÿ]" ) )
      {
        termPatternStart = new StringBuilder( "(?:Y|ÿ)" );
      }

      fullPattern = RegexUtils.createUnicodeWordRegex( term, termPatternStart.toString(), null );
    }
    else
    {
      fullPattern = RegexUtils.createUnicodeWordRegex( term, null, null );
    }

    return Pattern.compile( fullPattern, Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE );
  }

  /**
   * After terms have been checked, there may be a number of punctuation marks that are still
   * {@link CheckedClassTextPart#INVALID}; but punctuation marks between correct group text parts are also considered
   * correct. The same goes for punctuation marks between a correct group text part and a stopword.
   * 
   * @param checkedClassTextGroup The class text group whose punctuation marks are corrected
   */
  static void changeInlinePunctuationMarks( CheckedClassTextGroupVO checkedClassTextGroup )
  {
    for ( int i = 0; i < checkedClassTextGroup.getClassTextParts().size(); i++ )
    {
      ClassTextPartVO classTextPart = checkedClassTextGroup.getClassTextParts().get( i );
      String classTextPartString = classTextPart.getClassTextPart();
      // Punctuation marks are just 1 character long
      if ( classTextPartString.length() == 1 && classTextPart.getCheckFlag().equals( CheckedClassTextPart.UNKNOWN )
          && !Character.isLetterOrDigit( classTextPartString.charAt( 0 ) ) && classTextPartString.charAt( 0 ) != ' ' )
      {
        // The class text group can't start with a punctuation mark
        if ( i == 0 )
        {
          classTextPart.setCheckFlag( CheckedClassTextPart.INVALID );
          String classTextGroup = checkedClassTextGroup.getClassTextGroup();
          LOG.error( "Found punctuation mark '" + classTextPartString + "' at first class text position of '" + classTextGroup + "'" );
        }
        else
        {
          boolean isPreviousPartCorrect = false;
          CheckedClassTextPart previousCheckFlag = checkedClassTextGroup.getClassTextParts().get( i - 1 ).getCheckFlag();
          if ( previousCheckFlag.equals( CheckedClassTextPart.CORRECT ) || previousCheckFlag.equals( CheckedClassTextPart.STOPWORD ) )
          {
            isPreviousPartCorrect = true;
          }

          boolean isNextPartCorrect = false;
          // It's the last class text part anyway
          if ( i == checkedClassTextGroup.getClassTextParts().size() - 1 )
          {
            isNextPartCorrect = true;
          }
          else
          {
            CheckedClassTextPart nextCheckFlag = checkedClassTextGroup.getClassTextParts().get( i + 1 ).getCheckFlag();
            if ( nextCheckFlag.equals( CheckedClassTextPart.CORRECT ) || nextCheckFlag.equals( CheckedClassTextPart.STOPWORD ) )
            {
              isNextPartCorrect = true;
            }
          }

          // Punctuation marks surrounded by correct class text parts are also correct
          if ( isPreviousPartCorrect && isNextPartCorrect )
          {
            classTextPart.setCheckFlag( CheckedClassTextPart.CORRECT );
          }
        }
      }
    }
  }

  /**
   * After the terms have been checked, all brackets that are not {@link CheckedClassTextPart#CORRECT} are changed to
   * {@link CheckedClassTextPart#BRACKET}.
   * 
   * @param checkedClassTextGroup The class text group whose punctuation marks are corrected
   */
  static void changeBrackets( CheckedClassTextGroupVO checkedClassTextGroup )
  {
    String classTextGroup = checkedClassTextGroup.getClassTextGroup();
    for ( char bracket: BRACKETS )
    {
      int startPosition = -1;
      int bracketPosition = 0;
      do
      {
        bracketPosition = classTextGroup.indexOf( bracket, startPosition );
        if ( bracketPosition > -1 )
        {
          ClassTextPartVO bracketPart = checkedClassTextGroup.getPartAt( checkedClassTextGroup.new ModifiableInteger( bracketPosition ) );
          if ( !bracketPart.getCheckFlag().equals( CheckedClassTextPart.CORRECT ) )
          {
            checkedClassTextGroup.splitClassTextParts( bracketPosition, String.valueOf( bracket ), CheckedClassTextPart.BRACKET );
          }
        }
        startPosition = bracketPosition + 1;
      } while ( bracketPosition > -1 );
    }
  }

  /**
   * Shows whether the text starts with an uppercase character.
   * 
   * @param text The text to check
   * @return <code>true</code> if the text starts with an uppercase character, otherwise <code>false</code>
   */
  private static boolean startsWithUppercaseCharacter( String text )
  {
    if ( StringUtils.isEmpty( text ) )
    {
      return false;
    }
    else
    {
      String trimmedText = text.trim();
      if ( trimmedText.length() == 0 )
      {
        return false;
      }
      else
      {
        String firstCharacter = trimmedText.substring( 0, 1 );
        return firstCharacter.equals( firstCharacter.toUpperCase() );
      }
    }
  }

  /**
   * Gets the stopwords.
   * 
   * @param language The language to use
   * @return The stopwords
   * @throws SQLException if the query fails
   */
  private List<StopwordVO> getStopwords( Locale language ) throws SQLException
  {
    return this.stopwordDao.findAll( language );
  }

  /**
   * Gets the positive terms.
   * 
   * @param language The language to use
   * @param maximumLength The maximum length of term to look for
   * @return The terms with the class number as key
   * @throws SQLException if the query fails
   */
  private Map<Integer, List<AbstractTermVO>> getPositiveTerms( Locale language, Integer maximumLength ) throws SQLException
  {
    return this.generalTermDao.findPositiveTerms( language, maximumLength );
  }
}
