package ch.ige.wdltool.dao;

import java.util.HashSet;
import java.util.Locale;
import java.util.TreeSet;

import org.apache.commons.lang.RandomStringUtils;

import ch.ige.dbutils.sql.SQLWhereStatementCreator;
import ch.ige.util.date.DateConverter;
import ch.ige.wdltool.ParameterConstants;
import ch.ige.wdltool.dao.GeneralTermDAO.FullSqlStatement;
import ch.ige.wdltool.enums.DataSource;
import ch.ige.wdltool.enums.TermType;
import ch.ige.wdltool.vo.ExtendedSearchVO;

/**
 * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO} for its SQL statements.
 * <p>
 * Copyright (c) 2007, Swiss Federal Institute of Intellectual Property
 * </p>
 * @author Eric Lewis
 * @version $Id: GeneralTermDAOStatementTest.java,v 1.13 2008/12/09 09:39:32 heinzer Exp $
 */
public class GeneralTermDAOStatementTest extends AbstractGeneralTermDAOTestTemplate
{
  /**
   * The Integer value for a Boolean <code>true</code>
   */
  private static final Integer TRUE = 1;

  /**
   * The Integer value for a Boolean <code>false</code>
   */
  private static final Integer FALSE = 0;

  /**
   * The base <code>SELECT</code> statement with an alias
   */
  private static final String BASE_SELECT_WITH_ALIAS = "SELECT trm.term_id, trm.data_source_id AS data_source, trm.class_id AS class_number, trm.term_group_id, trm.term_type, "
      + "trm.language_sn AS language, trm.term, trm.search_term, trm.is_generic_bool AS is_generic, trm.is_public_bool AS is_public, trm.private_comment, trm.public_comment, "
      + "trm.is_active_bool AS is_active, trm.renewal_dat AS renewal, trm.usage_count, trm.latest_occurrence_dat AS latest_occurrence, "
      + "trm.created_user AS creation_user, trm.changed_user AS change_user, trm.changed_dat AS change_date, is_checker_only_bool AS is_checker_only";

  /**
   * The base <code>SELECT</code> statement
   */
  private static final String BASE_SELECT = BASE_SELECT_WITH_ALIAS.replaceAll( "trm\\.", "" ) + " FROM term";

  /**
   * The base <code>WHERE</code> statement
   */
  private static final String BASE_WHERE = " WHERE term_type != ? AND is_active_bool = ? AND ";

  /**
   * The only constructor
   * @param name The test's name
   */
  public GeneralTermDAOStatementTest( String name )
  {
    super( name );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#getDAOClass()
  @Override
  protected Class<GeneralTermDAO> getDAOClass()
  {
    return GeneralTermDAO.class;
  }

  /**
   * Tests
   * {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForFindTermWithinActiveTerms(Locale, Integer, String, Boolean)}.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForFindTermWithinActiveTerms() throws Exception
  {
    checkCreateWhereStatementForFindTermWithinActiveTerms( null );
    checkCreateWhereStatementForFindTermWithinActiveTerms( true );
    checkCreateWhereStatementForFindTermWithinActiveTerms( false );
  }

  private void checkCreateWhereStatementForFindTermWithinActiveTerms( Boolean isCheckerOnly ) throws Exception
  {
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForFindTermWithinActiveTerms( Locale.FRENCH, 1, "l'amour", isCheckerOnly );
    Object[] expectedParameters;
    String expectedQuery = " WHERE term_type != ? AND is_active_bool = ? AND language_sn = ?";
    if ( new Boolean( true ).equals( isCheckerOnly ) ) //isCheckerOnly can be null 
    {
      expectedQuery = expectedQuery + " AND is_checker_only_bool = ?";
      expectedParameters = new Object[] { TermType.IPI_TERM_CANDIDATE.toDB(), 1, "fr", 1, "l'amour", 1, TermType.NEGATIVE_IPI_TERM.toDB() };
    }
    else if ( new Boolean( false ).equals( isCheckerOnly ) )
    {
      expectedQuery = expectedQuery + " AND is_checker_only_bool = ?";
      expectedParameters = new Object[] { TermType.IPI_TERM_CANDIDATE.toDB(), 1, "fr", 0, "l'amour", 1, TermType.NEGATIVE_IPI_TERM.toDB() };
    }
    else
    {
      expectedParameters = new Object[] { TermType.IPI_TERM_CANDIDATE.toDB(), 1, "fr", "l'amour", 1, TermType.NEGATIVE_IPI_TERM.toDB() };
    }
    expectedQuery = expectedQuery + " AND term = ? AND (class_id = ? OR (class_id IS NULL AND term_type = ?))";
    checkWhereStatement( whereStatement, expectedQuery, expectedParameters );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createSqlStatementForSimpleFind(Locale, String, boolean, Boolean)}.
   * @throws Exception if the test fails
   */
  public void testCreateSqlStatementForSimpleFind() throws Exception
  {
    String germanWhereStart = "language_sn = ? AND ";
    String germanOrderBy = " ORDER BY class_id, NLSSORT(term, 'NLS_SORT = German')";

    String searchTerm1 = "Wölcken";
    String whereTest1 = "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? " + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? )";
    FullSqlStatement statement = getDAO().createSqlStatementForSimpleFind( Locale.GERMAN, searchTerm1, true, null );
    String expectedWhereStatement = BASE_WHERE + germanWhereStart + whereTest1 + germanOrderBy;
    assertEquals( "Checking statement for search term '" + searchTerm1 + "'", BASE_SELECT + expectedWhereStatement, statement.sqlStatement.toString() );
    checkWhereStatement( statement.whereStatement, expectedWhereStatement, 3, 1, "de", "%wölcken%", "%wolken%", "%wölcken%", "%wölcken%" );

    // Now with an SQL wildcard character
    String searchTerm2 = "Wöl%cken";
    String whereTest2 = "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? " + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? )";
    statement = getDAO().createSqlStatementForSimpleFind( Locale.GERMAN, searchTerm2, true, null );
    expectedWhereStatement = BASE_WHERE + germanWhereStart + whereTest2 + germanOrderBy;
    expectedWhereStatement = expectedWhereStatement.replace( "LIKE ?", "LIKE ? ESCAPE '~'" );
    assertEquals( "Checking statement for search term '" + searchTerm2 + "'", BASE_SELECT + expectedWhereStatement, statement.sqlStatement.toString() );
    checkWhereStatement( statement.whereStatement, expectedWhereStatement, 3, 1, "de", "%wöl~%cken%", "%wol~%ken%", "%wöl~%cken%", "%wöl~%cken%" );

    // Most of the WHERE statement is tested below
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForSimpleFind(Locale, String, boolean, Boolean)}.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForSimpleFind() throws Exception
  {
    String whereStart = "language_sn = ? AND ";
    String germanOrderBy = " ORDER BY class_id, NLSSORT(term, 'NLS_SORT = German')";
    String frenchOrderBy = " ORDER BY class_id, NLSSORT(term, 'NLS_SORT = French')";
    String italianOrderBy = " ORDER BY class_id, NLSSORT(term, 'NLS_SORT = Italian')";

    String searchTerm1 = "Wölcken";
    String whereTest1 = "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? )";
    SQLWhereStatementCreator statement = getDAO().createWhereStatementForSimpleFind( Locale.GERMAN, searchTerm1, true, null );
    String expectedSqlStatement = BASE_WHERE + whereStart + whereTest1 + germanOrderBy;
    checkWhereStatement( statement, expectedSqlStatement, 3, 1, "de", "%wölcken%", "%wolken%", "%wölcken%", "%wölcken%" );

    String searchTerm2 = "  Fratze Rheuma   Raaaabe  ";
    String whereTest2 = "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? " + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? OR "
        + "LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? " + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? OR "
        + "LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? " + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? )";
    statement = getDAO().createWhereStatementForSimpleFind( Locale.GERMAN, searchTerm2, true, null );
    expectedSqlStatement = BASE_WHERE + whereStart + whereTest2 + germanOrderBy;
    checkWhereStatement( statement, expectedSqlStatement, 3, 1, "de", "%fratze%", "%fraze%", "%fratze%", "%fratze%", "%rheuma%", "%rauma%", "%rheuma%",
        "%rheuma%", "%raaaabe%", "%rabe%", "%raaaabe%", "%raaaabe%" );

    // Exact phrase
    whereTest2 = "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? )";
    statement = getDAO().createWhereStatementForSimpleFind( Locale.FRENCH, searchTerm2, false, null );
    expectedSqlStatement = BASE_WHERE + whereStart + whereTest2 + frenchOrderBy;
    checkWhereStatement( statement, expectedSqlStatement, 3, 1, "fr", "%fratze rheuma   raaaabe%", "%fraze rauma   rabe%", "%fratze rheuma   raaaabe%",
        "%fratze rheuma   raaaabe%" );

    // Containing quotes
    String searchTerm3 = " Joe's Café";
    String whereTest3 = "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? )";
    statement = getDAO().createWhereStatementForSimpleFind( Locale.ITALIAN, searchTerm3, false, null );
    expectedSqlStatement = BASE_WHERE + whereStart + whereTest3 + italianOrderBy;
    checkWhereStatement( statement, expectedSqlStatement, 3, 1, "it", "%joe's café%", "%joe's cafe%", "%joe's café%", "%joe's café%" );

    // Without language
    String searchTerm4 = " Computer";
    String whereTest4 = "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? )";
    statement = getDAO().createWhereStatementForSimpleFind( null, searchTerm4, false, null );
    expectedSqlStatement = BASE_WHERE + whereTest4 + " ORDER BY class_id";
    checkWhereStatement( statement, expectedSqlStatement, 3, 1, "%computer%", "%computer%", "%computer%", "%computer%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with a faults.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithFaults() throws Exception
  {
    try
    {
      getDAO().createWhereStatementForComplexFind( null );
      fail( "Null search must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      ExtendedSearchVO search = new ExtendedSearchVO();
      search.isCheckerOnly = null;
      search.searchText = RandomStringUtils.randomAlphabetic( ParameterConstants.EXCEEDED_LIMIT_FIELD_TERM );
      getDAO().createWhereStatementForComplexFind( search );
      fail( "Too long search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with a search
   * text.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithSearchText() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchText = "l'amour";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND " + "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? "
        + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%l'amour%",
        "%l'amour%", "%l'amour%", "%l'amour%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the
   * language.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithLanguage() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchText = "l'amour";
    search.language = Locale.FRENCH;
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND language_sn = ? AND "
        + "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? "
        + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id, NLSSORT(term, 'NLS_SORT = French')",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, Locale.FRENCH.getLanguage(), "%l'amour%", "%l'amour%", "%l'amour%", "%l'amour%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the
   * sort language.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithSortLanguage() throws Exception
  {
    // Filter and sort by language
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchText = "l'amour";
    search.language = Locale.FRENCH;
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND language_sn = ? AND "
        + "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? "
        + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id, NLSSORT(term, 'NLS_SORT = French')",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, Locale.FRENCH.getLanguage(), "%l'amour%", "%l'amour%", "%l'amour%", "%l'amour%" );

    // Only sort by language
    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchText = "l'amour";
    search.language = null;
    search.sortLanguage = Locale.ITALIAN;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND " + "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? "
        + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id, NLSSORT(term, 'NLS_SORT = Italian')",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%l'amour%", "%l'amour%", "%l'amour%", "%l'amour%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the
   * 'is negative' flag.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithIsNegative() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.isNegative = Boolean.TRUE;
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND " + "term_type = ?" + " ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, TermType.NEGATIVE_IPI_TERM.toDB() );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.isNegative = Boolean.FALSE;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND " + "term_type != ?" + " ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, TermType.NEGATIVE_IPI_TERM.toDB() );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with a class
   * number.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithClassNumber() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.classNumbers = new HashSet<Integer>();
    search.classNumbers.add( 3 );
    search.searchText = "super";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND class_id = ? AND ( LOWER(term) LIKE ? OR LOWER(search_term)"
        + " LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, 3,
        "%super%", "%super%", "%super%", "%super%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.classNumbers = new HashSet<Integer>();
    search.classNumbers.add( ExtendedSearchVO.NO_CLASS );
    search.searchText = "duper";
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND class_id IS NULL AND ( LOWER(term) LIKE ? OR LOWER(search_term)"
        + " LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%duper%",
        "%duper%", "%duper%", "%duper%" );

    // Several class numbers
    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.classNumbers = new TreeSet<Integer>();
    search.classNumbers.add( 3 );
    search.classNumbers.add( 5 );
    search.classNumbers.add( 7 );
    search.searchText = "duper";
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND class_id IN (3, 5, 7) AND ( LOWER(term) LIKE ? OR LOWER(search_term)"
            + " LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE,
        "%duper%", "%duper%", "%duper%", "%duper%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the word
   * composition flag.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithWordComposition() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.compositionAnd = true;
    search.splitWords = true;
    search.searchText = "supertext mehrtext";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ( "
        + "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ?) AND "
        + "( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ?)" + " ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%supertext%", "%supertext%", "%supertext%", "%supertext%", "%mehrtext%", "%mehrtext%", "%mehrtext%",
        "%mehrtext%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.compositionAnd = false;
    search.splitWords = true;
    search.searchText = "supertext mehrtext";
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ( "
        + "LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? OR "
        + "LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ?" + " ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%supertext%", "%supertext%", "%supertext%", "%supertext%", "%mehrtext%", "%mehrtext%", "%mehrtext%",
        "%mehrtext%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with an exact
   * search.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithExactSearch() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.classNumbers = new HashSet<Integer>();
    search.classNumbers.add( 3 );
    search.exactSearch = true;
    search.searchComment = false;
    search.searchText = "exacttext";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND class_id = ? AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, 3, "%exacttext%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with split
   * words.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithSplitWords() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.compositionAnd = false;
    search.searchComment = true;
    search.splitWords = true;
    search.searchText = "supertext mehrtext";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ( LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? "
        + "OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%supertext%",
        "%supertext%", "%mehrtext%", "%mehrtext%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchComment = true;
    search.splitWords = false;
    search.searchText = "supertext mehrtext";
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ( LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) "
        + "ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%supertext mehrtext%", "%supertext mehrtext%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the
   * search comment flag.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithSearchComment() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchComment = null;
    search.searchText = "ultratext";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ( LOWER(term) LIKE ? OR LOWER(search_term)"
        + " LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE,
        "%ultratext%", "%ultratext%", "%ultratext%", "%ultratext%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchComment = true;
    search.searchText = "ultratext";
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ("
        + " LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%ultratext%",
        "%ultratext%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchComment = false;
    search.searchText = "ultratext";
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ("
        + " LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "%ultratext%", "%ultratext%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the 'is
   * public' flag.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithPublicFlag() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.isPublic = null;
    search.searchText = "supitext";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ( LOWER(term) LIKE ? OR LOWER(search_term)"
        + " LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE,
        "%supitext%", "%supitext%", "%supitext%", "%supitext%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.isPublic = true;
    search.searchText = "supitext";
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND is_public_bool = ? AND ( LOWER(term) LIKE ? OR LOWER(search_term)"
        + " LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, TRUE,
        "%supitext%", "%supitext%", "%supitext%", "%supitext%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.isPublic = false;
    search.searchText = "supitext";
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND is_public_bool = ? AND ( LOWER(term) LIKE ? OR LOWER(search_term)"
        + " LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, FALSE,
        "%supitext%", "%supitext%", "%supitext%", "%supitext%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the
   * term group number.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithTermGroupNumber() throws Exception
  {
    // Without term group number
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchText = "supitext";
    search.termGroupNumber = null;
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND ( LOWER(term) LIKE ? OR LOWER(search_term)"
        + " LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE,
        "%supitext%", "%supitext%", "%supitext%", "%supitext%" );

    // With term group number and search term
    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchText = "supitext";
    search.termGroupNumber = 123456;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND (tmg.nice_term_no = ? OR tmg.ipi_term_no = ?)"
        + " AND ( LOWER(term) LIKE ? OR LOWER(search_term) LIKE ? OR LOWER(private_comment) LIKE ? OR LOWER(public_comment) LIKE ? )" + " ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "123456", "123456", "%supitext%", "%supitext%", "%supitext%", "%supitext%" );

    // With term group number, but without search term
    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.searchText = null;
    search.termGroupNumber = 123456;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND (tmg.nice_term_no = ? OR tmg.ipi_term_no = ?)"
        + " ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "123456", "123456" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with a data
   * source.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithDataSource() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.dataSource = DataSource.NICE;
    search.searchText = "Kohlraben";
    search.searchComment = false;
    search.exactSearch = true;
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND data_source_id = ? AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, DataSource.NICE.toDB(), "%kohlraben%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.dataSource = DataSource.IPI;
    search.searchText = "Kohlraben";
    search.searchComment = false;
    search.exactSearch = true;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement, " WHERE term_type != ? AND is_active_bool = ? AND data_source_id = ? AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, DataSource.IPI.toDB(), "%kohlraben%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} without a data
   * source.
   * @throws Exception if the test fails
   */
  public void testCreateSqlStatementForComplexFindWithoutDataSource() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.dataSource = DataSource.NICE;
    search.searchText = "Kohlraben";
    search.searchComment = false;
    search.exactSearch = true;
    search.termGroupNumber = null;

    String expectedStatement = BASE_SELECT + " WHERE term_type != ? AND is_active_bool = ? AND data_source_id = ? AND ( LOWER(term) LIKE ? ) ORDER BY class_id";
    FullSqlStatement sqlStatement = getDAO().createSqlStatementForComplexFind( search );
    assertEquals( "Checking statement", expectedStatement, sqlStatement.sqlStatement.toString() );

    search.termGroupNumber = 123456;

    expectedStatement = BASE_SELECT.replace( "FROM term", "FROM term trm" ).replace( "term_group_id,", "trm.term_group_id," )
        + " LEFT JOIN term_group tmg ON trm.term_group_id = tmg.term_group_id"
        + " WHERE term_type != ? AND is_active_bool = ? AND data_source_id = ? AND (tmg.nice_term_no = ? OR tmg.ipi_term_no = ?)"
        + " AND ( LOWER(term) LIKE ? ) ORDER BY class_id";
    sqlStatement = getDAO().createSqlStatementForComplexFind( search );
    assertEquals( "Checking statement", expectedStatement, sqlStatement.sqlStatement.toString() );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the user that created the term.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithCreatedUser() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.classNumbers = new HashSet<Integer>();
    search.classNumbers.add( 3 );
    search.exactSearch = true;
    search.searchComment = false;
    search.searchText = "exacttext";
    search.createdUser = "bill_gates";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND class_id = ? AND created_user = ? AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, 3, "bill_gates", "%exacttext%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the user that changed the term.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithChangedUser() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.classNumbers = new HashSet<Integer>();
    search.classNumbers.add( 3 );
    search.exactSearch = true;
    search.searchComment = false;
    search.searchText = "exacttext";
    search.changedUser = "steve_ballmer";
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND class_id = ? AND changed_user = ? AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, 3, "steve_ballmer", "%exacttext%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the term's renewal date.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithRenewalDate() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.renewedFromDate = DateConverter.stringToTimestamp( "12.12.1988 11:00:00.000" );
    search.searchText = "Geburtstag";
    search.searchComment = false;
    search.exactSearch = true;
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND trunc(renewal_dat) >= to_date(?, 'dd.mm.yyyy') AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "12.12.1988", "%geburtstag%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.language = Locale.GERMAN;
    search.renewedToDate = DateConverter.stringToTimestamp( "11.11.1977 11:00:00.000" );
    search.classNumbers = new HashSet<Integer>();
    search.classNumbers.add( 3 );
    search.searchText = "Français";
    search.dataSource = DataSource.NICE;
    search.isGeneric = false;
    search.searchComment = false;
    search.exactSearch = true;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND language_sn = ? AND class_id = ? AND data_source_id = ? AND is_generic_bool = ? "
            + "AND trunc(renewal_dat) <= to_date(?, 'dd.mm.yyyy') " + "AND ( LOWER(term) LIKE ? ) ORDER BY class_id, NLSSORT(term, 'NLS_SORT = German')",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, Locale.GERMAN.getLanguage(), 3, DataSource.NICE.toDB(), FALSE, "11.11.1977", "%français%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.renewedFromDate = DateConverter.stringToTimestamp( "11.11.1977 11:00:00.000" );
    search.renewedToDate = DateConverter.stringToTimestamp( "12.12.1988 12:00:00.000" );
    search.searchText = "bla";
    search.searchComment = false;
    search.exactSearch = true;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND trunc(renewal_dat) >= to_date(?, 'dd.mm.yyyy') AND trunc(renewal_dat) <= to_date(?, 'dd.mm.yyyy') "
            + "AND ( LOWER(term) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "11.11.1977", "12.12.1988", "%bla%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the 'is
   * generic' flag.
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithGenericFlag() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.dataSource = DataSource.IPI;
    search.isGeneric = true;
    try
    {
      getDAO().createWhereStatementForComplexFind( search );
      fail( "Flag for generic term for data source IPI must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    search = new ExtendedSearchVO();
    search.dataSource = DataSource.IPI;
    search.isGeneric = false;
    try
    {
      getDAO().createWhereStatementForComplexFind( search );
      fail( "Flag for generic term for data source IPI must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.dataSource = DataSource.NICE;
    search.isGeneric = true;
    search.searchText = "Kohlraben";
    search.searchComment = false;
    search.exactSearch = true;
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND data_source_id = ? AND is_generic_bool = ? AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, DataSource.NICE.toDB(), TRUE, "%kohlraben%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.dataSource = DataSource.NICE;
    search.isGeneric = false;
    search.searchText = "Kohlraben";
    search.searchComment = false;
    search.exactSearch = true;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND data_source_id = ? AND is_generic_bool = ? AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, DataSource.NICE.toDB(), FALSE, "%kohlraben%" );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#createWhereStatementForComplexFind(ExtendedSearchVO)} with the
   * change date (start and end).
   * @throws Exception if the test fails
   */
  public void testCreateWhereStatementForComplexFindWithChangeDate() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.changedFromDate = DateConverter.stringToTimestamp( "12.12.1988 11:00:00.000" );
    search.searchText = "Geburtstag";
    search.searchComment = false;
    search.exactSearch = true;
    SQLWhereStatementCreator whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND trunc(changed_dat) >= to_date(?, 'dd.mm.yyyy') AND ( LOWER(term) LIKE ? ) ORDER BY class_id",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "12.12.1988", "%geburtstag%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.language = Locale.GERMAN;
    search.changedToDate = DateConverter.stringToTimestamp( "11.11.1977 11:00:00.000" );
    search.classNumbers = new HashSet<Integer>();
    search.classNumbers.add( 3 );
    search.searchText = "Français";
    search.dataSource = DataSource.NICE;
    search.isGeneric = false;
    search.searchComment = false;
    search.exactSearch = true;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND language_sn = ? AND class_id = ? AND data_source_id = ? AND is_generic_bool = ? "
            + "AND trunc(changed_dat) <= to_date(?, 'dd.mm.yyyy') " + "AND ( LOWER(term) LIKE ? ) ORDER BY class_id, NLSSORT(term, 'NLS_SORT = German')",
        TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, Locale.GERMAN.getLanguage(), 3, DataSource.NICE.toDB(), FALSE, "11.11.1977", "%français%" );

    search = new ExtendedSearchVO();
    search.isCheckerOnly = null;
    search.changedFromDate = DateConverter.stringToTimestamp( "11.11.1977 11:00:00.000" );
    search.changedToDate = DateConverter.stringToTimestamp( "12.12.1988 12:00:00.000" );
    search.searchText = "bla";
    search.searchComment = false;
    search.exactSearch = true;
    whereStatement = getDAO().createWhereStatementForComplexFind( search );
    checkWhereStatement( whereStatement,
        " WHERE term_type != ? AND is_active_bool = ? AND trunc(changed_dat) >= to_date(?, 'dd.mm.yyyy') AND trunc(changed_dat) <= to_date(?, 'dd.mm.yyyy') "
            + "AND ( LOWER(term) LIKE ? ) ORDER BY class_id", TermType.IPI_TERM_CANDIDATE.toDB(), TRUE, "11.11.1977", "12.12.1988", "%bla%" );
  }

  /**
   * Checks the <code>WHERE</code> statement.
   * @param actualStatement The actual statement
   * @param expectedQuery The expected SQL fragment
   * @param expectedParameters The expected parameters
   */
  public static void checkWhereStatement( SQLWhereStatementCreator actualStatement, String expectedQuery, Object... expectedParameters )
  {
    assertEquals( "Checking SQL fragment", expectedQuery, actualStatement.toString() );
    assertEquals( "Checking number of parameters", expectedParameters.length, actualStatement.getStatementParameters().length );
    for ( int i = 0; i < expectedParameters.length; i++ )
    {
      Object expectedParameter = expectedParameters[i];
      Object actualParameter = actualStatement.getStatementParameters()[i];
      assertEquals( "Checking parameter " + i, expectedParameter, actualParameter );
    }
  }
}