package ch.ige.wdltool.dao;

import java.sql.Timestamp;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.commons.lang.RandomStringUtils;
import org.apache.commons.lang.time.StopWatch;
import org.apache.log4j.Logger;

import ch.ige.dbutils.vo.VOUtils;
import ch.ige.util.collection.CollectionSize;
import ch.ige.util.date.DateConverter;
import ch.ige.wdltool.ParameterConstants;
import ch.ige.wdltool.enums.DataSource;
import ch.ige.wdltool.vo.AbstractTermVO;
import ch.ige.wdltool.vo.ExtendedSearchVO;
import ch.ige.wdltool.vo.IpiTermCandidateVO;
import ch.ige.wdltool.vo.IpiTermVO;
import ch.ige.wdltool.vo.NegativeIpiTermVO;
import ch.ige.wdltool.vo.NiceTermVO;
import ch.ige.wdlutil.text.SearchText;

/**
 * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO}. <p> Copyright (c) 2007, Swiss Federal Institute of Intellectual
 * Property </p>
 * @author Eric Lewis
 * @version $Id: GeneralTermDAOTest.java,v 1.63 2009/02/02 13:45:10 bandschapp Exp $
 */
public class GeneralTermDAOTest extends AbstractGeneralTermDAOTestTemplate
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( GeneralTermDAOTest.class );

  /**
   * The minimum number of terms in the DB
   */
  private static final int NUMBER_OF_TERMS = 24680;

  /**
   * The minimum number of German terms in the DB
   */
  private static final int NUMBER_OF_GERMAN_TERMS = 8100;

  /**
   * The only constructor
   * @param name The test's name
   */
  public GeneralTermDAOTest( String name )
  {
    super( name );
  }

  /**
   * Gets the latest migration date; all change dates must be later than this.
   * @return The latest migration date
   */
  public static Timestamp getLatestMigration()
  {
    try
    {
      return DateConverter.stringToTimestamp( "13.07.2007 11:00:00.000" );
    }
    catch ( ParseException e )
    {
      LOG.error( "Couldn't parse latest migration date: " + e.getMessage(), e );
      return null;
    }
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#getDAOClass()
  @Override
  protected Class<GeneralTermDAO> getDAOClass()
  {
    return GeneralTermDAO.class;
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#delete(Collection)}.
   * @throws Exception if the test fails
   */
  public void testDeleteSeveral() throws Exception
  {
    try
    {
      getDAO().delete( (Collection<Integer>) null );
      fail( "Null IDs must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().delete( new ArrayList<Integer>() );
      fail( "Empty IDs must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    AbstractTermVO term1 = createAndInsert();
    AbstractTermVO term2 = createAndInsert();

    List<Integer> termIds = new ArrayList<Integer>();
    termIds.add( term1.getIdentifier() );
    termIds.add( term2.getIdentifier() );

    getDAO().delete( termIds );

    AbstractTermVO loadedVO = read( term1.getIdentifier() );
    assertNull( "Checking that delete was successful", loadedVO );

    loadedVO = read( term2.getIdentifier() );
    assertNull( "Checking that delete was successful", loadedVO );

    clearSavedId();
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#updateClassNumber(List, Integer)}.
   * @throws Exception if the test fails
   */
  public void testUpdateClassNumber() throws Exception
  {
    List<AbstractTermVO> terms = new ArrayList<AbstractTermVO>();
    terms.add( new IpiTermVO() );

    try
    {
      getDAO().updateClassNumber( null, 3 );
      fail( "Null terms must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    // Empty terms does nothing
    getDAO().updateClassNumber( new ArrayList<AbstractTermVO>(), 3 );

    try
    {
      getDAO().updateClassNumber( terms, null );
      fail( "Null class number must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().updateClassNumber( terms, -1 );
      fail( "Negative class number must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    AbstractTermVO term1 = createAndInsert();
    AbstractTermVO term2 = createAndInsert();

    Integer oldClassNumber = term1.getClassNumber();
    Integer newClassNumber = oldClassNumber - 1;

    terms.clear();
    terms.add( term1 );
    terms.add( term2 );

    getDAO().updateClassNumber( terms, newClassNumber );

    AbstractTermVO loadedVO1 = read( term1.getIdentifier() );
    assertEquals( "Checking for class number update", newClassNumber, loadedVO1.getClassNumber() );

    AbstractTermVO loadedVO2 = read( term2.getIdentifier() );
    assertEquals( "Checking for class number update", newClassNumber, loadedVO2.getClassNumber() );

    // tearDown() only deletes one term, so let's do this ourselves
    getDAO().delete( VOUtils.extractIDs( terms ) );
    clearSavedId();
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findAllTerms()}.
   * @throws Exception if the test fails
   */
  public void testFindAllTerms() throws Exception
  {
    List<AbstractTermVO> allTerms = getDAO().findAllTerms();
    assertTrue( "Checking number of terms", allTerms.size() >= NUMBER_OF_TERMS );

    int numberOfGermanTerms = 0;
    for ( AbstractTermVO term: allTerms )
    {
      if ( term.getLanguage().equals( Locale.GERMAN ) )
      {
        numberOfGermanTerms++;
      }
    }
    assertTrue( "Checking number of German terms", numberOfGermanTerms >= NUMBER_OF_GERMAN_TERMS );

    checkForNoIpiTermCandidate( allTerms );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findAllTerms(Locale)}.
   * @throws Exception if the test fails
   */
  public void testFindAllTermsForLanguage() throws Exception
  {
    List<AbstractTermVO> allTerms = getDAO().findAllTerms( Locale.GERMAN );
    assertTrue( "Checking number of German terms", allTerms.size() >= NUMBER_OF_GERMAN_TERMS );

    checkForNoIpiTermCandidate( allTerms );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findPositiveTerms(Locale, Integer)}.
   * @throws Exception if the test fails
   */
  public void testFindPositiveTerms() throws Exception
  {
    final Integer MAXSIZE = 5;
    StopWatch stopWatch = new StopWatch();

    stopWatch.start();
    Map<Integer, List<AbstractTermVO>> someTerms = getDAO().findPositiveTerms( Locale.GERMAN, MAXSIZE );
    stopWatch.stop();
    LOG.info( "Getting some terms took " + stopWatch.toString() );
    checkForNoIpiTermCandidate( someTerms );

    stopWatch.reset();
    stopWatch.start();
    Map<Integer, List<AbstractTermVO>> allTerms = getDAO().findPositiveTerms( Locale.GERMAN, null );
    stopWatch.stop();
    int allTermsSize = CollectionSize.size( allTerms );

    LOG.info( "Getting all terms took " + stopWatch.toString() );
    LOG.debug( "Small term list size: " + someTerms.size() );
    LOG.debug( "Whole term list size: " + allTermsSize );
    assertTrue( "Checking list sizes", someTerms.size() <= allTermsSize );
    checkForNoIpiTermCandidate( allTerms );

    Map<Integer, Integer> lengthCounter = new HashMap<Integer, Integer>( MAXSIZE );
    for ( Integer classNumber: someTerms.keySet() )
    {
      for ( AbstractTermVO termVO: someTerms.get( classNumber ) )
      {
        Integer termLength = termVO.getTerm().length();
        if ( lengthCounter.containsKey( termLength ) )
        {
          lengthCounter.put( termLength, lengthCounter.get( termLength ) + 1 );
        }
        else
        {
          lengthCounter.put( termLength, 1 );
        }
        assertTrue( "Checking maximum length", termVO.getTerm().length() <= MAXSIZE );
      }
    }

    // Let's assume that at least one term with a length of 2 exists.
    for ( int i = 2; i <= MAXSIZE; i++ )
    {
      assertTrue( "Checking for terms with length " + i, lengthCounter.get( i ) != null && lengthCounter.get( i ) > 0 );
    }

    for ( Integer classNumber: allTerms.keySet() )
    {
      List<AbstractTermVO> classTerms = allTerms.get( classNumber );
      for ( AbstractTermVO termVO: classTerms )
      {
        assertTrue( "Checking for positive term", !( termVO instanceof NegativeIpiTermVO ) );
      }
    }
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findSameSearchTermWithinActiveTerms(AbstractTermVO)}.
   * @throws Exception if the test fails
   */
  public void testFindSameSearchTermWithinActiveTerms() throws Exception
  {
    IpiTermVO searcherVO = new IpiTermVO();
    try
    {
      searcherVO.setTerm( "bla" );
      getDAO().findSameSearchTermWithinActiveTerms( searcherVO );
      fail( "Null language must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      searcherVO.setTerm( null );
      searcherVO.setLanguage( Locale.GERMAN );
      getDAO().findSameSearchTermWithinActiveTerms( searcherVO );
      fail( "Null search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      searcherVO.setTerm( "" );
      getDAO().findSameSearchTermWithinActiveTerms( searcherVO );
      fail( "Empty search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    // Just to show the difference, start with a standard search
    String generalSearchTerm = "Früchte";
    String exactSearchTerm = "Früchte [gekocht]";

    List<AbstractTermVO> terms = getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, generalSearchTerm );
    assertTrue( "Checking for any word", terms.size() >= 5 );
    boolean foundExactSearchTerm = false;
    for ( AbstractTermVO term: terms )
    {
      if ( term.getTerm().equals( exactSearchTerm ) )
      {
        foundExactSearchTerm = true;
        break;
      }
    }
    assertTrue( "Checking that '" + exactSearchTerm + "' was found", foundExactSearchTerm );

    searcherVO.setTerm( exactSearchTerm );
    // Now search exactly for the search term
    terms = getDAO().findSameSearchTermWithinActiveTerms( searcherVO );
    assertEquals( "Checking for exact search term '" + exactSearchTerm + "'", 1, terms.size() );
    assertEquals( "Checking found term", exactSearchTerm, terms.get( 0 ).getTerm() );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findTermWithinActiveTerms(Locale, Integer, String)}. Only base
   * functionality is tested (no exception by query)
   * @throws Exception if the test fails
   */
  public void testFindTermWithinActiveTerms() throws Exception
  {
    try
    {
      getDAO().findTermWithinActiveTerms( null, 1, "bla" );
      fail( "Null language must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findTermWithinActiveTerms( Locale.GERMAN, 1, null );
      fail( "Null term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findTermWithinActiveTerms( Locale.GERMAN, 1, "" );
      fail( "Empty term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findTermWithinActiveTerms( Locale.GERMAN, 1, RandomStringUtils.randomAlphabetic( ParameterConstants.EXCEEDED_LIMIT_FIELD_TERM ) );
      fail( "Too long term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    String term = "Getränke auf der Basis von Tee";
    List<AbstractTermVO> terms = getDAO().findTermWithinActiveTerms( Locale.GERMAN, 30, term );
    assertNotNull( terms );
    int expectedNumberOfTerms = 1;
    assertTrue( "Checking number of exactly matching terms found", terms.size() >= expectedNumberOfTerms );

    term = "Some text that does surely not exist as term";
    terms = getDAO().findTermWithinActiveTerms( Locale.GERMAN, 1, term );
    assertNotNull( terms );
    assertNotNull( "Checking that find returned a a valid value", terms );
    assertEquals( "Checking number of terms", 0, terms.size() );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findAnyWordWithinActiveTerms(Locale, String)}.
   * @throws Exception if the test fails
   */
  public void testFindAnyWordWithinActiveTerms() throws Exception
  {
    // Several hard-to predict results
    String searchTerm = "Rad";
    int expectedNumberOfGermanTerms = 39;
    int numberOfGermanTerms = 0;
    int expectedNumberOfFrenchTerms = 51;
    int numberOfFrenchTerms = 0;
    int expectedNumberOfItalianTerms = 83;
    int numberOfItalianTerms = 0;

    List<AbstractTermVO> terms = getDAO().findAnyWordWithinActiveTerms( searchTerm );
    for ( AbstractTermVO term: terms )
    {
      if ( term.getLanguage().equals( Locale.GERMAN ) )
      {
        numberOfGermanTerms++;
      }
      else if ( term.getLanguage().equals( Locale.FRENCH ) )
      {
        numberOfFrenchTerms++;
      }
      else if ( term.getLanguage().equals( Locale.ITALIAN ) )
      {
        numberOfItalianTerms++;
      }
    }
    assertTrue( "Checking German terms", numberOfGermanTerms >= expectedNumberOfGermanTerms );
    assertTrue( "Checking French terms", numberOfFrenchTerms >= expectedNumberOfFrenchTerms );
    assertTrue( "Checking Italian terms", numberOfItalianTerms >= expectedNumberOfItalianTerms );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findAnyWordWithinActiveTerms(String)} with faults.
   * @throws Exception if the test fails
   */
  public void testFindAnyWordWithinActiveTermsWithFaults() throws Exception
  {
    try
    {
      getDAO().findAnyWordWithinActiveTerms( null );
      fail( "Null search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findAnyWordWithinActiveTerms( "" );
      fail( "Empty search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findAnyWordWithinActiveTerms( "           " );
      fail( "Blank search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findAnyWordWithinActiveTerms( RandomStringUtils.randomAlphabetic( ParameterConstants.EXCEEDED_LIMIT_FIELD_TERM ) );
      fail( "Too long search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findAnyWordWithinActiveTerms(Locale, String)}.
   * @throws Exception if the test fails
   */
  public void testFindAnyWordWithinActiveTermsWithLanguage() throws Exception
  {
    // One search result
    String searchTerm = "Katze";
    List<AbstractTermVO> terms = getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, searchTerm );
    checkSearchResult( terms, new String[] { "Kratzbäume für Katzen", "Katzenstreu" }, new String[] { "krazbaume fur kazen", "kazenstrau" }, new Integer[] {
        20, 31 } );

    // Several search results
    searchTerm = "Brontzze"; // Note the wrong spelling!
    terms = getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, searchTerm );
    checkSearchResult( terms, new String[] { "Bronzepulver", "Bronze", "Bronzen für Grabmäler", "Bronzen [Kunstgegenstände]",
        "Beschichtung und Veredelung von Stahl, Gusseisen, Kupfer, Messing, Bronze, Aluminium, Chrom, Kunststoffen und Magnesium" }, new String[] {
        "bronzepulver", "bronze", "bronzen fur grabmaler", "bronzen [kunstgegenstande]",
        "beschichtung und veredelung von stahl, guseisen, kupfer, mesing, bronze, aluminium, chrom, kunststofen und magnesium" }, new Integer[] { 2, 6, 6, 6,
        40 } );

    // Check correct German sorting order
    searchTerm = "Mittel";
    terms = getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, searchTerm );
    final String[] expectedTerms = { "Appreturmittel", "Ätzmittel", "Auffrischungsmittel", "Entfärbungsmittel", "Entformungsmittel", "Leimlösungsmittel",
        "Lösungsmittel" }; // Just a subset of class 1, but in the corrected order
    int foundIndex = 0;
    for ( AbstractTermVO term: terms )
    {
      if ( term.getClassNumber() != null && term.getClassNumber() == 1 && foundIndex < expectedTerms.length )
      {
        if ( term.getTerm().startsWith( expectedTerms[foundIndex] ) )
        {
          foundIndex++;
        }
      }
    }
    assertEquals( "Checking that all terms were found in order", foundIndex, expectedTerms.length );

    // Check quote
    searchTerm = "d'armes";
    terms = getDAO().findAnyWordWithinActiveTerms( Locale.FRENCH, searchTerm );
    checkSearchResult( terms, new String[] { "culasses d'armes à feu", "mires d'armes à feu", "tourillons d'armes lourdes" }, new String[] {
        "culases d'armes a fau", "mires d'armes a fau", "tourilons d'armes lourdes" }, new Integer[] { 13, 13, 13 } );

    // Negative term
    searchTerm = "Süssgetränke";
    terms = getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, searchTerm );
    checkSearchResult( terms, new String[] { "Süssgetränke" }, new String[] { "susgetranke" }, new Integer[] { null } );
    assertEquals( "Checking class for negative term", NegativeIpiTermVO.class, terms.get( 0 ).getClass() );

    // Nice and IPI terms
    searchTerm = "Sojamilch";
    terms = getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, searchTerm );
    checkSearchResult( terms, new String[] { "Sojamilch [Milchersatz]", "Sojamilch [Würze]" }, new String[] { "sojamilch [milchersaz]", "sojamilch [wurze]" },
        new Integer[] { 29, 30 } );
    assertEquals( "Checking class", NiceTermVO.class, terms.get( 0 ).getClass() );
    assertEquals( "Checking class", IpiTermVO.class, terms.get( 1 ).getClass() );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findAnyWordWithinActiveTerms(Locale, String)} with faults.
   * @throws Exception if the test fails
   */
  public void testFindAnyWordWithinActiveTermsWithLanguageWithFaults() throws Exception
  {
    try
    {
      getDAO().findAnyWordWithinActiveTerms( null, "bla" );
      fail( "Null language must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, null );
      fail( "Null search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, "" );
      fail( "Empty search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, "           " );
      fail( "Blank search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findAnyWordWithinActiveTerms( Locale.GERMAN, RandomStringUtils.randomAlphabetic( ParameterConstants.EXCEEDED_LIMIT_FIELD_TERM ) );
      fail( "Too long search term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }
  }

  /**
   * Returns a extended search configured for default search for a terms text
   * @param term The term to look for
   * @return The configured search
   */
  private ExtendedSearchVO getDefaultSearch( AbstractTermVO term )
  {
    ExtendedSearchVO searchDefinition = new ExtendedSearchVO();
    searchDefinition.searchText = term.getTerm();
    return searchDefinition;
  }

  /**
   * Checks for a single existing term.
   * @param hits The found terms
   * @param specificTerm The expected term
   */
  private void assertSingleSpecificTerm( List<AbstractTermVO> hits, AbstractTermVO specificTerm )
  {
    assertEquals( "extended search should find only the single term inserted by 'createAndInsert'", 1, hits.size() );
    compareVOs( hits.get( 0 ), specificTerm );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findTermWithinActiveTerms(ExtendedSearchVO searchDefinition)}.
   * Several but not all filters / configurations in different combinations are tested.
   * As base term the term inserted by <code>createAndInsert</code> is used - i.e. a Nice term (e.g. no renewal date)
   * @throws Exception if the test fails
   */
  public void testFindTermWithinActiveTermsForExtendedSearch() throws Exception
  {
    // insert an example term as defined by getSerializableVO
    AbstractTermVO niceTerm = createAndInsert( true );
    this.setSavedId( niceTerm.getIdentifier() );

    // default search
    ExtendedSearchVO searchDefinition = getDefaultSearch( niceTerm );
    List<AbstractTermVO> hits = getDAO().findTermWithinActiveTerms( searchDefinition );

    if ( hits.size() > 1 )
    { //delete all and try again...the test-term appears more than just once

      //take care: deleting one by one does not work from this DAO because GeneralTermDAO getTermType returns NULL -> NPExecption
      //batch-deletion works though

      List<Integer> termIds = new ArrayList<Integer>(); //required for batch-delete
      for ( AbstractTermVO abstractTermVO: hits )
      {
        termIds.add( abstractTermVO.getIdentifier() );
      }
      getDAO().delete( termIds );
      niceTerm = createAndInsert( true );
      this.setSavedId( niceTerm.getIdentifier() );

      hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    }

    assertSingleSpecificTerm( hits, niceTerm );

    Locale correctLanguage = niceTerm.getLanguage();
    Locale incorrectLanguage;
    if ( Locale.GERMAN == correctLanguage )
    {
      incorrectLanguage = Locale.FRENCH;
    }
    else
    {
      incorrectLanguage = Locale.GERMAN;
    }

    // set correct language
    searchDefinition.language = correctLanguage;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    // set different language
    searchDefinition.language = incorrectLanguage;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    // find by exact term text (no fuzzy search)
    searchDefinition = getDefaultSearch( niceTerm );
    searchDefinition.exactSearch = true;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    // find by exact term text (no fuzzy search) with wrong case: should match anyway
    searchDefinition = new ExtendedSearchVO();
    Character firstChar = niceTerm.getTerm().charAt( 0 );
    if ( Character.isUpperCase( firstChar ) )
    {
      firstChar = Character.toLowerCase( firstChar );
    }
    else
    {
      firstChar = Character.toUpperCase( firstChar );
    }
    searchDefinition.searchText = firstChar + niceTerm.getTerm().substring( 1 );
    searchDefinition.exactSearch = true;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    // find by exact term text (no fuzzy search) with different spelling: should not match
    searchDefinition.searchText = SearchText.toSearchText( niceTerm.getTerm(), Locale.GERMAN );
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    // and the same with fuzzy search: should match
    searchDefinition.exactSearch = false;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    // splitting words should hit at least our example term
    searchDefinition.splitWords = true;
    searchDefinition.compositionAnd = false;
    searchDefinition.exactSearch = false;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertTrue( hits.size() > 0 );

    // but searching with AND we should not find anything as the example term contains some very strange words
    searchDefinition.compositionAnd = true;
    searchDefinition.exactSearch = true;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    // back to default OR: comparing hits by language
    searchDefinition.compositionAnd = false;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertTrue( "There have to be many hits searching with OR and short parts", hits.size() > 10 );
    int allLanguagesHits = hits.size();
    // set different language
    searchDefinition.language = incorrectLanguage;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertTrue( "There should be less hits with restriction on a single language", hits.size() < allLanguagesHits );

    // reset language
    searchDefinition.language = null;

    // searching in comment only, as exactly as possible
    searchDefinition.searchComment = true;
    searchDefinition.compositionAnd = false;
    searchDefinition.splitWords = false;
    searchDefinition.searchText = niceTerm.getPublicComment();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    searchDefinition.searchText = niceTerm.getPrivateComment();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    // searching in term only
    searchDefinition.searchComment = false;
    searchDefinition.searchText = niceTerm.getPublicComment();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    searchDefinition.searchText = niceTerm.getPrivateComment();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    // searching in comment only but fuzzy search
    searchDefinition.exactSearch = false;
    searchDefinition.splitWords = true;
    searchDefinition.compositionAnd = false;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertTrue( "There have to be several hits searching with short parts", hits.size() > 1 );

    searchDefinition = getDefaultSearch( niceTerm );

    // filtering on correct datasource
    searchDefinition.dataSource = niceTerm.getDataSource();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    // filtering on different data sources

    DataSource correctDataSource = niceTerm.getDataSource();
    DataSource incorrectDataSource;
    if ( DataSource.NICE == correctDataSource )
    {
      incorrectDataSource = DataSource.IPI;
    }
    else
    {
      incorrectDataSource = DataSource.NICE;
    }

    searchDefinition.dataSource = correctDataSource;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    searchDefinition.dataSource = incorrectDataSource;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    searchDefinition = getDefaultSearch( niceTerm );

    // filtering on isGeneric: this works only for a nice Term!
    searchDefinition.dataSource = DataSource.NICE;
    searchDefinition.isGeneric = ( (NiceTermVO) niceTerm ).isGeneric();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    searchDefinition.isGeneric = !searchDefinition.isGeneric;
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    // filtering on changedUser
    searchDefinition = getDefaultSearch( niceTerm );
    searchDefinition.changedUser = niceTerm.getChangeUser();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    searchDefinition.changedUser = "un_known";
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    // filtering on createdUser
    searchDefinition = getDefaultSearch( niceTerm );
    searchDefinition.createdUser = niceTerm.getCreationUser();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    searchDefinition.createdUser = "un_known";
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    // filtering on changed Date
    searchDefinition = getDefaultSearch( niceTerm );

    Calendar calendar = Calendar.getInstance();
    calendar.setTime( niceTerm.getChangeDate() );
    calendar.add( Calendar.DATE, 5 );
    Date after = calendar.getTime();

    calendar.setTime( niceTerm.getChangeDate() );
    calendar.add( Calendar.DATE, -5 );
    Date before = calendar.getTime();

    searchDefinition.changedFromDate = before;
    searchDefinition.changedToDate = after;

    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    searchDefinition.changedFromDate = after;
    searchDefinition.changedToDate = new Date();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertEquals( 0, hits.size() );

    searchDefinition.changedFromDate = before;
    searchDefinition.changedToDate = new Date();

    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    // No search term at all
    searchDefinition = getDefaultSearch( niceTerm );
    searchDefinition.searchText = null;
    searchDefinition.createdUser = niceTerm.getCreationUser();
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    assertSingleSpecificTerm( hits, niceTerm );

    // Several class numbers
    searchDefinition = getDefaultSearch( niceTerm );
    searchDefinition.searchText = null;
    searchDefinition.classNumbers = new HashSet<Integer>();
    searchDefinition.classNumbers.add( 3 );
    searchDefinition.classNumbers.add( 5 );
    hits = getDAO().findTermWithinActiveTerms( searchDefinition );
    for ( AbstractTermVO term: hits )
    {
      assertTrue( "Checking class number", term.getClassNumber().equals( 3 ) || term.getClassNumber().equals( 5 ) );
    }
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findTermWithinActiveTerms(ExtendedSearchVO searchDefinition)}
   * with different sort languages, but no filtering language.
   * @throws Exception if the test fails
   */
  public void testFindTermWithinActiveTermsForExtendedSearchWithSortLanguage() throws Exception
  {
    ExtendedSearchVO search = new ExtendedSearchVO();
    // This is contained in terms like "Öl", but also "politiques".
    // Depending on the sort language, the result sequence must be different.
    search.searchText = "ol";
    search.sortLanguage = Locale.GERMAN;

    List<AbstractTermVO> termsSortedByGerman = getDAO().findTermWithinActiveTerms( search );

    search.sortLanguage = Locale.FRENCH;
    List<AbstractTermVO> termsSortedByFrench = getDAO().findTermWithinActiveTerms( search );

    assertEquals( "Checking number of terms", termsSortedByGerman.size(), termsSortedByFrench.size() );
    boolean sequenceIsTheSame = true;
    for ( int i = 0; i < termsSortedByGerman.size(); i++ )
    {
      AbstractTermVO termSortedByGerman = termsSortedByGerman.get( i );
      AbstractTermVO termSortedByFrench = termsSortedByFrench.get( i );
      LOG.debug( "Sorted by German: '" + termSortedByGerman.getTerm() + "', sorted by French: '" + termSortedByFrench.getTerm() + "'" );
      if ( !termSortedByGerman.getIdentifier().equals( termSortedByFrench.getIdentifier() ) )
      {
        sequenceIsTheSame = false;
        break;
      }
    }

    assertEquals( "Checking that the same hits sorted in different languages differ in sequence", false, sequenceIsTheSame );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findTermWithinActiveTerms(ExtendedSearchVO searchDefinition)}
   * for finding all terms within a term group.
   * @throws Exception if the test fails
   */
  public void testFindTermWithinActiveTermsForTermGroup() throws Exception
  {
    final int niceTermGroupNumber = 90590; // This may change!
    final int niceTermGroupId = 10692; // This may change!
    final int numberOfExpectedTerms = 3;
    ExtendedSearchVO search = new ExtendedSearchVO();
    search.termGroupNumber = niceTermGroupNumber;

    List<AbstractTermVO> terms = getDAO().findTermWithinActiveTerms( search );
    assertEquals( "Checking number of found terms in term group " + niceTermGroupNumber, numberOfExpectedTerms, terms.size() );
    for ( AbstractTermVO abstractTermVO: terms )
    {
      NiceTermVO niceTerm = (NiceTermVO) abstractTermVO;
      assertEquals( "Checking term group number", niceTermGroupId, niceTerm.getTermGroupId().intValue() );
      assertNotNull( "Checking term group terms", niceTerm.getTermGroupTerms() );
      // The current term must not be part of the term group terms
      assertEquals( "Checking number of term group terms", numberOfExpectedTerms - 1, CollectionSize.size( niceTerm.getTermGroupTerms() ) );
    }

    terms = getDAO().findTermWithinActiveTerms( search, false );
    assertEquals( "Checking number of found terms in term group " + niceTermGroupNumber, numberOfExpectedTerms, terms.size() );
    for ( AbstractTermVO abstractTermVO: terms )
    {
      NiceTermVO niceTerm = (NiceTermVO) abstractTermVO;
      assertTrue( "Checking term group terms", niceTerm.getTermGroupTerms().isEmpty() );
    }
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#findTermWithinActiveTerms( Locale language, Integer classNumber, String term )}
   * for finding all terms given the language, the class number and the term.
   * @throws Exception if the test fails
   */
  public void testFindTermWithinActiveTermsForAnalyzer() throws Exception
  {
    final Locale language = Locale.GERMAN;
    final int classnumber = 1;
    final String term = "Weinsteinsäure";
    final int numberOfExpectedTerms = 1; // This may change!

    //No results should be found with wrong class number
    List<AbstractTermVO> terms = getDAO().findTermWithinActiveTerms( language, 2, term );
    assertEquals( "Should not find term " + term + " in class 2", 0, terms.size() );

    //No results should be found with wrong language
    terms = getDAO().findTermWithinActiveTerms( Locale.FRENCH, classnumber, term );
    assertEquals( "Should not find term " + term + " in French", 0, terms.size() );

    //Now it should find it
    terms = getDAO().findTermWithinActiveTerms( language, classnumber, term );
    assertEquals( "Checking number of found terms ", numberOfExpectedTerms, terms.size() );
    for ( AbstractTermVO abstractTermVO: terms )
    {
      assertEquals( "Checking classnumber", classnumber, abstractTermVO.getClassNumber().intValue() );
      assertEquals( "Checking language", language, abstractTermVO.getLanguage() );
      assertEquals( "Checking term", term, abstractTermVO.getTerm() );
    }
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO#filterWholeWordOnly(ExtendedSearchVO, List)}.
   * @throws Exception if the test fails
   */
  public void testFilterWholeWordOnly() throws Exception
  {
    ExtendedSearchVO extendedSearch = new ExtendedSearchVO();
    extendedSearch.searchText = "ölkanne";
    extendedSearch.wholeWordOnly = false;
    extendedSearch.exactSearch = true;

    // No filtering
    List<AbstractTermVO> terms = createTerms();
    assertEquals( "Checking number of terms", 5, terms.size() );
    GeneralTermDAO.filterWholeWordOnly( extendedSearch, terms );
    assertEquals( "Checking number of terms after filtering", 5, terms.size() );

    // Filter
    extendedSearch.wholeWordOnly = true;
    terms = createTerms();
    assertEquals( "Checking number of terms", 5, terms.size() );
    GeneralTermDAO.filterWholeWordOnly( extendedSearch, terms );
    assertEquals( "Checking number of terms after filtering", 3, terms.size() );
    assertEquals( "Checking filtered term", "Ölkanne", terms.get( 0 ).getTerm() );
    assertEquals( "Checking filtered term", "Metallene Ölkanne", terms.get( 1 ).getTerm() );
    assertEquals( "Checking filtered term", "Ölkanne und -halter", terms.get( 2 ).getTerm() );

    // It would be this search term, but since exact search is turned on, it's not found
    extendedSearch.searchText = "olkanne";
    terms = createTerms();
    assertEquals( "Checking number of terms", 5, terms.size() );
    GeneralTermDAO.filterWholeWordOnly( extendedSearch, terms );
    assertEquals( "Checking number of terms after filtering", 0, terms.size() );

    // Now use the non-exact search, which converts the term to a search term and also
    // filters in search term.
    extendedSearch.searchText = "olkanne";
    extendedSearch.exactSearch = false;
    terms = createTerms();
    assertEquals( "Checking number of terms", 5, terms.size() );
    GeneralTermDAO.filterWholeWordOnly( extendedSearch, terms );
    assertEquals( "Checking number of terms after filtering", 3, terms.size() );
    assertEquals( "Checking filtered term", "Ölkanne", terms.get( 0 ).getTerm() );
    assertEquals( "Checking filtered term", "Metallene Ölkanne", terms.get( 1 ).getTerm() );
    assertEquals( "Checking filtered term", "Ölkanne und -halter", terms.get( 2 ).getTerm() );

    // Now split the words, search with OR.
    extendedSearch.searchText = "olkanne halter";
    extendedSearch.exactSearch = false;
    extendedSearch.splitWords = true;
    extendedSearch.compositionAnd = false;
    terms = createTerms();
    assertEquals( "Checking number of terms", 5, terms.size() );
    GeneralTermDAO.filterWholeWordOnly( extendedSearch, terms );
    assertEquals( "Checking number of terms after filtering", 3, terms.size() );
    assertEquals( "Checking filtered term", "Ölkanne", terms.get( 0 ).getTerm() );
    assertEquals( "Checking filtered term", "Metallene Ölkanne", terms.get( 1 ).getTerm() );
    assertEquals( "Checking filtered term", "Ölkanne und -halter", terms.get( 2 ).getTerm() );

    // Now split the words, search with AND.
    extendedSearch.searchText = "olkanne halter";
    extendedSearch.exactSearch = false;
    extendedSearch.splitWords = true;
    extendedSearch.compositionAnd = true;
    terms = createTerms();
    assertEquals( "Checking number of terms", 5, terms.size() );
    GeneralTermDAO.filterWholeWordOnly( extendedSearch, terms );
    assertEquals( "Checking number of terms after filtering", 1, terms.size() );
    assertEquals( "Checking filtered term", "Ölkanne und -halter", terms.get( 0 ).getTerm() );
  }

  /**
   * Creates some terms for {@link #testFilterWholeWordOnly()}.
   * @return A couple of terms, which all contain <code>Bett</code> somewhere
   */
  private List<AbstractTermVO> createTerms()
  {
    IpiTermVO term1 = new IpiTermVO();
    term1.setIdentifier( 1 );
    term1.setTerm( "Ölkanne" );
    term1.setSearchTerm( SearchText.toSearchText( term1.getTerm(), Locale.GERMAN ) );

    NiceTermVO term2 = new NiceTermVO();
    term2.setIdentifier( 2 );
    term2.setTerm( "Metallene Ölkanne" );
    term2.setSearchTerm( SearchText.toSearchText( term2.getTerm(), Locale.GERMAN ) );

    NegativeIpiTermVO term3 = new NegativeIpiTermVO();
    term3.setIdentifier( 3 );
    term3.setTerm( "Ölkannenhalter" );
    term3.setSearchTerm( SearchText.toSearchText( term3.getTerm(), Locale.GERMAN ) );

    IpiTermVO term4 = new IpiTermVO();
    term4.setIdentifier( 4 );
    term4.setTerm( "Ölkanne und -halter" );
    term4.setSearchTerm( SearchText.toSearchText( term4.getTerm(), Locale.GERMAN ) );

    IpiTermVO term5 = new IpiTermVO();
    term5.setIdentifier( 5 );
    term5.setTerm( "Orangensaft" );
    term5.setSearchTerm( SearchText.toSearchText( term5.getTerm(), Locale.GERMAN ) );

    List<AbstractTermVO> terms = new ArrayList<AbstractTermVO>();
    terms.add( term1 );
    terms.add( term2 );
    terms.add( term3 );
    terms.add( term4 );
    terms.add( term5 );

    return terms;
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.GeneralTermDAO} caching.
   * @throws Exception if the test fails
   */
  public void testCaching() throws Exception
  {
    defaultTestCache( getExpectedVO().getIdentifier() );
    defaultTestCacheSize( getExpectedVO().getIdentifier(), NUMBER_OF_TERMS );

    long[] runTimes = defaultTestCachePerformanceForRead( getExpectedVO(), getExpectedVO().getIdentifier() );
    LOG.info( "Reading one term without cache took " + runTimes[0] + " milliseconds." );
    LOG.info( "Reading one term with cache took " + runTimes[1] + " milliseconds." );

    runTimes = defaultTestCachePerformanceForReadCollection( getExpectedVO(), NUMBER_OF_GERMAN_TERMS, Locale.GERMAN, null );
    LOG.info( "Reading all terms without cache took " + runTimes[0] + " milliseconds." );
    LOG.info( "Reading all terms with cache took " + runTimes[1] + " milliseconds." );
  }

  /**
   * Checks the search results.
   * @param searchResults The search results
   * @param terms The expected terms
   * @param searchTerms The expected search terms
   * @param classNumbers The expected class numbers
   */
  private void checkSearchResult( List<AbstractTermVO> searchResults, String[] terms, String[] searchTerms, Integer[] classNumbers )
  {
    assertEquals( "Checking result size", terms.length, searchResults.size() );
    for ( int i = 0; i < terms.length; i++ )
    {
      AbstractTermVO searchResult = searchResults.get( i );
      assertEquals( "Checking term", terms[i], searchResult.getTerm() );
      assertEquals( "Checking search term", searchTerms[i], searchResult.getSearchTerm() );
      if ( classNumbers[i] == null )
      {
        assertNull( "Checking class number", searchResult.getClassNumber() );
      }
      else
      {
        assertEquals( "Checking class number", classNumbers[i], searchResult.getClassNumber() );
      }
    }
  }

  /**
   * Checks that the terms contain no IPI term candidate.
   * @param terms The terms to check
   */
  private void checkForNoIpiTermCandidate( List<AbstractTermVO> terms )
  {
    for ( AbstractTermVO term: terms )
    {
      assertTrue( "Checking that it's not an IPI term candidate", !( term instanceof IpiTermCandidateVO ) );
    }
  }

  /**
   * Checks that the terms contain no IPI term candidate.
   * @param classNumberTerms The terms to check, ordered by class number
   */
  private void checkForNoIpiTermCandidate( Map<Integer, List<AbstractTermVO>> classNumberTerms )
  {
    for ( Integer classNumber: classNumberTerms.keySet() )
    {
      checkForNoIpiTermCandidate( classNumberTerms.get( classNumber ) );
    }
  }
}
