package ch.ige.wdltool.dao;

import java.sql.SQLException;
import java.text.ParseException;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.RandomStringUtils;

import ch.ige.dbutils.dao.DAOPool;
import ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate;
import ch.ige.util.date.DateConverter;
import ch.ige.wdltool.ParameterConstants;
import ch.ige.wdltool.vo.IpiTermGroupVO;
import ch.ige.wdltool.vo.IpiTermVO;
import ch.ige.wdltool.vo.TermReplicationVO;

/**
 * Tests {@link ch.ige.wdltool.dao.IpiTermDAO}. <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual
 * Property</p>
 * @author Eric Lewis
 * @version $Id: IpiTermDAOTest.java,v 1.27 2009/04/17 09:37:43 bandschapp Exp $
 */
public class IpiTermDAOTest extends BaseDAOWithCRUDTestTemplate<IpiTermDAO, IpiTermVO, Integer>
{
  /**
   * The only constructor
   * @param name The test's name
   */
  public IpiTermDAOTest( String name )
  {
    super( name );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#getDAOClass()
  @Override
  protected Class<IpiTermDAO> getDAOClass()
  {
    return IpiTermDAO.class;
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.IpiTermDAO#find(java.lang.Integer)}.
   * @throws Exception if the test fails
   */
  public void testFind() throws Exception
  {
    IpiTermVO ipiTermVO = createAndInsert( true );
    IpiTermVO expectedVO = getSerializableVO();
    expectedVO.setTermGroupId( null );
    expectedVO.setIdentifier( ipiTermVO.getIdentifier() );
    defaultTestRead( expectedVO, -1, 999999999, ipiTermVO.getIdentifier() );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.IpiTermDAO#findAll()}.
   * @throws Exception if the test fails
   */
  public void testFindAll() throws Exception
  {
    List<IpiTermVO> terms = getDAO().findAll();
    assertNotNull( "Checking List", terms );
    assertTrue( "Checking for content", !terms.isEmpty() );
    assertTrue( "Checking number of terms", terms.size() > 300 ); // This may change!
    for ( int i = 0; i < terms.size(); i++ )
    {
      // This may look stupid, but it didn't work...
      assertEquals( "Checking class", IpiTermVO.class, terms.get( i ).getClass() );
    }
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.IpiTermDAO#findTermWithinInactiveTerms(Locale, Integer, String)}. Only base
   * functionality is tested (no exception by query)
   * @throws Exception if the test fails
   */
  public void testFindTermWithinInactiveTerms() throws Exception
  {
    try
    {
      getDAO().findTermWithinInactiveTerms( null, 1, "bla" );
      fail( "Null language must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findTermWithinInactiveTerms( Locale.GERMAN, 1, null );
      fail( "Null term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findTermWithinInactiveTerms( Locale.GERMAN, 1, "" );
      fail( "Empty term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    try
    {
      getDAO().findTermWithinInactiveTerms( Locale.GERMAN, 1,
          RandomStringUtils.randomAlphabetic( ParameterConstants.EXCEEDED_LIMIT_FIELD_TERM ) );
      fail( "Too long term must fail" );
    }
    catch ( IllegalArgumentException e )
    {
      // This is expected
    }

    String term = "Some text that does surely not exist as term";
    List<IpiTermVO> terms = getDAO().findTermWithinInactiveTerms( Locale.GERMAN, 1, term );
    assertNotNull( "Checking that find returned a a valid value", terms );
    assertEquals( "Checking number of terms", 0, terms.size() );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.IpiTermDAO#insert(ch.ige.wdltool.vo.IpiTermVO)}.
   * @throws Exception if the test fails
   */
  public void testInsert() throws Exception
  {
    defaultTestCreate( true );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.IpiTermDAO#update(ch.ige.wdltool.vo.IpiTermVO)}.
   * @throws Exception if the test fails
   */
  public void testUpdate() throws Exception
  {
    defaultTestUpdate();
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.IpiTermDAO#delete(java.lang.Integer)}.
   * @throws Exception if the test fails
   */
  public void testDelete() throws Exception
  {
    defaultTestDelete( -1, 999999999 );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.IpiTermDAO#findTermWithinInactiveTerms( Locale language, Integer classNumber, String term )}
   * for finding all terms given the language, the classnumber and the term.
   * @throws Exception if the test fails
   */
  public void testFindTermWithinInactiveTermsForAnalyzer() throws Exception
  {
    IpiTermVO inactiveTerm = new IpiTermVO();
    inactiveTerm.setActive( false );
    inactiveTerm.setClassNumber( 9 );
    inactiveTerm.setLanguage( Locale.GERMAN );
    inactiveTerm.setTerm( "Inaktivitäts_test" );
    inactiveTerm.setCreationUser( "JUnit test" );
    inactiveTerm.setChangeUser( "JUnit test" );

    insert( inactiveTerm );
    setSavedId( inactiveTerm.getIdentifier() ); // Will be deleted during tearDown()

    List<IpiTermVO> terms = getDAO()
        .findTermWithinInactiveTerms( inactiveTerm.getLanguage(), 2, inactiveTerm.getTerm() );
    assertEquals( "Should not find inactive term with class 2", 0, terms.size() );

    terms = getDAO().findTermWithinInactiveTerms( Locale.FRENCH, inactiveTerm.getClassNumber(), inactiveTerm.getTerm() );
    assertEquals( "Should not find inactive term in French", 0, terms.size() );

    // Now it should find it
    terms = getDAO().findTermWithinInactiveTerms( inactiveTerm.getLanguage(), inactiveTerm.getClassNumber(),
        inactiveTerm.getTerm() );
    assertEquals( "Checking number of found terms ", 1, terms.size() );
    for ( IpiTermVO term: terms )
    {
      assertEquals( "Checking classnumber", inactiveTerm.getClassNumber(), term.getClassNumber() );
      assertEquals( "Checking language", inactiveTerm.getLanguage(), term.getLanguage() );
      assertEquals( "Checking term", inactiveTerm.getTerm(), term.getTerm() );
    }
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#read(java.io.Serializable)
  @Override
  protected IpiTermVO read( Integer ipiTermId ) throws SQLException
  {
    return getDAO().find( ipiTermId );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#compareVOs(ch.ige.dbutils.vo.AbstractVO, ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected void compareVOs( IpiTermVO originalVO, IpiTermVO loadedVO )
  {
    assertEquals( "Checking ID", originalVO.getIdentifier(), loadedVO.getIdentifier() );
    assertEquals( "Checking class number", originalVO.getClassNumber(), loadedVO.getClassNumber() );
    assertEquals( "Checking language", originalVO.getLanguage(), loadedVO.getLanguage() );
    assertEquals( "Checking term", originalVO.getTerm(), loadedVO.getTerm() );
    assertEquals( "Checking private comment", originalVO.getPrivateComment(), loadedVO.getPrivateComment() );
    assertEquals( "Checking public comment", originalVO.getPublicComment(), loadedVO.getPublicComment() );
    assertEquals( "Checking change user", originalVO.getChangeUser(), loadedVO.getChangeUser() );
    assertEquals( "Checking creation user", originalVO.getCreationUser(), loadedVO.getCreationUser() );
    assertEquals( "Checking data source", originalVO.getDataSource(), loadedVO.getDataSource() );
    assertEquals( "Checking latest occurrence", originalVO.getLatestOccurrence(), loadedVO.getLatestOccurrence() );
    assertEquals( "Checking search term", originalVO.getSearchTerm(), loadedVO.getSearchTerm() );
    assertEquals( "Checking term group ID", originalVO.getTermGroupId(), loadedVO.getTermGroupId() );
    assertEquals( "Checking usage count", originalVO.getUsageCount(), loadedVO.getUsageCount() );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#validateVO(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected void validateVO( IpiTermVO loadedVO )
  {
    assertTrue( "Checking change date", loadedVO.getChangeDate().after( GeneralTermDAOTest.getLatestMigration() ) );
    //this accepts only renewal dates in the past
    assertTrue( "Checking renewal: upper bound", new Date( System.currentTimeMillis() + 1 ).after( loadedVO
        .getRenewal() ) );

    //this accepts only renewal dates after a specific inital date
    assertTrue( "Checking renewal: lower bound", loadedVO.getRenewal().after( GeneralTermDAOTest.getLatestMigration() ) );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#assertVOEqual(ch.ige.dbutils.vo.AbstractVO, ch.ige.dbutils.vo.AbstractVO)
  @Override
  public boolean assertVOEqual( IpiTermVO originalVO, IpiTermVO serializedVO )
  {
    return super.assertVOEqual( originalVO, serializedVO )
        && originalVO.getChangeDate().equals( serializedVO.getChangeDate() )
        && originalVO.getRenewal().equals( serializedVO.getRenewal() );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#compareVOsAndValidateLoaded(ch.ige.dbutils.vo.AbstractVO, ch.ige.dbutils.vo.AbstractVO)
  @Override
  public void compareVOsAndValidateLoaded( IpiTermVO originalVO, IpiTermVO loadedVO )
  {
    super.compareVOsAndValidateLoaded( originalVO, loadedVO );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#modifyVOBeforeInsert(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected void modifyVOBeforeInsert( IpiTermVO ipiTermVO )
  {
    // The term group must exist otherwise
    ipiTermVO.setTermGroupId( null );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#insert(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected Integer insert( IpiTermVO ipiTermVO )
  {
    return getDAO().insert( ipiTermVO );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#modifyVOBeforeUpdate(ch.ige.dbutils.vo.AbstractVO)
  @Override
  public void modifyVOBeforeUpdate( IpiTermVO ipiTermVO )
  {
    ipiTermVO.setClassNumber( 33 );
    ipiTermVO.setLanguage( Locale.ITALIAN );
    ipiTermVO.setTerm( "Use me" );
    ipiTermVO.setPrivateComment( "The private comment, modified" );
    ipiTermVO.setPublicComment( "The public comment, modified" );
    // Note that this depends on existing term groups
    try
    {
      List<IpiTermGroupVO> ipiTermGroups = DAOPool.get( IpiTermGroupDAO.class ).findAll();
      if ( !ipiTermGroups.isEmpty() )
      {
        ipiTermVO.setTermGroupId( ipiTermGroups.get( 0 ).getIdentifier() );
      }
    }
    catch ( SQLException e )
    {
      // Too bad...
    }
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#update(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected void update( IpiTermVO ipiTermVO )
  {
    getDAO().update( ipiTermVO );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#delete(java.io.Serializable)
  @Override
  protected void delete( Integer ipiTermId )
  {
    getDAO().delete( ipiTermId );
  }

  /**
   * Tests {@link ch.ige.wdltool.dao.IpiTermDAO#getReplicationTerms(String)}.
   * @throws Exception if the test fails
   */
  public void testGetReplicationTerms() throws Exception
  {
    TermReplicationVO ipiTermsToReplicate = getDAO().getReplicationTerms( "wdldmz" );
    assertNotNull( "Checking terms to replicate", ipiTermsToReplicate );
    assertNotNull( "Checking SQL statement", ipiTermsToReplicate.getInsertStatement() );
    assertNotNull( "Checking terms", ipiTermsToReplicate.getTermParameters() );
    assertTrue( "Checking number of terms", ipiTermsToReplicate.getTermParameters().size() > 0 );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#getSerializableVO()
  @Override
  public IpiTermVO getSerializableVO() throws SQLException
  {
    try
    {
      IpiTermVO ipiTerm = new IpiTermVO();
      ipiTerm.setIdentifier( 33 );
      ipiTerm.setClassNumber( 42 );
      ipiTerm.setLanguage( Locale.FRENCH );
      ipiTerm.setTerm( "Dont % use me" );
      ipiTerm.setPrivateComment( "The private comment" );
      ipiTerm.setPublicComment( "The public comment" );
      ipiTerm.setActive( true );
      ipiTerm.setChangeDate( DateConverter.stringToTimestamp( "03.02.2002 18:44:55.000" ) );
      ipiTerm.setChangeUser( "billg" );
      ipiTerm.setCreationUser( "steveb" );
      ipiTerm.setLatestOccurrence( DateConverter.stringToTimestamp( "03.05.2005 13:22:11.000" ) );
      ipiTerm.setPublic( false );
      ipiTerm.setRenewal( DateConverter.stringToTimestamp( "05.03.2003 13:11:22.000" ) );
      ipiTerm.setSearchTerm( "dont % use me" );
      ipiTerm.setTermGroupId( 15 );
      ipiTerm.setUsageCount( 4 );
      return ipiTerm;
    }
    catch ( ParseException e )
    {
      throw new SQLException( e.getMessage() );
    }
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#getExpectedVO()
  @Override
  public IpiTermVO getExpectedVO() throws ParseException
  {
    IpiTermVO ipiTerm = new IpiTermVO();
    ipiTerm.setIdentifier( 29988 );
    ipiTerm.setClassNumber( 5 );
    ipiTerm.setLanguage( Locale.GERMAN );
    ipiTerm.setTerm( "parapharmazeutische Produkte" );
    ipiTerm.setPrivateComment( "Heilmittel fallen generell in die Klasse 5" );
    ipiTerm.setPublicComment( null );
    ipiTerm.setActive( true );
    ipiTerm.setChangeDate( DateConverter.stringToTimestamp( "07.11.2007 14:34:38.000" ) );
    ipiTerm.setChangeUser( "benninger" );
    ipiTerm.setCreationUser( "jaeggi" );
    ipiTerm.setLatestOccurrence( null );
    ipiTerm.setPublic( false );
    ipiTerm.setRenewal( DateConverter.stringToTimestamp( "11.12.2007 00:00:00.000" ) );
    ipiTerm.setSearchTerm( "parafarmazautische produkte" );
    ipiTerm.setTermGroupId( null );
    ipiTerm.setUsageCount( null );
    return ipiTerm;
  }
}
