package ch.ige.wdltool.dao;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.ArrayUtils;
import org.apache.log4j.Logger;

import ch.ige.dbutils.dao.DAOPool;
import ch.ige.dbutils.dao.UnitOfWorkException;
import ch.ige.dbutils.sql.SQLStatement;
import ch.ige.dbutils.sql.SQLStatementCreator;
import ch.ige.dbutils.sql.SQLWhereStatementCreator;
import ch.ige.wdltool.enums.TermType;
import ch.ige.wdltool.vo.IpiTermCandidateVO;
import ch.ige.wdltool.vo.NiceTermVO;
import ch.ige.wdltool.vo.TermIdCleanupCandidateStatusVO;
import ch.ige.wdltool.vo.TermReplicationVO;

/**
 * Accesses the table <code>TERM</code> for Nice terms.
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   Eric Lewis
 * @version  $Id: NiceTermDAO.java,v 1.13 2008/12/09 16:09:26 bandschapp Exp $
 */
public class NiceTermDAO extends BaseTermDAO<NiceTermVO, Integer>
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( NiceTermDAO.class );

  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getVOType()
  @Override
  protected Class<NiceTermVO> getVOType()
  {
    return NiceTermVO.class;
  }

  // @see ch.ige.wdltool.dao.BaseTermDAO#getTermType()
  @Override
  protected TermType getTermType()
  {
    return TermType.NICE_TERM;
  }

  // @see ch.ige.wdltool.dao.BaseTermDAO#getTermTypeName()
  @Override
  protected String getTermTypeName()
  {
    return "Nice term";
  }

  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getValues(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected Object[] getValues( NiceTermVO niceTerm )
  {
    return new Object[] { niceTerm.getDataSource().toDB(), niceTerm.getClassNumber(), niceTerm.getTermGroupId(), TermType.NICE_TERM.toDB(),
        niceTerm.getLanguage().getLanguage(), niceTerm.getTerm(), niceTerm.getSearchTerm(), niceTerm.isGeneric(), niceTerm.isPublic(),
        niceTerm.getPrivateComment(), niceTerm.getPublicComment(), true, null, null, null, niceTerm.getCreationUser(), niceTerm.getChangeUser(),
        niceTerm.getChangeDate(), niceTerm.isCheckerOnly() };
  }

  /**
   * Finds a Nice term.
   * @param niceTermId The unique identifier for a Nice term
   * @return The Nice term or <code>null</code> if it doesn't exist
   * @throws SQLException if the query fails
   */
  public NiceTermVO find( final Integer niceTermId ) throws SQLException
  {
    return (NiceTermVO) super.find( niceTermId, getVOType() );
  }

  /**
   * Finds all Nice terms.
   * Note: This should not be made public!
   */
  protected List<NiceTermVO> findAll() throws SQLException
  {
    return super.findAll();
  }

  /**
   * Gets the generic Nice terms.
   * @return The generic Nice terms.
   */
  public List<NiceTermVO> findGenericTerms()
  {
    SQLWhereStatementCreator whereStatement = super.getFindAllWhereStatement();
    whereStatement.appendFragment( " AND is_generic_bool = '1'" ); //1 as in true: it is a generic Nice term
    try
    {
      return findAllRestricted( whereStatement );
    }
    catch ( SQLException e )
    {
      LOG.error( "Error when looking for Generic Nice Term" + e.getMessage(), e );
      return null;
    }
  }

  /**
   * Inserts a Nice term and simultaneously removes a corresponding term candidate if it exists.
   * @param term The term to insert
   * @return The new identifier
   * @see ch.ige.wdltool.dao.BaseTermDAO#insert(ch.ige.dbutils.vo.AbstractVO)
   */
  @Override
  public Integer insert( NiceTermVO term )
  {
    try
    {
      return ( insertAndRemoveIpiTermCandidate( term ) ).getNewTermId();
    }
    catch ( SQLException e )
    {
      LOG.error( "Error: could not insert NiceTerm" + e.getMessage(), e );
      return null;
    }
  }

  /**
   * Inserts a Nice term and simultaneously removes a corresponding term candidate if it exists  
   * @param term The term to insert
   * @return The new identifier an the flag if a term candidate had been removed
   * @throws SQLException if the query fails
   */
  public TermIdCleanupCandidateStatusVO insertAndRemoveIpiTermCandidate( NiceTermVO term ) throws SQLException
  {
    Integer newId = super.insert( term );

    IpiTermCandidateDAO ipiTermCandidateDAO = DAOPool.get( IpiTermCandidateDAO.class );
    List<IpiTermCandidateVO> candidates = ipiTermCandidateDAO.find( term.getLanguage(), term.getClassNumber(), term.getTerm() );
    if ( candidates.size() > 1 )
    {
      LOG.error( "multiple equal candidates have been found and all are going to be deleted: " + term.getTerm() + " in " + term.getLanguage() );
    }
    if ( candidates.size() >= 1 )
    {
      for ( IpiTermCandidateVO candidateVO: candidates )
      {
        ipiTermCandidateDAO.delete( candidateVO.getIdentifier() );
      }
      return new TermIdCleanupCandidateStatusVO( newId, true );
    }
    else
    {
      return new TermIdCleanupCandidateStatusVO( newId, false );
    }
  }

  /**
   * Updates a Nice term and simultaneously removes a corresponding term candidate if it exists.
   * @param term The term to update
   * @see ch.ige.wdltool.dao.BaseTermDAO#update(ch.ige.dbutils.vo.AbstractVO)
   */
  @Override
  public void update( NiceTermVO term )
  {
    try
    {
      updateAndRemoveIpiTermCandidate( term );
    }
    catch ( SQLException e )
    {
      LOG.error( "Error: could not update IpiTerm" + e.getMessage(), e );
      throw new UnitOfWorkException( e );
    }
  }

  /**
   * Updates an IPI term and simultaneously removes a corresponding term candidate if it exists  
   * @param term The term to update
   * @return The flag if a term candidate had been removed
   * @throws SQLException if the query fails
   */
  public boolean updateAndRemoveIpiTermCandidate( NiceTermVO term ) throws SQLException
  {
    super.update( term );
    IpiTermCandidateDAO ipiTermCandidateDAO = DAOPool.get( IpiTermCandidateDAO.class );
    List<IpiTermCandidateVO> candidates = ipiTermCandidateDAO.find( term.getLanguage(), term.getClassNumber(), term.getTerm() );
    if ( candidates.size() > 1 )
    {
      LOG.error( "multiple equal candidates have been found and all are going to be deleted: " + term.getTerm() + " in " + term.getLanguage() );
    }
    if ( candidates.size() >= 1 )
    {
      for ( IpiTermCandidateVO candidateVO: candidates )
      {
        ipiTermCandidateDAO.delete( candidateVO.getIdentifier() );
      }
      return true;
    }
    else
    {
      return false;
    }
  }

  @Override
  public TermReplicationVO getReplicationTerms( String schemaName ) throws SQLException
  {
    SQLStatementCreator<NiceTermVO> statementCreator = new SQLStatementCreator<NiceTermVO>( schemaName + "." + getTableName(), getIdColumnName(),
        getColumnNames(), getIdSequenceName() );
    SQLStatement<Integer> insertStatement = statementCreator.insert( false );

    // Replicates all Nice terms
    List<NiceTermVO> terms = findAll();

    List<Object[]> termValues = new ArrayList<Object[]>( terms.size() );
    for ( NiceTermVO niceTerm: terms )
    {
      Object[] insertValues = getValues( niceTerm );
      termValues.add( ArrayUtils.addAll( new Object[] { niceTerm.getIdentifier() }, insertValues ) );
    }

    return new TermReplicationVO( insertStatement, termValues );
  }
}
