package ch.ige.wdltool.dao;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import ch.ige.util.validation.Check;
import ch.ige.wdltool.vo.NiceTermGroupVO;
import ch.ige.wdltool.vo.TermGroupVO;

/**
 * Accesses the table <code>TERM_GROUP</code> and deals with the Nice term group data.
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   bandschapp
 * @version  $Id: NiceTermGroupDAO.java,v 1.13 2008/03/31 09:58:12 lewis_lee Exp $
 */
public class NiceTermGroupDAO extends BaseTermGroupDAO<NiceTermGroupVO>
{
  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getValues(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected Object[] getValues( NiceTermGroupVO vo )
  {
    return new Object[] { null, vo.getNiceTermNumber(), null };
  }

  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getVOType()
  @Override
  protected Class<NiceTermGroupVO> getVOType()
  {
    return NiceTermGroupVO.class;
  }

  /**
   * Inserts a Nice term group.
   * @param niceTermGroup The Nice term group to insert
   * @return The ID of the inserted Nice term group
   */
  public Integer insert( final NiceTermGroupVO niceTermGroup )
  {
    Integer id = super.insert( niceTermGroup );
    clearCache();
    return id;
  }

  /**
   * Finds one Nice term group.
   * @param niceTermGroupId The ID of this Nice term group
   * @return The Nice term group or <code>null</code> if none is found
   * @throws SQLException in case the database access is not successful
   */
  public NiceTermGroupVO find( final Integer niceTermGroupId ) throws SQLException
  {
    Check.notNegative( niceTermGroupId, "Nice term group ID" );
    return (NiceTermGroupVO) getTermGroups( NiceTermGroupVO.class ).get( niceTermGroupId );
  }

  /**
   * Finds one Nice term group.
   * @param niceTermNumber The number of this term group
   * @return The Nice term group or <code>null</code> if none is found
   * @throws SQLException in case the database access is not successful
   */
  public NiceTermGroupVO findByNumber( final Integer niceTermNumber ) throws SQLException
  {
    Check.notNegative( niceTermNumber, "Nice term number" );
    Collection<TermGroupVO> termGroups = getTermGroups( NiceTermGroupVO.class ).values();
    NiceTermGroupVO foundNiceTermGroup = null;
    for ( TermGroupVO termGroupVO: termGroups )
    {
      NiceTermGroupVO niceTermGroup = (NiceTermGroupVO) termGroupVO;
      if ( niceTermGroup.getNiceTermNumber().equals( niceTermNumber ) )
      {
        foundNiceTermGroup = niceTermGroup;
        break;
      }
    }
    return foundNiceTermGroup;
  }

  /**
   * Finds all the Nice term groups.
   * @return The Nice term groups ordered by ID
   * @throws SQLException in case the database access is not successful
   */
  public List<NiceTermGroupVO> findAll() throws SQLException
  {
    Collection<TermGroupVO> rawTermGroups = getTermGroups( NiceTermGroupVO.class ).values();
    List<NiceTermGroupVO> termGroups = new ArrayList<NiceTermGroupVO>( rawTermGroups.size() );
    for ( TermGroupVO rawTermGroup: rawTermGroups )
    {
      termGroups.add( (NiceTermGroupVO) rawTermGroup );
    }
    return termGroups;
  }

  /**
   * Updates a Nice term group.
   * @param termGroup The Nice term group to be updated
   */
  public void update( final NiceTermGroupVO termGroup )
  {
    super.update( termGroup );
    clearCache();
  }

  /**
   * Deletes a Nice term group.
   * @param niceTermGroupId The ID of the Nice term group to be deleted
   */
  public void delete( final Integer niceTermGroupId )
  {
    super.delete( niceTermGroupId );
    clearCache();
  }
}
