package ch.ige.wdltool.dao;

import java.sql.SQLException;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.dbutils.dao.AbstractWdlDAO;
import ch.ige.dbutils.dao.CachingDAO;
import ch.ige.dbutils.sql.SQLStatement;
import ch.ige.dbutils.sql.SQLStatementCreator;
import ch.ige.util.collection.CollectionSize;
import ch.ige.util.validation.Check;
import ch.ige.wdltool.vo.StopwordVO;
import ch.ige.wdltool.vo.TermReplicationVO;

/**
 * Accesses the table <code>STOPWORD</code>.
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   Eric Lewis
 * @version  $Id: StopwordDAO.java,v 1.24 2009/03/05 08:12:19 adanlessossi Exp $
 */
public class StopwordDAO extends AbstractWdlDAO<StopwordVO, Integer> implements CachingDAO
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( StopwordDAO.class );

  /**
   * All the stopwords are cached in here
   */
  private static Map<Locale, List<StopwordVO>> stopwords;

  /**
   * The only constructor.
   */
  public StopwordDAO()
  {
    super();
  }

  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getVOType()
  @Override
  protected Class<StopwordVO> getVOType()
  {
    return StopwordVO.class;
  }

  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getTableName()
  @Override
  protected String getTableName()
  {
    return "stopword";
  }

  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getColumnNames()
  @Override
  protected String getColumnNames()
  {
    return "language_sn, word";
  }

  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getAliases()
  @Override
  protected Map<String, String> getAliases()
  {
    return SQLStatementCreator.createAliases( "language_sn", "language" );
  }

  // @see ch.ige.dbutils.dao.AbstractWdlDAO#getValues(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected Object[] getValues( StopwordVO stopword )
  {
    return new Object[] { stopword.getLanguage().getLanguage(), stopword.getWord() };
  }

  /**
   * Gets the stopwords, which are cached after the first loading.
   * 
   * @return The stopwords by language
   * @throws SQLException if the query fails
   */
  protected synchronized Map<Locale, List<StopwordVO>> getStopwords() throws SQLException
  {
    if ( stopwords == null )
    {
      List<StopwordVO> allStopwords = super.findAllOrderBy( "language_sn ASC, LENGTH(word) DESC" );
      // Make sure the stopwords are stored in the right language order
      stopwords = new LinkedHashMap<Locale, List<StopwordVO>>( allStopwords.size() );

      for ( StopwordVO stopword: allStopwords )
      {
        if ( !stopwords.containsKey( stopword.getLanguage() ) )
        {
          stopwords.put( stopword.getLanguage(), new ArrayList<StopwordVO>() );
        }
        stopwords.get( stopword.getLanguage() ).add( stopword );
      }
    }
    else
    {
      LOG.debug( "Cache hit" );
    }
    return stopwords;
  }

  // @see ch.ige.dbutils.dao.CachingDAO#clearCache()
  public synchronized void clearCache()
  {
    stopwords = null;
    LOG.debug( "Cache cleared" );
  }

  // @see ch.ige.dbutils.dao.CachingDAO#describeCachedContent()
  public CharSequence describeCachedContent()
  {
    return stopwords == null ? "" : stopwords.toString();
  }

  // @see ch.ige.dbutils.dao.CachingDAO#getCacheSize()
  public int getCacheSize()
  {
    return stopwords == null ? 0 : CollectionSize.size( stopwords );
  }

  /**
   * Finds all stopwords.
   * @return The stopwords, ordered by language and by length
   * @throws SQLException if the query fails
   */
  public List<StopwordVO> findAll() throws SQLException
  {
    Map<Locale, List<StopwordVO>> cachedStopwords = getStopwords();
    List<StopwordVO> allStopwords = new ArrayList<StopwordVO>( cachedStopwords.size() );
    for ( Locale language: cachedStopwords.keySet() )
    {
      allStopwords.addAll( cachedStopwords.get( language ) );
    }
    return allStopwords;
  }

  /**
   * Finds all stopwords for a certain language.
   * @param language The language to find
   * @return The stopwords for this language
   * @throws SQLException if the query fails
   */
  public List<StopwordVO> findAll( Locale language ) throws SQLException
  {
    Map<Locale, List<StopwordVO>> cachedStopwords = getStopwords();
    if ( cachedStopwords.containsKey( language ) )
    {
      return cachedStopwords.get( language );
    }
    else
    {
      return Collections.emptyList();
    }
  }

  /**
   * Finds a stopword.
   * @param stopwordId The unique identifier for a stopword
   * @return The stopword or <code>null</code> if it doesn't exist
   * @throws SQLException if the query fails
   */
  public StopwordVO find( final Integer stopwordId ) throws SQLException
  {
    Check.notNegative( stopwordId, "stopword ID" );

    StopwordVO stopword = null;
    Map<Locale, List<StopwordVO>> cachedStopwords = getStopwords();
    for ( Locale language: cachedStopwords.keySet() )
    {
      List<StopwordVO> languageStopwords = cachedStopwords.get( language );
      for ( StopwordVO languageStopword: languageStopwords )
      {
        if ( languageStopword.getIdentifier().equals( stopwordId ) )
        {
          stopword = languageStopword;
          break;
        }
      }
    }
    return stopword;
  }

  /**
   * Finds a stopword for a certain language.
   * Note that unlike {@link #findAnyWord(Locale, String)}, this
   * searches for an exact match.
   * @param language The language to find
   * @param stopword The exact stopword to find
   * @return The stopword or <code>null</code> if it doesn't exist
   * @throws SQLException if the query fails
   */
  public StopwordVO find( final Locale language, final String stopword ) throws SQLException
  {
    Check.notEmpty( stopword, "stopword" );

    StopwordVO stopwordVO = null;
    Map<Locale, List<StopwordVO>> cachedStopwords = getStopwords();
    List<StopwordVO> languageStopwords = cachedStopwords.get( language );
    if ( languageStopwords != null )
    {
      for ( StopwordVO languageStopword: languageStopwords )
      {
        if ( languageStopword.getWord().equals( stopword ) )
        {
          stopwordVO = languageStopword;
          break;
        }
      }
    }
    return stopwordVO;
  }

  /**
   * Finds all stopwords (in any language) that correspond to any word within <code>searchStopword</code>.
   * This splits up <code>searchStopword</code> into separate words and tries to find
   * any of those words in the stopwords.
   * @param searchStopword The stopword to search for, which can contain any number of blank-separated words
   * @return The stopwords that were found or an empty <code>List</code> if none were found
   * @throws SQLException if the query fails
   */
  public List<StopwordVO> findAnyWord( String searchStopword ) throws SQLException
  {
    Check.notEmpty( searchStopword, "search stopword" );
    return findAnyWordFromCache( null, searchStopword );
  }

  /**
   * Finds all stopwords of a language that correspond to any word within <code>searchStopword</code>.
   * This splits up <code>searchStopword</code> into separate words and tries to find
   * any of those words in the stopwords.
   * @param language The language in which terms are searched for
   * @param searchStopword The stopword to search for, which can contain any number of blank-separated words
   * @return The stopwords that were found or an empty <code>List</code> if none were found
   * @throws SQLException if the query fails
   */
  public List<StopwordVO> findAnyWord( Locale language, String searchStopword ) throws SQLException
  {
    Check.notNull( language, "language" );
    Check.notEmpty( searchStopword, "search stopword" );
    return findAnyWordFromCache( language, searchStopword );
  }

  /**
   * Simulates finding any word within stopwords by going through the cache and
   * checking every stopword.
   * @param language The language in which stopwords are searched for; if it's <code>null</code>,
   *        all languages are taken
   * @param searchStopword The stopword to search for, which can contain any number of blank-separated words
   * @return The stopwords containing the search stopwords 
   * @throws SQLException if the query fails
   */
  private List<StopwordVO> findAnyWordFromCache( final Locale language, final String searchStopword ) throws SQLException
  {
    Check.notEmpty( searchStopword, "search stopword" );
    String[] searchWords = StringUtils.split( searchStopword ); // Use a light-weight split for whitespace
    if ( searchWords.length == 0 )
    {
      throw new IllegalArgumentException( "Found no search stopword words in " + searchStopword );
    }

    Map<Locale, List<StopwordVO>> cachedStopwords = getStopwords();

    // Since this is cached, simulate a search with 'LIKE'.
    // Note that a Set is used, because the same stopword can match several search words
    Set<StopwordVO> matchingStopwords = new HashSet<StopwordVO>();
    List<StopwordVO> allStopwords = new ArrayList<StopwordVO>();
    if ( language != null )
    {
      allStopwords = cachedStopwords.get( language );
    }
    else
    {
      for ( Locale stopwordLanguage: cachedStopwords.keySet() )
      {
        allStopwords.addAll( cachedStopwords.get( stopwordLanguage ) );
      }
    }

    if ( allStopwords != null )
    {
      for ( StopwordVO stopword: allStopwords )
      {
        for ( String searchWord: searchWords )
        {
          if ( stopword.getWord().contains( searchWord ) )
          {
            matchingStopwords.add( stopword );
          }
        }
      }
    }

    List<StopwordVO> resultingStopwords = new ArrayList<StopwordVO>( matchingStopwords );
    Collections.sort( resultingStopwords, new StopwordComparator( language ) );
    return resultingStopwords;
  }

  /**
   * Inserts a stopword and sets its ID.
   * <strong>Note that this clears the cache!</strong>
   * @param stopword The stopword to insert
   * @return The ID of the inserted stopword
   */
  public Integer insert( final StopwordVO stopword )
  {
    Integer id = super.insert( stopword );
    clearCache();
    return id;
  }

  /**
   * Updates a stopword.
   * <strong>Note that this clears the cache!</strong>
   * @param stopword The stopword to update
   */
  public void update( final StopwordVO stopword )
  {
    super.update( stopword );
    clearCache();
  }

  /**
   * Deletes a stopword.
   * <strong>Note that this clears the cache!</strong>
   * @param stopwordId The unique identifier for a stopword
   */
  public void delete( final Integer stopwordId )
  {
    super.delete( stopwordId );
    clearCache();
  }

  /**
   * Deletes a number of stopwords.
   * <strong>Note that this clears the cache!</strong>
   * @param stopwordIds The unique identifiers for the stopwords to delete
   */
  public void delete( final Collection<Integer> stopwordIds )
  {
    super.delete( stopwordIds );
    clearCache();
  }

  /**
   * Sorts stopwords language-specifically by their actual stopword.
   */
  private class StopwordComparator implements Comparator<StopwordVO>
  {
    /**
     * Used to sort by language
     */
    private Collator collator;

    /**
     * Creates this.
     * @param language The language to sort with, may be <code>null</code>;
     *        in this case {@link Locale#GERMAN} is taken as a default
     */
    public StopwordComparator( Locale language )
    {
      if ( language != null )
      {
        this.collator = Collator.getInstance( language );
      }
      else
      {
        this.collator = Collator.getInstance( Locale.GERMAN );
      }
    }

    // @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
    public int compare( StopwordVO stopword1, StopwordVO stopword2 )
    {
      return this.collator.compare( stopword1.getWord(), stopword2.getWord() );
    }
  }

  /**
   * Allows replicating stopwords.
   * 
   * @param schemaName
   * @return  The generic <code>TermReplicationVO</code>.
   * 
   * @throws SQLException
   */
  public TermReplicationVO getReplicationStopwords( String schemaName ) throws SQLException
  {
    SQLStatementCreator<StopwordVO> statementCreator = new SQLStatementCreator<StopwordVO>( schemaName + "." + getTableName(), getIdColumnName(),
        getColumnNames(), getIdSequenceName() );
    SQLStatement<Integer> insertStatement = statementCreator.insert( false );

    // Replicates all stopwords
    List<StopwordVO> stopwordList = findAll();

    List<Object[]> stopwordValues = new ArrayList<Object[]>( stopwordList.size() );
    for ( StopwordVO stopword: stopwordList )
    {
      Object[] insertValues = getValues( stopword );
      stopwordValues.add( ArrayUtils.addAll( new Object[] { stopword.getIdentifier() }, insertValues ) );
    }

    return new TermReplicationVO( insertStatement, stopwordValues );
  }

}
