package ch.ige.wdltool.ejb;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.commons.lang.RandomStringUtils;

import ch.ige.wdltool.ParameterConstants;
import ch.ige.wdltool.dao.IpiTermCandidateDAO;
import ch.ige.wdltool.dao.IpiTermCandidateDAOTest;
import ch.ige.wdltool.vo.AbstractTermVO;
import ch.ige.wdltool.vo.IpiTermCandidateVO;
import ch.ige.wdltool.vo.IpiTermVO;
import ch.ige.wdltool.vo.TermCleanupCandidateStatusVO;

/**
 * Tests {@link ch.ige.wdltool.ejb.WdlServices} for IPI term candidates. <p>Copyright (c) 2007, Swiss Federal Institute
 * of Intellectual Property</p>
 * @author Eric Lewis
 * @version $Id: WdlServicesForIpiTermCandidateTest.java,v 1.22 2009/02/02 13:45:10 bandschapp Exp $
 */
public class WdlServicesForIpiTermCandidateTest extends BaseServiceTestTemplate<IpiTermCandidateDAO, IpiTermCandidateVO, Integer, WdlServices>
{
  /**
   * The unit under test
   */
  private WdlServices wdlServices;

  /**
   * Used for utility stuff
   */
  private IpiTermCandidateDAOTest ipiTermCandidateDAOTest;

  /**
   * The only constructor.
   * @param name The test's name
   */
  public WdlServicesForIpiTermCandidateTest( String name )
  {
    super( name );
    this.wdlServices = WdlServicesFactory.getStandaloneWdlServices();
    this.ipiTermCandidateDAOTest = new IpiTermCandidateDAOTest( name );
  }

  /**
   * Tests {@link ch.ige.wdltool.ejb.WdlServices#activateIpiTermCandidate(IpiTermCandidateVO)}.
   * @throws Exception if the test fails
   */
  public void testActivateIpiTermCandidate() throws Exception
  {
    IpiTermCandidateVO newTerm = createAndInsert( true );
    setSavedId( null ); //we will take care of deletion

    Locale language = newTerm.getLanguage();
    String term = newTerm.getTerm();
    Integer classNumber = newTerm.getClassNumber();

    List<AbstractTermVO> activatedTerm = getServices().findTermWithinActiveTerms( language, classNumber, term );
    assertEquals( "corresponding term must not exist yet", 0, activatedTerm.size() );

    //make sure the candidate exists
    IpiTermCandidateVO candidateVO = getServices().findIpiTermCandidate( language, classNumber, term );
    assertNotNull( "candidate must exist", candidateVO );

    getServices().activateIpiTermCandidate( newTerm );

    //make sure the candidate does not exist any more
    candidateVO = getServices().findIpiTermCandidate( language, classNumber, term );
    assertNull( "candidate must have been activated", candidateVO );

    activatedTerm = getServices().findTermWithinActiveTerms( language, classNumber, term );
    assertNotNull( "corresponding term must exist", activatedTerm );
    assertEquals( "corresponding term must exist exactly once", 1, activatedTerm.size() );
    getServices().deleteIpiTerm( activatedTerm.get( 0 ).getIdentifier() );
  }

  /**
   * Tests that an existing candidate gets removed if a corresponding term is entered newly.
   * This is an integration test requiring correct IPI term and IPI term candidate behaviour for success.
   * @throws Exception if the test fails
   */
  public void testCandidateDeletionOnTermEntry() throws Exception
  {
    IpiTermCandidateVO newCandidateTerm = createAndInsert( true );
    setSavedId( null ); //deletion should happen automagically

    IpiTermVO newTerm = new IpiTermVO( newCandidateTerm );
    newTerm.setActive( true ); //active flag got copied from candidate - has to be true for the new term

    Locale language = newTerm.getLanguage();
    String term = newTerm.getTerm();
    Integer classNumber = newTerm.getClassNumber();

    assertNotNull( "candidate must exist", getServices().findIpiTermCandidate( language, classNumber, term ) );

    TermCleanupCandidateStatusVO<IpiTermVO> termCleanupCandidateStatusVO = getServices().insertIpiTermAndRemoveIpiTermCandidate( newTerm );
    assertEquals( "A candidate must have been removed", true, termCleanupCandidateStatusVO.isCandidateRemoved() );
    assertNull( "candidate must be gone", getServices().findIpiTermCandidate( language, classNumber, term ) );
    List<AbstractTermVO> activatedTerm = getServices().findTermWithinActiveTerms( language, classNumber, term );
    assertEquals( "corresponding term must exist exactly once", 1, activatedTerm.size() );

    getServices().deleteIpiTerm( termCleanupCandidateStatusVO.getNewTerm().getIdentifier() );
    activatedTerm = getServices().findTermWithinActiveTerms( language, classNumber, term );
    assertEquals( "corresponding term must be gone now", 0, activatedTerm.size() );
  }

  /**
   * Tests {@link ch.ige.wdltool.ejb.WdlServices#findIpiTermCandidate(Integer)}.
   * @throws Exception if the test fails
   */
  public void testFindIpiTermCandidate() throws Exception
  {
    IpiTermCandidateVO newTerm = createAndInsert( true );
    defaultTestRead( newTerm, MINUS_ONE, 999999999, newTerm.getIdentifier() );
  }

  /**
   * Tests {@link ch.ige.wdltool.ejb.WdlServices#insertIpiTermCandidate(IpiTermCandidateVO)}.
   * @throws Exception if the test fails
   */
  public void testInsertIpiTermCandidate() throws Exception
  {
    defaultTestCreate( true );
  }

  /**
   * Tests {@link ch.ige.wdltool.ejb.WdlServices#insertIpiTermCandidate(IpiTermCandidateVO)}
   * with existing stopword.
   * @throws Exception if the test fails
   */
  public void testInsertIpiTermCandidateWithExistingStopword() throws Exception
  {
    defaultTestCreateTermWithExistingStopword( Locale.FRENCH, "une" );
  }

  /**
   * Tests {@link ch.ige.wdltool.ejb.WdlServices#updateIpiTermCandidate(IpiTermCandidateVO)}.
   * @throws Exception if the test fails
   */
  public void testUpdateIpiTermCandidate() throws Exception
  {
    defaultTestUpdate();
  }

  /**
   * Tests {@link ch.ige.wdltool.ejb.WdlServices#updateIpiTermCandidate(IpiTermCandidateVO)}
   * with existing stopword.
   * @throws Exception if the test fails
   */
  public void testUpdateIpiTermCandidateWithExistingStopword() throws Exception
  {
    defaultTestUpdateTermWithExistingStopword( Locale.FRENCH, "pour" );
  }

  /**
   * Tests {@link ch.ige.wdltool.ejb.WdlServices#extractTermVariants(IpiTermCandidateVO)}.
   * @throws Exception if the test fails
   */
  public void testExtractTermVariants() throws Exception
  {
    final String termText = "Alle," /* Stopword */+ "  Kaiser," /* Not a term */+ " Skilifte," /* In class number 12 */+ "   MIT  ," /* Stopword */
        + " Hamburger," /* In class number 30 */
        + " café," /* In class number 30, but wrong language */+ "  eSSiG  " /* In class number 30 */;
    final String[] validTermVariants = { "Kaiser", "Skilifte", "café" };

    IpiTermCandidateVO termCandidate = new IpiTermCandidateVO();
    termCandidate.setLanguage( Locale.GERMAN );
    termCandidate.setTerm( termText );
    termCandidate.setClassNumber( 30 );

    List<String> termVariants = getServices().extractTermVariants( termCandidate );
    assertEquals( "Checking number of results", validTermVariants.length, termVariants.size() );
    for ( int i = 0; i < validTermVariants.length; i++ )
    {
      String validTermVariant = validTermVariants[i];
      assertEquals( "Checking term variant", validTermVariant, termVariants.get( i ) );
    }
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#getInvalidPropertyValues(ch.ige.dbutils.vo.AbstractVO, boolean)
  @Override
  @SuppressWarnings( "unused" )
  public Map<String, Object[]> getInvalidPropertyValues( IpiTermCandidateVO valueObject, boolean isUpdate ) throws IllegalStateException
  {
    Map<String, Object[]> properties = new HashMap<String, Object[]>();

    if ( isUpdate )
    {
      properties.put( "identifier", new Integer[] { null, -1, 0 } );
    }
    else
    {
      properties.put( "identifier", new Integer[] { -1, 0 } );
    }

    properties.put( "classNumber", new Integer[] { null, 0, 46 } );

    properties.put( "language", new Locale[] { null, Locale.CHINESE } );

    properties.put( "term", new String[] { null, "", " ", " bla", "bla ", RandomStringUtils.randomAlphabetic( ParameterConstants.EXCEEDED_LIMIT_FIELD_TERM ),
        "´", "bl´a", "`", "bl`a", "^", "bl^a", "¨", "bl¨a", "bla" + '\u2014' + "bla" /* long dash, created automatically by Word */, "bla  bla" } );

    properties.put( "privateComment", new String[] { "", RandomStringUtils.randomAlphabetic( 501 ), "bla   bla" } );

    properties.put( "publicComment", new String[] { "", RandomStringUtils.randomAlphabetic( 501 ), "bla bla  " } );

    properties.put( "latestOccurrence", new Timestamp[] { new Timestamp( System.currentTimeMillis() + 5000 ) } );

    properties.put( "usageCount", new Integer[] { -1, 0 } );

    properties.put( "creationUser", new String[] { null, "", RandomStringUtils.randomAlphabetic( 26 ) } );

    properties.put( "changeUser", new String[] { null, "", RandomStringUtils.randomAlphabetic( 26 ) } );

    return properties;
  }

  /**
   * Tests {@link ch.ige.wdltool.ejb.WdlServices#deleteIpiTermCandidate(Integer)}.
   * @throws Exception if the test fails
   */
  public void testDeleteIpiTermCandidate() throws Exception
  {
    defaultTestDelete( MINUS_ONE, 999999999 );
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#getServices()
  @Override
  protected WdlServices getServices()
  {
    return this.wdlServices;
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#getSerializableVO()
  @Override
  protected IpiTermCandidateVO getSerializableVO() throws SQLException
  {
    return this.ipiTermCandidateDAOTest.getSerializableVO();
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#compareVOsAndValidateLoaded(ch.ige.dbutils.vo.AbstractVO, ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected void compareVOsAndValidateLoaded( IpiTermCandidateVO originalVO, IpiTermCandidateVO loadedVO )
  {
    this.ipiTermCandidateDAOTest.compareVOsAndValidateLoaded( originalVO, loadedVO );
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#readWithService(java.io.Serializable)
  @Override
  public IpiTermCandidateVO readWithService( Integer id ) throws Exception
  {
    return getServices().findIpiTermCandidate( id );
  }

  // @see ch.ige.dbutils.dao.test.BaseDAOWithCRUDTestTemplate#modifyVOBeforeInsert(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected void modifyVOBeforeInsert( IpiTermCandidateVO vo )
  {
    vo.setIdentifier( null );
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#insertWithService(ch.ige.dbutils.vo.AbstractVO)
  @Override
  public Integer insertWithService( IpiTermCandidateVO vo ) throws Exception
  {
    IpiTermCandidateVO newTerm = getServices().insertIpiTermCandidate( vo );
    return newTerm.getIdentifier();
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#modifyVOBeforeUpdate(ch.ige.dbutils.vo.AbstractVO)
  @Override
  protected void modifyVOBeforeUpdate( IpiTermCandidateVO vo )
  {
    this.ipiTermCandidateDAOTest.modifyVOBeforeUpdate( vo );
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#updateWithService(ch.ige.dbutils.vo.AbstractVO)
  @Override
  public void updateWithService( IpiTermCandidateVO vo ) throws Exception
  {
    getServices().updateIpiTermCandidate( vo );
  }

  // @see ch.ige.wdltool.ejb.BaseServiceTestTemplate#deleteWithService(java.io.Serializable)
  @Override
  public void deleteWithService( Integer id ) throws Exception
  {
    getServices().deleteIpiTermCandidate( id );
  }
}
