package ch.ige.wdltool.ejb.impl;

import java.io.Serializable;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import ch.ige.dbutils.dao.DAO;
import ch.ige.dbutils.dao.DAOPool;
import ch.ige.dbutils.dao.UnitOfWorkException;
import ch.ige.dbutils.vo.AbstractVO;
import ch.ige.wdltool.exception.ExceptionThrower;
import ch.ige.wdltool.exception.WdlCreateException;
import ch.ige.wdltool.exception.WdlObjectNotFoundException;
import ch.ige.wdltool.exception.WdlRemoveException;
import ch.ige.wdltool.exception.WdlUpdateException;
import ch.ige.wdltool.exception.WdlValidationException;
import ch.ige.wdltool.exception.WrappedRuntimeException;
import ch.ige.wdltool.validator.Validator;

/**
 * The base class for services implementation.
 * <p>
 * Copyright (c) 2007, Swiss Federal Institute of Intellectual Property
 * </p>
 * 
 * @param <D> The DAO type to use
 * @param <V> The VO type to use
 * @param <I> The ID type to use
 * @author Eric Lewis
 * @version $Id: BaseServicesImpl.java,v 1.15 2008/05/08 10:43:57 lewis_lee Exp $
 */
abstract class BaseServicesImpl<D extends DAO, V extends AbstractVO, I extends Serializable>
{
  /**
   * Gets a instance of a DAO.
   * 
   * @return The instance of a DAO
   */
  protected D getDAOInstance()
  {
    return DAOPool.get( getDAOClass() );
  }

  /**
   * Gets the object's name, for instance <code>IPI term</code>.
   * 
   * @return The object's name
   */
  protected abstract String getObjectName();

  /**
   * Gets the DAO class to instantiate.
   * 
   * @return The DAO class
   */
  protected abstract Class<D> getDAOClass();

  /**
   * Finds an object.
   * 
   * @param dao The DAO to use
   * @param id The object's identifier
   * @return The value object
   * @throws RuntimeException if the ID is faulty
   * @throws SQLException if the query fails
   */
  @SuppressWarnings( "unused" )
  protected V find( D dao, I id ) throws RuntimeException, SQLException
  {
    throw new IllegalStateException( "Must be implemented" );
  }

  /**
   * Finds an object.
   * 
   * @param dao The DAO to use
   * @return The value object
   * @throws SQLException if the query fails
   */
  @SuppressWarnings( "unused" )
  protected List<V> findAll( D dao ) throws SQLException
  {
    throw new IllegalStateException( "Must be implemented" );
  }

  /**
   * Inserts an object.
   * 
   * @param dao The DAO to use
   * @param vo The VO to insert
   * @return The same <code>vo</code> with the ID inserted
   * @throws RuntimeException if the VO is faulty
   * @throws UnitOfWorkException if the insert fails
   * @throws WdlValidationException if the VO is invalid
   */
  @SuppressWarnings( "unused" )
  protected V insert( D dao, V vo ) throws RuntimeException, UnitOfWorkException, WdlValidationException
  {
    throw new IllegalStateException( "Must be implemented" );
  }

  /**
   * Updates an object.
   * 
   * @param dao The DAO to use
   * @param vo The VO to update
   * @throws RuntimeException if the VO is faulty
   * @throws UnitOfWorkException if the insert fails
   * @throws WdlValidationException if the VO is invalid
   */
  @SuppressWarnings( "unused" )
  protected void update( D dao, V vo ) throws RuntimeException, UnitOfWorkException, WdlValidationException
  {
    throw new IllegalStateException( "Must be implemented" );
  }

  /**
   * Deletes an object.
   * 
   * @param dao The DAO to use
   * @param id The identifier of the object to delete
   * @throws RuntimeException if the ID is faulty
   * @throws UnitOfWorkException if the delete fails
   */
  @SuppressWarnings( "unused" )
  protected void delete( D dao, I id ) throws RuntimeException, UnitOfWorkException
  {
    throw new IllegalStateException( "Must be implemented" );
  }

  /**
   * Deletes several objects.
   * 
   * @param dao The DAO to use
   * @param ids The identifiers of the objects to delete
   * @throws RuntimeException if the IDs are faulty
   * @throws UnitOfWorkException if the delete fails
   */
  @SuppressWarnings( "unused" )
  protected void delete( D dao, Collection<I> ids ) throws RuntimeException, UnitOfWorkException
  {
    throw new IllegalStateException( "Must be implemented" );
  }

  /**
   * Validates a value object.
   * 
   * @param vo The value object to validate
   * @param checkIdentifierExists If this is <code>true</code> and the value object is an
   *          {@link ch.ige.dbutils.vo.IdentifiableVO}, then its identifier must not be <code>null</code>
   * @throws WdlValidationException if the VO's data is invalid 
   */
  protected void validate( V vo, boolean checkIdentifierExists ) throws WdlValidationException
  {
    Validator.validate( vo, checkIdentifierExists );
  }

  /**
   * Finds an object.
   * 
   * @param id The object's identifier
   * @return The value object
   * @throws WrappedRuntimeException if an argument is illegal
   * @throws WdlObjectNotFoundException 
   */
  public final V find( I id ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    final String errorMessage = "Didn't find " + getObjectName() + " with ID " + id;
    V vo = null;
    try
    {
      vo = find( getDAOInstance(), id );
      if ( vo == null )
      {
        ExceptionThrower.throwObjectNotFoundException( errorMessage, id );
      }
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwObjectNotFoundException( errorMessage, e, id );
    }
    return vo;
  }

  /**
   * Finds all objects.
   * 
   * @return The value object
   * @throws WdlObjectNotFoundException if the objects can't be found
   */
  public final List<V> findAll() throws WdlObjectNotFoundException
  {
    List<V> vos = new ArrayList<V>();
    try
    {
      vos = findAll( getDAOInstance() );
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwObjectNotFoundException( "Failed to get all objects for " + getObjectName(), e, null );
    }
    return vos;
  }

  /**
   * Checks a value object, including special treatment for terms.
   * 
   * @param vo The value object to check
   * @param checkIdentifierExists If this is <code>true</code> and the value object is an
   *          {@link ch.ige.dbutils.vo.IdentifiableVO}, then its identifier must not be <code>null</code>
   * @throws WdlValidationException if the term already exists
   */
  protected void checkValueObject( V vo, boolean checkIdentifierExists ) throws WdlValidationException
  {
    validate( vo, checkIdentifierExists );
  }

  /**
   * Inserts an object.
   * 
   * @param vo The VO to insert
   * @return The same <code>vo</code> with the ID inserted
   * @throws WrappedRuntimeException if the VO is invalid 
   * @throws WdlCreateException if the VO can't be inserted
   * @throws WdlValidationException if the VO's data is invalid
   */
  public final V insert( V vo ) throws WrappedRuntimeException, WdlCreateException, WdlValidationException
  {
    try
    {
      checkValueObject( vo, false );
      return insert( getDAOInstance(), vo );
    }
    catch ( UnitOfWorkException e )
    {
      ExceptionThrower.throwCreateException( "Couldn't insert " + getObjectName() + " " + vo, e, vo );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
    return null; // This should never happen, since all caught Exceptions are re-thrown
  }

  /**
   * Updates an object.
   * 
   * @param vo The VO to update
   * @throws WrappedRuntimeException for illegal arguments 
   * @throws WdlUpdateException if the VO can't be updated
   * @throws WdlValidationException  if the VO's data is invalid
   */
  public final void update( V vo ) throws WrappedRuntimeException, WdlUpdateException, WdlValidationException
  {
    try
    {
      checkValueObject( vo, true );
      update( getDAOInstance(), vo );
    }
    catch ( UnitOfWorkException e )
    {
      ExceptionThrower.throwUpdateException( "Couldn't update " + getObjectName() + " " + vo, e, vo );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
  }

  /**
   * Deletes an object.
   * 
   * @param id The identifier of the object to delete
   * @throws WrappedRuntimeException 
   * @throws WdlRemoveException if the object can't be deleted
   */
  public final void delete( I id ) throws WrappedRuntimeException, WdlRemoveException
  {
    try
    {
      delete( getDAOInstance(), id );
    }
    catch ( UnitOfWorkException e )
    {
      ExceptionThrower.throwRemoveException( "Couldn't delete " + getObjectName() + " with ID " + id, e, id );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
  }

  /**
   * Deletes several objects.
   * 
   * @param ids The identifier of the object to delete
   * @throws WrappedRuntimeException 
   * @throws WdlRemoveException if the object can't be deleted
   */
  public final void delete( Collection<I> ids ) throws WrappedRuntimeException, WdlRemoveException
  {
    try
    {
      delete( getDAOInstance(), ids );
    }
    catch ( UnitOfWorkException e )
    {
      ExceptionThrower.throwRemoveException( "Couldn't delete " + getObjectName() + " with IDs " + StringUtils.join( ids.iterator(), ',' ), e, ids );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
  }
}
