package ch.ige.wdltool.ejb.impl;

import java.io.File;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.ejb.EJBException;

import org.apache.log4j.Logger;

import ch.ige.dbutils.dao.AbstractDAO;
import ch.ige.dbutils.dao.DAOPool;
import ch.ige.ipobjects.trademark.Trademark;
import ch.ige.util.enums.infrastructure.DBRealm;
import ch.ige.wdltool.BuildInfo;
import ch.ige.wdltool.analyzer.AnalysisResultVO;
import ch.ige.wdltool.analyzer.WdlAnalyzer;
import ch.ige.wdltool.dao.IpiTermCandidateDAO;
import ch.ige.wdltool.ejb.WdlLocalServices;
import ch.ige.wdltool.ejb.WdlServices;
import ch.ige.wdltool.enums.DataSource;
import ch.ige.wdltool.exception.ExceptionThrower;
import ch.ige.wdltool.exception.WdlCreateException;
import ch.ige.wdltool.exception.WdlObjectNotFoundException;
import ch.ige.wdltool.exception.WdlRemoveException;
import ch.ige.wdltool.exception.WdlUpdateException;
import ch.ige.wdltool.exception.WdlValidationException;
import ch.ige.wdltool.exception.WrappedRuntimeException;
import ch.ige.wdltool.util.Configuration;
import ch.ige.wdltool.vo.AbstractTermVO;
import ch.ige.wdltool.vo.ClassVO;
import ch.ige.wdltool.vo.DataSourceVO;
import ch.ige.wdltool.vo.ExtendedSearchVO;
import ch.ige.wdltool.vo.IpiTermCandidateVO;
import ch.ige.wdltool.vo.IpiTermGroupVO;
import ch.ige.wdltool.vo.IpiTermVO;
import ch.ige.wdltool.vo.NegativeIpiTermVO;
import ch.ige.wdltool.vo.NiceTermGroupVO;
import ch.ige.wdltool.vo.NiceTermVO;
import ch.ige.wdltool.vo.StopwordVO;
import ch.ige.wdltool.vo.TermCleanupCandidateStatusVO;
import ch.ige.wdltool.vo.TermGroupVO;
import ch.ige.wdltool.vo.WdlCheckResultVO;
import ch.ige.wdlutil.analyze.WdlSyntaxConfiguration;
import ch.ige.wdlutil.exception.WdlSyntaxException;

/**
 * Implements the WDL services.
 * Note: For methods with arguments <b>all</b> IllegalArgumentExceptions are wrapped into a WrappedRuntimeException.
 * This should not be the case for internally generated problems (not caused by the arguments given by the client of the service method).
 * e.g. getCleanerLogDirectory and setCleanerLogDirectory throw different exception types for the same internal problem (CHECKER_LOG_DIRECTORY not defined)
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   Eric Lewis
 * @version  $Id: WdlServicesImpl.java,v 1.60 2008/12/09 16:08:56 bandschapp Exp $
 */
public class WdlServicesImpl implements WdlServices, WdlLocalServices
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( WdlServicesImpl.class );

  /**
   * The stopword implementation.
   */
  private StopwordImpl stopwordImpl;

  /**
   * The data source implementation.
   */
  private DataSourceImpl dataSourceImpl;

  /**
   * The class implementation.
   */
  private ClassImpl classImpl;

  /**
   * The Nice term group implementation.
   */
  private NiceTermGroupImpl niceTermGroupImpl;

  /**
   * The general term implementation.
   */
  private GeneralTermImpl generalTermImpl;

  /**
   * The IPI term candidate implementation.
   */
  private IpiTermCandidateImpl ipiTermCandidateImpl;

  /**
   * The IPI term group implementation.
   */
  private IpiTermGroupImpl ipiTermGroupImpl;

  /**
   * The IPI term implementation.
   */
  private IpiTermImpl ipiTermImpl;

  /**
   * The negative IPI term implementation.
   */
  private NegativeIpiTermImpl negativeIpiTermImpl;

  /**
   * The Nice term implementation.
   */
  private NiceTermImpl niceTermImpl;

  /**
   * Used for checking WDLs
   */
  private WdlCheckerImpl wdlCheckerImpl;

  /**
   * The term conversion implementation.
   */
  private TermConversionImpl termConversionImpl;

  // @see ch.ige.wdltool.ejb.WdlServices#getWdlText(java.lang.Integer)
  public String getWdlText( Integer trademarkIpKey )
  {
    Trademark trademark = TrademarkImpl.getTrademarkForKey( trademarkIpKey );
    return TrademarkImpl.getCurrentWdlText( trademark );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#checkWdl(java.lang.Integer, boolean, java.util.Map)
  public WdlCheckResultVO checkWdl( Integer trademarkIpKey, boolean doStrictSyntaxCheck, Map<WdlSyntaxConfiguration, Boolean> wdlSyntaxConfiguration )
      throws WdlSyntaxException, WrappedRuntimeException
  {
    return getWdlCheckerImpl().checkWdl( trademarkIpKey, doStrictSyntaxCheck, wdlSyntaxConfiguration );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#checkWdl(java.lang.String, java.util.Locale, boolean, java.util.Map)
  public WdlCheckResultVO checkWdl( String wdlText, Locale language, boolean doStrictSyntaxCheck, Map<WdlSyntaxConfiguration, Boolean> wdlSyntaxConfiguration )
      throws WdlSyntaxException, WrappedRuntimeException
  {
    return getWdlCheckerImpl().checkWdl( wdlText, language, doStrictSyntaxCheck, wdlSyntaxConfiguration );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#analyzeWdls(java.util.Date, java.util.Date)
  public String analyzeWdls( Date from, Date to ) throws WrappedRuntimeException
  {
    try
    {
      //use local calls for each wdl      
      return getNewWdlAnalyzer().analyzeWdls( from, to );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
    return null; //will never happen
  }

  // @see ch.ige.wdltool.ejb.LocalWdlServices#analyzeWdl(int)
  public AnalysisResultVO analyzeWdl( int ipKey ) throws WrappedRuntimeException
  {
    try
    {
      return getNewWdlAnalyzer().analyzeWdl( ipKey );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
    return null; //will never happen
  }

  // @see ch.ige.wdltool.ejb.WdlServices#checkWdlForApplication(java.util.Date, java.util.Date)
  public String[] checkWdlForApplication( Date from, Date to ) throws WrappedRuntimeException
  {
    return getWdlCheckerImpl().checkWdlForApplication( from, to );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#checkWdlForRegistration(java.util.Date, java.util.Date)
  public String[] checkWdlForRegistration( Date from, Date to ) throws WrappedRuntimeException
  {
    return getWdlCheckerImpl().checkWdlForRegistration( from, to );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findStopword(java.lang.Integer)
  @SuppressWarnings( "unused" )
  public StopwordVO findStopword( Integer stopwordId ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    return getStopwordImpl().find( stopwordId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findAllStopwords()
  public List<StopwordVO> findAllStopwords() throws WdlObjectNotFoundException
  {
    return getStopwordImpl().findAll();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findAnyWordWithinStopwords(java.util.Locale, java.lang.String)
  public List<StopwordVO> findAnyWordWithinStopwords( Locale language, String searchStopword ) throws WdlObjectNotFoundException, WrappedRuntimeException
  {
    return getStopwordImpl().findAnyWordWithinStopwords( language, searchStopword );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#insertStopword(ch.ige.wdltool.vo.StopwordVO)
  public StopwordVO insertStopword( StopwordVO stopword ) throws EJBException, WrappedRuntimeException, WdlCreateException, WdlValidationException
  {
    return getStopwordImpl().insert( stopword );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#updateStopword(ch.ige.wdltool.vo.StopwordVO)
  public void updateStopword( StopwordVO stopword ) throws WrappedRuntimeException, WdlUpdateException, WdlValidationException
  {
    getStopwordImpl().update( stopword );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#deleteStopword(java.lang.Integer)
  public void deleteStopword( Integer stopwordId ) throws WrappedRuntimeException, WdlRemoveException
  {
    getStopwordImpl().delete( stopwordId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#deleteStopwords(java.util.Collection)
  public void deleteStopwords( Collection<Integer> stopwordIds ) throws WrappedRuntimeException, WdlRemoveException
  {
    getStopwordImpl().delete( stopwordIds );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findAllClasses()
  public List<ClassVO> findAllClasses() throws WdlObjectNotFoundException
  {
    return getClassImpl().findAll();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findAllDataSources()
  public List<DataSourceVO> findAllDataSources() throws WdlObjectNotFoundException
  {
    return getDataSourceImpl().findAll();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findTermGroup(java.lang.Integer)
  public TermGroupVO findTermGroup( Integer termGroupId ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    TermGroupVO result = null;
    try
    {
      result = getNiceTermGroupImpl().find( termGroupId );
    }
    catch ( WdlObjectNotFoundException e )
    {
      result = getIpiTermGroupImpl().find( termGroupId );
    }
    return result;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findAllNiceTermGroups()
  public List<NiceTermGroupVO> findAllNiceTermGroups() throws WdlObjectNotFoundException
  {
    return getNiceTermGroupImpl().findAll();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findByNiceTermNumber(java.lang.Integer)
  public NiceTermGroupVO findByNiceTermNumber( Integer niceTermNumber ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    return getNiceTermGroupImpl().findByNiceTermNumber( niceTermNumber );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#insertNiceTermGroup(NiceTermGroupVO)
  public NiceTermGroupVO insertNiceTermGroup( NiceTermGroupVO vo ) throws EJBException, WrappedRuntimeException, WdlCreateException, WdlValidationException
  {
    return getNiceTermGroupImpl().insert( vo );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findAllIpiTermGroups()
  public List<IpiTermGroupVO> findAllIpiTermGroups() throws WdlObjectNotFoundException
  {
    return getIpiTermGroupImpl().findAll();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findByIpiTermNumber(java.lang.Integer)
  public IpiTermGroupVO findByIpiTermNumber( Integer ipiTermNumber ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    return getIpiTermGroupImpl().findByIpiTermNumber( ipiTermNumber );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#insertIpiTermGroup(NiceTermGroupVO)
  public IpiTermGroupVO insertIpiTermGroup( IpiTermGroupVO vo ) throws EJBException, WrappedRuntimeException, WdlCreateException, WdlValidationException
  {
    return getIpiTermGroupImpl().insert( vo );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#createNextFreeIpiTermGroup()
  public IpiTermGroupVO createNextFreeIpiTermGroup() throws WdlCreateException
  {
    try
    {
      return getIpiTermGroupImpl().createNextFreeIpiTermGroup();
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwSQLException( "Couldn't create new IPI term group", e );
    }
    return null;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findAnyWordWithinActiveTerms(java.util.Locale, java.lang.String)
  public List<AbstractTermVO> findAnyWordWithinActiveTerms( Locale language, String searchTerm ) throws WdlObjectNotFoundException, WrappedRuntimeException
  {
    return getGeneralTermImpl().findAnyWordWithinActiveTerms( language, searchTerm, false );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findTermWithinActiveTerms(ExtendedSearchVO)
  public List<AbstractTermVO> findTermWithinActiveTerms( ExtendedSearchVO searchDefinition ) throws WdlObjectNotFoundException, WrappedRuntimeException
  {
    return getGeneralTermImpl().findTermWithinActiveTerms( searchDefinition );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findTermWithinActiveTerms(java.util.Locale, java.lang.Integer, java.lang.String)
  public List<AbstractTermVO> findTermWithinActiveTerms( Locale language, Integer classNumber, String term ) throws WdlObjectNotFoundException,
      WrappedRuntimeException
  {
    return getGeneralTermImpl().findTermWithinActiveTerms( language, classNumber, term, null );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findTermWithinActiveTerms(java.util.Locale, java.lang.Integer, java.lang.String, java.lang.Boolean)
  public List<AbstractTermVO> findTermWithinActiveTerms( Locale language, Integer classNumber, String term, Boolean isCheckerOnly )
      throws WdlObjectNotFoundException, WrappedRuntimeException
  {
    return getGeneralTermImpl().findTermWithinActiveTerms( language, classNumber, term, isCheckerOnly );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findTermWithinInactiveTerms(java.util.Locale, java.lang.Integer,
  // java.lang.String)
  public List<IpiTermVO> findTermWithinInactiveTerms( Locale language, Integer classNumber, String term ) throws WdlObjectNotFoundException,
      WrappedRuntimeException
  {
    return getIpiTermImpl().findTermWithinInactiveTerms( language, classNumber, term );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findSameSearchTermWithinActiveTerms(ch.ige.wdltool.vo.AbstractTermVO)
  public List<AbstractTermVO> findSameSearchTermWithinActiveTerms( AbstractTermVO term ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    return getGeneralTermImpl().findSameSearchTermWithinActiveTerms( term );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findIpiTermCandidate(java.util.Locale, java.lang.Integer, java.lang.String)
  public IpiTermCandidateVO findIpiTermCandidate( Locale language, Integer classNumber, String searchTerm ) throws WdlObjectNotFoundException,
      WrappedRuntimeException
  {
    return getIpiTermCandidateImpl().findIpiTermCandidate( language, classNumber, searchTerm );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findNiceTerm(java.lang.Integer)
  public NiceTermVO findNiceTerm( Integer niceTermId ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    return getNiceTermImpl().find( niceTermId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findGenericTerms()
  public List<NiceTermVO> findGenericTerms() throws WrappedRuntimeException
  {
    return getNiceTermImpl().findGenericTerms();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#insertNiceTermAndRemoveIpiTermCandidate(ch.ige.wdltool.vo.NiceTermVO)
  public TermCleanupCandidateStatusVO<NiceTermVO> insertNiceTermAndRemoveIpiTermCandidate( NiceTermVO niceTerm ) throws WrappedRuntimeException,
      WdlCreateException, WdlValidationException
  {
    try
    {
      return getNiceTermImpl().insertAndRemoveIpiTermCandidate( niceTerm );
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwCreateException( "error during insertAndRemoveIpiTermCandidate", e, niceTerm );
    }
    return null;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#updateNiceTermAndRemoveIpiTermCandidate(ch.ige.wdltool.vo.NiceTermVO)
  public boolean updateNiceTermAndRemoveIpiTermCandidate( NiceTermVO niceTerm ) throws WrappedRuntimeException, WdlUpdateException, WdlValidationException
  {
    try
    {
      return getNiceTermImpl().updateAndRemoveIpiTermCandidate( niceTerm );
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwUpdateException( "error during updateAndRemoveIpiTermCandidate", e, niceTerm );
    }
    return false;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#deleteNiceTerm(java.lang.Integer)
  public void deleteNiceTerm( Integer niceTermId ) throws WrappedRuntimeException, WdlRemoveException
  {
    getNiceTermImpl().delete( niceTermId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findIpiTerm(java.lang.Integer)
  public IpiTermVO findIpiTerm( Integer ipiTermId ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    return getIpiTermImpl().find( ipiTermId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#insertIpiTerm(ch.ige.wdltool.vo.IpiTermVO)
  public TermCleanupCandidateStatusVO<IpiTermVO> insertIpiTermAndRemoveIpiTermCandidate( IpiTermVO ipiTerm ) throws WrappedRuntimeException,
      WdlCreateException, WdlValidationException
  {
    try
    {
      return getIpiTermImpl().insertAndRemoveIpiTermCandidate( ipiTerm );
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwCreateException( "error during insertAndRemoveIpiTermCandidate", e, ipiTerm );
    }
    return null;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#updateIpiTermAndRemoveIpiTermCandidate(ch.ige.wdltool.vo.IpiTermVO)
  public boolean updateIpiTermAndRemoveIpiTermCandidate( IpiTermVO ipiTerm ) throws WrappedRuntimeException, WdlUpdateException, WdlValidationException
  {
    try
    {
      return getIpiTermImpl().updateAndRemoveIpiTermCandidate( ipiTerm );
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwUpdateException( "error during updateIpiTermAndRemoveIpiTermCandidate", e, ipiTerm );
    }
    return false;
  }

  // @see
  // ch.ige.wdltool.ejb.WdlServices#convertNegativeIpiTermToIpiTermAndRemoveIpiTermCandidate(ch.ige.wdltool.vo.IpiTermVO)
  public boolean convertNegativeIpiTermToIpiTermAndRemoveIpiTermCandidate( IpiTermVO ipiTerm ) throws WrappedRuntimeException, WdlUpdateException,
      WdlValidationException
  {
    return getTermConversion().convertNegativeIpiTermToIpiTermAndRemoveIpiTermCandidate( ipiTerm );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#deleteIpiTerm(java.lang.Integer)
  public void deleteIpiTerm( Integer ipiTermId ) throws WrappedRuntimeException, WdlRemoveException
  {
    getIpiTermImpl().delete( ipiTermId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findIpiTermCandidate(java.lang.Integer)
  public IpiTermCandidateVO findIpiTermCandidate( Integer ipiTermCandidateId ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    return getIpiTermCandidateImpl().find( ipiTermCandidateId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findIpiTermCandidates(java.util.Locale)
  public List<IpiTermCandidateVO> findIpiTermCandidates( Locale language ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    List<IpiTermCandidateVO> resultList = null;
    try
    {
      resultList = getIpiTermCandidateImpl().findByLanguage( DAOPool.get( IpiTermCandidateDAO.class ), language );
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwObjectNotFoundException( "Didn't find any candidate", e, null );
    }
    return resultList;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#insertIpiTermCandidate(ch.ige.wdltool.vo.IpiTermCandidateVO)
  public IpiTermCandidateVO insertIpiTermCandidate( IpiTermCandidateVO ipiTermCandidate ) throws EJBException, WrappedRuntimeException, WdlCreateException,
      WdlValidationException
  {
    return getIpiTermCandidateImpl().insert( ipiTermCandidate );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#updateIpiTermCandidate(ch.ige.wdltool.vo.IpiTermCandidateVO)
  public void updateIpiTermCandidate( IpiTermCandidateVO ipiTermCandidate ) throws WrappedRuntimeException, WdlUpdateException, WdlValidationException
  {
    getIpiTermCandidateImpl().update( ipiTermCandidate );
  }

  // @see
  // ch.ige.wdltool.ejb.WdlServices#convertNegativeIpiTermToIpiTermAndRemoveIpiTermCandidate(ch.ige.wdltool.vo.IpiTermVO)
  public void convertInactiveTermToIpiTermCandidate( IpiTermCandidateVO ipiTermCandidate ) throws WrappedRuntimeException, WdlUpdateException,
      WdlValidationException
  {
    getTermConversion().convertInactiveTermToIpiTermCandidate( ipiTermCandidate );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#deleteIpiTermCandidate(java.lang.Integer)
  public void deleteIpiTermCandidate( Integer ipiTermCandidateId ) throws WrappedRuntimeException, WdlRemoveException
  {
    getIpiTermCandidateImpl().delete( ipiTermCandidateId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#activateIpiTermCandidate(ch.ige.wdltool.vo.IpiTermCandidateVO)
  public void activateIpiTermCandidate( IpiTermCandidateVO ipiTermCandidate ) throws WdlUpdateException, WdlValidationException, WrappedRuntimeException
  {
    getIpiTermCandidateImpl().activate( DAOPool.get( IpiTermCandidateDAO.class ), ipiTermCandidate );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#extractTermVariants(ch.ige.wdltool.vo.IpiTermCandidateVO)
  public List<String> extractTermVariants( IpiTermCandidateVO ipiTermCandidate ) throws WdlObjectNotFoundException, WrappedRuntimeException
  {
    return getIpiTermCandidateImpl().extractTermVariants( ipiTermCandidate );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findNegativeIpiTerm(java.lang.Integer)
  public NegativeIpiTermVO findNegativeIpiTerm( Integer negativeNegativeIpiTermId ) throws WrappedRuntimeException, WdlObjectNotFoundException
  {
    return getNegativeIpiTermImpl().find( negativeNegativeIpiTermId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#insertNegativeIpiTerm(ch.ige.wdltool.vo.NegativeIpiTermVO)
  public NegativeIpiTermVO insertNegativeIpiTerm( NegativeIpiTermVO negativeNegativeIpiTerm ) throws EJBException, WrappedRuntimeException, WdlCreateException,
      WdlValidationException
  {
    return getNegativeIpiTermImpl().insert( negativeNegativeIpiTerm );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#updateNegativeIpiTerm(ch.ige.wdltool.vo.NegativeIpiTermVO)
  public void updateNegativeIpiTerm( NegativeIpiTermVO negativeIpiTerm ) throws WrappedRuntimeException, WdlUpdateException, WdlValidationException
  {
    getNegativeIpiTermImpl().update( negativeIpiTerm );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#convertIpiTermToNegativeIpiTerm(ch.ige.wdltool.vo.NegativeIpiTermVO)
  public void convertIpiTermToNegativeIpiTerm( NegativeIpiTermVO negativeIpiTerm ) throws WrappedRuntimeException, WdlUpdateException,
      WdlObjectNotFoundException, WdlValidationException
  {
    getTermConversion().convertIpiTermToNegativeIpiTerm( negativeIpiTerm );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#deleteNegativeIpiTerm(java.lang.Integer)
  public void deleteNegativeIpiTerm( Integer negativeNegativeIpiTermId ) throws WrappedRuntimeException, WdlRemoveException
  {
    getNegativeIpiTermImpl().delete( negativeNegativeIpiTermId );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#deleteTerms(java.util.Collection)
  public void deleteTerms( Collection<Integer> termIds ) throws WrappedRuntimeException, WdlRemoveException
  {
    getGeneralTermImpl().delete( termIds );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findExpiredTerms()
  public List<IpiTermVO> findInactiveTerms()
  {
    try
    {
      return getIpiTermImpl().findInactiveTerms();
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwSQLException( "Couldn't find inactive terms", e );
    }
    return null;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findExpiredTerms()
  public List<IpiTermVO> findExpiredTerms()
  {
    try
    {
      return getIpiTermImpl().findExpiredTerms();
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwSQLException( "Couldn't find expired terms", e );
    }
    return null;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#findExpiringTerms()
  public List<IpiTermVO> findExpiringTerms()
  {
    try
    {
      return getIpiTermImpl().findExpiringTerms();
    }
    catch ( SQLException e )
    {
      ExceptionThrower.throwSQLException( "Couldn't find expiring terms", e );
    }
    return null;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#renewTerms(List<IpiTermVO>)
  public List<Integer> renewTerms( List<IpiTermVO> terms ) throws WrappedRuntimeException
  {
    return getIpiTermImpl().renewTerms( terms );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#updateClassNumber(java.util.List, java.lang.Integer)
  public void updateClassNumber( List<AbstractTermVO> terms, Integer classNumber ) throws WdlUpdateException, WrappedRuntimeException, WdlValidationException
  {
    getGeneralTermImpl().updateClassNumber( terms, classNumber );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getDbURL()
  public String getDbUrl()
  {
    Connection tempConnection = null;
    String dbUrl = null;
    try
    {
      tempConnection = AbstractDAO.getDirectConnection( DBRealm.WDL );
      dbUrl = tempConnection.getMetaData().getURL();
    }
    catch ( SQLException e )
    {
      LOG.error( "Error when getting the db url: " + e.getMessage(), e );
    }
    finally
    {
      if ( tempConnection != null )
      {
        try
        {
          tempConnection.close();
        }
        catch ( SQLException e )
        {
          LOG.error( "Error when closing the db connection: " + e.getMessage(), e );
        }
      }
    }
    return dbUrl;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getVersion()
  public String getServerVersion()
  {
    return BuildInfo.CVS_TAG;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getAnalyzerLogDirectory()
  public File getReplicationLogDirectory()
  {
    return Configuration.getReplicationLogDirectory();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getAnalyzerLogDirectory()
  public File getAnalyzerLogDirectory()
  {
    return Configuration.getAnalyzerLogDirectory();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#setAnalyzerLogDirectory(java.io.File)
  public void setAnalyzerLogDirectory( File logDirectory ) throws WrappedRuntimeException
  {
    try
    {
      Configuration.setAnalyzerLogDirectory( logDirectory );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getNewAnalyzerLogFileName()
  public File getNewAnalyzerLogFileName()
  {
    return Configuration.getNewAnalyzerLogFileName();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getCheckerLogDirectory()
  public File getCheckerLogDirectory()
  {
    return Configuration.getCheckerLogDirectory();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#setCheckerLogDirectory(java.io.File)
  public void setCheckerLogDirectory( File logDirectory ) throws WrappedRuntimeException
  {
    try
    {
      Configuration.setCheckerLogDirectory( logDirectory );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getNewCheckerLogFileName()
  public File getNewCheckerLogFileName()
  {
    return Configuration.getNewCheckerLogFileName();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getNewCheckerSummaryFileName()
  public File getNewCheckerSummaryFileName()
  {
    return Configuration.getNewCheckerSummaryFileName();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getCleanerLogDirectory()
  public File getCleanerLogDirectory()
  {
    return Configuration.getCleanerLogDirectory();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#setCleanerLogDirectory(java.io.File)
  public void setCleanerLogDirectory( File logDirectory ) throws WrappedRuntimeException
  {
    try
    {
      Configuration.setCleanerLogDirectory( logDirectory );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getNewCleanerLogFileName()
  public File getNewCleanerLogFileName()
  {
    return Configuration.getNewCleanerLogFileName();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#createLinksForCorrectWords()
  public boolean createLinksForCorrectWords()
  {
    return Configuration.createLinksForCorrectWords();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#setCreateLinksForCorrectWords(boolean)
  public void setCreateLinksForCorrectWords( boolean createLinksForCorrectWords )
  {
    Configuration.setCreateLinksForCorrectWords( createLinksForCorrectWords );
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getTermConfirmationPeriod(ch.ige.wdltool.enums.DataSource)
  public Integer getTermConfirmationPeriod( DataSource dataSource ) throws WrappedRuntimeException
  {
    try
    {
      return Configuration.getTermConfirmationPeriod( dataSource );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
    return null;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#setTermConfirmationPeriod(ch.ige.wdltool.enums.DataSource, java.lang.Integer)
  public void setTermConfirmationPeriod( DataSource dataSource, Integer days ) throws WrappedRuntimeException
  {
    try
    {
      Configuration.setTermConfirmationPeriod( dataSource, days );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
  }

  // @see ch.ige.wdltool.ejb.WdlServices#getTermValidityPeriod(ch.ige.wdltool.enums.DataSource)
  public Integer getTermValidityPeriod( DataSource dataSource ) throws WrappedRuntimeException
  {
    try
    {
      return Configuration.getTermValidityPeriod( dataSource );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
    return null;
  }

  // @see ch.ige.wdltool.ejb.WdlServices#setTermValidityPeriod(ch.ige.wdltool.enums.DataSource, java.lang.Integer)
  public void setTermValidityPeriod( DataSource dataSource, Integer days ) throws WrappedRuntimeException
  {
    try
    {
      Configuration.setTermValidityPeriod( dataSource, days );
    }
    catch ( RuntimeException e )
    {
      ExceptionThrower.throwRuntimeException( e.getMessage(), e );
    }
  }

  // @see ch.ige.wdltool.ejb.WdlServices#inactivateExpiredTerms()
  public List<IpiTermVO> inactivateExpiredTerms()
  {
    return getIpiTermImpl().inactivateExpiredTerms();
  }

  // @see ch.ige.wdltool.ejb.WdlServices#clearCache()
  public void clearCache()
  {
    DAOPool.INSTANCE.clearCachingDAOs();
  }

  /**
   * Gets the stopword implementation.
   * @return The stopword implementation
   */
  protected StopwordImpl getStopwordImpl()
  {
    if ( this.stopwordImpl == null )
    {
      this.stopwordImpl = new StopwordImpl();
    }
    return this.stopwordImpl;
  }

  /**
   * Gets the data source implementation.
   * @return The data source implementation
   */
  protected DataSourceImpl getDataSourceImpl()
  {
    if ( this.dataSourceImpl == null )
    {
      this.dataSourceImpl = new DataSourceImpl();
    }
    return this.dataSourceImpl;
  }

  /**
   * Gets the class implementation.
   * @return The class implementation
   */
  protected ClassImpl getClassImpl()
  {
    if ( this.classImpl == null )
    {
      this.classImpl = new ClassImpl();
    }
    return this.classImpl;
  }

  /**
   * Gets the Nice term group implementation.
   * @return The Nice term group implementation
   */
  protected NiceTermGroupImpl getNiceTermGroupImpl()
  {
    if ( this.niceTermGroupImpl == null )
    {
      this.niceTermGroupImpl = new NiceTermGroupImpl();
    }
    return this.niceTermGroupImpl;
  }

  /**
   * Gets the general term implementation.
   * @return The general term implementation
   */
  protected GeneralTermImpl getGeneralTermImpl()
  {
    if ( this.generalTermImpl == null )
    {
      this.generalTermImpl = new GeneralTermImpl();
    }
    return this.generalTermImpl;
  }

  /**
   * Gets the IPI term candidate implementation.
   * @return The IPI term candidate implementation
   */
  protected IpiTermCandidateImpl getIpiTermCandidateImpl()
  {
    if ( this.ipiTermCandidateImpl == null )
    {
      this.ipiTermCandidateImpl = new IpiTermCandidateImpl();
    }
    return this.ipiTermCandidateImpl;
  }

  /**
   * Gets the IPI term group implementation.
   * @return The IPI term group implementation
   */
  protected IpiTermGroupImpl getIpiTermGroupImpl()
  {
    if ( this.ipiTermGroupImpl == null )
    {
      this.ipiTermGroupImpl = new IpiTermGroupImpl();
    }
    return this.ipiTermGroupImpl;
  }

  /**
   * Gets the IPI term implementation.
   * @return The IPI term implementation
   */
  protected IpiTermImpl getIpiTermImpl()
  {
    if ( this.ipiTermImpl == null )
    {
      this.ipiTermImpl = new IpiTermImpl();
    }
    return this.ipiTermImpl;
  }

  /**
   * Gets the negative IPI term implementation.
   * @return The negative IPI term implementation
   */
  protected NegativeIpiTermImpl getNegativeIpiTermImpl()
  {
    if ( this.negativeIpiTermImpl == null )
    {
      this.negativeIpiTermImpl = new NegativeIpiTermImpl();
    }
    return this.negativeIpiTermImpl;
  }

  /**
   * Gets the Nice term implementation.
   * @return The Nice term implementation
   */
  protected NiceTermImpl getNiceTermImpl()
  {
    if ( this.niceTermImpl == null )
    {
      this.niceTermImpl = new NiceTermImpl();
    }
    return this.niceTermImpl;
  }

  /**
   * Gets the WDL checker implementation.
   * @return The WDL checker implementation
   */
  protected WdlCheckerImpl getWdlCheckerImpl()
  {
    if ( this.wdlCheckerImpl == null )
    {
      this.wdlCheckerImpl = new WdlCheckerImpl();
    }
    return this.wdlCheckerImpl;
  }

  /**
   * Gets the WDL checker implementation.
   * @return The WDL checker implementation
   */
  protected TermConversionImpl getTermConversion()
  {
    if ( this.termConversionImpl == null )
    {
      this.termConversionImpl = new TermConversionImpl( this );
    }
    return this.termConversionImpl;
  }

  /**
   * Returns a new WDL analyzer.
   * Note as WdlAnalyzer tracks global book keeping information, a single instance is not enough - we really need a new one.
   * @return The WDL analyzer
   */
  private WdlAnalyzer getNewWdlAnalyzer()
  {
    return new WdlAnalyzer( this );
  }
}
