package ch.ige.wdltool.util;

import java.text.NumberFormat;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import ch.ige.ipobjects.trademark.Trademark;
import ch.ige.util.date.DateFormatter;
import ch.ige.util.format.ReportIpNumberFormatter;

/**
 * Utilities for batch jobs or long-running methods that create a log.
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   Eric Lewis
 * @version  $Id: BatchUtils.java,v 1.5 2008/02/19 10:20:32 lewis_lee Exp $
 */
public class BatchUtils
{
  /**
   * The log
   */
  private static final Logger LOG = Logger.getLogger( BatchUtils.class );

  /**
   * Length of any separator
   */
  private static final int SEPARATOR_LENGTH = 80;

  /**
   * Separates two trademarks
   */
  public static final String TRADEMARK_SEPARATOR = StringUtils.repeat( "*", SEPARATOR_LENGTH );

  /**
   * Separates the trademark header from the details
   */
  public static final String DETAILS_SEPARATOR = StringUtils.repeat( "-", SEPARATOR_LENGTH );

  /**
   * Shows where the value starts in the header
   */
  public static final int HEADER_PADDING = 19;

  /**
   * Used to format percentages.
   */
  public static NumberFormat PERCENTAGE_FORMATTER = NumberFormat.getPercentInstance( Locale.GERMAN );

  /**
   * The log is written with Windows line endings, because it will be
   * browsed with a Windows editor.
   */
  private static final String LINE_ENDING = "\r\n";

  /**
   * Adds lines to the start of the log.
   * @param log The log to be added to
   * @param lines The lines to add
   */
  public static void insertLine( StringBuilder log, List<String> lines )
  {
    // Note: This has to run backwards, since we're always inserting at the start
    for ( int i = lines.size() - 1; i >= 0; i-- )
    {
      log.insert( 0, LINE_ENDING );
      log.insert( 0, lines.get( i ) );
    }
  }

  /**
   * Adds a line to the log.
   * @param log The log to be added to
   * @param line The line to add
   */
  public static void addLine( StringBuilder log, String line )
  {
    log.append( line );
    log.append( LINE_ENDING );
  }

  /**
   * Adds an Exception to the log and
   * logs that Exception as ERROR in the class's log.
   * @param log The log to be added to
   * @param e The Exception to add
   */
  public static void addException( StringBuilder log, Exception e )
  {
    LOG.error( "Exception during batch processing", e );
    addLine( log, "Es ist ein Fehler aufgetreten: " + e.getMessage() );
  }

  /**
   * Adds a line to the log.
   * @param log The log to be added to
   * @param label The label of the line
   * @param value The value of the line
   * @param padding The position where the value should start
   */
  public static void addLine( StringBuilder log, String label, String value, int padding )
  {
    log.append( label );
    log.append( ":" );
    log.append( StringUtils.repeat( " ", padding - label.length() - 1 ) );
    log.append( value );
    log.append( LINE_ENDING );
  }

  /**
   * Adds a line to the log.
   * @param log The log to be added to
   * @param indentationLevel The requested indentation level, which translates to the number of tabs
   * @param line The line to add
   */
  public static void addLine( StringBuilder log, int indentationLevel, String line )
  {
    addLine( log, getIndentedLine( indentationLevel, line ) );
  }

  /**
   * Gets a log line with indentation.
   * @param indentationLevel The requested indentation level, which translates to the number of tabs
   * @param line The line to add
   * @return The line with indentation
   */
  public static String getIndentedLine( int indentationLevel, String line )
  {
    return StringUtils.repeat( "\t", indentationLevel ) + line;
  }

  /**
   * Adds a line to a CSV text.
   * @param csvText The CSV text to be added to
   * @param columns The columns to add to <code>csvText</code>, each will be surrounded by double quotes <code>&quot;...&quot;</code> 
   */
  public static void addCSVLine( StringBuilder csvText, Object... columns )
  {
    StringBuilder line = new StringBuilder();
    for ( Object column: columns )
    {
      if ( line.length() > 0 )
      {
        line.append( ';' );
      }
      if ( column instanceof String )
      {
        line.append( '"' + column.toString() + '"' );
      }
      else
      {
        line.append( column.toString() );
      }
    }
    addLine( csvText, line.toString() );
  }

  /**
   * Gets a formatted percentage.
   * @param number The number to convert to a percentage
   * @param total The total
   * @param ofAllExtractable If this is <code>true</code>, the percentage is take from all extractable terms
   * @return The percentage of <code>number</code> compared to <code>total</code>, as in <code> (35%)</code> 
   */
  public static String getPercentage( int number, int total, boolean ofAllExtractable )
  {
    if ( total == 0 )
    {
      return getPercentage( number, total );
    }
    else
    {
      String percentage = getPercentage( number, total );
      percentage = " (" + percentage;
      if ( ofAllExtractable )
      {
        percentage += " aller extrahierbaren Begriffe";
      }
      percentage += ")";
      return percentage;
    }
  }

  /**
   * Gets a formatted percentage.
   * @param number The number to convert to a percentage
   * @param total The total
   * @return The percentage of <code>number</code> compared to <code>total</code>, as in <code>35%</code> 
   */
  public static String getPercentage( int number, int total )
  {
    if ( total == 0 )
    {
      return "";
    }
    else
    {
      return PERCENTAGE_FORMATTER.format( ( (float) number ) / total );
    }
  }

  /**
   * Adds the trademark's base information to the log, like name, trademark number, IP key etc.
   * @param log The log to be added to
   * @param trademark The trademark to log
   */
  public static void addBaseInformation( StringBuilder log, Trademark trademark )
  {
    String trademarkName = "(leer)";
    if ( trademark.getDescription().length() > 0 )
    {
      trademarkName = trademark.getDescription();
    }
    BatchUtils.addLine( log, "Markenname", trademarkName, BatchUtils.HEADER_PADDING );

    String trademarkNumber = "(keine)";
    if ( trademark.getIPNumber() != null )
    {
      trademarkNumber = ReportIpNumberFormatter.formatTrademark( trademark.getIPNumber(), trademark.getBAGISType(), trademark.getProlongationCount() );
    }
    BatchUtils.addLine( log, "Markennummer", trademarkNumber, BatchUtils.HEADER_PADDING );

    BatchUtils.addLine( log, "Schutit-Key", trademark.getIPKey().toString(), BatchUtils.HEADER_PADDING );

    String registrationDate = "(keines)";
    if ( trademark.getRegistrationEvent() != null )
    {
      registrationDate = DateFormatter.shortFormat( trademark.getRegistrationEvent().getDate() );
    }
    BatchUtils.addLine( log, "Eintragungsdatum", registrationDate, BatchUtils.HEADER_PADDING );
  }
}
