package ch.ige.wdltool.validator;

import junit.framework.TestCase;

import org.hibernate.validator.ClassValidator;
import org.hibernate.validator.InvalidValue;

import ch.ige.wdltool.exception.WdlValidationException;
import ch.ige.wdltool.vo.NiceTermVO;

/**
 * Tests {@link ch.ige.wdltool.validator.Validator}. 
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   Eric Lewis
 * @version  $Id: ValidatorTest.java,v 1.4 2008/02/14 16:03:36 lewis_lee Exp $
 */
public class ValidatorTest extends TestCase
{
  /**
   * The only constructor
   * @param name The test's name
   */
  public ValidatorTest( String name )
  {
    super( name );
  }

  /**
   * Tests  {@link ch.ige.wdltool.validator.Validator#getValidator(java.lang.Class)}.
   */
  @SuppressWarnings( "unchecked" )
  public void testGetValidator()
  {
    NiceTermVO term = new NiceTermVO();
    ClassValidator<NiceTermVO> validator1 = (ClassValidator<NiceTermVO>) Validator.getValidator( term.getClass() );
    ClassValidator<NiceTermVO> validator2 = (ClassValidator<NiceTermVO>) Validator.getValidator( term.getClass() );
    assertSame( "Checking that class validators are cached", validator1, validator2 );
  }

  /**
   * Tests  {@link ch.ige.wdltool.validator.Validator#validateTermGroupNumber(NiceTermVO)}.
   * @throws Exception if the test fails
   */
  public void testValidateTermGroupNumber() throws Exception
  {
    NiceTermVO niceTerm = new NiceTermVO();

    // Nothing set, that's ok.
    Validator.validateTermGroupNumber( niceTerm );

    try
    {
      niceTerm.setTermGroupId( 999 );
      niceTerm.setNiceTermGroupNumber( null );
      Validator.validateTermGroupNumber( niceTerm );
      fail( "Missing term group number must fail" );
    }
    catch ( WdlValidationException e )
    {
      checkWdlValidationException( e, null );
    }

    try
    {
      niceTerm.setTermGroupId( null );
      niceTerm.setNiceTermGroupNumber( 429999 );
      fail( "Setting term group number without term group ID must fail" );
    }
    catch ( IllegalStateException e )
    {
      // This is expected
    }

    niceTerm.setTermGroupId( 7788 );

    try
    {
      niceTerm.setNiceTermGroupNumber( 429999 );
      niceTerm.setClassNumber( null );
      Validator.validateTermGroupNumber( niceTerm );
      fail( "Missing class number must fail" );
    }
    catch ( WdlValidationException e )
    {
      checkWdlValidationException( e, 429999 );
    }

    try
    {
      niceTerm.setNiceTermGroupNumber( 9999 );
      niceTerm.setClassNumber( 1 );
      Validator.validateTermGroupNumber( niceTerm );
      fail( "Too low group number must fail" );
    }
    catch ( WdlValidationException e )
    {
      checkWdlValidationException( e, 9999 );
    }

    // Valid case
    niceTerm.setNiceTermGroupNumber( 19999 );
    niceTerm.setClassNumber( 1 );

    try
    {
      niceTerm.setNiceTermGroupNumber( 20000 );
      niceTerm.setClassNumber( 1 );
      Validator.validateTermGroupNumber( niceTerm );
      fail( "Too high group number must fail" );
    }
    catch ( WdlValidationException e )
    {
      checkWdlValidationException( e, 20000 );
    }

    try
    {
      niceTerm.setNiceTermGroupNumber( 449999 );
      niceTerm.setClassNumber( 45 );
      Validator.validateTermGroupNumber( niceTerm );
      fail( "Too low group number must fail" );
    }
    catch ( WdlValidationException e )
    {
      checkWdlValidationException( e, 449999 );
    }

    // Valid case
    niceTerm.setNiceTermGroupNumber( 459999 );
    niceTerm.setClassNumber( 45 );

    try
    {
      niceTerm.setNiceTermGroupNumber( 460000 );
      niceTerm.setClassNumber( 45 );
      Validator.validateTermGroupNumber( niceTerm );
      fail( "Too high group number must fail" );
    }
    catch ( WdlValidationException e )
    {
      checkWdlValidationException( e, 460000 );
    }

    try
    {
      niceTerm.setNiceTermGroupNumber( 440001 );
      niceTerm.setClassNumber( 45 );
      Validator.validateTermGroupNumber( niceTerm );
      fail( "Valid group number for wrong class number must fail" );
    }
    catch ( WdlValidationException e )
    {
      checkWdlValidationException( e, 440001 );
    }
  }

  /**
   * Checks a validation exception.
   * @param e The validation exception
   * @param expectedInvalidTermGroupNumber The term group number to expect
   */
  private void checkWdlValidationException( WdlValidationException e, Integer expectedInvalidTermGroupNumber )
  {
    assertTrue( "Checking for Validator invalid values", e.getInvalidValues().size() == 1 );
    InvalidValue firstInvalidValue = e.getInvalidValues().get( 0 );
    assertEquals( "Checking for Validator message", "validator.invalidTermGroupNumber", firstInvalidValue.getMessage() );
    assertEquals( "Checking invalid term group number", expectedInvalidTermGroupNumber, firstInvalidValue.getValue() );
  }
}
