package ch.ige.wdltool.validator;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Map;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.ClassUtils;
import org.apache.commons.lang.StringUtils;

import ch.ige.dbutils.vo.AbstractVO;
import ch.ige.wdltool.ejb.BaseServiceTestTemplate;
import ch.ige.wdltool.exception.WdlValidationException;

/**
 * Utilities for testing value objects with the validator.
 * Note that since this is not a {@link junit.framework.TestCase}, it can't
 * use <code>fail()</code> - instead it throws an {@link java.lang.IllegalStateException}.
 *
 * <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author   Eric Lewis
 * @version  $Id: ValidatorTestUtils.java,v 1.6 2008/01/23 11:06:17 lewis_lee Exp $
 */
public class ValidatorTestUtils
{
  /**
   * The possible suffix of a value object
   */
  private static final String VO_SUFFIX = "VO";

  /**
   * The method name for inserting an object
   */
  private static final String INSERT_METHOD = "insertWithService";

  /**
   * The method name for updating an object
   */
  private static final String UPDATE_METHOD = "updateWithService";

  /**
   * For all properties, this checks all the faulty values which must be caught by the Validator for a method that inserts the value object.
   * Note that the method name is guessed from the value object's class: For instance, a <code>UserVO</code> results
   * in a method called <code>insertUser()</code>.
   * @param servicesTest The services test
   * @param valueObject The value object to check
   * @param properties A <code>Map</code> where the key is the name of the property, for instance <code>identifier</code>, and
   *        the value is an array of all the values that provoke an error 
   * @throws IllegalStateException if the validation is wrong
   */
  public static void checkValidatedValuesForInsert( BaseServiceTestTemplate<?, ?, ?, ?> servicesTest, AbstractVO valueObject, Map<String, Object[]> properties )
      throws IllegalStateException
  {
    checkValidatedValues( servicesTest, INSERT_METHOD, valueObject, properties );
  }

  /**
   * For all properties, this checks all the faulty values which must be caught by the Validator for a method that updates the value object.
   * Note that the method name is guessed from the value object's class: For instance, a <code>UserVO</code> results
   * in a method called <code>updateUser()</code>.
   * @param servicesTest The services test
   * @param valueObject The value object to check
   * @param properties A <code>Map</code> where the key is the name of the property, for instance <code>identifier</code>, and
   *        the value is an array of all the values that provoke an error 
   * @throws IllegalStateException if the validation is wrong
   */
  public static void checkValidatedValuesForUpdate( BaseServiceTestTemplate<?, ?, ?, ?> servicesTest, AbstractVO valueObject, Map<String, Object[]> properties )
      throws IllegalStateException
  {
    checkValidatedValues( servicesTest, UPDATE_METHOD, valueObject, properties );
  }

  /**
   * For all properties, this checks all the faulty values which must be caught by the Validator.
   * @param servicesTest The services test
   * @param methodName The method to call, for instance the one that inserts the value object into the DB
   * @param valueObject The value object to check
   * @param properties A <code>Map</code> where the key is the name of the property, for instance <code>identifier</code>, and
   *        the value is an array of all the values that provoke an error 
   * @throws IllegalStateException if the validation is wrong
   */
  public static void checkValidatedValues( BaseServiceTestTemplate<?, ?, ?, ?> servicesTest, String methodName, AbstractVO valueObject,
      Map<String, Object[]> properties ) throws IllegalStateException
  {
    for ( String property: properties.keySet() )
    {
      checkValidatedValues( servicesTest, methodName, valueObject, property, properties.get( property ) );
    }
  }

  /**
   * For one property, this checks all the faulty values which must be caught by the Validator for a method that updates the value object.
   * Note that the method name is guessed from the value object's class: For instance, a <code>UserVO</code> results
   * in a method called <code>updateUser()</code>.
   * @param servicesTest The services test
   * @param valueObject The value object to check
   * @param propertyName The name of the value object's property to check, for instance <code>identifier</code>
   * @param faultyValues All the values that provoke an error
   * @throws IllegalStateException if the validation is wrong
   */
  public static void checkValidatedValuesForInsert( BaseServiceTestTemplate<?, ?, ?, ?> servicesTest, AbstractVO valueObject, String propertyName,
      Object... faultyValues ) throws IllegalStateException
  {
    checkValidatedValues( servicesTest, INSERT_METHOD, valueObject, propertyName, faultyValues );
  }

  /**
   * For one property, this checks all the faulty values which must be caught by the Validator for a method that inserts the value object.
   * Note that the method name is guessed from the value object's class: For instance, a <code>UserVO</code> results
   * in a method called <code>insertUser()</code>.
   * @param servicesTest The services test
   * @param valueObject The value object to check
   * @param propertyName The name of the value object's property to check, for instance <code>identifier</code>
   * @param faultyValues All the values that provoke an error
   * @throws IllegalStateException if the validation is wrong
   */
  public static void checkValidatedValuesForUpdate( BaseServiceTestTemplate<?, ?, ?, ?> servicesTest, AbstractVO valueObject, String propertyName,
      Object... faultyValues ) throws IllegalStateException
  {
    checkValidatedValues( servicesTest, UPDATE_METHOD, valueObject, propertyName, faultyValues );
  }

  /**
   * For one property, this checks all the faulty values which must be caught by the Validator.
   * @param servicesTest The services test
   * @param methodName The method to call, for instance the one that inserts the value object into the DB
   * @param valueObject The value object to check
   * @param propertyName The name of the value object's property to check, for instance <code>identifier</code>
   * @param faultyValues All the values that provoke an error
   * @throws IllegalStateException if the validation is wrong
   */
  public static void checkValidatedValues( BaseServiceTestTemplate<?, ?, ?, ?> servicesTest, String methodName, AbstractVO valueObject, String propertyName,
      Object... faultyValues ) throws IllegalStateException
  {
    try
    {
      Object valueBackup = PropertyUtils.getProperty( valueObject, propertyName );
      for ( Object faultyValue: faultyValues )
      {
        try
        {
          PropertyUtils.setProperty( valueObject, propertyName, faultyValue );
          Method serviceMethod = servicesTest.getClass().getMethod( methodName, new Class[] { valueObject.getClass() } );
          serviceMethod.invoke( servicesTest, new Object[] { valueObject } );
          throw new IllegalStateException( "Calling " + methodName + " with a property '" + propertyName + "' containing '" + faultyValue + "' must fail." );
        }
        catch ( Exception e )
        {
          Throwable causingException = null;
          if ( e instanceof InvocationTargetException )
          {
            causingException = e.getCause();
          }
          if ( causingException != null && causingException instanceof WdlValidationException )
          {
            String illegalPropertyName = ( (WdlValidationException) causingException ).getInvalidValues().get( 0 ).getPropertyName();
            if ( !illegalPropertyName.equals( propertyName ) )
            {
              throw new IllegalStateException( "The validation was wrong for property '" + illegalPropertyName + "' instead of property '" + propertyName + "'" );
            }
            String errorMessage = ( (WdlValidationException) causingException ).getInvalidValues().get( 0 ).getMessage();
            if ( !errorMessage.startsWith( "validator." ) )
            {
              throw new IllegalStateException( "The validation message must contain a properties key, but contains '" + errorMessage + "'" );
            }
          }
          else
          {
            e.printStackTrace();
            throw e;
          }
        }
      }
      PropertyUtils.setProperty( valueObject, propertyName, valueBackup );
    }
    catch ( Exception e )
    {
      e.printStackTrace();
      throw new IllegalStateException( e );
    }
  }

  /**
   * Gets the object's name from its class name, without {@link #VO_SUFFIX}.
   * @param valueObject The value object whose name is taken
   * @return The name of the value object
   */
  protected static String getObjectName( AbstractVO valueObject )
  {
    String valueObjectName = ClassUtils.getShortClassName( valueObject.getClass() );
    if ( valueObjectName.endsWith( VO_SUFFIX ) )
    {
      valueObjectName = StringUtils.chomp( valueObjectName, VO_SUFFIX );
    }
    return valueObjectName;
  }
}
