package ch.ige.wdltool.vo;

import java.sql.Timestamp;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.hibernate.validator.Length;
import org.hibernate.validator.Min;
import org.hibernate.validator.NotNull;
import org.hibernate.validator.Range;

import ch.ige.dbutils.vo.AbstractVO;
import ch.ige.dbutils.vo.IdentifiableVO;
import ch.ige.wdltool.enums.DataSource;
import ch.ige.wdltool.enums.TermType;
import ch.ige.wdltool.validator.Language;
import ch.ige.wdltool.validator.NoMultipleBlanks;

/**
 * Contains a term with all member variables. Note that because of the Validator, certain getters whose validation
 * changes depending on the inheriting class are only protected. These must be overridden in the inheriting class and
 * that class must defined its validation. <p>Copyright (c) 2007, Swiss Federal Institute of Intellectual Property</p>
 * @author Eric Lewis
 * @version $Id: AbstractTermVO.java,v 1.30 2009/02/02 13:45:10 bandschapp Exp $
 */
public abstract class AbstractTermVO extends AbstractVO implements IdentifiableVO<Integer>
{
  /**
   * The unique ID which identifies this
   */
  protected Integer termId;

  /**
   * The data source owning this
   */
  protected DataSource dataSource;

  /**
   * The Nice class number for this
   */
  protected Integer classNumber;

  /**
   * The ID of the term group to which this belongs
   */
  protected Integer termGroupId;

  /**
   * The term type, which defines whether this is a
   * {@link ch.ige.wdltool.vo.AbstractTermVO} or a
   * {@link ch.ige.wdltool.vo.IpiTermVO} or a
   * {@link ch.ige.wdltool.vo.IpiTermCandidateVO}.
   */
  protected TermType termType;

  /**
   * The term's language
   */
  protected Locale language;

  /**
   * The term itself
   */
  protected String term;

  /**
   * The search term, a simplified version of {@link #term}
   */
  protected String searchTerm;

  /**
   * Shows whether this is a generic term within the Nice class definition
   */
  protected boolean isGeneric;

  /**
   * Shows whether this is visible for customers
   */
  protected boolean isPublic;

  /**
   * The comment only used within IPI
   */
  protected String privateComment;

  /**
   * The comment which is visible for customers
   */
  protected String publicComment;

  /**
   * The NT user ID of the user who last changed this
   */
  protected String changeUser;

  /**
   * The date/time of the latest change
   */
  protected Timestamp changeDate;

  /**
   * Shows whether the term is active at the moment.
   */
  protected boolean isActive;

  /**
   * The latest renewal date/time of the term
   */
  protected Timestamp renewal;

  /**
   * The number of times the term was found in a WDL
   */
  protected Integer usageCount;

  /**
   * The latest time when this was found in a WDL
   */
  protected Timestamp latestOccurrence;

  /**
   * The NT user ID of the user that created this
   */
  protected String creationUser;

  /**
   * Shows whether the term is relevant for the checker only or should be found when searching for terms. 
   */
  protected boolean isCheckerOnly;

  /**
   * If this term is part of a term group, then this contains a <code>Map</code> of all
   * other term group terms (excluding this term!) separated by term language.
   */
  protected Map<Locale, List<AbstractTermVO>> termGroupTerms;

  /**
   * A copy of the term group number which is common to all terms in {@link #termGroupTerms}.
   */
  protected Integer termGroupNumber;

  /**
   * The empty constructor
   */
  public AbstractTermVO()
  {
    super();
    initTermTypeAndDataSource();
  }

  /**
   * Creates this from an abstract VO.
   * @param vo The VO whose values are copied
   */
  public AbstractTermVO( AbstractTermVO vo )
  {
    this();
    this.changeDate = vo.changeDate;
    this.changeUser = vo.changeUser;
    this.classNumber = vo.classNumber;
    this.creationUser = vo.creationUser;
    this.dataSource = vo.dataSource;
    this.isActive = vo.isActive;
    this.isGeneric = vo.isGeneric;
    this.isPublic = vo.isPublic;
    this.language = vo.language;
    this.latestOccurrence = vo.latestOccurrence;
    this.privateComment = vo.privateComment;
    this.publicComment = vo.publicComment;
    this.renewal = vo.renewal;
    this.searchTerm = vo.searchTerm;
    this.term = vo.term;
    this.termGroupId = vo.termGroupId;
    this.termGroupNumber = vo.termGroupNumber;
    this.termId = vo.termId;
    this.termType = vo.termType;
    this.usageCount = vo.usageCount;
    if ( vo.termGroupTerms != null && ( !vo.termGroupTerms.isEmpty() ) )
    {
      this.termGroupTerms = new HashMap<Locale, List<AbstractTermVO>>( vo.termGroupTerms.size() );
      this.termGroupTerms.putAll( vo.termGroupTerms );
    }
    this.isCheckerOnly = vo.isCheckerOnly;
  }

  /**
   * Initializes the term type and the data source
   * for the concrete type of term.
   */
  protected abstract void initTermTypeAndDataSource();

  /**
   * Gets the unique identifier.
   * @return The unique identifier
   * @see ch.ige.dbutils.vo.IdentifiableVO#getIdentifier()
   */
  @NotNull
  // See the Validator's checkIdentifierExists flag
  @Min( value = 1 )
  public Integer getIdentifier()
  {
    return this.termId;
  }

  /**
   * Sets the unique identifier
   * @param id The unique identifier
   */
  public void setIdentifier( Integer id )
  {
    this.termId = id;
  }

  /**
   * Gets the data source owning this.
   * @return The data source owning this
   */
  @NotNull
  public DataSource getDataSource()
  {
    return this.dataSource;
  }

  /**
   * Gets the Nice class number for this.
   * @return the Nice class number for this
   */
  @Range( min = 1, max = 45 )
  public Integer getClassNumber()
  {
    return this.classNumber;
  }

  /**
   * Sets the Nice class number for this.
   * @param classNumber The Nice class number for this
   */
  public void setClassNumber( Integer classNumber )
  {
    this.classNumber = classNumber;
  }

  /**
   * Gets the ID of the term group to which this belongs.
   * A term group is either a standard WDL, an IPI term group
   * (used for translation) or a Nice term group (used for classification
   * and translation).
   * @return the ID of the term group owning this or <code>null</code> if
   *         it's not in a term group
   */
  @Min( value = 1 )
  protected Integer getTermGroupId()
  {
    return this.termGroupId;
  }

  /**
   * Sets the ID of the term group to which this belongs.
   * A term group is either a standard WDL, an IPI term group
   * (used for translation) or a Nice term group (used for classification
   * and translation).
   * @param termGroupId The ID of the term group to which this belongs
   */
  protected void setTermGroupId( Integer termGroupId )
  {
    this.termGroupId = termGroupId;
  }

  /**
   * Gets the term's language.
   * @return the term's language
   */
  @NotNull
  @Language( isoCodes = "de,fr,it" )
  public Locale getLanguage()
  {
    return this.language;
  }

  /**
   * Sets the term's language.
   * @param language The term's language to set
   */
  public void setLanguage( Locale language )
  {
    this.language = language;
  }

  /**
   * Gets the term.
   * @return the term
   */
  @NotNull
  @Length( min = 1, max = 1000 )
  @NoMultipleBlanks
  public String getTerm()
  {
    return this.term;
  }

  /**
   * Sets the term.
   * @param term The term to set
   */
  public void setTerm( String term )
  {
    this.term = term;
  }

  /**
   * Gets the search term, a simplified version of the term.
   * @return the search term used for searching this
   */
  @Length( min = 1, max = 1000 )
  @NoMultipleBlanks
  public String getSearchTerm()
  {
    return this.searchTerm;
  }

  /**
   * Sets the search term, a simplified version of the term.
   * @param searchTerm The search term used for searching this
   */
  public void setSearchTerm( String searchTerm )
  {
    this.searchTerm = searchTerm;
  }

  /**
   * Shows whether this is a generic term within the Nice class definition.
   * @return <code>true</code> if this is a generic term, otherwise <code>false</code>
   */
  protected boolean isGeneric()
  {
    return this.isGeneric;
  }

  /**
   * Sets the flag showing whether this is a generic term within the Nice class definition.
   * @param isGeneric <code>true</code> if this is a generic term, otherwise <code>false</code>
   */
  protected void setGeneric( boolean isGeneric )
  {
    this.isGeneric = isGeneric;
  }

  /**
   * Shows whether this is visible for customers.
   * @return <code>true</code> if this is visible for customers, <code>false</code> if it's
   *         only visible within IPI
   */
  public boolean isPublic()
  {
    return this.isPublic;
  }

  /**
   * Sets the flag that shows whether this is visible for customers.
   * @param isPublic <code>true</code> if this is visible for customers, <code>false</code> if it's
   *         only visible within IPI
   */
  protected void setPublic( boolean isPublic )
  {
    this.isPublic = isPublic;
  }

  /**
   * Gets the comment only used within IPI.
   * @return the comment only used within IPI
   */
  @Length( min = 1, max = 500 )
  @NoMultipleBlanks
  public String getPrivateComment()
  {
    return this.privateComment;
  }

  /**
   * Sets the comment only used within IPI.
   * @param privateComment The comment only used within IPI
   */
  public void setPrivateComment( String privateComment )
  {
    this.privateComment = privateComment;
  }

  /**
   * Gets the comment which is visible for customers.
   * @return the comment which is visible for customers
   */
  @Length( min = 1, max = 500 )
  @NoMultipleBlanks
  public String getPublicComment()
  {
    return this.publicComment;
  }

  /**
   * Sets the comment which is visible for customers.
   * @param publicComment The comment which is visible for customers
   */
  public void setPublicComment( String publicComment )
  {
    this.publicComment = publicComment;
  }

  /**
   * Gets the user ID (NT user ID) of the user who last changed this.
   * @return The NT user ID of the user who last changed this
   */
  @NotNull
  @Length( min = 1, max = 25 )
  public String getChangeUser()
  {
    return this.changeUser;
  }

  /**
   * Sets the user ID (NT user ID) of the user who last changed this.
   * @param changeUser The NT user ID of the user who last changed this
   */
  public void setChangeUser( String changeUser )
  {
    this.changeUser = changeUser;
  }

  /**
   * Gets the date/time of the latest change.
   * @return The date/time of the latest change
   */
  public Timestamp getChangeDate()
  {
    return this.changeDate;
  }

  /**
   * Sets the date/time of the latest change.
   * @param changeDate The date/time of the latest change
   */
  public void setChangeDate( Timestamp changeDate )
  {
    this.changeDate = changeDate;
  }

  /**
   * Gets the flag showing whether the term is active at the moment.
   * @return <code>true</code> if the term is active, otherwise <code>false</code>
   */
  protected boolean isActive()
  {
    return this.isActive;
  }

  /**
   * Sets the flag showing whether the term is active at the moment.
   * @param isActive <code>true</code> if the term is active, otherwise <code>false</code>
   */
  protected void setActive( boolean isActive )
  {
    this.isActive = isActive;
  }

  /**
   * Gets the latest time when this was found in a WDL.
   * @return The latest time when this was found in a WDL
   */
  protected Timestamp getLatestOccurrence()
  {
    return this.latestOccurrence;
  }

  /**
   * Sets the latest time when this was found in a WDL.
   * @param latestOccurrence The latest time when this was found in a WDL
   */
  protected void setLatestOccurrence( Timestamp latestOccurrence )
  {
    this.latestOccurrence = latestOccurrence;
  }

  /**
   * Gets the latest renewal date/time of the term.
   * @return The latest renewal date/time of the term
   */
  protected Timestamp getRenewal()
  {
    return this.renewal;
  }

  /**
   * Sets the latest renewal date/time of the term.
   * @param renewal The latest renewal date/time of the term
   */
  protected void setRenewal( Timestamp renewal )
  {
    this.renewal = renewal;
  }

  /**
   * Gets the number of times the term was found in a WDL.
   * @return The number of times the term was found in a WDL
   */
  protected Integer getUsageCount()
  {
    return this.usageCount;
  }

  /**
   * Sets the number of times the term was found in a WDL.
   * @param usageCount The number of times the term was found in a WDL
   */
  protected void setUsageCount( Integer usageCount )
  {
    this.usageCount = usageCount;
  }

  /**
   * Gets the user ID (NT user ID) of the user that created this.
   * @return The NT user ID of the user that created this
   */
  @NotNull
  @Length( min = 1, max = 25 )
  public String getCreationUser()
  {
    return this.creationUser;
  }

  /**
   * Sets the user ID (NT user ID) of the user that created this.
   * @param creationUser The NT user ID of the user that created this
   */
  public void setCreationUser( String creationUser )
  {
    this.creationUser = creationUser;
  }

  /**
   * If this term is part of a term group, then this returns a <code>Map</code> of all
   * other term group terms (excluding this term!) separated by term language.
   * <strong>Note: This is not always set, check the documentation of the method that creates this!</strong>
   * @return the other terms from the term group that this term belongs to
   */
  protected Map<Locale, List<AbstractTermVO>> getTermGroupTerms()
  {
    if ( this.termGroupTerms == null )
    {
      return Collections.emptyMap();
    }
    else
    {
      return this.termGroupTerms;
    }
  }

  /**
   * Sets the other terms from the term group that this term belongs to.
   * Note that this <strong>is never saved in the database!</strong>
   * @param termGroupTerms The other terms from the term group that this term belongs to,
   *        separated by term language
   */
  protected void setTermGroupTerms( Map<Locale, List<AbstractTermVO>> termGroupTerms )
  {
    this.termGroupTerms = termGroupTerms;
  }

  /**
   * Gets the term group number which is common to all terms in {@link #getTermGroupTerms()}.
   * <strong>Note: This is not always set, check the documentation of the method that creates this!</strong>
   * @return the term group number
   */
  protected Integer getTermGroupNumber()
  {
    return this.termGroupNumber;
  }

  /**
   * Sets the term group number.
   * @param termGroupNumber The term group number to set
   */
  protected void setTermGroupNumber( Integer termGroupNumber )
  {
    if ( this.termGroupId == null )
    {
      throw new IllegalStateException( "Term group number can't be set without term group ID" );
    }
    this.termGroupNumber = termGroupNumber;
  }

  /**
   * Gets the flag showing whether the term is to be used for checker only or also for searching.
   * @return <code>true</code> if the term is only for checking
   */
  public boolean isCheckerOnly()
  {
    return this.isCheckerOnly;
  }

  /**
   * Sets the flag showing whether the term is to be used only for checking.
   * @param isCheckerOnly <code>true</code> if the term is only for checking, otherwise <code>false</code>
   */
  public void setCheckerOnly( boolean isCheckerOnly )
  {
    this.isCheckerOnly = isCheckerOnly;
  }
}
